/* Copyright (C) 2000-2015 Lavtech.com corp. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
*/

#include "udm_config.h"

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <string.h>

#include "udm_common.h"
#include "udm_boolean.h"
#include "udm_utils.h"

/* #define DEBUG_BOOL */

#define UDM_MAXSTACK 128

static udm_rc_t
UdmBoolStackInit(UDM_BOOLSTACK *s)
{
  s->ncstack= 0;
  s->nastack= 0;
  s->mcstack= s->mastack= UDM_MAXSTACK;
  if (!(s->cstack= (udm_boolcmd_t*) UdmMalloc(UDM_MAXSTACK * sizeof(udm_boolcmd_t))))
    return UDM_ERROR;
  if (!(s->astack= (unsigned long*)UdmMalloc(UDM_MAXSTACK * sizeof(unsigned long))))
  {
    UDM_FREE(s->cstack);
    return UDM_ERROR;
  }
  return UDM_OK;
}


static void
UdmBoolStackFree(UDM_BOOLSTACK *s)
{
  UDM_FREE(s->cstack);
  UDM_FREE(s->astack);
}


#ifdef DEBUG_BOOL
static const char *
item_type(udm_boolcmd_t cmd, unsigned long arg)
{
  switch(cmd)
  {
    case UDM_BOOLCMD_LEFT:   return "(";
    case UDM_BOOLCMD_RIGHT:  return ")";
    case UDM_BOOLCMD_OR:     return "|";
    case UDM_BOOLCMD_AND:    return "&";
    case UDM_BOOLCMD_NOT:    return "~";
    case UDM_BOOLCMD_PHRASE: return "\"";
    case UDM_BOOLCMD_WORD:   return (arg ? "1" : "0");
    case UDM_BOOLCMD_BOT:    return "BOT";
    case UDM_BOOLCMD_STOP:   return "STOP";
  }
  return "?";
}
#endif

static int
TOPCMD(UDM_BOOLSTACK *s)
{
  if (s->ncstack)
    return s->cstack[s->ncstack-1];
  else
    return UDM_BOOLCMD_BOT;
}

static int
POPARG(UDM_BOOLSTACK *s)
{
  if (s->nastack>0)
  {
    s->nastack--;
    return s->astack[s->nastack];
  }
  else
  {
    return UDM_BOOLCMD_BOT;
  }
}


static udm_boolcmd_t
POPCMD(UDM_BOOLSTACK *s)
{
  if(s->ncstack>0)
  {
    s->ncstack--;
    return s->cstack[s->ncstack];
  }
  else
  {
    return UDM_BOOLCMD_BOT;
  }
}


static void
PUSHARG(UDM_BOOLSTACK *s, unsigned long arg)
{
  s->astack[s->nastack]= arg;
  s->nastack++;
  if (s->nastack >= s->mastack)
  {
    s->mastack+= UDM_MAXSTACK;
    s->astack= (unsigned long*)UdmRealloc(s->astack, s->mastack * sizeof(unsigned long));
  }
}

static void
PUSHCMD(UDM_BOOLSTACK *s, udm_boolcmd_t arg)
{
  s->cstack[s->ncstack]= arg;
  s->ncstack++;
  if (s->ncstack >= s->mcstack)
  {
    s->mcstack+= UDM_MAXSTACK;
    s->cstack= (udm_boolcmd_t*) UdmRealloc(s->cstack, s->mcstack * sizeof(udm_boolcmd_t));
  }
}

/* Compute one operation and store result */
static int
perform(UDM_BOOLSTACK *s, udm_boolcmd_t com)
{
  unsigned long res,x1,x2;
  switch(com)
  {
    case UDM_BOOLCMD_OR:
      /* This is to avoid compiler      */
      /* optimization in || calculation */
      x1=POPARG(s);
      x2=POPARG(s);
      res=(x1||x2);
#ifdef DEBUG_BOOL
      fprintf(stderr,"Perform %lu | %lu\n",x1,x2);
#endif
      PUSHARG(s,res);
      break;
    case UDM_BOOLCMD_AND:
      x1=POPARG(s);
      x2=POPARG(s);
      res=(x1&&x2);
#ifdef DEBUG_BOOL
      fprintf(stderr,"Perform %lu & %lu\n",x1,x2);
#endif
      PUSHARG(s,res);
      break;
    case UDM_BOOLCMD_NOT:
      x1=POPARG(s);
      res=x1?0:1;
#ifdef DEBUG_BOOL
      fprintf(stderr,"Perform ~%lu\n",x1);
#endif
      PUSHARG(s,res);
      break;
    case UDM_BOOLCMD_LEFT:
    case UDM_BOOLCMD_RIGHT:
    case UDM_BOOLCMD_BOT:
    case UDM_BOOLCMD_PHRASE:
    case UDM_BOOLCMD_WORD:
    case UDM_BOOLCMD_STOP:
      break;
  }
  return(0);
}


/* Main function to calculate items sequence */
int
UdmBoolExprExec(const UDM_BOOLEXPR *L, char *count)
{
  UDM_BOOLSTACK s;
  int res;
  UDM_BOOLEXPR_ITEM *Item, *Last;

  if (UDM_OK != UdmBoolStackInit(&s)) /* TODO34: prealloc this! */
    return 0; /*TODO34: expose error to the caller */

#ifdef DEBUG_BOOL
  for (Item= L->items, Last= &L->items[L->nitems]; Item < Last; Item++)
  {
    fprintf(stderr, "[%d]cmd=%d arg=%d\n",
            (int) (Item - L->items), (int) Item->cmd, (int) Item->arg);
  }
#endif

#ifdef DEBUG_BOOL
  fprintf(stderr,"--------\n");
#endif
  for (Item= L->items, Last= &L->items[L->nitems]; Item < Last; Item++)
  {
    int c;
#ifdef DEBUG_BOOL
    fprintf(stderr,".[%d] %s\n", (int) (Item - L->items), item_type(Item->cmd, Item->arg));
#endif
    switch ((c= Item->cmd))
    {
      case UDM_BOOLCMD_RIGHT:
        /* Perform till LEFT bracket */
        while((TOPCMD(&s)!=UDM_BOOLCMD_LEFT)&&(TOPCMD(&s)!=UDM_BOOLCMD_BOT))
          perform(&s,POPCMD(&s));
        /* Pop LEFT bracket itself */
        if(TOPCMD(&s)==UDM_BOOLCMD_LEFT)
          POPCMD(&s);
        break;
      case UDM_BOOLCMD_OR:
      case UDM_BOOLCMD_AND:
        while(c<=TOPCMD(&s))
          perform(&s,POPCMD(&s));
        /* IMPORTANT! No break here! That's OK*/
        /*   ! */
      case UDM_BOOLCMD_LEFT:
      case UDM_BOOLCMD_NOT:
        PUSHCMD(&s,c);
        break;
      case UDM_BOOLCMD_PHRASE:
        PUSHARG(&s, (count[Item[1].arg]) ? 1UL : 0UL);
        for (Item++; Item->cmd != UDM_BOOLCMD_PHRASE; Item++)
#ifdef DEBUG_BOOL
    fprintf(stderr,"[%d] %s\n", (int) (Item - L->items), item_type(Item->cmd, count[Item->arg]));
#endif
        ;
        break;
      case UDM_BOOLCMD_STOP:
        PUSHARG(&s, 1UL);
        break;
      case UDM_BOOLCMD_WORD:
        PUSHARG(&s, (count[Item->arg]) ? 1UL : 0UL);
        break;
      default:
        UDM_ASSERT(0);
    }
  }
  while(TOPCMD(&s) != UDM_BOOLCMD_BOT)
    perform(&s, POPCMD(&s));

  res= POPARG(&s);
  UdmBoolStackFree(&s);
#ifdef DEBUG_BOOL
  fprintf(stderr, "result: %d\n", res);
#endif
  return res;
}


void
UdmBoolExprFree(UDM_BOOLEXPR *L)
{
  UDM_FREE(L->items);
  bzero((void*) L, sizeof(*L));
}


udm_rc_t
UdmBoolExprAddItem(UDM_BOOLEXPR *List, const UDM_BOOLEXPR_ITEM *item)
{
  if (List->nitems >= List->mitems)
  {
    List->mitems+= UDM_MAXSTACK;
    if (!(List->items= (UDM_BOOLEXPR_ITEM*)UdmRealloc(List->items,
                                                      List->mitems *
                                                      sizeof(UDM_BOOLEXPR_ITEM))))
      return UDM_ERROR;
  }
  List->items[List->nitems]= item[0];
  List->nitems++;
  return UDM_OK;
}


typedef struct udm_stack_stat_st
{
  size_t n_left;
  size_t n_right;
  size_t n_or;
  size_t n_and;
  size_t n_not;
  size_t n_phrase;
  size_t n_word;
  size_t n_stop;
} UDM_BOOLEXPR_STAT;


static void
UdmBoolExprStatAdd(UDM_BOOLEXPR_STAT *S, const UDM_BOOLEXPR_ITEM *item)
{
  switch (item->cmd)
  {
    case UDM_BOOLCMD_LEFT:   S->n_left++;   break;
    case UDM_BOOLCMD_RIGHT:  S->n_right++;  break;
    case UDM_BOOLCMD_OR:     S->n_or++;     break;
    case UDM_BOOLCMD_AND:    S->n_and++;    break;
    case UDM_BOOLCMD_NOT:    S->n_not++;    break;
    case UDM_BOOLCMD_PHRASE: S->n_phrase++; break;
    case UDM_BOOLCMD_WORD:   S->n_word++;   break;
    case UDM_BOOLCMD_STOP:   S->n_stop++;   break;
    case UDM_BOOLCMD_BOT:    break;
    default:
      UDM_ASSERT(0);
      break;
  }
}

static udm_rc_t
UdmBoolExprAddItem_with_Stat(UDM_BOOLEXPR *List,
                             UDM_BOOLEXPR_STAT *Stat,
                             const UDM_BOOLEXPR_ITEM *item)
{
  UdmBoolExprStatAdd(Stat, item);
  return UdmBoolExprAddItem(List, item);
}


udm_rc_t
UdmBoolExprCopy(UDM_BOOLEXPR *Dst,
                const UDM_BOOLEXPR *Src,
                udm_search_mode_t search_mode)
{
  size_t i;
  udm_rc_t rc;
  int inphrase;
  udm_boolcmd_t add_cmd= UDM_BOOLCMD_AND;
  UDM_BOOLEXPR_STAT Stat;
  
  bzero((void*) &Stat, sizeof(Stat));
  bzero((void*) Dst, sizeof(*Dst));
  
  switch(search_mode)
  {
    case UDM_MODE_ANY:
      add_cmd= UDM_BOOLCMD_OR;
      break;
    case UDM_MODE_BOOL:
    case UDM_MODE_ALL:
    case UDM_MODE_ALL_MINUS:
    case UDM_MODE_ALL_MINUS_HALF:
    case UDM_MODE_PHRASE:
      add_cmd= UDM_BOOLCMD_AND;
      break;
  }

  if (UDM_OK != (rc= UdmBoolExprAddItem_with_Stat(Dst, &Stat, &Src->items[0])))
    return rc;
  
  inphrase= (Src->items[0].cmd == UDM_BOOLCMD_PHRASE) ? 1 : 0;

  for (i= 1; i < Src->nitems; i++)
  {
    /*
      If previous item is WORD or PHRASE or RIGHT or STOPWORD
      and next item is WORD or PHRASE or LEFT or STOPWORD or NOT
      and we are not in phrase
      we have to insert search mode dependent operator.
     */
    if ((Src->items[i - 1].cmd == UDM_BOOLCMD_WORD ||
         Src->items[i - 1].cmd == UDM_BOOLCMD_STOP ||
         Src->items[i - 1].cmd == UDM_BOOLCMD_PHRASE ||
         Src->items[i - 1].cmd == UDM_BOOLCMD_RIGHT) &&
        (Src->items[i].cmd == UDM_BOOLCMD_WORD ||
         Src->items[i].cmd == UDM_BOOLCMD_STOP ||
         Src->items[i].cmd == UDM_BOOLCMD_PHRASE ||
         Src->items[i].cmd == UDM_BOOLCMD_LEFT ||
         Src->items[i].cmd == UDM_BOOLCMD_NOT) &&
         !inphrase)
    {
      UDM_BOOLEXPR_ITEM add;
      add.cmd= add_cmd;
      add.arg= 0;
      if (UDM_OK != (rc= UdmBoolExprAddItem_with_Stat(Dst, &Stat, &add)))
        return rc;
    }
    if (Src->items[i].cmd == UDM_BOOLCMD_PHRASE)
      inphrase= !inphrase;
    if (UDM_OK != (rc= UdmBoolExprAddItem_with_Stat(Dst, &Stat, &Src->items[i])))
      return rc;
  }
  /*
  fprintf(stderr, "paren=%d/%d or=%d and=%d not=%d phrase=%d word=%d stop=%d\n",
          Stat.n_left, Stat.n_right, Stat.n_or, Stat.n_and, Stat.n_not,
          Stat.n_phrase, Stat.n_word, Stat.n_stop);
  */       
  /* Ignore phrases in loose modes */
  if (search_mode == UDM_MODE_ALL_MINUS ||
      search_mode == UDM_MODE_ALL_MINUS_HALF)
  {
    Stat.n_phrase= 0;
  }
  
  /*
    If no commands other than UDM_BOOLCMD_PHRASE and UDM_BOOLCMD_AND and parenthesis,
    then no needs to calculate boolean expression.
  */
  if (!Stat.n_phrase && !Stat.n_or && !Stat.n_not)
  {
    UdmBoolExprFree(Dst);
  }
  return UDM_OK;
}
