/* Copyright (C) 2000-2015 Lavtech.com corp. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
*/

#include "udm_config.h"
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <string.h>
#include <errno.h>

#include "udm_common.h"
#include "udm_ctype.h"
#include "udm_spell.h"
#include "udm_proto.h"
#include "udm_url.h"
#include "udm_parser.h"
#include "udm_conf.h"
#include "udm_log.h"
#include "udm_hrefs.h"
#include "udm_robots.h"
#include "udm_utils.h"
#include "udm_host.h"
#include "udm_server.h"
#include "udm_alias.h"
#include "udm_search_tl.h"
#include "udm_env.h"
#include "udm_match.h"
#include "udm_stopwords.h"
#include "udm_guesser.h"
#include "udm_unicode.h"
#include "udm_synonym.h"
#include "udm_vars.h"
#include "udm_db.h"
#include "udm_agent.h"
#include "udm_chinese.h"
#include "udm_contentencoding.h"
#include "udm_indexcache.h"

static udm_rc_t EnvLoad(UDM_CFG *Cfg,const char *cname);

/****************************  Load Configuration **********************/

udm_search_mode_t
UdmSearchMode(const char *mode)
{
  if(!mode)return(UDM_MODE_ALL);
  if(!strcmp(mode,"all-minus"))return(UDM_MODE_ALL_MINUS);
  if(!strcmp(mode,"all-minus-half"))return(UDM_MODE_ALL_MINUS_HALF);
  if(!strcmp(mode,"all"))return(UDM_MODE_ALL);
  if(!strcmp(mode,"any"))return(UDM_MODE_ANY);
  if(!strcmp(mode,"bool"))return(UDM_MODE_BOOL);
  if(!strcmp(mode,"phrase"))return(UDM_MODE_PHRASE);
  return(UDM_MODE_ALL);
}


udm_match_mode_t
UdmMatchMode(const char * mode)
{
  if(!mode)return(UDM_MATCH_FULL);
  if(!strcmp(mode,"wrd"))return(UDM_MATCH_FULL);
  if(!strcmp(mode,"full"))return(UDM_MATCH_FULL);
  if(!strcmp(mode,"beg"))return(UDM_MATCH_BEGIN);
  if(!strcmp(mode,"end"))return(UDM_MATCH_END);
  if(!strcmp(mode,"sub"))return(UDM_MATCH_SUBSTR);
  return(UDM_MATCH_FULL);
}


UDM_API(const char *)
UdmFollowStr(udm_webspace_t method)
{
  switch(method)
  {
    case UDM_WEBSPACE_PAGE:  return "Page";
    case UDM_WEBSPACE_PATH:  return "Path";
    case UDM_WEBSPACE_SITE:  return "Site";
    case UDM_WEBSPACE_WORLD: return "World";
    case UDM_WEBSPACE_URLLIST: return "URLList";
    case UDM_WEBSPACE_UNKNOWN: break;
  }
  return "<Unknown follow type>";
}


udm_webspace_t UdmFollowType(const char * follow)
{
  UDM_ASSERT(follow != NULL);
  if (!strcasecmp(follow,"page")) return UDM_WEBSPACE_PAGE;
  if (!strcasecmp(follow,"path")) return UDM_WEBSPACE_PATH;
  if (!strcasecmp(follow,"site")) return UDM_WEBSPACE_SITE;
  if (!strcasecmp(follow,"world")) return UDM_WEBSPACE_WORLD;
  if (!strcasecmp(follow,"urllist")) return UDM_WEBSPACE_URLLIST;
  return UDM_WEBSPACE_UNKNOWN;
}

const char *UdmMethodStr(udm_method_t method)
{
  switch(method)
  {
    case UDM_METHOD_DISALLOW:     return "Disallow";
    case UDM_METHOD_GET:          return "Allow";
    case UDM_METHOD_CHECKMP3ONLY: return "CheckMP3Only";
    case UDM_METHOD_CHECKMP3:     return "CheckMP3";
    case UDM_METHOD_HEAD:         return "CheckOnly";
    case UDM_METHOD_HREFONLY:     return "HrefOnly";
    case UDM_METHOD_VISITLATER:   return "Skip";
    case UDM_METHOD_INDEX:        return "IndexIf";
    case UDM_METHOD_NOINDEX:      return "NoIndexIf";
    case UDM_METHOD_IMPORTONLY:   return "ImportOnly";
    case UDM_METHOD_UNKNOWN:      break;
  }
  return "<Unknown method>";
}


udm_method_t UdmMethod(const char *s)
{
  if (s == NULL)
    return UDM_METHOD_UNKNOWN;
  if (!strcasecmp(s,"Disallow"))     return UDM_METHOD_DISALLOW;
  if (!strcasecmp(s,"Allow"))        return UDM_METHOD_GET;
  if (!strcasecmp(s,"CheckMP3Only")) return UDM_METHOD_CHECKMP3ONLY;
  if (!strcasecmp(s,"CheckMP3"))     return UDM_METHOD_CHECKMP3;
  if (!strcasecmp(s,"CheckOnly"))    return UDM_METHOD_HEAD;
  if (!strcasecmp(s,"HrefOnly"))     return UDM_METHOD_HREFONLY;
  if (!strcasecmp(s,"Skip"))         return UDM_METHOD_VISITLATER;
  if (!strcasecmp(s,"SkipIf"))       return UDM_METHOD_VISITLATER;
  if (!strcasecmp(s,"IndexIf"))      return UDM_METHOD_INDEX;
  if (!strcasecmp(s,"NoIndexIf"))    return UDM_METHOD_NOINDEX;
  if (!strcasecmp(s,"ImportOnly"))   return UDM_METHOD_IMPORTONLY;
  return UDM_METHOD_UNKNOWN;
}


/*
  Convert 0..9A..Z into integer.
  TODO34: using WF with big values (e.g. 'Q') overflows
*/
static int
UdmBase_09AZ_to_int(int h)
{
  if((h>='0')&&(h<='9'))return(h-'0');
  if((h>='A')&&(h<='Z'))return(h-'A'+10);
  if((h>='a')&&(h<='Z'))return(h-'a'+10);
  return 0;
}


static void
UdmWeightFactorsInit(char *res, const char *wf, size_t num)
{
  size_t len;
  int sn;
  
  for(sn=0;sn<256;sn++)
    res[sn]=1;
  
  len=strlen(wf);
  if((len>0)&&(len<256))
  {
    const char *sec;
    int secno;
    
    for(sec= wf + len - 1, secno= 1; sec >= wf; sec--)
    {
      if (*sec != '-' && *sec != '.')
      {
        res[secno]= UdmBase_09AZ_to_int(*sec);
        secno++;
      }
    }
  }
  
  for (sn= num + 1 ; sn < 256; sn++)
    res[sn]= 0;
}


size_t UdmWeightFactorsInit2(char *res,
                             UDM_VARLIST *V1,
                             UDM_VARLIST *V2,
                             const char *name)
{
  size_t numsections= UdmVarListFindInt(V1, "NumSections", 256);
  const char *wf1= UdmVarListFindStr(V1, name, "");
  const char *wf2= UdmVarListFindStr(V2, name, "");
  const char *wf3= wf2[0] ? wf2 : wf1;
  UdmWeightFactorsInit(res, wf3, numsections);
  return wf3[0] ? numsections : 0;
}

/* hold the path of the current config file */
static char current[1024]= ""; /* TODO34: get rid of this */

static void update_current(const char *name)
{
  char *slash;
  strcpy(current, name);
  if ((slash= strrchr(current, '/')))
  {
   *slash= 0;
  }
  else
  {
    *current= 0;
  }
}

static size_t rel_cur_name(char *res, size_t maxlen, const char *name)
{
  size_t n = 0;
  if (*current)
  {
#ifdef WIN32
    n= udm_snprintf(res, maxlen, "%s", name);
#else
    n= udm_snprintf(res, maxlen, "%s%s%s", current, UDMSLASHSTR, name);
#endif
  }
  else
  {
    strncpy(res, name, maxlen);
  }
  res[maxlen]= 0;
  return n;
}


static size_t
rel_name(UDM_ENV *Env, char *res, size_t maxlen,
         const char *varname, const char *dirname, const char *name)
{
  size_t    n;
  const char  *dir= UdmVarListFindStr(&Env->Vars, varname, dirname);
  if(name[0]=='/' || (name[0] && name[1] == ':'))
    n= udm_snprintf(res, maxlen, "%s", name);
  else
    n= udm_snprintf(res,maxlen,"%s%s%s",dir,UDMSLASHSTR,name);
  res[maxlen]='\0';
  return n;
}


/* Relative name for included .conf files */
static size_t
rel_etc_name(UDM_ENV *Env, char *res, size_t maxlen, const char *fname)
{
  return rel_name(Env, res, maxlen, "ConfDir", UDM_CONF_DIR, fname);
}


/*
  Relative name for langmap, stopwords, synonym files.
  Their position depends on --enable-fhs-layot.
*/
static size_t
rel_etc2_name(UDM_ENV *Env, char *res, size_t maxlen, const char *fname)
{
#ifdef HAVE_FHS_LAYOUT
  /* Better FHS layout */
  return rel_name(Env, res, maxlen, "ShareDir", UDM_SHARE_DIR, fname);
#else
  /* Traditional mnoGoSearch layout */
  return rel_name(Env, res, maxlen, "ConfDir", UDM_CONF_DIR, fname);
#endif
}


static size_t
rel_langmap_name(UDM_ENV *Env, char *res, size_t maxlen, const char *fname)
{
  return rel_etc2_name(Env, res, maxlen, fname);
}


static size_t
rel_stopwords_name(UDM_ENV *Env, char *res, size_t maxlen, const char *fname)
{
  return rel_etc2_name(Env, res, maxlen, fname);
}


static size_t
rel_synonym_name(UDM_ENV *Env, char *res, size_t maxlen, const char *fname)
{
  return rel_etc2_name(Env, res, maxlen, fname);
}


/*
  *.freq files is a special case.
  It's installed to /share/freq/xxx.freq with --enable-fhs-layout,
  and to /etc/xxx.freq otherwise.
*/
static size_t
rel_freq_name(UDM_ENV *Env, char *res, size_t maxlen, const char *fname)
{
#ifdef HAVE_FHS_LAYOUT
  char freqname[128];
  if (fname[0] == '/')
    udm_snprintf(freqname, sizeof(freqname), "%s", fname);
  else
    udm_snprintf(freqname, sizeof(freqname), "freq/%s", fname);
  return rel_name(Env, res, maxlen, "ShareDir", UDM_SHARE_DIR, freqname);
#else
  return rel_name(Env, res, maxlen, "ConfDir", UDM_CONF_DIR, fname);
#endif
}


/* Relative name for /var files */
static size_t rel_var_name(UDM_ENV *Env,char *res,size_t maxlen,
                           const char *name)
{
  size_t    n;
  const char  *dir=UdmVarListFindStr(&Env->Vars,"VarDir",UDM_VAR_DIR);
  if(name[0]=='/')n = udm_snprintf(res, maxlen, "%s", name);
  else    n = udm_snprintf(res,maxlen,"%s%s%s",dir,UDMSLASHSTR,name);
  res[maxlen]='\0';
  return n;
}

size_t
UdmGetArgs(char *str, const char **av, size_t max)
{
  size_t  ac=0;
  char  *lt;
  char  *tok;
  
  bzero((void*)av, max * sizeof(*av));
  tok=UdmGetStrToken(str,&lt);
  
  while (tok && (ac<max))
  {
    av[ac]=tok;
    ac++;
    tok=UdmGetStrToken(NULL,&lt);
  }
  return ac;
}


static udm_bool_t
UdmMatchParamModeRegexOrString(UDM_MATCH_PARAM *Param, const char *prm)
{
  if (!strcasecmp(prm, "string"))
  {
    Param->match_mode= UDM_MATCH_WILD;
    return UDM_FALSE;
  }
  if (!strcasecmp(prm, "regex"))
  {
    Param->match_mode= UDM_MATCH_REGEX;
    return UDM_FALSE;
  }
  if (!strcasecmp(prm, "regexp"))
  {
    Param->match_mode= UDM_MATCH_REGEX;
    return UDM_FALSE;
  }
  return UDM_TRUE;
}


static udm_bool_t
UdmMatchParamStrToMode(UDM_MATCH_PARAM *Param, const char *prm)
{
  /* Match modes */
  if (!UdmMatchParamModeRegexOrString(Param, prm))
    return UDM_FALSE;
  if (!strcasecmp(prm, "prefix"))
  {
    Param->match_mode= UDM_MATCH_BEGIN;
    return UDM_FALSE;
  }
  return UDM_TRUE;
}


static udm_bool_t
UdmMatchParamStrToCaseSensitivity(UDM_MATCH_PARAM *Param, const char *prm)
{
  /* Case sensitivity */
  if (!strcasecmp(prm, "nocase"))
  {
    UdmMatchParamSetCaseInsensitive(Param, UDM_FALSE);
    return UDM_FALSE;
  }
  if (!strcasecmp(prm, "case"))
  {
    UdmMatchParamSetCaseInsensitive(Param, UDM_TRUE);
    return UDM_FALSE;
  }
  return UDM_TRUE;
}


static udm_bool_t
UdmMatchParamStrToNegative(UDM_MATCH_PARAM *Param, const char *prm)
{
  /* Negative condition */
  if (!strcasecmp(prm, "match"))
  {
    UdmMatchParamSetNegative(Param, UDM_FALSE);
    return UDM_FALSE;
  }
  if (!strcasecmp(prm, "nomatch"))
  {
    UdmMatchParamSetNegative(Param, UDM_TRUE);
    return UDM_FALSE;
  }
  return UDM_TRUE;
}


static udm_bool_t
UdmMatchParamSetParam(UDM_MATCH_PARAM *Param, const char *prm)
{
  if (!UdmMatchParamStrToMode(Param, prm))
    return UDM_FALSE;
  if (!UdmMatchParamStrToCaseSensitivity(Param, prm))
    return UDM_FALSE;
  if (!UdmMatchParamStrToNegative(Param, prm))
    return UDM_FALSE;
  return UDM_TRUE;
}


static udm_rc_t
add_srv(UDM_CFG *C, size_t ac, const char **av)
{
  UDM_ENV  *Conf=C->Indexer->Conf;
  UDM_AGENT *Indexer = C->Indexer;
  size_t  i;
  int  has_alias=0;
  
  if(!(C->flags & UDM_FLAG_ADD_SERV))
    return UDM_OK;
  
  C->Srv->command = 'S';
  C->Srv->webspace= UDM_WEBSPACE_DEFAULT;
  C->Srv->ordre = ++C->ordre;
  C->Srv->Filter.method= UDM_METHOD_DEFAULT;
  UdmMatchParamSetNegative(&C->Srv->Filter.Match.Param, UDM_FALSE);
  UdmMatchParamSetCaseInsensitive(&C->Srv->Filter.Match.Param, UDM_TRUE);
  UdmMatchParamSetOptimization(&C->Srv->Filter.Match.Param,
                               !UDM_TEST((C->flags & UDM_FLAG_DONT_ADD_TO_DB)));

  if(!strcasecmp(av[0],"Server"))
  {
    C->Srv->Filter.Match.Param.match_mode= UDM_MATCH_BEGIN;
  }
  else if(!strcasecmp(av[0],"Subnet"))
  {
    C->Srv->Filter.Match.Param.match_mode= UDM_MATCH_SUBNET;
    Conf->Servers.have_subnets=1;
  }
  else
  {
    C->Srv->Filter.Match.Param.match_mode= UDM_MATCH_WILD;
  }
  
  
  for(i=1; i<ac; i++)
  {
    int  o;
    
    if (UDM_WEBSPACE_UNKNOWN!= (o= UdmFollowType(av[i])))
      C->Srv->webspace= (udm_webspace_t) o;
    else if (UDM_METHOD_UNKNOWN!= (o= UdmMethod(av[i])))
      C->Srv->Filter.method= (udm_method_t) o;
    else if (UdmMatchParamSetParam(&C->Srv->Filter.Match.Param, av[i]))
    {
      if (!UdmMatchPatternConstStr(&C->Srv->Filter.Match))
        UdmMatchSetPattern(&C->Srv->Filter.Match, av[i]);
      else if(!has_alias)
      {
        has_alias=1;
        UdmVarListReplaceStr(&C->Srv->Vars,"Alias",av[i]);
      }
      else
      {
        sprintf(Conf->errstr,"too many argiments: '%s'",av[i]);
        return UDM_ERROR;
      }
    }
  }
  if (!UdmMatchPatternConstStr(&C->Srv->Filter.Match))
  {
    sprintf(Conf->errstr,"too few argiments in '%s' command", av[0]);
    return UDM_ERROR;
  }
  if(UDM_OK != UdmServerAdd(Indexer, C->Srv, C->flags))
  {
    char * s_err;
    s_err = (char*)UdmStrdup(Conf->errstr);
    sprintf(Conf->errstr,"%s",s_err);
    UDM_FREE(s_err);
    UdmMatchFreeAndInit(&C->Srv->Filter.Match);
    return UDM_ERROR;
  }
  if ((C->Srv->Filter.Match.Param.match_mode == UDM_MATCH_BEGIN) &&
      (UdmMatchPatternConstStr(&C->Srv->Filter.Match)[0])&&
      (C->flags&UDM_FLAG_ADD_SERVURL))
  {
    UDM_HREFPARAM HrefParam;
    UdmHrefParamInit(&HrefParam);
    HrefParam.server_id= C->Srv->site_id;
    HrefParam.hops= (uint4) UdmVarListFindInt(&C->Srv->Vars, "StartHops", 0);
    HrefParam.link_source= UDM_LINK_SOURCE_CONF;
    UdmHrefListAddConst(&Conf->Hrefs, &HrefParam,
                        UdmMatchPatternConstStr(&C->Srv->Filter.Match));
  }
  UdmMatchFreeAndInit(&C->Srv->Filter.Match);
  UdmVarListDel(&C->Srv->Vars,"AuthBasic");
  UdmVarListDel(&C->Srv->Vars,"Alias");
  return UDM_OK;
}


static udm_rc_t
UdmFilterListAddWithServer(UDM_AGENT *A,
                           UDM_FILTERLIST *L, UDM_MATCH_PARAM *Param,
                           udm_method_t method,
                           const UDM_CONST_STR *Pattern,
                           char *err, size_t errsize, int ordre)
{
  if (UDM_OK != UdmFilterListAdd(L, Param, method, Pattern, err, errsize))
    return UDM_ERROR;
  
  if (A != NULL)
  {
    UDM_SERVERLIST S;
    UDM_SERVER n;
    udm_rc_t rc;
    
    bzero((void*)&n, sizeof(n));
    S.Server= &n;
    n.command= 'F';
    UdmMatchSetPattern(&n.Filter.Match, Pattern->str);
    n.Filter.Match.Param= *Param;
    n.Filter.method= method;
    n.ordre= ordre;
    
    rc= UdmSrvAction(A, &S, UDM_SRV_ACTION_ADD);
    UdmVarListFree(&n.Vars);
    
    if (rc != UDM_OK) return rc;
  }
  return UDM_OK;
}


static udm_rc_t
add_alias(UDM_CFG *C, size_t ac, const char **av)
{
  UDM_ENV *Conf = C->Indexer->Conf;
  UDM_MATCH_PARAM Param;
  UDM_CONST_STR Pattern;
  size_t    i;

  UdmConstStrInit(&Pattern);
  UdmMatchParamInit(&Param);
  Param.match_mode= UDM_MATCH_BEGIN;
  UdmMatchParamSetCaseInsensitive(&Param, UDM_TRUE);
  UdmMatchParamSetOptimization(&Param,
                               !UDM_TEST(C->flags & UDM_FLAG_DONT_ADD_TO_DB));
  for (i= 1; i < ac; i++)
  {
    if (!UdmMatchParamStrToMode(&Param, av[i]) ||
        !UdmMatchParamStrToCaseSensitivity(&Param, av[i]))
      continue;

    if (!Pattern.str)
    {
      UdmConstStrSetStr(&Pattern, av[i]);
    }
    else
    {
      char err[120]= "";
      UDM_REPLACELIST  *L= NULL;
      UDM_CONST_STR Alias;

      UdmConstStrSetStr(&Alias, av[i]);
      if (!strcasecmp(av[0], "Alias")) L= &Conf->Aliases;
      if (!strcasecmp(av[0], "ReverseAlias")) L= &Conf->ReverseAliases;

      if (UDM_OK != UdmReplaceListAdd(L, &Param, &Pattern, &Alias, err, sizeof(err)))
      {
        udm_snprintf(Conf->errstr, sizeof(Conf->errstr), "%s", err);
        return UDM_ERROR;
      }
    }
  }
  if (!Pattern.str)
  {
    udm_snprintf(Conf->errstr, sizeof(Conf->errstr), "too few arguments");
    return UDM_ERROR;
  }
  return UDM_OK;
}


static udm_rc_t
UdmExcerptQualityParse(UDM_ENV *Env,
                       UDM_EXCERPT_FRAGMENT *Fragment, const char *str)
{
  if (!strcasecmp(str, "ignore"))
  {
    Fragment->quality= 0;
    return UDM_OK;
  }
  if (udm_isdigit(str[0]))
  {
    Fragment->quality= atoi(str);
    return UDM_OK;
  }
  udm_snprintf(Env->errstr, sizeof(Env->errstr),
               "ExcerptFragment: syntax error near '%s'; "
               "Expected 'ignore' or an unsigned number.", str);
  return UDM_ERROR;
}


/*
  ExcerptFragment  [Case | NoCase] [String | Regex] quality pattern
*/
static udm_rc_t
add_excerpt_fragment(UDM_CFG *C, size_t ac, const char **av)
{
  UDM_ENV *Conf= C->Indexer->Conf;
  UDM_EXCERPT_FRAGMENT Specific, *VarSpecific;
  UDM_VAR *Var;
  char err[128];
  char name[64];
  size_t i;
  udm_rc_t rc;
  UDM_VALUE_HANDLER *ha= &UdmValueHandlerExcerptFragment;

  UdmExcerptFragmentInit(&Specific);
  Specific.Match.Param.match_mode= UDM_MATCH_WILD;
  UdmMatchParamSetCaseInsensitive(&Specific.Match.Param, UDM_TRUE);

  for (i= 1; i < ac; i++)
  {
    if (UdmMatchParamModeRegexOrString(&Specific.Match.Param, av[i]) &&
        UdmMatchParamStrToCaseSensitivity(&Specific.Match.Param, av[i]))
      break;
  }

  if (ac - i != 2)
  {
    udm_snprintf(Conf->errstr, sizeof(Conf->errstr) - 1,
                 "too %s (%d) required paramenters",
                 ac - i > 2 ? "many" : "few", (int) (ac - i));
    return UDM_ERROR;
  }

  if (UDM_OK != UdmExcerptQualityParse(Conf, &Specific, av[i]))
  {
    UdmExcerptFragmentFree(&Specific);
    return UDM_ERROR;
  }

  if (UDM_OK != UdmMatchSetPattern(&Specific.Match, av[i + 1]))
    return UDM_ERROR;

  if (UDM_OK != UdmMatchComp(&Specific.Match, err, sizeof(err)))
  {
    udm_snprintf(Conf->errstr, sizeof(Conf->errstr) - 1, "%s", err);
    UdmExcerptFragmentFree(&Specific);
    return UDM_ERROR;
  }

  udm_snprintf(name, sizeof(name), "ExcerptFragment%04d",
               (int) C->excerpt_fragments_count++);
  if (UDM_OK != (rc= UdmVarCreate(ha, &Var, name, NULL, 0)))
  {
    UdmExcerptFragmentFree(&Specific);
    return rc;
  }

  VarSpecific= (UDM_EXCERPT_FRAGMENT *) UdmVarDataPtr(Var);
  VarSpecific[0]= Specific;
  bzero((void *) &Specific, sizeof(Specific));

  if (UDM_OK != (rc= UdmVarListReplaceVar(&Conf->Vars, Var)))
  {
    UdmVarFree(Var);
    return rc;
  }
  return UDM_OK;
}


static udm_rc_t
add_filter(UDM_CFG *C, size_t ac, const char **av)
{
  UDM_ENV    *Conf= C->Indexer->Conf;
  UDM_MATCH_PARAM Param;
  size_t     i;
  udm_method_t method= UdmMethod(av[0]);

  if (method == UDM_METHOD_UNKNOWN)
  {
    udm_snprintf(Conf->errstr, sizeof(Conf->errstr), "Unknown method %s", av[0]);
    return UDM_ERROR;
  }
  if (!(C->flags & UDM_FLAG_ADD_SERV))
    return UDM_OK;

  UdmMatchParamInit(&Param);
  Param.match_mode= UDM_MATCH_WILD;
  UdmMatchParamSetCaseInsensitive(&Param, UDM_TRUE);
  UdmMatchParamSetOptimization(&Param, !UDM_TEST(C->flags & UDM_FLAG_DONT_ADD_TO_DB));

  C->ordre++;
  for (i= 1; i < ac ; i++)
  {
    if (UdmMatchParamSetParam(&Param, av[i]))
    {
      char err[120]= "";
      UDM_CONST_STR Pattern;
      UdmConstStrSetStr(&Pattern, av[i]);

      if (UDM_OK != UdmFilterListAddWithServer(NULL, &Conf->Filters,
                                               &Param, method, &Pattern,
                                               err, sizeof(err), ++C->ordre))
      {
        udm_snprintf(Conf->errstr,sizeof(Conf->errstr)-1,"%s",err);
        return UDM_ERROR;
      }
    }
  }
  return UDM_OK;
}


static udm_rc_t
add_section_filter(UDM_CFG *C, size_t ac, const char **av)
{
  UDM_ENV *Conf= C->Indexer->Conf;
  UDM_MATCH_PARAM MatchParam;
  size_t    i;
  const char *section= NULL;
  udm_method_t method= UdmMethod(av[0]);

  UdmMatchParamInit(&MatchParam);
  MatchParam.match_mode= UDM_MATCH_WILD;
  UdmMatchParamSetCaseInsensitive(&MatchParam, UDM_TRUE);
  UdmMatchParamSetOptimization(&MatchParam, !UDM_TEST(C->flags & UDM_FLAG_DONT_ADD_TO_DB));

  C->ordre++;
  for(i=1; i<ac ; i++)
  {
    if (!UdmMatchParamSetParam(&MatchParam, av[i]))
      continue;
    if (!section)
    {
      section= av[i];
    }
    else
    {
      char err[120]="";
      UDM_CONST_STR Pattern;

      UdmConstStrSetStr(&Pattern, av[i]);
      if (UDM_OK != UdmSectionFilterListAdd(&Conf->SectionFilters,
                                       &MatchParam, method, &Pattern, section,
                                       err, sizeof(err)))
      {
        udm_snprintf(Conf->errstr,sizeof(Conf->errstr)-1,"%s",err);
        return UDM_ERROR;
      }
    }
  }

  if (!section)
  {
    udm_snprintf(Conf->errstr, sizeof(Conf->errstr) - 1,
                 "No section given for %s", av[0]);
    return UDM_ERROR;
  }
  return UDM_OK;
}


static udm_rc_t
add_type_internal(UDM_CFG *C, size_t ac, const char **av, UDM_REPLACELIST *Lst)
{
  UDM_ENV    *Conf=C->Indexer->Conf;
  UDM_MATCH_PARAM MatchParam;
  UDM_CONST_STR Type;
  size_t    i;
  udm_rc_t rc= UDM_OK;
  char err[128];

  UdmConstStrInit(&Type);
  UdmMatchParamInit(&MatchParam);
  MatchParam.match_mode= UDM_MATCH_WILD;
  UdmMatchParamSetCaseInsensitive(&MatchParam, UDM_TRUE);
  UdmMatchParamSetOptimization(&MatchParam, !UDM_TEST(C->flags & UDM_FLAG_DONT_ADD_TO_DB));

  for (i=1; i<ac; i++)
  {
    if (!UdmMatchParamSetParam(&MatchParam, av[i]))
      continue;
    if (!Type.str)
      UdmConstStrSetStr(&Type, av[i]);
    else
    {
      UDM_CONST_STR Pattern;
      UdmConstStrSetStr(&Pattern, av[i]);
      if(UDM_OK != (rc = UdmReplaceListAdd(Lst ,&MatchParam, &Pattern, &Type,
                                           err,sizeof(err))))
      {
        udm_snprintf(Conf->errstr, sizeof(Conf->errstr), "%s", err);
        return rc;
      }
    }
  }
  return rc;
}


static udm_rc_t
add_type(UDM_CFG *C, size_t ac, const char **av)
{
  UDM_ENV *Conf=C->Indexer->Conf;
  return add_type_internal(C, ac, av, &Conf->MimeTypes);
}


static udm_rc_t
add_encoding(UDM_CFG *C, size_t ac, const char **av)
{
  UDM_ENV *Conf= C->Indexer->Conf;
  return add_type_internal(C, ac, av, &Conf->Encodings);
}


static udm_rc_t
add_parser(UDM_CFG *C, size_t ac, const char **av)
{
  UDM_ENV  *Conf=C->Indexer->Conf;
  UDM_PARSER P;
  P.from_mime= UdmStrdup(av[1]);
  P.to_mime= UdmStrdup(av[2]);
  P.cmd= UdmStrdup(av[3] ? av[3] : "");
  P.src= av[4] ? UdmStrdup(av[4]) : NULL;
  UdmParserAdd(&Conf->Parsers,&P);
  return UDM_OK;
}


static udm_rc_t
add_separator(UDM_VARLIST *Vars, const char *name, const char *val)
{
  UDM_DSTR buf;
  UdmDSTRInit(&buf, 128);
  UdmDSTRReset(&buf);
  UdmDSTRAppendf(&buf, "separator.%s", name);
  UdmVarListAddStr(Vars, UdmDSTRPtr(&buf), val);
  UdmDSTRFree(&buf);
  return UDM_OK;
}


static udm_rc_t
add_maxlen(UDM_VARLIST *Vars, const char *name, size_t maxlen)
{
  UDM_DSTR buf;
  UdmDSTRInit(&buf, 128);
  UdmDSTRReset(&buf);
  UdmDSTRAppendf(&buf, "maxlen.%s", name);
  UdmVarListReplaceInt(Vars, UdmDSTRPtr(&buf), (int) maxlen);
  UdmDSTRFree(&buf);
  return UDM_OK;
}


static udm_rc_t
add_section(UDM_CFG *C, size_t ac, const char **av)
{
  UDM_ENV  *Conf=C->Indexer->Conf;
  UDM_SECTION_PARAM Param;
  int      cdon, noindex= 0;
  UDM_USERSECTIONLIST *SectionMatch= &Conf->SectionMatch;
  const char *name= av[1];
  int secno= 0, maxlen= 0;

  UdmSectionParamInit(&Param, 0);
  if (!strncasecmp(name, UDM_CSTR_WITH_LEN("Raw.")))
    Param.flags= (udm_var_flag_t) (Param.flags | UDM_VARFLAG_RAW);
  /*
    Do not use "url.*" sections in
    clone detection by default
  */
  cdon= strncasecmp(av[1], "url", 3) ? 1 : 0;
  if ((!(secno= atoi(av[2])) && av[2][0] != '0') || secno < 0 || secno > 255)
  {
    sprintf(Conf->errstr,"Section ID is not a valid number: %s",av[2]);
    return UDM_ERROR;
  }
  if ((ac > 3 && !(maxlen= atoi(av[3])) && av[3][0] != '0') || maxlen < 0)
  {
    sprintf(Conf->errstr,"Section length is not a valid number: %s",av[3]);
    return UDM_ERROR;
  }

  if (ac > 3)
  {
    av++;
    ac--;
  }
  av+= 3;
  ac-= 3;
  
  for ( ; ac ; ac--, av++)
  {
    if (!strcasecmp(av[0], "cdon") || !strcasecmp(av[0], "DetectClones"))
      cdon= 1;
    else if (!strcasecmp(av[0], "cdoff") || !strcasecmp(av[0], "NoDetectClones"))
      cdon= 0;
    else if (!strcasecmp(av[0], "html"))
      Param.flags= (udm_var_flag_t) (Param.flags| UDM_VARFLAG_HTMLSOURCE);
    else if (!strcasecmp(av[0], "decimal"))
      Param.flags= (udm_var_flag_t) (Param.flags | UDM_VARFLAG_DECIMAL);
    else if (!strcasecmp(av[0], "wiki"))
      Param.flags= (udm_var_flag_t) (Param.flags | UDM_VARFLAG_HTMLSOURCE | UDM_VARFLAG_WIKI);
    else if (!strcasecmp(av[0], "noindex"))
      noindex= 1;
    else if (!strcasecmp(av[0], "index"))
      noindex= 0;
    else if (!strcasecmp(av[0], "text"))
      /* do nothing */;
    else if (!strcasecmp(av[0], "afterheaders"))
      SectionMatch= &Conf->SectionHdrMatch;
    else if (!strcasecmp(av[0], "afterguesser"))
      SectionMatch= &Conf->SectionGsrMatch;
    else if (!strcasecmp(av[0], "afterparser"))
      SectionMatch= &Conf->SectionMatch;
    else
      break;
  }
  if (secno > 0) /* Don't set maxlen for secno=0 */
    Conf->SectionParam.maxlen[secno]= (size_t) maxlen;
  Param.secno= (udm_secno_t) secno;
  Param.flags= (udm_var_flag_t) (Param.flags | (cdon ? 0 : UDM_VARFLAG_NOCLONE));
  Param.flags= (udm_var_flag_t) (Param.flags | (noindex ? UDM_VARFLAG_NOINDEX : 0));

  if (!secno)
    add_maxlen(&Conf->Vars, name, maxlen);
  if (ac == 0)
  {
    /* no optional arguments */
  }
  else if (ac == 1)
  {
    /* <sep> */
    add_separator(&Conf->Vars, name, av[0]);
  }
  else if (ac >= 2 && ac <= 4)
  {
    /* 
       <expr> <repl>
       <sep> <expr> <repl>
       <sep> <src> <expr> <repl>
    */

    UDM_MATCH_PARAM MatchParam;
    char err[120]= "";
    UDM_CONST_STR Name, Source, Pattern, Replacement;

    UdmMatchParamInit(&MatchParam);
    UdmConstStrSetStr(&Name, name);
    UdmConstStrInit(&Source);
    UdmConstStrInit(&Pattern);
    UdmConstStrInit(&Replacement);
    MatchParam.match_mode= UDM_MATCH_REGEX;
    UdmMatchParamSetCaseInsensitive(&MatchParam, UDM_TRUE);
    UdmMatchParamSetOptimization(&MatchParam, !UDM_TEST(C->flags & UDM_FLAG_DONT_ADD_TO_DB));

    switch (ac)
    {
      case 2:
        UdmConstStrSetStr(&Pattern, av[0]);
        UdmConstStrSetStr(&Replacement, av[1]);
        break;

      case 3:
        add_separator(&Conf->Vars, name, av[0]);
        UdmConstStrSetStr(&Pattern, av[1]);
        UdmConstStrSetStr(&Replacement, av[2]);
        break;

      case 4:
        add_separator(&Conf->Vars, name, av[0]);
        UdmConstStrSetStr(&Source, av[1]);
        UdmConstStrSetStr(&Pattern, av[2]);
        UdmConstStrSetStr(&Replacement, av[3]);
        break;
    }

    if(UDM_OK != UdmUserSectionListAdd(SectionMatch, &MatchParam,
                                       &Name, &Source, &Pattern, &Replacement,
                                       err, sizeof(err)))
    {
      udm_snprintf(Conf->errstr,sizeof(Conf->errstr)-1,"%s",err);
      return UDM_ERROR;
    }
    Param.flags= (udm_var_flag_t) (Param.flags | UDM_VARFLAG_USERDEF);
  }
  else
  {
    sprintf(Conf->errstr,"too many argiments: '%s'", av[0]);
    return UDM_ERROR;
  }

  UdmVarListReplaceStrnWithParam(&Conf->Sections, &Param, name, NULL, 0);
  return UDM_OK;
}


static udm_rc_t
do_include(UDM_CFG *C, size_t ac, const char **av)
{
  FILE *test;
  char save[1024];
  if(C->level<5)
  {
    udm_rc_t rc;
    char  fname[1024];
    rel_cur_name(fname, sizeof(fname)-1, av[1]);
    if ((test= fopen(fname, "r")))
      fclose(test);
    else
      rel_etc_name(C->Indexer->Conf, fname, sizeof(fname)-1, av[1]);
    strcpy(save, current);
    C->level++;
    rc= EnvLoad(C,fname);
    strcpy(current, save);
    C->level--;
    return rc;
  }
  else
  {
    sprintf(C->Indexer->Conf->errstr,"too big (%d) level in included files",C->level);
    return UDM_ERROR;
  }
  return UDM_OK;
}


static udm_rc_t
add_affix(UDM_CFG *C, size_t ac, const char **av)
{
  UDM_ENV  *Conf=C->Indexer->Conf;
  
  if(C->flags&UDM_FLAG_SPELL)
  {
    char  fname[1024];
    rel_etc_name(Conf, fname, sizeof(fname) - 1, av[3]);
    if(UdmAffixListListAdd(&Conf->Affixes,av[1],av[2],fname))
    {
      sprintf(Conf->errstr,"Can't add affix :%s",fname);
      return UDM_ERROR;
    }
  }
  return UDM_OK;
}


static udm_rc_t
add_spell(UDM_CFG *C, size_t ac, const char **av)
{
  UDM_ENV  *Conf=C->Indexer->Conf;
  
  if(C->flags&UDM_FLAG_SPELL)
  {
    char  fname[1024];
    rel_etc_name(Conf, fname, sizeof(fname) - 1, av[3]);
    if(UdmSpellListListAdd(&Conf->Spells,av[1],av[2],fname))
    {
     sprintf(Conf->errstr,"Can't load dictionary :%s",fname);
      return UDM_ERROR;
    }
  }
  return UDM_OK;
}


static udm_rc_t
add_stoplist(UDM_CFG *C, size_t ac, const char **av)
{
  UDM_ENV  *Conf=C->Indexer->Conf;
  char  fname[1024];
  rel_stopwords_name(Conf, fname, sizeof(fname) - 1, av[1]);
  return UdmStopListLoad(Conf,fname);
}


static udm_rc_t
add_langmap(UDM_CFG *C, size_t ac, const char **av)
{
  UDM_ENV  *Conf=C->Indexer->Conf;
  udm_rc_t rc= UDM_OK;
  if (C->flags&UDM_FLAG_LOAD_LANGMAP)
  {
    char  fname[1024];
    rel_langmap_name(Conf, fname, sizeof(fname) - 1, av[1]);
    rc= UdmLoadLangMapFile(&Conf->LangMaps, fname);
  }
  return rc;
}


static udm_rc_t
add_synonym(UDM_CFG *C, size_t ac, const char **av)
{
  UDM_ENV  *Conf=C->Indexer->Conf;
  udm_rc_t rc= UDM_OK;
  if(C->flags&UDM_FLAG_SPELL)
  {
    char  fname[1024];
    rel_synonym_name(Conf, fname, sizeof(fname) - 1, av[1]);
    rc= UdmSynonymListLoad(Conf,fname);
  }
  return rc;
}


static udm_rc_t
add_chinese(UDM_CFG *C, size_t ac, const char **av)
{
  UDM_ENV *Conf=C->Indexer->Conf;
  
  /*
    This line was wrong: ChinesList was not really loaded
    from search.cgi
  */
  /* if(C->flags & UDM_FLAG_ADD_SERV)*/
  
  {
    char fname[1024];
    rel_freq_name(Conf, fname, sizeof(fname)-1, av[2] ? av[2] : "mandarin.freq");
    return UdmChineseListLoad(C->Indexer, &Conf->Chi, 
                              av[1] ? av[1] : "GB2312", fname);
  }
  return UDM_OK;
}


static udm_rc_t
add_thai(UDM_CFG *C, size_t ac, const char **av)
{
  UDM_ENV *Conf= C->Indexer->Conf;
  if (C->flags & UDM_FLAG_ADD_SERV)
  {
    char fname[1024];
    rel_freq_name(Conf, fname, sizeof(fname)-1, av[2] ? av[2] : "thai.freq");
    return UdmChineseListLoad(C->Indexer, &Conf->Thai, 
                              av[1] ? av[1] : "tis-620", fname);
  }
  return UDM_OK;
}
        

static udm_rc_t
add_url(UDM_CFG *C, size_t ac, const char **av)
{
  UDM_AGENT *A= C->Indexer;

  if (C->flags&UDM_FLAG_ADD_SERV)
  {
    char    *al = NULL;
    UDM_SERVER  *Srv;
    if ((Srv= UdmServerFind(A, &A->Conf->Servers, av[1], &al)))
    {
      UDM_HREFPARAM HrefParam;
      UdmHrefParamInit(&HrefParam);
      HrefParam.link_source= UDM_LINK_SOURCE_CONF;
      UdmHrefListAddConst(&A->Conf->Hrefs, &HrefParam, av[1]);
    }
    UDM_FREE(al);
  }
  return UDM_OK;
}


static udm_rc_t
add_srv_table(UDM_CFG *C, size_t ac, const char **av)
{
  UDM_ENV *Conf=C->Indexer->Conf;
  udm_rc_t rc= UDM_OK;
  UDM_DBLIST dbl;

  /*
    Skip ServerTable when loading for search, not for indexing.
    Useful when the ServerTable options are written in a shared
    include.conf file together with DBAddr options, and this file
    is included from both indexer.conf and search.htm
  */
  if (!(C->flags & UDM_FLAG_ADD_SERV))
    return UDM_OK;

  UdmDBListInit(&dbl);
  if (UDM_OK != (rc= UdmDBListAdd(&dbl, av[1], Conf->errstr, sizeof(Conf->errstr))))
    goto ex;
  UDM_ASSERT(dbl.nitems == 1);

  if (UDM_OK != (rc= dbl.Item[0].dbhandler->ServerAction(C->Indexer, &dbl.Item[0], &Conf->Servers, UDM_SRV_ACTION_TABLE)))
    UdmEnvCopyErrMsgFromDB(Conf, &dbl.Item[0]);
ex:
  UdmDBListFree(&dbl);
  return rc;
}


static udm_rc_t
add_limit(UDM_CFG *C, size_t ac, const char **av)
{
  UDM_ENV  *Conf=C->Indexer->Conf;
  char * sc;
  char * nm;

  if (ac == 2)
  {
    if((sc = strchr(av[1],':')))
    {
      *sc++='\0';
      nm=(char*)UdmMalloc(strlen(av[1])+8);
      sprintf(nm,"Limit-%s",av[1]);
      UdmVarListReplaceStr(&Conf->Vars, nm, sc);
      UDM_FREE(nm);
    }
  }
  else if (ac == 3)
  {
    char name[128];
    udm_snprintf(name, sizeof(name), "Limit.%s", av[1]);
    UdmVarListReplaceStr(&Conf->Vars, name, av[2]);
  }
  return UDM_OK;
}


static udm_rc_t
add_user_score(UDM_CFG *C, size_t ac, const char **av)
{
  UDM_ENV  *Conf=C->Indexer->Conf;
  char name[128];
  UDM_ASSERT(ac == 3);
  udm_snprintf(name, sizeof(name), "Score.%s", av[1]);
  UdmVarListReplaceStr(&Conf->Vars, name, av[2]);
  return UDM_OK;
}


static udm_rc_t
add_user_site_score(UDM_CFG *C, size_t ac, const char **av)
{
  UDM_ENV  *Conf=C->Indexer->Conf;
  char name[128];
  udm_snprintf(name, sizeof(name), "SiteScore.%s", av[1]);
  UdmVarListReplaceStr(&Conf->Vars, name, av[2]);
  return UDM_OK;
}


static udm_rc_t
add_user_order(UDM_CFG *C, size_t ac, const char **av)
{
  UDM_ENV *Conf= C->Indexer->Conf;
  char name[128];
  udm_snprintf(name, sizeof(name), "Order.%s", av[1]);
  UdmVarListReplaceStr(&Conf->Vars, name, av[2]);
  return UDM_OK;
}


static udm_rc_t
flush_srv_table(UDM_CFG *C, size_t ac, const char **av)
{
  UDM_ENV *Conf=C->Indexer->Conf;
  udm_rc_t rc= UDM_OK;
  if(C->flags&UDM_FLAG_ADD_SERV)
  {
    UDM_AGENT A;
    A.Conf= Conf;
    rc= UdmSrvAction(&A, &Conf->Servers, UDM_SRV_ACTION_FLUSH);
  }
  return rc;
}


static udm_rc_t
dblist_free(UDM_CFG *C, size_t ac, const char **av)
{
  UDM_ENV  *Conf=C->Indexer->Conf;
  UdmDBListFree(&Conf->DBList);
  return UDM_OK;
}


static udm_rc_t
env_rpl_casefolding(UDM_CFG *C, size_t ac, const char **av)
{
  UDM_ENV *Conf = C->Indexer->Conf;
  UDM_UNIDATA *unidata;
  if (!(unidata= UdmUnidataGetByName(av[1])))
  {
    sprintf(Conf->errstr,"CaseFolding '%s' is not supported", av[1]);
    return UDM_ERROR;
  }
  Conf->unidata= unidata;
  return UDM_OK;
}


static udm_rc_t
env_rpl_charset(UDM_CFG *C, size_t ac, const char **av)
{
  UDM_ENV *Conf= C->Indexer->Conf;
  UDM_CHARSET *cs;
  if (!(cs= UdmGetCharSet(av[1])))
  {
    sprintf(Conf->errstr,"charset '%s' is not supported",av[1]);
    return UDM_ERROR;
  }
  if (!strcasecmp(av[0],"LocalCharset"))
  {
    if (!cs->cset->septoken)
    {
      sprintf(Conf->errstr,
              "charset '%s' is not supported as LocalCharset", av[1]);
      return UDM_ERROR;
    }
    Conf->lcs= cs;
    UdmVarListReplaceStr(&Conf->Vars,av[0],av[1]);
  }
  else if(!strcasecmp(av[0],"BrowserCharset")){
    Conf->bcs=cs;
    UdmVarListReplaceStr(&Conf->Vars,av[0],av[1]);
  }
  return UDM_OK;
}


static udm_rc_t
srv_rpl_charset(UDM_CFG *C, size_t ac, const char **av)
{
  UDM_ENV *Conf = C->Indexer->Conf;
  UDM_CHARSET *cs;
  if (!(cs= UdmGetCharSet(av[1])))
  {
    sprintf(Conf->errstr,"charset '%s' is not supported",av[1]);
    return UDM_ERROR;
  }
  UdmVarListReplaceStr(&C->Srv->Vars,av[0],av[1]);
  return UDM_OK;
}


static udm_rc_t
srv_rpl_mirror(UDM_CFG *C, size_t ac, const char **av)
{
  if (!strcasecmp(av[0],"MirrorRoot") || !strcasecmp(av[0],"MirrorHeadersRoot"))
  {
    char fname[1024];
    rel_var_name(C->Indexer->Conf, fname, sizeof(fname)-1, av[1]);
    UdmVarListReplaceStr(&C->Srv->Vars,av[0],fname);
  }
  else if(!strcasecmp(av[0],"MirrorPeriod"))
  {
    int tm= Udm_dp2time_t(av[1]);
    UdmVarListReplaceInt(&C->Srv->Vars,"MirrorPeriod",tm);
  }
  return UDM_OK;
}
    

static udm_rc_t
srv_rpl_auth(UDM_CFG *C, size_t ac, const char **av)
{
  char name[128];
  udm_snprintf(name, sizeof(name) - 1, "%s", av[0]);
  name[sizeof(name)-1]= '\0';
  if (av[1])
  {
    size_t  len= strlen(av[1]);
    char *auth=(char*)UdmMalloc(BASE64_LEN(strlen(av[1])));
    udm_base64_encode(av[1],auth,len);
    UdmVarListReplaceStr(&C->Srv->Vars,name,auth);
    UDM_FREE(auth);
  }
  else
  {
    UdmVarListReplaceStr(&C->Srv->Vars,name,"");
  }
  return UDM_OK;
}


char *
UdmParseEnvVar(UDM_ENV *Conf, const char *str)
{
  const char *p1= str, *p2= str;
  UDM_DSTR rc;
  UDM_STR tmp;
  UdmDSTRInit(&rc, 256);
  while ((p1= strstr(p1, "$(")))
  {
    const char *p3;
    UdmDSTRAppend(&rc, p2, p1 - p2);
    if ((p3= strchr(p1 + 2, ')')))
    {
      const char *s;
      char varname[128];
      udm_snprintf(varname, sizeof(varname), "%.*s", (int) (p3 - p1 - 2), p1 + 2);
      if ((s= UdmVarListFindStr(&Conf->Vars, varname, NULL)))
        UdmDSTRAppendSTR(&rc, s);
      p1= p2= p3 + 1;
    }
    else
    {
      UdmDSTRFree(&rc);
      return(NULL);
    }
  }
  UdmDSTRAppendSTR(&rc, p2);
  UdmDSTRGiveValue(&rc, &tmp);
  return tmp.str;
}


static udm_rc_t
env_rpl_env_var(UDM_CFG *C, size_t ac, const char **av)
{
  UDM_ENV *Conf= C->Indexer->Conf;
  char *p= getenv(av[1]);
  if (!p)
  {
    sprintf(Conf->errstr, "ImportEnv '%s': no such variable.", av[1]);
    return UDM_ERROR;
  }
  UdmVarListReplaceStr(&Conf->Vars, av[1], p);
  return UDM_OK;
}


static udm_rc_t
env_rpl_encoding(UDM_CFG *C, size_t ac, const char **av)
{
  UDM_ENV  *Conf=C->Indexer->Conf; 
  udm_content_encoding_t ce;
  UDM_ASSERT(ac == 2);

  UdmVarListReplaceStr(&Conf->Vars, av[0], av[1]);
  ce= UdmContentEncodingID(av[1]);
  switch (ce)
  {
    case UDM_CONTENT_ENCODING_IDENTITY:
    case UDM_CONTENT_ENCODING_DEFLATE:
      return UDM_OK;
    default:
      udm_snprintf(Conf->errstr, sizeof(Conf->errstr),
                   "Bad value for '%s': '%s'", av[0], av[1]);
      return UDM_ERROR;
  }
  return UDM_OK;
}


static udm_rc_t
env_dbaddr(UDM_CFG *C, size_t ac, const char **av)
{
  UDM_ENV  *Conf=C->Indexer->Conf;
  udm_rc_t rc;
  size_t i;
  for (i= 1; i < ac; i++)
  {
    if (UDM_OK != (rc= UdmEnvDBListAdd(Conf, av[i] ? av[i] : "")))
      return rc;
  }
  return UDM_OK;
}


static udm_rc_t
env_rpl_var(UDM_CFG *C, size_t ac, const char **av)
{
  UDM_ENV  *Conf=C->Indexer->Conf;
  if (!strcasecmp(av[0], "Segmenter"))
  {
    int seg= 0;
#ifdef CHASEN
    if (!strcasecmp(av[1], "Chasen"))
      seg= 1;
#endif
#ifdef MECAB
    if (!strcasecmp(av[1], "Mecab"))
      seg= 1;
#endif
    if (!strcasecmp(av[1], "Freq"))
      seg= 1;
    if (!strcasecmp(av[1], "CJK"))
      seg= 1;
    if (!seg)
    {
      sprintf(Conf->errstr, "Unsupported segmenter method: '%s'", av[1]);
      return UDM_ERROR;
    }
  }
  if (!strcasecmp(av[0], "Log2Stderr"))
    Conf->Log.logFD= udm_strntobool(av[1], strlen(av[1])) ? stderr : NULL;
  UdmVarListReplaceStr(&Conf->Vars,av[0],av[1]);
  return UDM_OK;
}


static udm_rc_t
env_rpl_named_var(UDM_CFG *C, size_t ac, const char **av)
{
  UDM_ENV *Conf= C->Indexer->Conf;
  UDM_SECTION_PARAM Param;
  UdmSectionParamInit(&Param, 0);
  if (!strncasecmp(av[1], UDM_CSTR_WITH_LEN("Raw.")))
    Param.flags= (udm_var_flag_t) (Param.flags | UDM_VARFLAG_RAW);
  UdmVarListReplaceStrnWithParam(&Conf->Vars, &Param, av[1], av[2], strlen(av[2]));
  return UDM_OK;
}


static udm_rc_t
rpl_xml_hook(UDM_CFG *C, size_t ac, const char **av)
{
  UDM_ENV *Conf= C->Indexer->Conf;
  UDM_VARLIST *Vars= !strcasecmp(av[0], "XMLEnterHook") ?
                     &Conf->XMLEnterHooks : 
                     !strcasecmp(av[0], "XMLDataHook") ? 
                     &Conf->XMLDataHooks : &Conf->XMLLeaveHooks;
  UdmVarListReplaceStr(Vars,av[1],av[2]);
  return UDM_OK;
}


static udm_rc_t
srv_rpl_var(UDM_CFG *C, size_t ac, const char **av)
{
  UdmVarListReplaceStr(&C->Srv->Vars,av[0],av[1]);
  return UDM_OK;
}


static udm_rc_t
srv_add_proxy(UDM_CFG *C, const char *cmd, const char *arg)
{
  UDM_SERVER *Server= C->Srv;
  udm_rc_t rc= UDM_OK;
  UDM_URL url;
  size_t auth_length;

  if (!strcasecmp(arg, "none") ||
      !strcasecmp(arg, ""))
  {
    UdmURLListFree(&Server->ProxyList);
    return UDM_OK;
  }

  UdmURLInit(&url);
  if (UDM_OK != UdmURLParse(&url, arg) ||
      !url.schema || strcasecmp(url.schema, "http") ||
      url.filename || (url.path && strcmp(url.path, "/")) ||
      (url.auth && (auth_length= strlen(url.auth)) > 64))
  {
    UDM_ENV *Conf= C->Indexer->Conf;
    udm_snprintf(Conf->errstr, sizeof(Conf->errstr),
                 "Bad '%s' value: '%s'", cmd, arg);
    rc= UDM_ERROR;
  }
  if (rc == UDM_OK && url.auth)
  {
    /* Wrap the authorization part into Base64 */
    char auth[128];
    udm_base64_encode(url.auth, auth, auth_length);
    UdmFree(url.auth);
    url.auth= UdmStrdup(auth);
  }
  UdmURLListAdd(&Server->ProxyList, &url);
  UdmURLFree(&url);
  return rc;
}


static udm_rc_t
srv_proxy_var(UDM_CFG *C, size_t ac, const char **av)
{
  size_t i;
  UdmURLListFree(&C->Srv->ProxyList);
  for (i= 1; i < ac; i++)
  {
    if (UDM_OK != srv_add_proxy(C, av[0], av[i]))
      return UDM_ERROR;
  }
  return UDM_OK;
}


static int
collect_links_destination(const char *str)
{
  if (!strcasecmp(str, "yes"))
    return UDM_COLLECT_LINKS_YES;
  if (!strcasecmp(str, "all"))
    return UDM_COLLECT_LINKS_ALL_DST;
  if (!strcasecmp(str, "inner"))
    return UDM_COLLECT_LINKS_INNER;
  if (!strcasecmp(str, "outer"))
    return UDM_COLLECT_LINKS_OUTER;
  if (!strcasecmp(str, "site"))
    return UDM_COLLECT_LINKS_SITE;
  if (!strcasecmp(str, "page"))
    return UDM_COLLECT_LINKS_PAGE;
  if (!strcasecmp(str, "badscheme"))
    return UDM_COLLECT_LINKS_BADSCHEME;
  if (!strcasecmp(str, "bad"))
    return UDM_COLLECT_LINKS_BAD;
  if (!strcasecmp(str, "hops"))
    return UDM_COLLECT_LINKS_HOPS;
  if (!strcasecmp(str, "filter"))
    return UDM_COLLECT_LINKS_FILTER;
  if (!strcasecmp(str, "persite"))
    return UDM_COLLECT_LINKS_PERSITE;
  if (!strcasecmp(str, "no"))
    return UDM_COLLECT_LINKS_NONE;
  return -1;
}


static int
collect_links_format(const char *str)
{
  if (!strcasecmp(str, "asis"))
    return UDM_COLLECT_LINKS_ASIS;
  if (!strcasecmp(str, "absolute"))
    return UDM_COLLECT_LINKS_ABSOLUTE;
  return -1;
}


static udm_rc_t
srv_rpl_collect_links(UDM_CFG *C, size_t ac, const char **av)
{
  size_t i;
  int destination= UDM_COLLECT_LINKS_NONE;
  int format= UDM_COLLECT_LINKS_ASIS;
  for (i= 1; i < ac; i++)
  {
    int flag;
    if ((flag= collect_links_destination(av[i])) >= 0)
      destination|= flag;
    else if ((flag= collect_links_format(av[i])) >= 0)
      format= flag;
    else
    {
      UDM_ENV  *Conf= C->Indexer->Conf;
      udm_snprintf(Conf->errstr, sizeof(Conf->errstr),
                   "%s: unknown flag: '%s'", av[0], av[i]);
      return UDM_ERROR;
    }
  }
  UdmVarListReplaceInt(&C->Srv->Vars,av[0], destination | format);
  return UDM_OK;
}


static udm_rc_t
srv_rpl_follow_links(UDM_CFG *C, size_t ac, const char **av)
{
  size_t i;
  int follow= 0;
  for (i= 1; i < ac; i++)
  {
    udm_link_source_t source;
    if (!strcasecmp(av[i], "yes"))
      follow= UDM_LINK_SOURCES_YES;
    else if (!strcasecmp(av[i], "no"))
      follow= 0;
    else if (av[i][0] == '-' &&
             (source= UdmLinkSourceByName(av[i] + 1)) != UDM_LINK_SOURCE_UNKNOWN)
    {
      follow&= ~(1 << source);
    }
    else if ((source= UdmLinkSourceByName(av[i])) != UDM_LINK_SOURCE_UNKNOWN)
    {
      follow|= (1 << source);
    }
    else
    {
      UDM_ENV  *Conf= C->Indexer->Conf;
      udm_snprintf(Conf->errstr, sizeof(Conf->errstr),
                   "%s: unknown link source: '%s'", av[0], av[i]);
      return UDM_ERROR;
    }
  }
  UdmVarListReplaceInt(&C->Srv->Vars, av[0], follow);
  return UDM_OK;
}


static int
robots_flag(const char *str)
{
  if (!strcasecmp(str, "yes"))
    return UDM_ROBOTS_ALL;
  if (!strcasecmp(str, "xrobotstag"))
    return UDM_ROBOTS_HEADER;
  if (!strcasecmp(str, "robotstxt"))
    return UDM_ROBOTS_TXT;
  if (!strcasecmp(str, "meta"))
    return UDM_ROBOTS_META;
  if (!strcasecmp(str, "rel"))
    return UDM_ROBOTS_REL;
  if (!strcasecmp(str, "no"))
    return UDM_ROBOTS_NONE;
  return -1;
}


static udm_rc_t
srv_rpl_robots(UDM_CFG *C, size_t ac, const char **av)
{
  size_t i;
  int flags= UDM_ROBOTS_NONE;
  for (i= 1; i < ac; i++)
  {
    int flag;
    if ((flag= robots_flag(av[i])) >= 0)
      flags|= flag;
    else
    {
      UDM_ENV  *Conf= C->Indexer->Conf;
      udm_snprintf(Conf->errstr, sizeof(Conf->errstr),
                   "%s: unknown flag: '%s'", av[0], av[i]);
      return UDM_ERROR;
    }
  }
  UdmVarListReplaceInt(&C->Srv->Vars, av[0], flags);
  return UDM_OK;
}


static udm_rc_t
srv_rpl_hdr(UDM_CFG *C, size_t ac, const char **av)
{
  const char *semicolon= NULL;
  char  name[128];

  switch(ac)
  {
    case 3:
      udm_snprintf(name, sizeof(name), "Request.%s", av[1]);
      return UdmVarListReplaceStr(&C->Srv->Vars, name, av[2]);

    case 2:
      if ((semicolon= strchr(av[1],':')))
      {
        UDM_CONST_STR val;
        size_t namelength= semicolon - av[1];
        udm_snprintf(name, sizeof(name), "Request.%.*s", (int) namelength, av[1]);
        UdmConstStrSetStr(&val, semicolon + 1);
        UdmConstStrTrim(&val," \t");
        return UdmVarListReplaceStrn(&C->Srv->Vars, name, val.str, val.length);
      }
      udm_snprintf(C->Indexer->Conf->errstr, sizeof(C->Indexer->Conf->errstr),
                   "No semicolon found in %s\n", av[0]);
      return UDM_ERROR;
      break;
  }
  UDM_ASSERT(0);
  return UDM_ERROR;
}


static udm_rc_t
env_rpl_bool_var(UDM_CFG *C, size_t ac, const char **av)
{
  UDM_ENV *Conf= C->Indexer->Conf;
  int res= !strcasecmp(av[1],"yes") || atoi(av[1]) == 1;
  if(!strcasecmp(av[0], "CVSIgnore")) Conf->CVS_ignore= res;
  UdmVarListReplaceInt(&Conf->Vars,av[0],res);
  return UDM_OK;
}


static udm_rc_t
srv_rpl_bool_var(UDM_CFG *C, size_t ac, const char **av)
{
  int res= !strcasecmp(av[1],"yes") || atoi(av[1]) == 1;
  UdmVarListReplaceInt(&C->Srv->Vars,av[0],res);
  return UDM_OK;
}


static udm_rc_t
env_rpl_size(UDM_CFG *C, size_t ac, const char **av)
{
  UDM_ENV *Conf= C->Indexer->Conf;
  int error;
  char *endptr;
  unsigned long long res= UdmStrToSize(av[1], &endptr, &error);
  if (error || endptr[0])
  {
    udm_snprintf(Conf->errstr, sizeof(Conf->errstr),
                 "Bad size value: %s", av[1]);
    return UDM_ERROR;
  }
  UdmVarListReplaceULongLong(&Conf->Vars, av[0], res);
  return UDM_OK;
}


static udm_rc_t
env_rpl_num_var(UDM_CFG *C, size_t ac, const char **av)
{
  UDM_ENV  *Conf=C->Indexer->Conf;
  int  res=atoi(av[1]);
  if(!strcasecmp(av[0], "DocSizeWeight"))
  {
    UdmVarListReplaceInt(&Conf->Vars, "MaxCoordFactor" ,res);
    return UDM_OK;
  }
  if(!strcasecmp(av[0],"MinWordLength"))Conf->WordParam.min_word_len=res;
  if(!strcasecmp(av[0],"MaxWordLength"))Conf->WordParam.max_word_len=res;
  if (!strcasecmp(av[0],"IndexerThreads") &&
      (res < 1 || res > UDM_INDEXER_THREADS_MAX))
  {
    udm_snprintf(Conf->errstr, sizeof(Conf->errstr),
                 "%s: bad value: %d. Valid values: 1..%d.",
                 av[0], res, UDM_INDEXER_THREADS_MAX);
    return UDM_ERROR;
  }
  if (!strcasecmp(av[0], "LogLevel"))
    Conf->Log.level= atoi(av[1]);
  UdmVarListReplaceInt(&Conf->Vars,av[0],res);
  return UDM_OK;
}    


static udm_rc_t
srv_rpl_num_var(UDM_CFG *C, size_t ac, const char **av)
{
  int res=atoi(av[1]);
  UdmVarListReplaceInt(&C->Srv->Vars,av[0],res);
  if (strcasecmp(av[0], "MaxHops") == 0) C->Srv->MaxHops = (uint4) res;
  if (strcasecmp(av[0], "ServerWeight") == 0) C->Srv->weight = atof(av[1]);
  return UDM_OK;
}


static udm_rc_t
srv_rpl_time_var(UDM_CFG *C, size_t ac, const char **av)
{
  UDM_ENV *Conf=C->Indexer->Conf;
  int res= Udm_dp2time_t(av[1]);
  udm_rc_t rc= UDM_OK;
  if (res == -1)
  {
    sprintf(Conf->errstr,"bad time interval: %s",av[1]);
    rc= UDM_ERROR;
  }
  UdmVarListReplaceInt(&C->Srv->Vars,av[0],res);
  return rc;
}


typedef struct conf_cmd_st
{
  const char  *name;
  size_t    argmin;
  size_t    argmax;
  udm_rc_t  (*action)(UDM_CFG *a, size_t n, const char **av);
} UDM_CONFCMD;


static UDM_CONFCMD commands[] = 
{
  {"Include",              1,1,   do_include},      /* Documented */

  {"ImportEnv",            1,1,   env_rpl_env_var}, /* Documented */
  {"DBAddr",               1,100, env_dbaddr},      /* Documented */
  {"DefaultContentType",   1,1,   env_rpl_var},     /* Documented */
  {"ResultContentType",    1,1,   env_rpl_var},     /* Documented */
  {"Listen",               1,1,   env_rpl_var},     /* TODO       */
  {"UseRemoteContentType", 1,1,   env_rpl_bool_var},/* Documented */
  {"UsePopularity",        1,1,   env_rpl_bool_var},/* Documented */
  {"UseCRC32URLId",        1,1,   env_rpl_var},     /* Documented */
  {"NewsExtensions",       1,1,   env_rpl_var},     /* Documented */
  {"SyslogFacility",       1,1,   env_rpl_var},     /* Documented */
  {"AliasProg",            1,1,   env_rpl_var},     /* Documented */
  {"ForceIISCharset1251",  1,1,   env_rpl_var},     /* Documented */
  {"GroupBySite",          1,1,   env_rpl_bool_var},/* Documented */
  {"wf",                   1,1,   env_rpl_var},     /* Documented */
  {"StrictModeThreshold",  1,1,   env_rpl_var},     /* Documented */
/*{"PopRankSkipSameSite",  1,1,   env_rpl_var},*/     /* TODO34: restore */
  {"VarDir",               1,1,   env_rpl_var},     /* Documented */
  {"DocMemCacheSize",      1,1,   env_rpl_var},     /* Documented */
  {"IspellUsePrefixes",    1,1,   env_rpl_var},     /* Documented */
  {"URLSelectCacheSize",   1,1,   env_rpl_var},     /* Documented */
  {"URLSelectSkipLock",    1,1,   env_rpl_bool_var},/* Documented */
  {"MaxDocSize",           1,1,   env_rpl_var},     /* Documented */
  {"ParserTimeOut",        1,1,   env_rpl_var},     /* Documented */
  {"NumSections",          1,1,   env_rpl_var},     /* Documented */
  {"DateFormat",           1,1,   env_rpl_var},     /* Documented */
  {"GuesserUseMeta",       1,1,   env_rpl_var},     /* Documented */
  {"ResultsLimit",         1,1,   env_rpl_var},     /* Documented */
  {"Segmenter",            1,1,   env_rpl_var},     /* TODO       */
  {"Log2stderr",           1,1,   env_rpl_var},     /* Documented */
  {"LogFlags",             1,1,   env_rpl_num_var}, /* TODO     */
  {"SQLClearDBHook",       1,1,   env_rpl_var},     /* TODO       */
  {"UserCacheQuery",       1,1,   env_rpl_var},     /* Documented */
#ifdef HAVE_SETVBUF
  {"StdoutBufferSize",     1,1,   env_rpl_var},     /* TODO       */
#endif
  {"AlwaysFoundWord",      1,1,   env_rpl_var},     /* Documented */
  {"CustomLog",            1,1,   env_rpl_var},     /* Documented */
  {"CurrentTime",          1,1,   env_rpl_var},     /* TODO       */
  {"Locale",               1,1,   env_rpl_var},     /* Documented */
  {"WordDistanceWeight",   1,1,   env_rpl_num_var}, /* Documented */
  {"MaxCoordFactor",       1,1,   env_rpl_num_var}, /* TODO       */
  {"PopularityFactor",     1,1,   env_rpl_num_var}, /* Documented */
  {"IDFFactor",            1,1,   env_rpl_num_var}, /* Documented */
  {"MinCoordFactor",       1,1,   env_rpl_num_var}, /* Documented */
  {"NumWordFactor",        1,1,   env_rpl_var},     /* Documented */
  {"NumDistinctWordFactor",1,1,   env_rpl_num_var}, /* Documented */
  {"UserScoreFactor",      1,1,   env_rpl_num_var}, /* Documented */
  {"WordDensityFactor"    ,1,1,   env_rpl_num_var}, /* Documented */
  {"WordFormFactor"       ,1,1,   env_rpl_num_var}, /* Documented */
  {"URLDataThreshold",     1,1,   env_rpl_num_var}, /* Documented */
  {"DocSizeWeight",        1,1,   env_rpl_num_var}, /* Documented */
  {"RelevancyFactor",      1,1,   env_rpl_num_var}, /* TODO       */
  {"Phrase2CountFactor",   1,1,   env_rpl_num_var}, /* Documented */
  {"Phrase3CountFactor",   1,1,   env_rpl_num_var}, /* Documented */
  {"DateFactor",           1,1,   env_rpl_num_var}, /* Documented */
  {"MinWordLength",        1,1,   env_rpl_num_var}, /* Documented */
  {"MaxWordLength",        1,1,   env_rpl_num_var}, /* Documented */
  {"SubstringMatchMinWordLength", 1,1, env_rpl_num_var}, /* Documented */

  {"ExcerptSize",          1,1,   env_rpl_num_var}, /* Documented */
  {"ExcerptPadding",       1,1,   env_rpl_num_var}, /* Documented */
  {"ExcerptFragment",      2,100, add_excerpt_fragment}, /* TODO */

  {"LogLevel",             1,1,   env_rpl_num_var}, /* Documented */
  {"CrawlerThreads",       1,1,   env_rpl_num_var}, /* Documented */
  {"IndexerThreads",       1,1,   env_rpl_num_var}, /* Documented */
  {"WordCacheSize",        1,1,   env_rpl_num_var}, /* Documented */
  {"IPRequestPerMinLimit", 1,1,   env_rpl_num_var}, /* Documented */
  {"CVSIgnore",            1,1,   env_rpl_bool_var},/* Documented */
  {"UseHTDBURLId",         1,1,   env_rpl_bool_var},/* TODO       */
  {"Suggest",              1,1,   env_rpl_bool_var},/* Documented */
  {"IndexTime",            1,1,   env_rpl_bool_var},/* Documented */
  {"ExcerptStopword",      1,1,   env_rpl_bool_var},/* Documented */
  {"UseCookie",            1,1,   env_rpl_bool_var},/* Documented */
  {"UseSitemap",           1,1,   env_rpl_bool_var},/* Documented */
  {"UseNumericOperators",  1,1,   env_rpl_bool_var},/* Documented */
  {"UseRangeOperators",    1,1,   env_rpl_bool_var},/* Documented */
  {"SaveSectionSize",      1,1,   env_rpl_bool_var},/* Documented */
  {"Dehyphenate",          1,1,   env_rpl_bool_var},/* Documented */
  {"HyphenateNumbers",     1,1,   env_rpl_bool_var},/* TODO       */
  {"StripAccents",         1,1,   env_rpl_bool_var},/* Documented */
  {"LoadURLInfo",          1,1,   env_rpl_bool_var},/* Documented */
  {"LoadURLBasicInfo",     1,1,   env_rpl_bool_var},/* Documented */
  {"LoadTagInfo",          1,1,   env_rpl_bool_var},/* Documented */
  {"ComplexSynonyms",      1,1,   env_rpl_bool_var},/* Documented */

  {"ReplaceVar",           2,2,   env_rpl_named_var},/* Documented */

  {"LocalCharset",         1,1,   env_rpl_charset},  /* Documented */
  {"BrowserCharset",       1,1,   env_rpl_charset},  /* Documented */
  {"CaseFolding",          1,1,   env_rpl_casefolding},/* Documented */

  {"XMLEnterHook",         2,2,   rpl_xml_hook},     /* TODO       */
  {"XMLLeaveHook",         2,2,   rpl_xml_hook},     /* TODO       */
  {"XMLDataHook",          2,2,   rpl_xml_hook},     /* TODO       */
  
  {"HTDBAddr",             1,1,   srv_rpl_var},      /* Documented */
  {"HTDBList",             1,1,   srv_rpl_var},      /* Documented */
  {"HTDBDoc",              1,1,   srv_rpl_var},      /* Documented */
  {"HTDBLimit",            1,1,   srv_rpl_var},      /* Documented */
  {"SQLImportSection",     1,1,   srv_rpl_var},      /* TODO       */
  {"SQLExportHref",        1,1,   srv_rpl_var},      /* TODO       */
  {"SQLExportPopularityTable",1,1,env_rpl_var},      /* TODO       */
  {"SQLWordForms",         1,1,   env_rpl_var},      /* Documented */
  {"DefaultLang",          1,1,   srv_rpl_var},      /* Documented */
  {"Tag",                  1,1,   srv_rpl_var},      /* Documented */
  {"Proxy",                0,100, srv_proxy_var},    /* Documented */
  {"VaryLang",             1,1,   srv_rpl_var},      /* Documented */
  {"UseRobotsTxtURL",      1,1,   srv_rpl_var},      /* TODO       */
  {"MaxNetErrors",         1,1,   srv_rpl_num_var},  /* Documented */
  {"CrawlDelay",           1,1,   srv_rpl_num_var},  /* Documented */
  {"MaxHops",              1,1,   srv_rpl_num_var},  /* Documented */
  {"StartHops",            1,1,   srv_rpl_num_var},  /* Documented */
  {"MaxDocPerSite",        1,1,   srv_rpl_num_var},  /* Documented */
  {"ServerWeight",         1,1,   srv_rpl_num_var},  /* Documented */
  {"Robots",               1,4,   srv_rpl_robots},   /* Documented */
  {"DetectClones",         1,1,   srv_rpl_bool_var}, /* Documented */
  {"CollectLinks",         1,100, srv_rpl_collect_links},/*Documented*/
  {"Index",                1,1,   srv_rpl_bool_var}, /* Documented */
  {"FollowLinks",          1,100, srv_rpl_follow_links},/*Documented*/
  {"FollowSymLinks",       1,1,   srv_rpl_bool_var}, /* Documented */
  {"AjaxLinks",            1,1,   srv_rpl_bool_var}, /* TODO */
  {"NetErrorDelayTime",    1,1,   srv_rpl_time_var}, /* Documented */
  {"ReadTimeOut",          1,1,   srv_rpl_time_var}, /* Documented */
  {"DocTimeOut",           1,1,   srv_rpl_time_var}, /* Documented */
  {"Period",               1,1,   srv_rpl_time_var}, /* Documented */
  {"HoldBadHrefs",         1,1,   srv_rpl_time_var}, /* Documented */
  {"DNSCacheTimeout",      1,1,   srv_rpl_time_var}, /* Documented */
  {"HTTPHeader",           1,2,   srv_rpl_hdr},      /* Documented */
  {"AuthBasic",            1,1,   srv_rpl_auth},     /* Documented */
  {"MirrorRoot",           1,1,   srv_rpl_mirror},   /* Documented */
  {"MirrorHeadersRoot",    1,1,   srv_rpl_mirror},   /* Documented */
  {"MirrorPeriod",         1,1,   srv_rpl_mirror},   /* Documented */
  {"RemoteCharset",        1,1,   srv_rpl_charset},  /* Documented */
  {"RemoteFileNameCharset",1,1,   srv_rpl_charset},  /* Documented */
  {"RobotsDisallowAction", 1,1,   srv_rpl_num_var},  /* TODO */
  
  {"Disallow",             1,100, add_filter},       /* Documented */
  {"Allow",                1,100, add_filter},       /* Documented */
  {"CheckMP3Only",         1,100, add_filter},       /* Documented */
  {"CheckMP3",             1,100, add_filter},       /* Documented */
  {"CheckOnly",            1,100, add_filter},       /* Documented */
  {"HrefOnly",             1,100, add_filter},       /* Documented */
  {"ImportOnly",           1,100, add_filter},       /* TODO       */
  {"Skip",                 1,100, add_filter},       /* Documented */

  {"IndexIf",              1,100, add_section_filter},/* Documented */
  {"NoIndexIf",            1,100, add_section_filter},/* Documented */
  {"SkipIf",               1,100, add_section_filter},/* Documented */

  {"Server",               1,100, add_srv},          /* Documented */
  {"Realm",                1,100, add_srv},          /* Documented */
  {"Subnet",               1,100, add_srv},          /* Documented */
  {"URL",                  1,1,   add_url},          /* Documented */
  
  {"Alias",                1,100, add_alias},        /* Documented */
  {"ReverseAlias",         1,100, add_alias},        /* Documented */
  
  {"AddType",              1,100, add_type},         /* Documented */
  {"AddEncoding",          1,100, add_encoding},     /* Documented */
  {"CachedCopyEncoding",   1,1,   env_rpl_encoding}, /* Documented */
  {"Mime",                 2,4,   add_parser},       /* Documented */
  {"Section",              2,10,  add_section},      /* Documented */ /* TODO: index/noindex */
  {"Affix",                3,3,   add_affix},        /* Documented */
  {"Spell",                3,3,   add_spell},        /* Documented */
  {"StopwordFile",         1,1,   add_stoplist},     /* Documented */
  {"LangMapFile",          1,1,   add_langmap},      /* Documented */
  {"LangMapUpdate",        1,1,   env_rpl_var},      /* Documented */
  {"Synonym",              1,1,   add_synonym},      /* Documented */
  {"LoadChineseList",      0,2,   add_chinese},      /* Documented */
  {"LoadThaiList",         0,2,   add_thai},         /* Documented */
  {"Limit",                1,2,   add_limit},        /* Documented */
  {"UserScore",            2,2,   add_user_score},   /* Documented */
  {"UserSiteScore",        2,2,   add_user_site_score},/* Documented */
  {"UserOrder",            2,2,   add_user_order},   /* Documented */
  {"ServerTable",          1,1,   add_srv_table},    /* Documented */
  {"FlushServerTable",     0,0,   flush_srv_table},  /* Documented */
  {"DBListFree",           0,0,   dblist_free},      /* TODO       */

  {"IndexCacheSize",       1,1,   env_rpl_size},     /* Documented */
  {NULL,0,0,0}
};


UDM_API(udm_rc_t)
UdmEnvAddLine(UDM_CFG *C, char *str)
{
  UDM_ENV    *Conf=C->Indexer->Conf;
  UDM_CONFCMD  *Cmd;
  const char *av[255];
  size_t  ac= UdmGetArgs(str, av, 255);

  for (Cmd= commands ; Cmd->name ; Cmd++)
  {
    if(!strcasecmp(Cmd->name,av[0]))
    {
      udm_rc_t rc= UDM_OK;
      int argc=ac;
      size_t i;
      char *p;
      char *tmpav[255];
      
      argc--;
      if(ac<Cmd->argmin+1)
      {
        sprintf(Conf->errstr,"too few (%d) arguments for command '%s'",
                argc,Cmd->name);
        return UDM_ERROR;
      }
      
      if(ac>Cmd->argmax+1)
      {
        sprintf(Conf->errstr,"too many (%d) arguments for command '%s'",
                argc,Cmd->name);
        return UDM_ERROR;
      }

      for (i= 1; i < ac; i++)
      {
        if (!av[i])
        {
          tmpav[i]= NULL;
          continue;
        }
        if (!(p = UdmParseEnvVar(Conf, av[i])))
        {
          sprintf(Conf->errstr, "An error occured while parsing '%s'", av[i]);
          return UDM_ERROR;
        }
        av[i]= tmpav[i]= p;
      }

      if (Cmd->action)
        rc= Cmd->action(C, ac, av);

      for (i= 1; i < ac; i++)
        UDM_FREE(tmpav[i]);

      if (Cmd->action)
        return rc;
    }
  }
  udm_snprintf(Conf->errstr, sizeof(Conf->errstr),
               "Unknown command: %s", av[0]);
  return UDM_ERROR;
}


UDM_API(udm_rc_t)
UdmAgentAddLine(UDM_AGENT *Agent, const char *line)
{
  UDM_CFG Cfg;
  char str[1024];
  bzero((void*) &Cfg, sizeof(Cfg));
  Cfg.Indexer= Agent;
  udm_snprintf(str, sizeof(str) - 1, "%s", line);
  return UdmEnvAddLine(&Cfg, str);
}


static udm_rc_t
EnvLoad(UDM_CFG *Cfg, const char *cname)
{
  char  *str0 = NULL;  /* Unsafe copy - will be used in strtok  */
  char  str1[1024]="";  /* To concatenate lines      */
  FILE  *config;  /* File struct */
  udm_rc_t rc= UDM_OK;
  size_t  line = 0, str0len = 0, str1len, str0size = 4096;
  
  if ((str0 = (char*)UdmMalloc(str0size)) == NULL)
  {
    sprintf(Cfg->Indexer->Conf->errstr,
            "Can't alloc %d bytes at '%s': %d",
            (int) str0size, __FILE__, __LINE__);
    return UDM_ERROR;
  }
  str0[0]=0;
  
  /* Open config file */
  if(!(config=fopen(cname,"r")))
  {
    sprintf(Cfg->Indexer->Conf->errstr,
            "Can't open config file '%s': %s", cname, strerror(errno));
    UDM_FREE(str0);
    return UDM_ERROR;
  }

  update_current(cname);
  
  /*  Read lines and parse */
  while(fgets(str1,sizeof(str1),config))
  {
    char  *end;
    
    line++;
    
    if(str1[0]=='#')continue;
    for (end = str1 + (str1len = strlen(str1)) - 1 ;
         (end>=str1) && (*end=='\r'||*end=='\n'||*end==' '||*end=='\t') ;
         *end--='\0');
    if(!str1[0])continue;
    
    if(*end=='\\')
    {
      *end=0;
      if (str0len + str1len >= str0size)
      {
        str0size += 4096 + str1len;
        if ((str0 = (char*)UdmRealloc(str0, str0size)) == NULL)
        {
          sprintf(Cfg->Indexer->Conf->errstr,
                  "Can't realloc %d bytes at '%s': %d",
                  (int) str0size, __FILE__, __LINE__);
          return UDM_ERROR;
        }
      }
      strcat(str0,str1);
      str0len += str1len;
      continue;
    }
    strcat(str0,str1);
    str0len += str1len;
    
    if (UDM_OK != (rc= UdmEnvAddLine(Cfg,str0)))
    {
      char  err[2048];
      strcpy(err,Cfg->Indexer->Conf->errstr);
      sprintf(Cfg->Indexer->Conf->errstr, "%s:%d: %s", cname, (int) line, err);
      break;
    }
    
    str0[0]=0;
    str0len = 0;
  }
  UDM_FREE(str0);
  fclose(config);
  return rc;
}


UDM_API(udm_rc_t)
UdmEnvLoad(UDM_AGENT *Indexer, const char *cname, int lflags)
{
  UDM_CFG    Cfg;
  UDM_SERVER  Srv;
  udm_rc_t rc= UDM_OK;
  const char  *dbaddr=NULL;
  
  UdmServerInit(&Srv);
  bzero((void*)&Cfg, sizeof(Cfg));
  Cfg.Indexer = Indexer;
  Indexer->Conf->Cfg_Srv = Cfg.Srv = &Srv;
  Cfg.flags=lflags;
  Cfg.level=0;
  
  /* Set DBAddr if for example passed from environment */
  if((dbaddr=UdmVarListFindStr(&Indexer->Conf->Vars,"DBAddr",NULL)))
  {
    if(UDM_OK != (rc= UdmEnvDBListAdd(Indexer->Conf, dbaddr)))
      goto freeex;
  }
  
  if(UDM_OK == (rc=EnvLoad(&Cfg,cname)))
  {
    UDM_ENV *Env= Indexer->Conf;
    
    if (UDM_OK != (rc= UdmEnvPrepare(Env)))
      goto freeex;

    UdmVarListInsStr(&Env->Vars, "Request.User-Agent", UDM_USER_AGENT);
  }

freeex:
  UdmServerFree(&Srv);
  return rc;
}



static size_t
UdmMatchToStr(char *str, size_t size,
              const UDM_MATCH *M, const char *arg, const char *cmd)
{
  if (cmd)
    return udm_snprintf(str, size, "%s %s%s%s \"%s\" \"%s\"",
                        cmd,
                        M->Param.match_mode == UDM_MATCH_REGEX ? " regex" : "",
                        UdmMatchIsNegative(M) ? " nomatch" : "",
                        UdmMatchIsCaseInsensitive(M) ? "" : " NoCase",
                        arg, UdmMatchPatternConstStr(M));
  else
    return udm_snprintf(str, size, "%s %s%s%s \"%s\"",
                        arg,
                        M->Param.match_mode == UDM_MATCH_REGEX ? " regex" : "",
                        UdmMatchIsNegative(M) ? " nomatch" : "",
                        UdmMatchIsCaseInsensitive(M) ? "" : " NoCase",
                        UdmMatchPatternConstStr(M));

}


/*
static udm_rc_t
UdmMatchListPrint(FILE *f, UDM_MATCHLIST *L, const char *cmd)
{
  size_t i;
  char str[128];
  for (i= 0; i < L->nmatches; i++)
  {
    UDM_MATCH *M= &L->Match[i];
    UdmMatchToStr(str, sizeof(str), M, M->arg, cmd);
    fprintf(f, "%s\n", str);
  }
  return UDM_OK;
}
*/

static udm_rc_t
UdmDBListPrint(FILE *f, UDM_DBLIST *L)
{
  size_t i;
  for (i= 0; i < L->nitems; i++)
  {
    UDM_DB *db= &L->Item[i];
    char dbaddr[128]= "<noaddr>";
    size_t nbytes;
    db->dbhandler->Info(db, dbaddr, sizeof(dbaddr), &nbytes, UDM_DBINFO_ADDR);
    fprintf(f, "DBAddr %s\n", dbaddr);
  }
  return UDM_OK;
}


static const char *
UdmMatchTypeToServerCommand(int match_type)
{
  switch (match_type)
  {
    case UDM_MATCH_WILD:   return "Realm";
    case UDM_MATCH_REGEX:  return "Realm regex";
    case UDM_MATCH_SUBNET: return "Subnet";
    case UDM_MATCH_BEGIN:  return "Server";
    default: return "<UnknownMatchType>";
  }
}


static udm_rc_t
UdmServerOptionsPrint(FILE *f, UDM_SERVER *Prev, UDM_SERVER *Curr)
{
  UDM_CONFCMD *cmd;
  for (cmd= commands; cmd->name; cmd++)
  {
    if (cmd->action == srv_rpl_var      ||
        cmd->action == srv_rpl_num_var  ||
        cmd->action == srv_rpl_bool_var ||
        cmd->action == srv_rpl_time_var || 
     /* cmd->action ==  srv_rpl_hdr     || */
        cmd->action ==  srv_rpl_auth    ||
     /* cmd->action ==  srv_rpl_mirror  || */
        cmd->action ==  srv_rpl_charset ||
        0)
    {
      const char *cval= UdmVarListFindStr(&Curr->Vars, cmd->name, "");
      if (cmd->action == srv_rpl_auth)
      {
        if (cval[0])
        {
          char encoded[128], decoded[128];
          udm_snprintf(encoded, sizeof(encoded), "%s", cval);
          udm_base64_decode(decoded, encoded, sizeof(decoded));
          fprintf(f, "%s '%s'\n", cmd->name, decoded);
        }
      }
      else
      {
        const char *pval= Prev ? UdmVarListFindStr(&Prev->Vars, cmd->name, "") : "";
        if (strcmp(pval, cval))
          fprintf(f, "%s '%s'\n", cmd->name, cval);
      }
    }
  }
  return UDM_OK;
}


static size_t
UdmServerToStr(char *str, size_t size, UDM_SERVER *S)
{
  const char *method= UdmMethodStr(S->Filter.method);
  udm_bool_t case_sense= UdmVarListFindBool(&S->Vars, "case_sense", UDM_TRUE);
  udm_bool_t nomatch= UdmVarListFindBool(&S->Vars, "nomatch", UDM_FALSE);
  const char *case_str= case_sense ? "" : "NoCase";
  const char *match_str= nomatch ? "NoMatch" : "";
  const char *follow_str= UdmFollowStr(S->webspace);
  const char *command= UdmMatchTypeToServerCommand(S->Filter.Match.Param.match_mode);
  const char *alias= UdmVarListFindStr(&S->Vars, "Alias", "");
  
  /* TODO: Server site: cuts directory name */
  switch (S->Filter.Match.Param.match_mode)
  {
    case UDM_MATCH_WILD:
    case UDM_MATCH_REGEX:
    case UDM_MATCH_SUBNET:
      follow_str= "";
      break;
    case UDM_MATCH_BEGIN:
    case UDM_MATCH_FULL:
    case UDM_MATCH_SUBSTR:
    case UDM_MATCH_END:
    case UDM_MATCH_NUMERIC_LT:
    case UDM_MATCH_NUMERIC_GT:
    case UDM_MATCH_RANGE:
      break;
  }
  
  return udm_snprintf(str, size,
                      "%s %s %s %s %s '%s'%s%s",
                      command, follow_str, method,
                      case_str, match_str,
                      UdmMatchPatternConstStr(&S->Filter.Match),
                      alias[0] ? " " : "", alias);
}


static udm_rc_t
UdmServerListPrint(FILE *f, UDM_SERVERLIST *L)
{
  size_t i;
  char str[128];
  for (i= 0; i < L->nservers; i++)
  {
    UDM_SERVER *S= &L->Server[i];
    UDM_SERVER *P= i ? &L->Server[i-1] : NULL;
/*    UdmVarListPrint(f, &S->Vars);*/
    UdmServerOptionsPrint(f, P, S);
    UdmServerToStr(str, sizeof(str), S);
    fprintf(f, "%s\n", str);
  }
  return UDM_OK;
}


static udm_rc_t
UdmAliasListPrint(FILE *f, UDM_REPLACELIST *L, const char *cmd)
{
  size_t i;
  char str[128];
  for (i= 0; i < L->nitems; i++)
  {
    UdmMatchToStr(str, sizeof(str), &L->Item[i].Match, L->Item[i].Replace.str, cmd);
    fprintf(f, "%s\n", str);
  }
  return UDM_OK;
}


static udm_rc_t
UdmParserListPrint(FILE *f, UDM_PARSERLIST *L)
{
  size_t i;
  char str[1024];
  for (i= 0; i < L->nparsers; i++)
  {
    UDM_PARSER *P= &L->Parser[i];
    udm_snprintf(str, sizeof(str),
                 "Mime \"%s\" \"%s\" '%s'%s%s%s",
                 P->from_mime, P->to_mime, P->cmd,
                 P->src ? " \"" : "",
                 P->src ? P->src : "",
                 P->src ? "\"" : "");
    fprintf(f, "%s\n", str);
  }
  return UDM_OK;
}


static udm_rc_t
UdmAffixListListPrint(FILE *f, UDM_AFFIXLISTLIST *L)
{
  size_t i;
  char str[256];
  for (i= 0; i < L->nitems; i++)
  {
    UDM_AFFIXLIST *A= &L->Item[i];
    udm_snprintf(str, sizeof(str),
                 "Affix %s %s '%s'", A->lang, A->cset, A->fname);
    fprintf(f, "%s\n", str);
  }
  return UDM_OK;
}


static udm_rc_t
UdmSpellListListPrint(FILE *f, UDM_SPELLLISTLIST *L)
{
  size_t i;
  char str[256];
  for (i= 0; i < L->nitems; i++)
  {
    UDM_SPELLLIST *S= &L->Item[i];
    udm_snprintf(str, sizeof(str),
                 "Spell %s %s '%s'", S->lang, S->cset, S->fname);
    fprintf(f, "%s\n", str);
  }
  return UDM_OK;
}


static udm_rc_t
UdmLangmapListPrint(FILE *f, UDM_LANGMAPLIST *L)
{
  size_t i;
  char str[256];
  for (i= 0; i < L->nmaps; i++)
  {
    UDM_LANGMAP *M= &L->Map[i];
    udm_snprintf(str, sizeof(str),
                 "LangmapFile '%s'", M->filename);
    fprintf(f, "%s\n", str);
  }
  return UDM_OK;
}


static udm_rc_t
UdmStopListListPrint(FILE *f, UDM_STOPLISTLIST *L)
{
  size_t i;
  char str[256];
  for (i= 0; i < L->nitems; i++)
  {
    UDM_STOPLIST *S= &L->Item[i];
    udm_snprintf(str, sizeof(str),
                 "StopwordFile '%s'", S->fname);
    fprintf(f, "%s\n", str);
  }
  return UDM_OK;
}


static udm_rc_t
UdmSynonymListListPrint(FILE *f, UDM_SYNONYMLISTLIST *L)
{
  size_t i;
  char str[256];
  for (i= 0; i < L->nitems; i++)
  {
    UDM_SYNONYMLIST *S= &L->Item[i];
    udm_snprintf(str, sizeof(str),
                 "Synonym '%s'", S->fname);
    fprintf(f, "%s\n", str);
  }
  return UDM_OK;
}

                    
UDM_API(udm_rc_t)
UdmEnvSave(UDM_AGENT *Indexer, const char *cname, int lflags)
{
  FILE *f;
  UDM_ENV *E= Indexer->Conf;

  if (!strcmp(cname, "-"))
    f= stdout;
  else if (!(f= fopen(cname,"w")))
  {
    sprintf(Indexer->Conf->errstr,
            "Can't open output file '%s': %s", cname, strerror(errno));
    return UDM_ERROR;
  }

  /* TODO: put interpreter line */
  
  UdmDBListPrint(f, &E->DBList);

  fprintf(f, "LocalCharset %s\n", E->lcs->name);
  fprintf(f, "BrowserCharset %s\n", E->bcs->name);

  UdmSectionListPrint(&E->Sections, f);
  UdmVarListPrint(&E->Vars, f);

  /*UdmMatchListPrint(f, &E->MimeTypes, "AddType");*/
  UdmParserListPrint(f, &E->Parsers);
  /*UdmMatchListPrint(f, &E->Filters, NULL);*/
  
  /*
  UDM_MATCHLIST	SectionFilters;
  UDM_MATCHLIST SectionHdrMatch;
  UDM_MATCHLIST SectionGsrMatch;
  UDM_MATCHLIST SectionMatch;
  */

  UdmStopListListPrint(f, &E->StopWord);
  UdmSynonymListListPrint(f, &E->Synonym);
  UdmAffixListListPrint(f, &E->Affixes);
  UdmSpellListListPrint(f, &E->Spells);
  UdmLangmapListPrint(f, &E->LangMaps);
  
  /*
  int		CVS_ignore;
  UDM_WORDPARAM	WordParam;
  UDM_CHINALIST   Chi;
  UDM_CHINALIST   Thai;
#ifdef MECAB
  mecab_t         *mecab;
#endif
  UDM_UNIDATA *unidata;
  */


  /*
  UDM_VARLIST	XMLEnterHooks;
  UDM_VARLIST	XMLLeaveHooks;
  UDM_VARLIST	XMLDataHooks;
  */

  UdmAliasListPrint(f, &E->Aliases, "Alias");
  UdmAliasListPrint(f, &E->ReverseAliases, "ReverseAlias");
  UdmServerListPrint(f, &E->Servers);

  if (f != stdout)
    fclose(f);
  return UDM_OK;
}
