/* Copyright (C) 2000-2015 Lavtech.com corp. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
*/

#include "udm_config.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#ifdef HAVE_WINSOCK_H
#include <winsock.h>
#endif
#ifdef HAVE_SYS_TIME_H
#include <sys/time.h>
#endif
#ifdef HAVE_SYS_SOCKET_H
#include <sys/socket.h>
#endif
#ifdef HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif
#ifdef HAVE_ARPA_INET_H
#include <arpa/inet.h>
#endif
#ifdef HAVE_ARPA_NAMESER_H
#include <arpa/nameser.h>
#endif
#ifdef HAVE_RESOLV_H
#include <resolv.h>
#endif
#ifdef HAVE_NETDB_H
#include <netdb.h>
#endif


#include "udm_common.h"
#include "udm_utils.h"
#include "udm_db.h"
#include "udm_url.h"
#include "udm_sdp.h"
#include "udm_vars.h"
#include "udm_query.h"
#include "udm_searchtool.h"
#include "udm_result.h"
#include "udm_log.h"
#include "udm_agent.h"
#include "udm_proto.h"
#include "udm_host.h"
#include "udm_hash.h"
#include "udm_doc.h"
#include "udm_services.h"
#include "udm_xmalloc.h"
#include "udm_store.h"
#include "udm_match.h"
#include "udm_word.h"


/* Allocate an array for documents information */
static udm_rc_t
UdmResAllocDoc(UDM_RESULT *Res)
{
  size_t i;
  if (!Res->Doc && UdmResultNumRows(Res) > 0)
  {
    Res->Doc= (UDM_DOCUMENT*) UdmMalloc(sizeof(UDM_DOCUMENT) * 
                                        UdmResultNumRows(Res));
    for (i= 0; i < UdmResultNumRows(Res); i++)
      UdmDocInit(&Res->Doc[i]);
  }
  return UDM_OK;
}


static udm_rc_t
UdmFindWordsDB(UDM_AGENT *A, UDM_QUERY *Query, UDM_DB *db)
{
  udm_rc_t rc= UDM_OK;
  char dbaddr[128];
  size_t nbytes;
  udm_timer_t ticks= UdmStartTimer();

  db->dbhandler->Info(db, dbaddr, sizeof(dbaddr), &nbytes, UDM_DBINFO_ADDR);
  UdmLog(A, UDM_LOG_DEBUG, "Start FindWordsDB for %s", dbaddr);

  rc= db->dbhandler->QueryAction(A, db, Query, UDM_QUERYCMD_FINDWORDS);

  UdmLog(A, UDM_LOG_DEBUG, "%-30s%.2f", "Stop  FindWordsDB:", UdmStopTimer(&ticks));
  return rc;
}

typedef struct search_param_st
{
  UDM_AGENT *Agent;
  UDM_QUERY *Query;
  UDM_DB *db;
  udm_rc_t rc;
  void *thd;
  size_t num_best_rows;
} UDM_SPRM;


static
void *SearchThread(void *arg)
{
  UDM_SPRM *sprm= (UDM_SPRM*) arg;
  sprm->rc= UdmFindWordsDB(sprm->Agent, sprm->Query, sprm->db);
  return NULL;
}


static udm_rc_t
UdmResultSetMachineNumber(UDM_URLDATALIST *List, size_t num)
{
  size_t i;
  UDM_URLDATA *Item= List->Item;

  for(i= 0; i < List->nitems; i++)
  {
    /* 
      We use (256-i) to sort a document from the first database
      before the same document from the second database.
    */
    Item[i].score = (Item[i].score << 8) + (255 - (num & 255));
   }
  return UDM_OK;
}


static udm_rc_t
UdmResultJoin(UDM_AGENT *A,
              UDM_QUERY *TmpQuery, size_t nresults,
              UDM_QUERY *Query)
{
  size_t i, max_total_found= 0;
  udm_timer_t ticks;
  
  for (Query->URLData.nitems= 0,
       Query->Res.num_rows= 0,
       Query->stats.total_found= 0, i= 0; i < nresults; i++)
  {
    size_t j;
    if (max_total_found < TmpQuery[i].stats.total_found)
      max_total_found= TmpQuery[i].stats.total_found;
    Query->stats.total_found+= TmpQuery[i].stats.total_found;
    Query->Res.num_rows+= TmpQuery[i].Res.num_rows;
    Query->URLData.nitems+= TmpQuery[i].URLData.nitems;
    for (j= 0; j < TmpQuery[i].Res.WWList.nwords; j++)
      UdmWideWordListAddForStat(&Query->Res.WWList, &TmpQuery[i].Res.WWList.Word[j]);
  }

  ticks= UdmStartTimer();
  UdmLog(A, UDM_LOG_DEBUG, "Start joining results from %d dbs", (int) nresults);

  if (Query->URLData.nitems > 0)
  {
    size_t offs;

    Query->URLData.Item= (UDM_URLDATA*)UdmMalloc(sizeof(UDM_URLDATA)
                                                 * Query->URLData.nitems);
    for (offs= 0, i= 0; i < nresults; offs+= TmpQuery[i].URLData.nitems, i++)
    {
      UDM_QUERY *CurQuery= &TmpQuery[i];
      UDM_URLDATA *CurData= TmpQuery[i].URLData.Item;
      size_t ncoords= CurQuery->URLData.nitems;
      if(CurData)
      {
        UdmResultSetMachineNumber(&CurQuery->URLData, i);
        memcpy(Query->URLData.Item + offs, CurData,
               sizeof(CurData[0]) * ncoords);
      }
      UDM_FREE(CurQuery->URLData.Item);
    }
  }

  if (nresults > 1 && Query->URLData.nitems > 0)
  {
    int group_by_site= UdmVarListFindBool(&A->Conf->Vars, "GroupBySite", UDM_FALSE) &&
                       (UdmVarListFindStr(&A->Conf->Vars, "site", "")[0] == '\0');
    int group_by_site_rank=  !strcmp(UdmVarListFindStr(&A->Conf->Vars, "GroupBySite", "no"), "rank");
    if (group_by_site)
    {
      UdmURLDataSortBySite(&Query->URLData);
      UdmURLDataGroupBySite(&Query->URLData);
      /*
        TODO: add better approximation of total results
        found with cluster. This number will be between
        max_total_found and Res->total_found, and will
        depend on overlap between the sites:
        (Res->URLData.nitems / Res->total_found)
      */
      Query->stats.total_found= Query->URLData.nitems;
    }
    else if (group_by_site_rank)
    {
      UdmURLDataSortBySite(&Query->URLData);
      UdmURLDataApplySiteRank(A, &Query->URLData, 1);
    }
    if (Query->URLData.nitems)
      UdmURLDataSortByPattern(&Query->URLData,
                              UdmVarListFindStr(&A->Conf->Vars, "s", "R"));
  }

  UdmLog(A, UDM_LOG_DEBUG, "Stop  joining results:\t%.2f", UdmStopTimer(&ticks));

  return UDM_OK;
}


static udm_rc_t
UdmFindWordsUseThreads(UDM_AGENT *A, udm_bool_t *use_threads)
{
  size_t i;
  if (!A->Conf->THDHandler.ThreadCreate ||
      !A->Conf->THDHandler.ThreadJoin ||
      A->Conf->DBList.nitems <= 1)
  {
    *use_threads= UDM_FALSE;
    return UDM_OK;
  }

  /* Check if all drivers are thread safe */
  for (i= 0; i < A->Conf->DBList.nitems; i++)
  {
    UDM_DB *db= &A->Conf->DBList.Item[i];
    udm_bool_t is_thread_safe;
    size_t bytes_written;
    udm_rc_t rc;
    if (UDM_OK != (rc= db->dbhandler->Info(db,
                                           &is_thread_safe, sizeof(is_thread_safe),
                                           &bytes_written, UDM_DBINFO_IS_THREAD_SAFE)))
      return rc;
    if (!is_thread_safe)
    {
      *use_threads= UDM_FALSE;
      return UDM_OK;
    }
  }
  *use_threads= UDM_TRUE;
  return UDM_OK;
}


static udm_rc_t
UdmFindWordsMulDB(UDM_AGENT *A, UDM_QUERY *Query, UDM_QUERY *TmpQuery)
{
  size_t i, ndatabases= A->Conf->DBList.nitems;
  UDM_SPRM search_param[256];
  udm_bool_t use_threads;
  udm_rc_t rc= UDM_OK;
  
  if (UDM_OK != (rc= UdmFindWordsUseThreads(A, &use_threads)))
    return rc;

  for (i= 0; i < ndatabases; i++)
  {
    UDM_QUERY *CurQuery= &TmpQuery[i];
    UDM_DB *db = &A->Conf->DBList.Item[i];
    CurQuery[0]= Query[0];    /* Copy bool items */
    UdmWideWordListCopy(&CurQuery->Res.WWList, &Query->Res.WWList);

    if (use_threads)
    {
      void *thd;
      UdmLog(A, UDM_LOG_DEBUG, "Starting thread[%d]", (int) i + 1);
      search_param[i].Agent= (UDM_AGENT*) UdmMalloc(sizeof(UDM_AGENT));
      UdmAgentInit(search_param[i].Agent, A->Conf, i + 1);
      search_param[i].db= db;
      search_param[i].Query= CurQuery;
      A->Conf->THDHandler.ThreadCreate(&thd, SearchThread, (void*) &search_param[i]);
      search_param[i].thd= thd;
    }
    else
    {
      if (UDM_OK != (rc= UdmFindWordsDB(A, CurQuery, db)))
        return rc;
    }
  }
    
  if (use_threads)
  {
    udm_timer_t ticks= UdmStartTimer();
    UdmLog(A, UDM_LOG_DEBUG, "Creating threads");
    for (i= 0; i < ndatabases; i++)
    {
      void *thd= search_param[i].thd;
      UdmLog(A, UDM_LOG_DEBUG, "Joining thread[%d]", (int) i + 1);
      A->Conf->THDHandler.ThreadJoin(thd);
    }
    UdmLog(A, UDM_LOG_DEBUG, "Threads finished: %.2f", UdmStopTimer(&ticks));
  }
  
  if (use_threads)
  {
    for (i= 0; i < ndatabases; i++)
    {
      UdmAgentFree(search_param[i].Agent);
      UdmFree(search_param[i].Agent);
    }
  }

  UdmResultJoin(A, TmpQuery, ndatabases, Query);
  
  return rc;
}


static udm_rc_t
UdmFindWords(UDM_AGENT *A, UDM_QUERY *Query, UDM_QUERY *TmpQuery)
{
  if (A->Conf->DBList.nitems > 1)
  {
    udm_rc_t rc;
    if (UDM_OK  != (rc= UdmFindWordsMulDB(A, Query, TmpQuery)))
      return rc;
  }
  else
  {
    udm_rc_t rc;
    UDM_DB *db = &A->Conf->DBList.Item[0];
    if (UDM_OK != (rc= UdmFindWordsDB(A, Query, db)))
      return rc;
    UdmResultSetMachineNumber(&Query->URLData, 0);
  }
  return UDM_OK;
}


#define SECTIONS_IN_VECTOR 64

/*
  Doc      - document
  Query    - the aggregate point of single database query
  TmpQuery - the array of partitions (in case of a multi database query)
*/
static void
UdmMakeWordAndSectionHitStatistics(UDM_AGENT *A,
                                   UDM_DOCUMENT *Doc,
                                   UDM_QUERY *Query,
                                   UDM_QUERY *TmpQuery,
                                   UDM_URLDATA *Data)
{
  char unique_word_vec[UDM_MAXWORDPERQUERY + 1];
  char section_vec[SECTIONS_IN_VECTOR + 1];
  unsigned long long unique_word= 0;
  UDM_QUERY *Q= A->Conf->DBList.nitems > 1 ?
                &TmpQuery[UDM_COORD2DBNUM(Data->score)] : Query;
  UDM_SEARCHSECTIONLIST *SectionList= &Q->SectionList;
  UDM_SEARCHSECTION *Sec= UdmSearchSectionListFind(SectionList, Data->url_id);
  UDM_SEARCHSECTION *End= SectionList->Section + SectionList->nsections;
  if (!Sec)
    return;
  memset(unique_word_vec, '0', sizeof(unique_word_vec) - 1);
  unique_word_vec[sizeof(unique_word_vec)-1]= '\0';
  memset(section_vec, '0', sizeof(section_vec) - 1);
  section_vec[sizeof(section_vec) - 1]= '\0';
  for ( ; Sec < End && Sec->url_id == Data->url_id; Sec++)
  {
    UDM_ASSERT(Sec->order < UDM_MAXWORDPERQUERY);
    unique_word|= 1ULL << Sec->order;
    unique_word_vec[Sec->order]= '1';
    if (Sec->secno > 0 && Sec->secno < SECTIONS_IN_VECTOR)
      section_vec[Sec->secno - 1]= '1';
  }
  UdmVarListReplaceULongLong(&Doc->Sections, "UniqueWordHitFlags", unique_word);
  UdmVarListReplaceStr(&Doc->Sections, "UniqueWordHitVector", unique_word_vec);
  UdmVarListReplaceStr(&Doc->Sections, "SectionHitVector", section_vec);
}


static udm_rc_t
CreateAlias(UDM_ENV *Conf, UDM_DOCUMENT *Doc)
{
  char *alcopy, *aliastr;
  const UDM_REPLACE *Alias;
  UDM_MATCH_PART Parts[10];

  /* Create "Alias" variable */
  alcopy= UdmRemoveHiLightDup(UdmVarListFindStr(&Doc->Sections,"URL",""));
  if ((Alias= UdmReplaceListFind(&Conf->Aliases, alcopy, 10, Parts)))
  {
    size_t nbytes= Alias->Replace.length + strlen(alcopy) + 64;
    aliastr= (char*)UdmMalloc(nbytes);
    UdmMatchApply(aliastr, nbytes, alcopy,
                  Alias->Replace.str, Alias->Replace.length,
                  &Alias->Match, 10, Parts);
  }
  else
  {
    aliastr = (char*)UdmStrdup(alcopy);
  }
  UdmVarListReplaceStr(&Doc->Sections,"Alias", aliastr);
  UdmFree(aliastr);
  UdmFree(alcopy);
  return UDM_OK;
}


static udm_rc_t
UdmQueryAddCloneList(UDM_AGENT *A, UDM_QUERY *Query)
{
  size_t i;
  udm_timer_t ticks= UdmStartTimer();
  udm_rc_t rc= UDM_OK;

  UdmLog(A, UDM_LOG_DEBUG, "Start Clones");
  for (i= 0; i < A->Conf->DBList.nitems && rc == UDM_OK; i++)
  {
    UDM_DB *db= &A->Conf->DBList.Item[i];
    rc= db->dbhandler->QueryAction(A, db, Query, UDM_QUERYCMD_CLONES);
  }
  UdmLog(A, UDM_LOG_DEBUG, "%-30s%.2f", "Stop  Clones:", UdmStopTimer(&ticks));
  return rc;
}


static udm_rc_t
UdmTrack(UDM_AGENT * query, UDM_QUERY *Query)
{
  udm_rc_t rc= UDM_OK;
  size_t i;
  const char *env= getenv("REMOTE_ADDR");
  UdmVarListAddStr(&query->Conf->Vars, "IP", env ? env : "");

  for (i= 0; i < query->Conf->DBList.nitems; i++)
  {
    UDM_DB *db= &query->Conf->DBList.Item[i];
    rc= db->dbhandler->QueryAction(query, db, Query, UDM_QUERYCMD_TRACK);
  }
  return rc;
}


static udm_rc_t
UdmResultAddCachedCopyLinks(UDM_AGENT *Agent, UDM_QUERY *Query)
{
  UDM_ENV *Env= Agent->Conf;
  char *searchwords= NULL, *storedstr= NULL; 
  size_t i, swlen, storedlen;
  
  for (i= swlen= 0; i < Query->Res.WWList.nwords; i++)
    swlen+= (8 * Query->Res.WWList.Word[i].Word.length) + 2;
  
  if ((searchwords= (char*) UdmXmalloc(swlen)) != NULL)
  {
    int z;
    for (z= i= 0; i < Query->Res.WWList.nwords; i++)
    {
      if (Query->Res.WWList.Word[i].Param.count > 0)
      {
        sprintf(UDM_STREND(searchwords), (z)?"+%s":"%s", Query->Res.WWList.Word[i].Word.str);
        z++;
      }
    }
  }
  storedstr= (char*) UdmRealloc(storedstr, storedlen= 1024 + 10 * swlen);
  
  for(i= 0; i < UdmResultNumRows(&Query->Res); i++)
  {
    UDM_DOCUMENT *Doc= &Query->Res.Doc[i];
    const char   *stored_href;
    int dbnum= UdmVarListFindInt(&Doc->Sections, "dbnum", 0);
    
    if ((!(stored_href= UdmVarListFindStr(&Doc->Sections, "stored_href", NULL)) ||
         !stored_href[0]) /*&&
         UdmVarListFindStr(&Doc->Sections, "CachedCopy", NULL) TODO34 */)
    {
      const char *u= UdmVarListFindStr(&Doc->Sections, "URL", "");
      size_t ulen= strlen(u);
      char *eu= (char*)UdmMalloc(ulen * 10 + 30);
      UdmURLEncode(eu, u, ulen);
      if (dbnum)
      {
        udm_snprintf(storedstr, storedlen, "?cc=1&amp;dbnum=%d&amp;URL=%s&amp;q=%s&amp;wm=%s",
                     dbnum, eu, searchwords,
                     UdmVarListFindStr(&Env->Vars, "wm", "wrd"));
      }
      else
      {
        udm_snprintf(storedstr, storedlen, "?cc=1&amp;URL=%s&amp;q=%s&amp;wm=%s",
                     eu, searchwords,
                     UdmVarListFindStr(&Env->Vars, "wm", "wrd"));
      }
      UdmFree(eu);
      UdmVarListReplaceStr(&Doc->Sections, "stored_href",  storedstr);
    }
    else
    {
      /* add dbnum to stored_href returned from searchd */
      if (stored_href && dbnum)
      {
        char tmp[256];
        udm_snprintf(tmp, sizeof(tmp), "?dbnum=%d&%s", dbnum, stored_href + 1);
        UdmVarListReplaceStr(&Doc->Sections, "stored_href", tmp);
      }
    }
  }
  UDM_FREE(searchwords);
  UDM_FREE(storedstr);
  
  return UDM_OK;
}



UDM_API(udm_rc_t)
UdmFind2(UDM_AGENT *Agent, UDM_RESULT *Res, const char *query_string)
{
  UdmVarListDelBySection(&Agent->Conf->Vars, UDM_VARSRC_QSTRING);
  UdmVarListAddQueryString(&Agent->Conf->Vars, "query.", query_string);
  return UdmFind(Agent, Res);
}


UDM_API(udm_rc_t)
UdmEnvFind(UDM_ENV *Env, UDM_RESULT *Res, const char *query_string)
{
  udm_rc_t rc;
  UDM_AGENT Agent;
  UdmAgentInit(&Agent, Env, 0);
  UdmVarListDelBySection(&Env->Vars, UDM_VARSRC_QSTRING);
  UdmVarListAddQueryString(&Env->Vars, "query.", query_string);
  rc= UdmFind(&Agent, Res);
  UdmAgentFree(&Agent);
  return rc;
}


UDM_API(udm_rc_t)
UdmFind(UDM_AGENT *A, UDM_RESULT *Result)
{
  UDM_QUERY *TmpQuery;
  udm_rc_t rc= UDM_OK;
  udm_timer_t ticks= UdmStartTimer(), ticks_;
  size_t  i, orig_num_rows, nbytes, numdatabases=  A->Conf->DBList.nitems;
  size_t  page_number= (size_t) UdmVarListFindInt(&A->Conf->Vars, "np", 0);
  size_t  page_size=   (size_t) UdmVarListFindInt(&A->Conf->Vars, "ps", 10);
  size_t  offs=        (size_t) UdmVarListFindInt(&A->Conf->Vars, "offs", 0);
  size_t  ExcerptSize= (size_t)UdmVarListFindInt(&A->Conf->Vars, "ExcerptSize", 256);
  size_t  ExcerptPadding = (size_t)UdmVarListFindInt(&A->Conf->Vars, "ExcerptPadding", 40);
  /*int     UseLocalCachedCopy= UdmVarListFindBool(&A->Conf->Vars, "UseLocalCachedCopy", 0);*/
  size_t  ResultsLimit= 1000;
  const UDM_VAR *ResultsLimitVar= UdmVarListFind(&A->Conf->Vars, "ResultsLimit");
  char    str[128];
  UDM_QUERY Query;

  UdmQueryInit(&Query);

  /*
    Make sure we use only search.htm command,
    do not allow to overwrite in query string.
    FIXME: Using template command and query string parameter
    at the same time makes limit equal to 1000.
    This should be fixed to use the template value instead.
    This needed when the template value is smaller than 1000.
  */
  if (ResultsLimitVar && (UdmVarSecno(ResultsLimitVar) == 0))
  {
    ResultsLimit= UdmVarIntVal(ResultsLimitVar);
  }

  UdmLog(A, UDM_LOG_EXTRA, "Start UdmFind");
  
  ticks_= UdmStartTimer();
  UdmLog(A, UDM_LOG_DEBUG, "Start Prepare");  
  nbytes= sizeof(UDM_QUERY) * numdatabases;
  TmpQuery= (UDM_QUERY*) UdmMalloc(nbytes);
  bzero((void*) TmpQuery, sizeof(UDM_QUERY) * numdatabases);
  if (UDM_OK != (rc= UdmQueryPrepare(A, &Query)))
    goto conv;
  UdmVarListAddStr(&A->Conf->Vars, "orig_m", UdmVarListFindStr(&A->Conf->Vars, "m", "all"));
  UdmLog(A, UDM_LOG_DEBUG, "%-30s%.2f", "Stop  Prepare", UdmStopTimer(&ticks_));

  Query.stats.first = offs ? offs : page_number * page_size;
  if (ResultsLimit && (Query.stats.first + page_size > ResultsLimit))
  {
    UdmLog(A, UDM_LOG_DEBUG, "Too large ps or np: offset=%d limit=%d\n",
           (int) (Query.stats.first + page_size), (int) ResultsLimit);
    goto conv;
  }

  ticks_= UdmStartTimer();
  UdmLog(A, UDM_LOG_DEBUG, "Start FindWords");  
  Query.num_best_rows= Query.stats.first + page_size;
  if (!Query.Res.WWList.nwords ||
      UDM_OK != (rc= UdmFindWords(A, &Query, TmpQuery)))
    goto conv;
  UdmLog(A, UDM_LOG_DEBUG, "%-30s%.2f", "Stop  FindWords", UdmStopTimer(&ticks_));

  if (!Query.stats.total_found)
  {
    udm_bool_t suggest= UdmVarListFindBool(&A->Conf->Vars, "Suggest", UDM_FALSE);
    if (suggest)
    {
      if (UDM_OK != (rc= UdmQueryAction(A, &Query, UDM_QUERYCMD_SUGGEST)))
        goto conv;
    }
  }
  
  UdmVarListReplaceStr(&A->Conf->Vars, "m", UdmVarListFindStr(&A->Conf->Vars, "orig_m", "all"));
  UdmVarListDel(&A->Conf->Vars, "orig_m");
  UdmVarListReplaceInt(&A->Conf->Vars, "CurrentTimestamp", (int) time(0));


  if (Query.stats.first >= Query.URLData.nitems)
  {
    Query.stats.last= Query.stats.first;
    Query.Res.num_rows= 0;
    goto conv; /* jump to converting variables into BrowserCharset */
  }
  if (Query.stats.first + page_size > Query.URLData.nitems)
  {
    orig_num_rows= Query.Res.num_rows= Query.URLData.nitems -
                                       Query.stats.first;
  }
  else
  {
    orig_num_rows= Query.Res.Doc ? Query.Res.num_rows : page_size;
    Query.Res.num_rows= page_size;
  }
  Query.stats.last= Query.stats.first + UdmResultNumRows(&Query.Res) - 1;
  ticks_= UdmStartTimer();
  UdmLog(A, UDM_LOG_DEBUG, "Start AddDocInfo");

  UdmResAllocDoc(&Query.Res);

  for (i= 0; i < numdatabases; i++)
  {
    UDM_DB *db= &A->Conf->DBList.Item[i];
    rc= db->dbhandler->QueryAction(A, db, &Query, UDM_QUERYCMD_DOCINFO);
  }

  /*
    Cluster: the original response had more documents than we need.
    Free the unused part of the documents: UdmResultFree()
    will not know that they existed.
  */
  if (orig_num_rows > UdmResultNumRows(&Query.Res))
  {
    for (i= UdmResultNumRows(&Query.Res); i < orig_num_rows ; i++)
    {
      UdmDocFree(&Query.Res.Doc[i]);
    }
  }

  /* Copy url_id and coord to result */
  for (i= 0; i < UdmResultNumRows(&Query.Res); i++)
  {
    UDM_URLDATA *Data= &Query.URLData.Item[i + Query.stats.first];
    uint4 score= Data->score;
    UDM_VARLIST *Sections= &Query.Res.Doc[i].Sections;
    UdmVarListReplaceUnsigned(Sections, "PerSite", Data->per_site);
    udm_snprintf(str, 128, "%.3f%%", ((double)(score >> 8)) / 1000);
    UdmVarListReplaceStr(Sections, "Score", str);
    UdmVarListReplaceInt(Sections,"Order",(int)(i + Query.stats.first + 1));
    UdmVarListReplaceInt(Sections, "dbnum", UDM_COORD2DBNUM(score));
    UdmMakeWordAndSectionHitStatistics(A, &Query.Res.Doc[i],
                                       &Query, TmpQuery, Data);
    /*
      Doc->lcs is needed for UdmResultMethodDocumentProperty in value-result.c.
      We currently set Doc->lcs to A->Conf->lcs.
      UdmQueryConvert() will reset it to A->Conf->bcs later.
    */
    Query.Res.Doc[i].lcs= A->Conf->lcs;
    /* TODO34: Add a fake cached copy to have search.cgi display stored_href */
    /*
    if (UseLocalCachedCopy && !UdmVarListFind(Sections, "CachedCopy")) TODO34
      UdmVarListAddStr(Sections, "CachedCopy", "Fake-CachedCopy");     TODO34
    */
  }
  
  
  for (i= 0; i < UdmResultNumRows(&Query.Res); i++)
    CreateAlias(A->Conf, &Query.Res.Doc[i]);
  
  UdmLog(A, UDM_LOG_DEBUG, "%-30s%.2f", "Stop  AddDocInfo", UdmStopTimer(&ticks_));
  
  if (UdmVarListFindInt(&A->Conf->Vars, "DetectClones", 0))
    rc= UdmQueryAddCloneList(A, &Query);

  /* first and last begins from 0, make it begin from 1 */
  UdmVarListAddInt(&Query.Res.Vars, "first", (int) Query.stats.first + 1);
  UdmVarListAddInt(&Query.Res.Vars, "last", (int) Query.stats.last + 1);
  UdmVarListAddInt(&Query.Res.Vars, "total", (int) Query.stats.total_found);

conv:

  UdmVarListAddInt(&Query.Res.Vars, "nwords", (int) Query.Res.WWList.nwords);
  UdmVarListAddInt(&Query.Res.Vars, "nuniq", (int) Query.Res.WWList.nuniq);

#ifndef NO_ADVANCE_CONV
  ticks_=UdmStartTimer();
  UdmLog(A, UDM_LOG_DEBUG, "Start UdmQueryConvert");
  UdmQueryConvert(A, &Query, A->Conf->lcs, A->Conf->bcs);
  UdmLog(A, UDM_LOG_DEBUG, "%-30s%.2f", "Stop  UdmQueryConvert:", UdmStopTimer(&ticks_));
#endif

  ticks_=UdmStartTimer();
  UdmLog(A, UDM_LOG_DEBUG, "Start Excerpts");
  for (i= 0; i < UdmResultNumRows(&Query.Res); i++)
  {
    char *Excerpt;
    if ((Excerpt= UdmExcerptDoc(A, &Query, &Query.Res.Doc[i], ExcerptSize, ExcerptPadding)))
    {
      UdmVarListReplaceStr(&Query.Res.Doc[i].Sections, "body", Excerpt);
      UdmFree(Excerpt);
    }
  }
  UdmLog(A, UDM_LOG_DEBUG, "%-30s%.2f", "Stop  Excerpts:", UdmStopTimer(&ticks_));

  UdmLog(A, UDM_LOG_DEBUG, "Start WordInfo");
  ticks_=UdmStartTimer();
  UdmQueryWordInfo(&Query);
  UdmResultAddCachedCopyLinks(A, &Query);
  UdmLog(A, UDM_LOG_DEBUG, "%-30s%.2f", "Stop  WordInfo:", UdmStopTimer(&ticks_));

  ticks= UdmStartTimer() - ticks;
  UdmVarListReplaceInt(&Query.Res.Vars, "SearchTime", (int) ticks);

  UdmTrack(A, &Query);
  UdmLog(A, UDM_LOG_EXTRA, "%-30s%.2f", "Stop  UdmFind:", (float)ticks/1000);

  /*
    Free the remaining parts of TmpRes.
    Note:
    - TmpRes[i].URLData is already freed in UdmResultJoin.
    - TmpRes[i].ItemList.items is not allocated, it points to Res.ItemList.items
    - TmpRes[i].Doc is allocated only in case of cluster
    - TmpRes[i].WWList is allocated.
    - TmpRes[i].where is allocated
    - TmpRes[i].from is allocated
  */
  for (i= 0; i < numdatabases; i++)
  {
    size_t j;
    UdmWideWordListFree(&TmpQuery[i].Res.WWList);
    UdmSearchSectionListFree(&TmpQuery[i].SectionList);
    for (j= 0; j < UdmResultNumRows(&TmpQuery[i].Res); j++)
      UdmDocFree(&TmpQuery[i].Res.Doc[j]);
    UdmFree(TmpQuery[i].Res.Doc);
    UDM_FREE(TmpQuery[i].where);
    UDM_FREE(TmpQuery[i].from);
  }

  UdmFree(TmpQuery);
  if (rc != UDM_OK)
  {
    UdmResultFree(&Query.Res);
    UdmResultInit(Result);
  }
  else
  {
    *Result= Query.Res;
    /* Clear Query.Res, to prevent UdmResultFree() in UdmQueryFree() */
    bzero(&Query.Res, sizeof(Query.Res));
  }
  UdmQueryFree(&Query);
  return rc;
}
