/* Copyright (C) 2000-2015 Lavtech.com corp. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
*/

#include "udm_config.h"

#include <string.h>

#include "udm_common.h"
#include "udm_log.h"
#include "udm_utils.h"
#include "udm_searchtool.h"
#include "udm_coords.h"
#include "udm_db.h"
#include "udm_hash.h"
#include "udm_url.h"

/*******************  GroupBySite  *********************************/


void
UdmURLDataApplySiteRank(UDM_AGENT *A, UDM_URLDATALIST *DataList,
                        int is_aggregation_point)
{
  size_t i, in_site_rank;
  urlid_t prev_site_id;
  for (i= 0, prev_site_id= in_site_rank= 1; i < DataList->nitems; i++)
  {
    UDM_URLDATA *D= &DataList->Item[i];
    if (prev_site_id != D->site_id)
    {
      in_site_rank= 0;
      prev_site_id= D->site_id;
    }
    in_site_rank++;
    if (is_aggregation_point)
    {
      if (in_site_rank > 1)
      {
        /* Keep machine number in the low byte */
        D->score= (D->score & 0xFF) + ((D->score / in_site_rank) & 0x7FFFFF00);
      }
    }
    else
      D->score/= in_site_rank;
  }
}


udm_rc_t
UdmURLDataListGroupBySiteUsingSort(UDM_AGENT *A,
                                   UDM_URLDATALIST *R,
                                   UDM_DB *db)
{
  udm_timer_t ticks;
  UDM_URLDATA *Dat, *End;

  /* Initialize per_site */
  for (Dat= R->Item, End= R->Item + R->nitems;
       Dat < End; Dat++)
    Dat->per_site= 1;

  UdmLog(A,UDM_LOG_DEBUG,"Start sort by site_id %d docs", (int) R->nitems);
  ticks=UdmStartTimer();
  UdmURLDataSortBySite(R);
  UdmLog(A,UDM_LOG_DEBUG,"Stop sort by site_id:\t%.2f",UdmStopTimer(&ticks));
  
  UdmLog(A,UDM_LOG_DEBUG,"Start group by site_id %d docs", (int) R->nitems);
  ticks=UdmStartTimer();
  UdmURLDataGroupBySite(R);
  UdmLog(A,UDM_LOG_DEBUG,"Stop group by site_id:\t%.2f", UdmStopTimer(&ticks));

  return UDM_OK;
}


void UdmURLDataGroupBySite(UDM_URLDATALIST *List)
{
  UDM_URLDATA *src= List->Item + 1;
  UDM_URLDATA *dst= List->Item;
  UDM_URLDATA *end= List->Item + List->nitems;
  uint4 count;
  
  if (!List->nitems)
    return;

  for(count= List->Item[0].per_site; src < end; src++)
  {
    UDM_ASSERT(src->url != dst->url || src->url == NULL);
    /* Group by site_id */
    if(dst->site_id == src->site_id)
    {
      count+= src->per_site;
      if (dst->score > src->score)
      {
        UdmURLDataFree(src);
        continue;
      }
      else if (dst->score == src->score)
      {
        if (dst->pop_rank > src->pop_rank)
        {
          UdmURLDataFree(src);
          continue;
        }
        else if (dst->pop_rank == src->pop_rank)
        {
          if (dst->url_id < src->url_id)
          {
            UdmURLDataFree(src);
            continue;
          }
        }
      }
      UdmURLDataFree(dst);
      dst->url_id=        src->url_id;
      dst->score=         src->score;
      dst->last_mod_time= src->last_mod_time;
      dst->pop_rank=      src->pop_rank;
      dst->url=           src->url;
      dst->section=       src->section;
    }
    else
    {
      /* Next site */
      dst->per_site= count;
      *++dst= *src;
      count= src->per_site;
    }
  }
  dst->per_site= count;
  List->nitems= dst - List->Item + 1;
  return;
}


typedef struct udm_str_with_url_id_st
{
  char *str;
  urlid_t url_id;
} UDM_STR_WITH_URL_ID;

typedef struct udm_url_with_url_id_list
{
  size_t nitems;
  UDM_STR_WITH_URL_ID *Item;
} UDM_STR_WITH_URL_ID_LIST;


static void
UdmStrWithURLIdListInit(UDM_STR_WITH_URL_ID_LIST *List)
{
  bzero((void*) List, sizeof(*List));
}

static udm_rc_t
UdmStrWithURLIdListAlloc(UDM_STR_WITH_URL_ID_LIST *List, size_t n)
{
  if (!((List->Item= (UDM_STR_WITH_URL_ID*) UdmMalloc(n * sizeof(UDM_STR_WITH_URL_ID)))))
    return UDM_ERROR;
  return UDM_OK;
}

static void
UdmStrWithURLIdListFree(UDM_STR_WITH_URL_ID_LIST *List)
{
  size_t i;
  for (i= 0; i < List->nitems; i++)
    UdmFree(List->Item[i].str);
  UdmFree(List->Item);
}

static int
UdmStrWithURLIdCmp(const UDM_STR_WITH_URL_ID *a, const UDM_STR_WITH_URL_ID *b)
{
  int rc= strcmp(a->str, b->str);
  if (rc)
    return rc;
  if (a->url_id < b->url_id)
    return -1;
  if (a->url_id > b->url_id)
    return 1;
  return 0;
}

static void
UdmStrWithURLIdListSort(UDM_STR_WITH_URL_ID_LIST *List)
{
  UdmSort(List->Item, List->nitems, sizeof(UDM_STR_WITH_URL_ID),
          (udm_qsort_cmp) UdmStrWithURLIdCmp);
}

static udm_rc_t
UdmStrWithURLIdListInitFromURLDataList(UDM_STR_WITH_URL_ID_LIST *Dst,
                                       UDM_URLDATALIST *Src)
{
  size_t i;
  UdmStrWithURLIdListInit(Dst);
  if (UDM_OK != UdmStrWithURLIdListAlloc(Dst, Src->nitems))
    return UDM_ERROR;
  for (i= 0; i < Src->nitems; i++)
  {
    UDM_URLDATA *Item= &Src->Item[i];
    UDM_STR_WITH_URL_ID *S= &Dst->Item[i];
    size_t sitelen= UdmAbsoluteURLSiteLength(Item->url);
    S->url_id= Item->url_id;
    if (!(S->str= udm_strndup(Item->url, sitelen)))
      return UDM_ERROR;
  }
  Dst->nitems= Src->nitems;
  return UDM_OK;
}


static size_t
UdmStrWithURLIdListSameSite(UDM_STR_WITH_URL_ID_LIST *List, size_t offs)
{
  size_t i;
  const char *first= List->Item[offs].str;
  for (i= offs; i < List->nitems; i++)
  {
    if (strcmp(first, List->Item[i].str))
      break;
  }
  return i - offs;
}


static udm_rc_t
UdmStrWithURLIdListToDSTR(UDM_STR_WITH_URL_ID_LIST *List, UDM_DSTR *dstr)
{
  size_t i;
  for (i= 0; i < List->nitems; )
  {
    UDM_STR_WITH_URL_ID *Item= &List->Item[i];
    size_t sitelen= strlen(Item->str);
    size_t j, n= UdmStrWithURLIdListSameSite(List, i);
    if (!n)
      break;
    /*fprintf(stderr, "[%d][%d]curlen=%d URL=%s\n", Item->url_id, (int) n, (int) dstr->size_data, Item->str);*/
    if (UDM_OK != UdmDSTRAppendCoord(dstr, sitelen) ||
        !UdmDSTRAppend(dstr, Item->str, sitelen) ||
        UDM_OK != UdmDSTRAppendCoord(dstr, n))
      return UDM_ERROR;
    for (j= 0; j < n; j++)
    {
      UDM_STR_WITH_URL_ID *Item2= &List->Item[i +  j];
      size_t delta= (j == 0) ? Item2->url_id : (Item2->url_id - Item2[-1].url_id);
      if (UDM_OK != UdmDSTRAppendCoord(dstr, delta))
        return UDM_ERROR;
      /*fprintf(stderr, "  [%d][%d][%d]\n", Item2->url_id, (int) delta, (int) dstr->size_data);*/
    }
    i+= n;
  }
  return UDM_OK;
}


/*
  Pack '##site' record from URLDataList.
  dstr must be initialized.
*/
udm_rc_t
UdmURLDataListPackSite(UDM_URLDATALIST *List, UDM_DSTR *dstr)
{
  udm_rc_t rc= UDM_OK;
  UDM_STR_WITH_URL_ID_LIST Site;
  bzero((void*) &Site, sizeof(Site));
  if (UDM_OK != (rc= UdmStrWithURLIdListInitFromURLDataList(&Site, List)))
   return UDM_ERROR;
  if (Site.nitems)
    UdmStrWithURLIdListSort(&Site);
  rc= UdmStrWithURLIdListToDSTR(&Site, dstr);
  UdmStrWithURLIdListFree(&Site);
  return UDM_OK;
}


static inline size_t
udm_coord_get_s(size_t *coord, const char *str, const char *end)
{
  return udm_coord_get(coord, (const unsigned char *) str,
                              (const unsigned char *) end);
}


static int
UdmURLDataComp(const UDM_URLDATA *a, const UDM_URLDATA *b)
{
  if (a->url_id < b->url_id)
    return -1;
  if (a->url_id > b->url_id)
    return 1;
  return 0;
}


udm_rc_t
UdmURLDataListUnpackSite(UDM_AGENT *A, UDM_URLDATALIST *List,
                         const UDM_CONST_STR *str_arg)
{
  const char *str= str_arg->str;
  const char *end= str + str_arg->length;
  UDM_URLDATA key;

  bzero((void*) &key, sizeof(key));
  UDM_ASSERT(List->nitems);

  for ( ; str < end ; )
  {
    UDM_CONST_STR site;
    size_t coord, nbytes, nurls;
    /* Get site name length */
    str+= (nbytes= udm_coord_get_s(&coord, str, end));
    if (!nbytes || !coord || (((size_t)(end - str)) < coord))
      goto err;
    UdmConstStrSet(&site, str, coord); /* Set site name */
    str+= coord;
    /* Get number of URLs in this site */
    str+= (nbytes= udm_coord_get_s(&nurls, str, end));
    if (!nbytes || !nurls || (((size_t)(end - str)) < nurls))
      goto err;
    /* Get all URL ids */
    for (key.url_id= 0; nurls; nurls--)
    {
      UDM_URLDATA *d;
      str+= (nbytes= udm_coord_get_s(&coord, str, end));
      if (!nbytes || !coord)
        goto err;
      key.url_id+= coord;
      if ((d= (UDM_URLDATA*) UdmBSearch(&key, List->Item, List->nitems,
                                        sizeof(UDM_URLDATA),
                                        (udm_qsort_cmp) UdmURLDataComp)))
      {
        /* TODO34: change to in-memory hash to resolve collisions */
        d->site_id= UdmHash32(site.str, site.length);
      }
    }
  }
  return UDM_OK;
err:
  UdmLog(A, UDM_LOG_ERROR, "Bad format for '##site' string");
  return UDM_ERROR;
}
