/* Copyright (C) 2000-2015 Lavtech.com corp. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
*/

#include "udm_config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>


#ifdef HAVE_WINSOCK_H
#include <winsock.h>
#endif
#ifdef HAVE_NETDB_H
#include <netdb.h>
#endif
#ifdef HAVE_SYS_SOCKET_H
#include <sys/socket.h>
#endif
#ifdef HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif
#ifdef HAVE_ARPA_INET_H
#include <arpa/inet.h>
#endif

#include "udm_common.h"
#include "udm_socket.h"
#include "udm_host.h"
#include "udm_proto.h"
#include "udm_utils.h"
#include "udm_crc32.h"

#ifndef INADDR_NONE
#define INADDR_NONE ((unsigned long) -1)
#endif

#define MAX_HOST_ADDR_SIZE     512
#define HOST_ADDR_REALLOC_SIZE  32

static int cmphost(const void * v1, const void * v2)
{
  return(strcasecmp(((const UDM_HOST_ADDR*)v1)->hostname,
                    ((const UDM_HOST_ADDR*)v2)->hostname));
}


static int
host_addr_add(UDM_HOSTLIST *List,
              const char *hostname, struct in_addr *addr,
              int dns_cache_timeout)
{
  UDM_HOST_ADDR *Host;

  if (List->nhost_addr < List->mhost_addr)
  {
    Host= &List->host_addr[List->nhost_addr];
    List->nhost_addr++;
  }
  else if (List->mhost_addr < MAX_HOST_ADDR_SIZE)
  {
    List->mhost_addr+= HOST_ADDR_REALLOC_SIZE;
    List->host_addr= (UDM_HOST_ADDR*) UdmRealloc(List->host_addr, List->mhost_addr*sizeof(UDM_HOST_ADDR));
    bzero((void*) &List->host_addr[List->nhost_addr], sizeof(UDM_HOST_ADDR));
    Host= &List->host_addr[List->nhost_addr];
    List->nhost_addr++;
  }
  else
  {
    size_t i;
    Host= &List->host_addr[0];
    /* find the last used host and replace it */
    for (i= 0; i < List->nhost_addr; i++)
    {
      if (List->host_addr[i].last_used < Host->last_used)
        Host= &List->host_addr[i];
    }
    UdmFree(Host->hostname);
  }

  Host->hostname= (char*)UdmStrdup(hostname);
  if (addr)
    memcpy(&Host->addr, addr, sizeof(struct in_addr));
  else
    bzero((void*) &Host->addr, sizeof(struct in_addr));
  Host->net_errors= 0;
  Host->last_used= time(NULL);
  Host->expires= Host->last_used + dns_cache_timeout;

  UdmSort(List->host_addr, List->nhost_addr, sizeof(UDM_HOST_ADDR), cmphost);
  return 0;
}


static void
host_addr_del(UDM_HOSTLIST *List, UDM_HOST_ADDR *Host)
{
  UDM_HOST_ADDR *End= List->host_addr + List->nhost_addr;
  UDM_ASSERT(Host >= List->host_addr && Host < End);
  UdmFree(Host->hostname);
  if (Host + 1 < End)
    memmove(Host, Host + 1, (End - Host - 1) * sizeof(UDM_HOST_ADDR));
  List->nhost_addr--;
  if (List->nhost_addr)
    UdmSort(List->host_addr, List->nhost_addr, sizeof(UDM_HOST_ADDR), cmphost);
}


static UDM_HOST_ADDR *
UdmHostFind(UDM_AGENT *A, UDM_HOSTLIST *List,const char *hostname)
{
  int l,c,r,res;

  UDM_LOCK_CHECK_OWNER(A, UDM_LOCK_HOST_CACHE);

  if (!hostname)
    return NULL;
  
  /* Find current URL in sorted part of list */
  l=0;r=List->nhost_addr-1;
  while(l<=r)
  {
    c=(l+r)/2;
    if(!(res= strcasecmp(List->host_addr[c].hostname, hostname)))
      return(&List->host_addr[c]);
    
    if(res<0)
      l=c+1;
    else
      r=c-1;
  }
  return NULL;
}


void
UdmHostErrorReset(UDM_AGENT *A, UDM_HOSTLIST *List, const char *hostname)
{
  UDM_HOST_ADDR *h;
  UDM_GETLOCK(A, UDM_LOCK_HOST_CACHE);
  if ((h= UdmHostFind(A, List, hostname)))
    h->net_errors= 0;
  UDM_RELEASELOCK(A, UDM_LOCK_HOST_CACHE);
}


void
UdmHostErrorIncrement(UDM_AGENT *A, UDM_HOSTLIST *List, const char *hostname)
{
  UDM_HOST_ADDR *h;
  UDM_GETLOCK(A, UDM_LOCK_HOST_CACHE);
  if ((h= UdmHostFind(A, List, hostname)))
    h->net_errors++;
  UDM_RELEASELOCK(A, UDM_LOCK_HOST_CACHE);
}


#ifdef HAVE_GETHOSTBYNAME_R_6
#define GETHOSTBYNAME_IS_REENTRANT
static udm_rc_t
HostLookupR(UDM_AGENT *A, UDM_CONN *connp)
{
  int err, rc;
  char buf[2048];
  struct hostent *he, res;
  if (!(rc= gethostbyname_r(connp->hostname, &res,
                            buf, sizeof(buf), &he, &err)) && he)
  {
    memcpy((char *) &(connp->sin.sin_addr), he->h_addr, (size_t)he->h_length);
    return UDM_OK;
  }
  return UDM_ERROR;
}

#else
static udm_rc_t
HostLookup(UDM_AGENT *A, UDM_CONN *connp)
{
  struct hostent *he;
  UDM_LOCK_CHECK_OWNER(A, UDM_LOCK_HOST_CACHE); /*  gethostbyname() isn't reentrant */
  if ((he= gethostbyname(connp->hostname)))
  {
    memcpy((char *)&(connp->sin.sin_addr), he->h_addr, (size_t)he->h_length);
    return UDM_OK;
  }
  return UDM_ERROR;
}
#ifdef WIN32
#define GETHOSTBYNAME_IS_REENTRANT
#define HostLookupR(A,C) HostLookup(A,C)
#endif /* WIN32 */
#endif /* HAVE_GETHOSTBYNAME_R_6 */



#ifdef GETHOSTBYNAME_IS_REENTRANT
static int
HostMutexNo(const char *host)
{
  return UDM_LOCK_HOST_FIRST + (UdmStrCRC32(host) % UDM_LOCK_HOST_COUNT);
}
#endif


/*
  Resolve host name and put into cache.
  Note, caching is needed for two purposes:
  - Speed up DNS lookups
  - Perform CrawlDelay
*/
static udm_rc_t
UdmHostLookupCached(UDM_AGENT *A,
                    UDM_HOSTLIST *List, UDM_CONN *connp,
                    int dns_cache_timeout)
{
  /* Check if hostname is in the standard numbers-and-dots notation */ 
  if ((connp->sin.sin_addr.s_addr= inet_addr(connp->hostname)) == INADDR_NONE)
  {
    /* Not numbers-and-dots notation */
    UDM_HOST_ADDR *Host;
    udm_rc_t rc= UDM_OK;

    UDM_GETLOCK(A, UDM_LOCK_HOST_CACHE);
    if ((Host= UdmHostFind(A, List, connp->hostname)))
    {
      Host->last_used= time(NULL);
      connp->net_errors= Host->net_errors;
      connp->host_last_used= Host->last_used;
      if (Host->addr.s_addr)
      {
        memcpy(&connp->sin.sin_addr, &Host->addr, sizeof(struct in_addr));
      }
      else
      {
        rc= UDM_ERROR;
      }
      if (Host->expires < time(NULL))
      {
        host_addr_del(List, Host);
        Host= NULL;
      }
    }
    UDM_RELEASELOCK(A, UDM_LOCK_HOST_CACHE);

    if (!Host)
    {
#ifdef GETHOSTBYNAME_IS_REENTRANT
      {
        int mutex= HostMutexNo(connp->hostname);
        /*printf("Mutex=%d\n", mutex);*/
        UDM_GETLOCK(A, mutex);
        rc= HostLookupR(A, connp);
        UDM_GETLOCK(A, UDM_LOCK_HOST_CACHE);
        host_addr_add(List, connp->hostname,
                      rc == UDM_OK ? &connp->sin.sin_addr : NULL,
                      dns_cache_timeout);
        UDM_RELEASELOCK(A, UDM_LOCK_HOST_CACHE);
        UDM_RELEASELOCK(A, mutex);
      }
#else
      {
        UDM_GETLOCK(A, UDM_LOCK_HOST_CACHE);
        rc= HostLookup(A, connp);
        host_addr_add(List, connp->hostname,
                      rc == UDM_OK ? &connp->sin.sin_addr : NULL,
                      dns_cache_timeout);
        UDM_RELEASELOCK(A, UDM_LOCK_HOST_CACHE);
      }
#endif
    }
    connp->err= (rc == UDM_OK) ? 0 : UDM_NET_CANT_RESOLVE;
    return rc;
  }
  else
  {
    /* Numbers-and-dots notation */
    UDM_HOST_ADDR *Host;
    UDM_GETLOCK(A, UDM_LOCK_HOST_CACHE);
    if ((Host= UdmHostFind(A, List, connp->hostname)))
    {
      Host->last_used= time(NULL);
      connp->net_errors= Host->net_errors;
      connp->host_last_used= Host->last_used;
    }
    else
    {
      /* Put into cache, for CrawlDelay */
      host_addr_add(List, connp->hostname, &connp->sin.sin_addr,
                    dns_cache_timeout);
    }
    UDM_RELEASELOCK(A, UDM_LOCK_HOST_CACHE);
  }
  return UDM_OK;
}


udm_rc_t
UdmHostLookupEx(UDM_AGENT *A, UDM_HOSTLIST *List, UDM_CONN *connp,
                int dns_cache_timeout)
{
  connp->net_errors= 0;

  if (!connp->hostname)
    return UDM_ERROR;

  bzero((void*)&connp->sin, sizeof(struct sockaddr_in));
  
  /*  Set port  */
  if (connp->port) 
    connp->sin.sin_port= htons(connp->port);
  else
  {
    connp->err= UDM_NET_ERROR;
    return UDM_ERROR;
  }
  return UdmHostLookupCached(A, List, connp, dns_cache_timeout);
}


udm_rc_t
UdmHostLookup2(UDM_AGENT *A, UDM_HOSTLIST *List, UDM_CONN *connp)
{
  return UdmHostLookupEx(A, List, connp, UDM_DNSCACHE_TIMEOUT);
}


void UdmHostListFree(UDM_HOSTLIST *List)
{
  size_t i;
  
  for(i=0; i<List->nhost_addr; i++)
    UDM_FREE(List->host_addr[i].hostname);
  
  UDM_FREE(List->host_addr);
  List->nhost_addr= List->mhost_addr= 0;
  return;
}


/********* IP address cache ****************/

void
UdmInAddrInit(UDM_INADDR *Item)
{
  bzero((void*) Item, sizeof(*Item));
}


void
UdmInAddrListInit(UDM_INADDRLIST *List)
{
  bzero((void*)List, sizeof(*List));
}


void
UdmInAddrListFree(UDM_INADDRLIST *List)
{
  UdmFree(List->Item);
}


static int
addrcmp(const UDM_INADDR *a, const UDM_INADDR *b)
{
  if (a->addr.s_addr < b->addr.s_addr)
    return -1;
  if (a->addr.s_addr > b->addr.s_addr)
    return 1;
  return 0;
}


UDM_INADDR *
UdmInAddrListFind(UDM_INADDRLIST *List, struct in_addr addr)
{
  UDM_INADDR Item;
  Item.addr= addr;
  if (!List->nitems)
    return NULL;
  return (UDM_INADDR*) UdmBSearch(&Item, List->Item, List->nitems,
                                  sizeof(UDM_INADDR), (udm_qsort_cmp) addrcmp);
}


udm_rc_t
UdmInAddrListAdd(UDM_INADDRLIST *List, const UDM_INADDR *Item)
{
  UDM_INADDR *Addr;
  if (List->mitems >= List->nitems)
  {
    List->mitems+= 1024;
    List->Item= (UDM_INADDR*) UdmRealloc(List->Item, List->mitems * sizeof(List->Item[0]));
    if (!List->Item)
      return UDM_ERROR;
  }
  Addr= &List->Item[List->nitems++];
  *Addr= *Item;
  UdmSort(List->Item, List->nitems, sizeof(UDM_INADDR), (udm_qsort_cmp) addrcmp);
  return UDM_OK;
}
