/* Copyright (C) 2000-2015 Lavtech.com corp. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
*/

#include "udm_config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>

#include "udm_common.h"
#include "udm_socket.h"
#include "udm_host.h"
#include "udm_utils.h"
#include "udm_xmalloc.h"
#include "udm_http.h"
#include "udm_conf.h"
#include "udm_contentencoding.h"
#include "udm_url.h"
#include "udm_vars.h"
#include "udm_hrefs.h"
#include "udm_crc32.h"


udm_content_type_t
UdmContentTypeByName(const char *str)
{
  if (!strncasecmp(str, UDM_CSTR_WITH_LEN("text/plain")) ||
      !strncasecmp(str, UDM_CSTR_WITH_LEN("text/tab-separated-values")) ||
      !strncasecmp(str, UDM_CSTR_WITH_LEN("text/x-diff")) ||
      !strncasecmp(str, UDM_CSTR_WITH_LEN("text/x-patch")) ||
      !strncasecmp(str, UDM_CSTR_WITH_LEN("application/x-patch")) ||
      !strncasecmp(str, UDM_CSTR_WITH_LEN("text/css")))
    return UDM_CONTENT_TYPE_TEXT_PLAIN;

  if (!strncasecmp(str,"text/html", 9))
    return UDM_CONTENT_TYPE_TEXT_HTML;

  /*
    application/atom+xml
    text/xml
    application/xml
    application/rsd+xml
    application/rdf+xml
    application/xml
    application/opensearchdescription+xml
    text/xml+rss
    application/rss+xml
    text/vnd.wap.wml
  */
  if (!strncasecmp(str, UDM_CSTR_WITH_LEN("text/xml")) ||
      !strncasecmp(str, UDM_CSTR_WITH_LEN("text/vnd.wap.wml")) ||
      !strncasecmp(str, UDM_CSTR_WITH_LEN("application/xml")) ||
      strstr(str, "+xml") ||
      strstr(str, "rss"))
    return UDM_CONTENT_TYPE_TEXT_XML;

  if (!strncasecmp(str, UDM_CSTR_WITH_LEN("message/rfc822")))
    return UDM_CONTENT_TYPE_MESSAGE_RFC822;

  if (!strncasecmp(str, "audio/mpeg", 10))
    return UDM_CONTENT_TYPE_AUDIO_MPEG;

  if (!strncasecmp(str, "mnogosearch/htdb", 16))
    return UDM_CONTENT_TYPE_HTDB;

  if (!strncasecmp(str, UDM_CSTR_WITH_LEN("application/vnd.openxmlformats-officedocument.wordprocessingml.document")))
    return UDM_CONTENT_TYPE_DOCX;

  /* Note: text/richtext and text/enriched is not rtf, it's something else */
  if (!strncasecmp(str, UDM_CSTR_WITH_LEN("text/rtf")) ||
      !strncasecmp(str, UDM_CSTR_WITH_LEN("application/rtf")) ||
      !strncasecmp(str, UDM_CSTR_WITH_LEN("application/x-rtf")))
    return UDM_CONTENT_TYPE_TEXT_RTF;

  return UDM_CONTENT_TYPE_UNKNOWN;
}


size_t
UdmHTTPFindContent(const char *src, size_t srclen)
{
  const char *src0= src, *srcend= src + srclen;
  for ( ; src < srcend; src++)
  {
    if (*src == '\n' && src + 1 < srcend && src[1] == '\n')
      return src - src0 + 2;
    if (*src == '\r' && src + 3 < srcend && !memcmp(src, "\r\n\r\n", 4))
      return src - src0 + 4;
  }
  return 0;
}


size_t
UdmHTTPBufFindContent(UDM_HTTPBUF *Buf)
{
  char *token;
  for (token= Buf->buf; *token; token++)
  {
    if(!strncmp(token,"\r\n\r\n",4))
    {
      Buf->content= token + 4;
      return token - Buf->buf;
      break;
    }
    else if(!strncmp(token,"\n\n",2))
    {
      Buf->content= token + 2;
      return token - Buf->buf;
      break;
    }
  }
  return 0;
}


static void
UdmSplitContentTypeAndCharset(UDM_VARLIST *Vars,
                              const char *ctname,
                              const char *csname, char *val)
{
  char *v;
  if ((v= strstr(val, "charset=")))
  {
    const char *cs;
    char *semicolon= strchr(v + 8, ';');
    if (semicolon)
      *semicolon= '\0';
    cs= UdmCharsetCanonicalName(v + 8);
    *v= '\0';
    UdmRTrim(val, "; ");
    UdmVarListReplaceStr(Vars, csname, cs ? cs : v + 8);
  }
  UdmVarListReplaceStr(Vars, ctname, val);
}


/* TODO34: unify with UdmParseHTTPResponse */
void
UdmHTTPBufParseHeadersForCachedCopy(UDM_VARLIST *Vars,
                                    const char *src, size_t srclen)
{
  char *headers, *token, *lt;
  udm_bool_t x_orig_content_type_found= UDM_FALSE;

  if (!(headers= (char*) udm_strndup(src, srclen)))
    return;

  for (token= udm_strtok_r(headers,"\r\n",&lt) ; token;
       token= udm_strtok_r(NULL,"\r\n",&lt))
  {
    char *val;
    if ((val= strchr(token,':')))
    {
      char name[128];
      *val++='\0';
      val= UdmTrim(val," \t:");
      udm_snprintf(name, sizeof(name), "Cached.%s", token);
      if (!strcasecmp(token, "Content-Type") ||
          !strcasecmp(token, "Content-Encoding") ||
          !strcasecmp(token, "X-Orig.Content-Type"))
      {
        char *v;
        for(v=val ; *v ; v++) 
          *v= tolower(*v);
        /* TODO34: UdmNormalizeContenType*/
        if (!strcasecmp(token, "Content-Type"))
          UdmSplitContentTypeAndCharset(Vars, name, "Cached.Charset", val);
        else if (!strcasecmp(token, "X-Orig.Content-Type"))
        {
          UdmSplitContentTypeAndCharset(Vars, "Content-Type", "Charset", val);
          x_orig_content_type_found= UDM_TRUE;
        }
        else
          UdmVarListReplaceStr(Vars, name, val);
      }
      else
        UdmVarListReplaceStr(Vars, name, val);
    }
  }

  /*
     If there was no "X-Orig.Content-Type" header, then copy
     "Content-Type" and "Charset" from
     "Cached.Content-Type" and "Cached.Charset",
     to display in search template.
   */
  if (!x_orig_content_type_found)
  {
    const char *val;
    if ((val= UdmVarListFindStr(Vars, "Cached.Content-Type", NULL)))
      UdmVarListReplaceStr(Vars, "Content-Type", val);
    if ((val= UdmVarListFindStr(Vars, "Cached.Charset", NULL)))
      UdmVarListReplaceStr(Vars, "Charset", val);
  }
  UdmFree(headers);
}


void UdmParseHTTPResponse(UDM_AGENT * Indexer,UDM_DOCUMENT * Doc){      
  char  *token, *lt, *headers;
  int     oldstatus, htdb_content_length= 0;
  size_t content_offset;
  
  UdmHTTPBufResetContent(&Doc->Buf);
  oldstatus = UdmVarListFindInt(&Doc->Sections, "Status", 0);
  UdmVarListReplaceInt(&Doc->Sections, "ResponseSize", (int) UdmHTTPBufSize(&Doc->Buf));
  UdmVarListDel(&Doc->Sections, "Content-Length");
  UdmVarListDel(&Doc->Sections, "Last-Modified");
  
  /* Cut HTTP response header first */
  if (!(content_offset= UdmHTTPBufFindContent(&Doc->Buf)))
  {
    /* Put content pointer to the very end of the buffer */
    Doc->Buf.content= Doc->Buf.buf + Doc->Buf.size;
    return;
  }
  
  /* Copy headers not to break them */
  headers= (char*) udm_strndup(UdmHTTPBufPtr(&Doc->Buf), content_offset);
  
  /* Now lets parse response header lines */
  if (!(token = udm_strtok_r(headers,"\r\n",&lt)))
    return;
  
  if(!strncmp(token,"HTTP/",5))
  {
    int  status = atoi(token + 8);
    UdmVarListReplaceStr(&Doc->Sections,"ResponseLine",token);
    UdmVarListReplaceInt(&Doc->Sections, "Status", (oldstatus > status) ? oldstatus : status );
  }
  else
  {
    return;
  }
  
  for (token= udm_strtok_r(NULL,"\r\n",&lt) ; token;
       token = udm_strtok_r(NULL,"\r\n",&lt))
  {
    char *val;
    
    if((val=strchr(token,':')))
    {
      *val++='\0';
      val = UdmTrim(val," \t:");
      
      if (!strcasecmp(token, "Content-Type") ||
          !strcasecmp(token, "Content-Encoding"))
      {
        char *v;
        for(v=val ; *v ; v++) 
          *v = tolower(*v);
      }
      /*
      else if (!strcasecmp(token, "HTDB-Content-Length"))
      {
        htdb_content_length= atoi(val);
      }
      */
      else if (!strcasecmp(token, "Set-Cookie"))
      {
        char fullname[256];
        char *part, *lpart;
        char *name= NULL;
        char *value= NULL;
        const char *domain= NULL;
        const char *path= NULL;
        for (part= udm_strtok_r(val, ";" , &lpart) ; part;
             part= udm_strtok_r(NULL, ";", &lpart))
        {
          char *arg;
          part= UdmTrim(part, " ");
          if ((arg= strchr(part, '=')))
          {
            *arg++= '\0';
            if (!name)
            {
              name= part;
              value= arg;
              continue;
            }
            if (!strcasecmp(part, "path"))
            {
              path= arg;
              continue;
            }
            if (!strcasecmp(part, "domain"))
            {
              domain= arg;
              continue;
            }
          }
        }
        if (name && value)
        {
          if (domain && domain[0] == '.')
          {
            domain++;
          }
          else
          {
            domain= Doc->CurURL.hostname ? Doc->CurURL.hostname : "localhost";
          }
          if (!path)
          {
            path= Doc->CurURL.path ? Doc->CurURL.path : "/";
          }
          udm_snprintf(fullname, sizeof(fullname), "Set-Cookie.%s@%s%s", name, domain, path);
          UdmVarListReplaceStr(&Doc->Sections, fullname, value);
        }
        continue;
      }
    }
    UdmVarListReplaceStr(&Doc->Sections,token,val?val:"<NULL>");
  }
  UDM_FREE(headers);
  
  UdmVarListInsInt(&Doc->Sections,"Content-Length",
                   UdmHTTPBufContentSize(&Doc->Buf) + htdb_content_length);
}


void
UdmHTTPBufInit(UDM_HTTPBUF *Buf)
{
  bzero(Buf, sizeof(*Buf));
}


const char*
UdmHTTPBufPtr(const UDM_HTTPBUF *Buf)
{
  /*return Buf->buf + header_size;*/
  return Buf->buf;
}


const char*
UdmHTTPBufContent(const UDM_HTTPBUF *Buf)
{
  /*return Buf->buf + header_size;*/
  return Buf->content;
}


udm_bool_t
UdmHTTPBufContentToConstStr(const UDM_HTTPBUF *Buf, UDM_CONST_STR *str)
{
  if (!Buf->content)
    return UDM_TRUE;
  str->str= UdmHTTPBufContent(Buf);
  str->length= UdmHTTPBufContentSize(Buf);
  return UDM_FALSE;
}

void
UdmHTTPBufReset(UDM_HTTPBUF *Buf)
{
  Buf->size= 0;
  Buf->content= NULL;
  Buf->buf[0]= '\0';
}


size_t
UdmHTTPBufAllocedSize(const UDM_HTTPBUF *Buf)
{
  return Buf->alloced_size;
}


size_t
UdmHTTPBufAvailableSize(const UDM_HTTPBUF *Buf)
{
  UDM_ASSERT(Buf->alloced_size >= Buf->size);
  return Buf->alloced_size - Buf->size;
}


size_t
UdmHTTPBufSize(const UDM_HTTPBUF *Buf)
{
  return Buf->size;
}


size_t
UdmHTTPBufContentSize(const UDM_HTTPBUF *Buf)
{
  return Buf->size - (Buf->content - Buf->buf);
}


/*
  Header size, including trailing "\r\n"
*/
size_t
UdmHTTPBufHeaderSize(const UDM_HTTPBUF *Buf)
{
  return Buf->content - Buf->buf;
}


udmcrc32_t
UdmHTTPBufCRC32(const UDM_HTTPBUF *Buf)
{
  UDM_CONST_STR content;
  if (UdmHTTPBufContentToConstStr(Buf, &content))
    return 0;
  return UdmCRC32(content.str, content.length);
}


/*
  "dst" must have enough space to store UdmHTTPBufContentSize() bytes.
*/
void
UdmHTTPBufContentExport(const UDM_HTTPBUF *Buf, char *dst)
{
  UDM_CONST_STR content;
  if (UdmHTTPBufContentToConstStr(Buf, &content))
    return;
  memcpy(dst, content.str, content.length);
}


void
UdmHTTPBufPutContent(UDM_HTTPBUF *Buf, const char *src, size_t len)
{
  size_t available_len= UdmHTTPBufAvailableSize(Buf);
  if (!available_len)
    return;
  available_len--;  /* For trailing 0 */
  if (len > available_len)
    len= available_len;
  memcpy(Buf->content, src, len);
  Buf->size= UdmHTTPBufHeaderSize(Buf) + len;
  Buf->buf[Buf->size]= '\0';
}


void
UdmHTTPBufTruncateContent(UDM_HTTPBUF *Buf)
{
  Buf->size= UdmHTTPBufHeaderSize(Buf);
  Buf->buf[Buf->size]= '\0';
}

void
UdmHTTPBufResetContent(UDM_HTTPBUF *Buf)
{
  Buf->content= NULL;
}


udm_rc_t
UdmHTTPBufAlloc(UDM_HTTPBUF *Buf, size_t size)
{
  UDM_ASSERT(size > 0 && Buf->alloced_size == 0 && Buf->content == NULL);
  if (!(Buf->buf= (char*) UdmMalloc(size)))
    return UDM_ERROR;
  Buf->alloced_size= size;
  Buf->buf[0]= '\0';
  return UDM_OK;
}


udm_rc_t
UdmHTTPBufRealloc(UDM_HTTPBUF *Buf, size_t size)
{
  UDM_ASSERT(size > 0);
  if (!(Buf->buf= (char*) UdmRealloc(Buf->buf, size)))
    return UDM_ERROR;
  Buf->alloced_size= size;
  return UDM_OK;
}


udm_rc_t
UdmHTTPBufShrink(UDM_HTTPBUF *Buf)
{
  return UdmHTTPBufRealloc(Buf, Buf->size);
}


void
UdmHTTPBufFree(UDM_HTTPBUF *Buf)
{
  UDM_FREE(Buf->buf);
  Buf->content= NULL;
  Buf->size= 0;
  Buf->alloced_size= 0;
}


static udm_rc_t
UdmHTTPBufAppendVA(UDM_HTTPBUF *Buf, const char *fmt, va_list ap)
{
  int nc;
  size_t bytes_left= Buf->alloced_size - Buf->size - 1;
  UDM_ASSERT(Buf->alloced_size);
  UDM_ASSERT(Buf->alloced_size > Buf->size);

  if (bytes_left < 2)
    return UDM_ERROR;

  nc= vsnprintf(Buf->buf + Buf->size, bytes_left, fmt, ap);
  if (nc > -1 && nc < bytes_left)
    Buf->size+= nc;
  Buf->buf[Buf->size]= '\0';
  return UDM_OK;
}


udm_rc_t
UdmHTTPBufPrintf(UDM_HTTPBUF *Buf, const char *fmt, ...)
{
  udm_rc_t rc;
  va_list ap;
  UdmHTTPBufReset(Buf);
  va_start(ap, fmt);
  rc= UdmHTTPBufAppendVA(Buf, fmt, ap);
  va_end(ap);
  return rc;
}


udm_rc_t
UdmHTTPBufAppendf(UDM_HTTPBUF *Buf, const char *fmt, ...)
{
  udm_rc_t rc;
  va_list ap;
  va_start(ap, fmt);
  rc= UdmHTTPBufAppendVA(Buf, fmt, ap);
  va_end(ap);
  return rc;
}


udm_rc_t
UdmHTTPBufAppendFromFile(UDM_HTTPBUF *Buf, int fd)
{
  ssize_t bytes;
  /* Note: we need a loop, in case fd is a pipe (e.g. popen) */
  while ((bytes= read(fd, Buf->buf + Buf->size,
                          Buf->alloced_size - 1 - Buf->size)))
  {
    if (bytes == -1)
      return UDM_ERROR;
    Buf->size+= bytes;
  }
  Buf->buf[Buf->size]= '\0';
  return UDM_OK;
}

udm_rc_t
UdmHTTPBufAppend(UDM_HTTPBUF *Buf, const char *src, size_t len)
{
  size_t bytes_left= Buf->alloced_size - Buf->size - 1;
  if (bytes_left > 0)
  {
    size_t nbytes= UDM_MIN(bytes_left, len);
    memcpy(Buf->buf + Buf->size, src, nbytes);
    Buf->size+= nbytes;
    Buf->buf[Buf->size]= '\0';
    return UDM_OK;
  }
  return UDM_ERROR;
}


udm_rc_t
UdmHTTPBufAppendHTTPTime(UDM_HTTPBUF *Buf, time_t tm)
{
  char str[UDM_MAXTIMESTRLEN];
  UdmTime_t2HttpStr(tm, str, UDM_MAXTIMESTRLEN);
  return UdmHTTPBufAppend(Buf, str, strlen(str));
}


void
UdmHTTPBufShiftLeft(UDM_HTTPBUF *Buf, size_t offs)
{
  UDM_ASSERT(offs <= Buf->size);
  memmove(Buf->buf, Buf->buf + offs, Buf->size - offs);
  Buf->content= Buf->buf;
  Buf->size-= offs;
  Buf->buf[Buf->size]= '\0';
}


udm_rc_t
UdmHTTPBufInflate(UDM_HTTPBUF *Buf, const char *src, size_t srclen)
{
  udm_rc_t rc;
  Buf->size= UdmInflate(Buf->buf, Buf->alloced_size - 1, src, srclen, &rc);
  Buf->content= Buf->buf;
  Buf->buf[Buf->size]= '\0';
  return rc;
}


udm_rc_t
UdmHTTPBufInflateAppend(UDM_HTTPBUF *Buf, const char *src, size_t srclen)
{
  udm_rc_t rc;
  size_t dstlen;
  if (Buf->size + 1 > Buf->alloced_size)
    return UDM_ERROR;
  /* TODO34: Check why the old code (in 3.3) had this:
  if ((content[0] == (char) 0x1f) && (content[1] == (char)0x8b))
  {
    memcpy(zstream.next_in, content + 2, csize - 2);
    // 2 bytes - header, 4 bytes - trailing CRC
    zstream.avail_in= csize - 6;   
  }
  else
  {
    memcpy(zstream.next_in, UdmHTTPBufContent (&Doc->Buf), csize);
    zstream.avail_in= csize;
  }
  */
  dstlen= UdmInflate(Buf->buf + Buf->size, Buf->alloced_size - Buf->size - 1,
                     src, srclen, &rc);
  Buf->size+= dstlen;
  Buf->buf[Buf->size]= '\0';
  return rc;
}


udm_rc_t
UdmHTTPBufDeflateAppend(UDM_HTTPBUF *Buf, const char *src, size_t srclen)
{
  udm_rc_t rc;
  size_t dstlen;
  if (Buf->size + 1 > Buf->alloced_size)
    return UDM_ERROR;
  dstlen= UdmDeflate(Buf->buf + Buf->size, Buf->alloced_size - Buf->size - 1,
                     src, srclen, &rc);
  Buf->size+= dstlen;
  Buf->buf[Buf->size]= '\0';
  return rc;
}


udm_rc_t
UdmHTTPBufUnGzipAppend(UDM_HTTPBUF *Buf, const char *src, size_t srclen)
{
  udm_rc_t rc;
  size_t dstlen;
  if (Buf->size + 1 > Buf->alloced_size)
    return UDM_ERROR;
  dstlen= UdmUnGzip(Buf->buf + Buf->size, Buf->alloced_size - Buf->size - 1,
                    src, srclen, &rc);
  Buf->size+= dstlen;
  Buf->buf[Buf->size]= '\0';
  return rc;
}


udm_rc_t
UdmHTTPBufUncompressAppend(UDM_HTTPBUF *Buf, const char *src, size_t srclen)
{
  size_t dstlen;
  udm_rc_t rc;
  if (Buf->size + 1 > Buf->alloced_size)
    return UDM_ERROR;
  dstlen= UdmUncompress(Buf->buf + Buf->size, Buf->alloced_size - Buf->size - 1,
                        src, srclen, &rc);
  Buf->size+= dstlen;
  Buf->buf[Buf->size]= '\0';
  return rc;
}


udm_rc_t
UdmHTTPBufAppendEncodingDecode(UDM_HTTPBUF *Buf,
                               const char *src, size_t length,
                               udm_content_encoding_t encoding)
{
  udm_rc_t rc= UDM_OK;
  switch (encoding)
  {
    case UDM_CONTENT_ENCODING_GZIP:
      rc= UdmHTTPBufUnGzipAppend(Buf, src, length);
      break;
    case UDM_CONTENT_ENCODING_DEFLATE:
      rc= UdmHTTPBufInflateAppend(Buf, src, length);
      break;
    case UDM_CONTENT_ENCODING_ZLIB_COMPRESS:
      rc= UdmHTTPBufUncompressAppend(Buf, src, length);
      break;
    case UDM_CONTENT_ENCODING_IDENTITY:
      UdmHTTPBufAppend(Buf, src, length);
      rc= UDM_OK;
      break;
    case UDM_CONTENT_ENCODING_UNKNOWN:
      rc= UDM_ERROR;
  }
  return rc;
}


udm_rc_t
UdmHTTPBufAppendBase64Decode(UDM_HTTPBUF *Buf,
                             const char *src, size_t srclen)
{
  int res;

  if (UdmHTTPBufAvailableSize(Buf) < srclen + 1 ||
      (res= UdmBase64Decode(src, srclen, Buf->buf + Buf->size, NULL, 0)) < 0)
    return UDM_ERROR;
  Buf->size+= (size_t) res;
  Buf->buf[Buf->size]= '\0';
  return UDM_OK;
}


udm_rc_t
UdmHTTPBufBase64Decode(UDM_HTTPBUF *Buf,
                       const char *src, size_t len)
{
  UdmHTTPBufReset(Buf);
  Buf->content= Buf->buf; /*  No headers */
  return UdmHTTPBufAppendBase64Decode(Buf, src, len);
}


udm_rc_t
UdmHTTPBufAppendQuotedPrintableDecode(UDM_HTTPBUF *Buf,
                                      const char *src, size_t srclen)
{
  size_t res, available_size;

  if ((available_size= UdmHTTPBufAvailableSize(Buf)) < srclen + 1)
    return UDM_ERROR;
  res= udm_quoted_printable_decode(src, srclen, Buf->buf + Buf->size, available_size);
  Buf->size+= (size_t) res;
  Buf->buf[Buf->size]= '\0';
  return UDM_OK;
}


udm_rc_t
UdmHTTPBufQuotedPrintableDecode(UDM_HTTPBUF *Buf,
                                const char *src, size_t len)
{
  UdmHTTPBufReset(Buf);
  Buf->content= Buf->buf; /*  No headers */
  return UdmHTTPBufAppendQuotedPrintableDecode(Buf, src, len);
}


int
UdmHTTPBufAppendRecv(UDM_HTTPBUF *Buf, int fd, size_t len, int flags)
{
  int status;
  if (UdmHTTPBufAvailableSize(Buf) < 2)
    return 0;
  if (UdmHTTPBufAvailableSize(Buf) < len + 1)
    len= UdmHTTPBufAvailableSize(Buf) - 1;
  if ((status= recv(fd, Buf->buf + Buf->size, len, flags)) > 0)
  {
    Buf->size+= status;
    Buf->buf[Buf->size]= '\0';
  }
  return status;
}

