/* Copyright (C) 2000-2015 Lavtech.com corp. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
*/

#include "udm_config.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "udm_common.h"
#include "udm_utils.h"
#include "udm_db.h"
#include "udm_log.h"
#include "udm_indexcache.h"
#include "udm_vars.h"
#include "udm_stopwords.h"
#include "udm_coords.h"

/** Words ************************************/


static void
UdmInvertedIndexWordListFree(UDM_INVERTED_INDEX_WORD_LIST *List)
{
  UdmFree(List->Item);
}


static inline udm_rc_t
UdmInvertedIndexWordListRealloc(UDM_INVERTED_INDEX_WORD_LIST *List)
{
  if (List->nitems >= List->mitems)
  {
    List->mitems+= 16*1024;
    List->Item= (UDM_INVERTED_INDEX_WORD*)
                UdmRealloc(List->Item, List->mitems * sizeof(UDM_INVERTED_INDEX_WORD));
    if (!List->Item)
    {
      List->nitems= 0;
      List->mitems= 0;
      return UDM_ERROR;
    }
  }
  return UDM_OK;

}


static udm_rc_t
UdmInvertedIndexWordListAdd(UDM_INVERTED_INDEX_WORD_LIST *List,
                            const UDM_INVERTED_INDEX_WORD *Item)
{
  if (UDM_OK != UdmInvertedIndexWordListRealloc(List))
    return UDM_ERROR;
  List->Item[List->nitems]= *Item;
  List->nitems++;
  return UDM_OK;
}


typedef struct
{
  UDM_CONST_STR str;
  udmhash32_t crc;
} UDM_INVERTED_INDEX_HASH_WORD_REC;


typedef struct
{
  udmhash32_t crc;                /*  4 */
  uint4 word_id;                  /*  4 */
} UDM_INVERTED_INDEX_HASH_WORD_KEY;


#define WORD_RESERVE_SIZE 128

static udm_rc_t
UdmWordStore(UDM_HASH *hash, void *ofs, void *item)
{
  UDM_CONSTWORD_HASH_DATA *hash_data= (UDM_CONSTWORD_HASH_DATA *) hash->user_data;
  UDM_ENV *Env= hash_data->cache->param.Env;
  UDM_INVERTED_INDEX_HASH_WORD_KEY *dst= (UDM_INVERTED_INDEX_HASH_WORD_KEY*) ofs;
  UDM_INVERTED_INDEX_HASH_WORD_REC *src= (UDM_INVERTED_INDEX_HASH_WORD_REC*) item;
  UDM_INVERTED_INDEX_WORD W;

  if (!(W.str= UdmMemrootAlloc(&hash_data->cache->Words.root,
                               WORD_RESERVE_SIZE + 1)))
    return UDM_ERROR;
  W.length= UdmConvLCase(hash_data->cache->param.unidata,
                         &hash_data->cnv,
                         hash_data->cache->param.cnvflags | UDM_RECODE_HTML_IN,
                         W.str, WORD_RESERVE_SIZE,
                         src->str.str, src->str.length);
  W.str[W.length]= '\0';
  hash_data->cache->Words.root.last_page_used_size-=
    (WORD_RESERVE_SIZE - W.length);
  /*W.count= 1;*/
  W.last_url_id= hash_data->url_id;
  W.last_url_id_count= 1;
  W.crc= src->crc;
  W.is_stopword= Env->StopWord.nitems && 
                 UdmStopListListFind(&Env->StopWord, W.str, "") ?
                 UDM_TRUE : UDM_FALSE;
  dst->crc= src->crc;
  dst->word_id= (uint4) hash_data->cache->Words.list.nitems;
  if (UDM_OK != UdmInvertedIndexWordListAdd(&hash_data->cache->Words.list, &W))
    return UDM_ERROR;

/*
  fprintf(stderr, "%s-%s '%.*s'-'%.*s'\n",
   hash_data->cnv.from->name,
   hash_data->cnv.to->name,
   (int) src->str.length, src->str.str,
   (int) lcword_length, lcword);
*/    
  /*memcpy(ofs, item, sizeof(UDM_INVERTED_INDEX_HASH_WORD));*/
  return UDM_OK;
}


static udm_rc_t
UdmWordJoin(UDM_HASH *hash, void *ofs, void *item)
{
  UDM_CONSTWORD_HASH_DATA *hash_data= (UDM_CONSTWORD_HASH_DATA *) hash->user_data;
  UDM_INVERTED_INDEX_WORD_LIST *WL= &hash_data->cache->Words.list;
  UDM_INVERTED_INDEX_HASH_WORD_KEY *W1= ((UDM_INVERTED_INDEX_HASH_WORD_KEY *) ofs);
  UDM_INVERTED_INDEX_WORD *W= &WL->Item[W1->word_id];
  /*W->count++;*/
  if (W->last_url_id == hash_data->url_id)
  {
    W->last_url_id_count++;
  }
  else
  {
    W->last_url_id_count= 1;
    W->last_url_id= hash_data->url_id;
  }
  return UDM_OK;
}


static int
UdmWordCmp(UDM_HASH *hash, void *W1, void *W2)
{
  /* TODO34: add more thorough comparision */
  return (((const UDM_INVERTED_INDEX_HASH_WORD_KEY *) W1)->crc !=
          ((const UDM_INVERTED_INDEX_HASH_WORD_REC *) W2)->crc);
}


static udmcrc32_t
UdmWordKeyKey(UDM_HASH *hash, const void *key)
{
  return ((const UDM_INVERTED_INDEX_HASH_WORD_KEY *) key)->crc;
}


static udmcrc32_t
UdmWordRecKey(UDM_HASH *hash, const void *rec)
{
  return ((const UDM_INVERTED_INDEX_HASH_WORD_REC *) rec)->crc;
}


UDM_HASH_HANDLER word_hash_handler=
{
  UdmWordStore,  /* store  */
  UdmWordJoin,   /* join   */
  UdmWordCmp,    /* cmp    */
  UdmWordKeyKey, /* keykey */
  UdmWordRecKey  /* reckey */
};


static int
bc2cmp(UDM_INVERTED_INDEX_CACHE_ITEM *w1, UDM_INVERTED_INDEX_CACHE_ITEM *w2)
{
  int rc;
  if ((rc= strcmp(w1->ptr, w2->ptr)))
    return rc;
  if ((rc= ((int) w1->secno - (int) w2->secno)))
    return rc;
  return w1->url_id < w2->url_id ? -1 : 1;
}


static void
UdmInvertedIndexWordDictionaryInit(UDM_INVERTED_INDEX_WORD_DICTIONARY *D)
{
  bzero((void*) &D->list, sizeof(UDM_INVERTED_INDEX_WORD_LIST));
  UdmHashInit(&D->hash, &word_hash_handler, NULL, 1024, sizeof(UDM_INVERTED_INDEX_HASH_WORD_KEY));
  UdmMemrootInit(&D->root, 1024*1024);
}


static void
UdmInvertedIndexWordDictionaryFree(UDM_INVERTED_INDEX_WORD_DICTIONARY *D)
{
  UdmHashFree(&D->hash);
  UdmInvertedIndexWordListFree(&D->list);
  UdmMemrootFree(&D->root);
}


static void
UdmInvertedIndexWordDictionaryReset(UDM_INVERTED_INDEX_WORD_DICTIONARY *D)
{
  UdmInvertedIndexWordDictionaryFree(D);
  UdmInvertedIndexWordDictionaryInit(D);
}


/*****************************************************************/

void
UdmInvertedIndexCachePartSort(UDM_INVERTED_INDEX_CACHE_PART *part)
{
  if (part->nitems)
    UdmSort(part->Item, part->nitems,
            sizeof(UDM_INVERTED_INDEX_CACHE_ITEM), (udm_qsort_cmp) bc2cmp);
}


void
UdmInvertedIndexCachePartInit(UDM_INVERTED_INDEX_CACHE_PART *part)
{
  bzero(part, sizeof(*part));
}


void
UdmInvertedIndexCachePartFree(UDM_INVERTED_INDEX_CACHE_PART *part)
{
  UdmFree(part->Item);
  bzero(part, sizeof(*part));
}


static inline udm_rc_t
UdmInvertedIndexCachePartRealloc(UDM_AGENT *A,
                                 UDM_INVERTED_INDEX_CACHE_PART *part)
{
  if (part->nitems >= part->mitems)
  {
    size_t mitems2= part->mitems + INVERTED_INDEX_CACHE_PART_SIZE;
    size_t nbytes= mitems2 * sizeof(UDM_INVERTED_INDEX_CACHE_ITEM);
    if (!(part->Item= (UDM_INVERTED_INDEX_CACHE_ITEM*) UdmRealloc(part->Item, nbytes)))
    {
      part->nitems= part->mitems= 0;
      UdmLog(A, UDM_LOG_ERROR,
             "UdmBlobCache2Realloc failed: %d bytes needed", (int) nbytes);
      return UDM_ERROR;
    }
    part->mitems= mitems2;
  }
  return UDM_OK;
}


static udm_rc_t
UdmInvertedIndexCacheItemEncode(UDM_INVERTED_INDEX_CACHE_ITEM *Item,
                                urlid_t url_id,
                                const UDM_WORD *W, size_t ncoords)
{
  size_t wordlen, datalen, i, nbytes;
  char *d, *de;
  Item->url_id= url_id;
  Item->secno= W->coord.secno;
  datalen= (wordlen= strlen(W->word) + 1) + ncoords * 4 + 4;
  if (!(Item->ptr= (char*) UdmMalloc(datalen)))
    return UDM_ERROR;

  /*TODO34-STAT: cache->nbytes+= datalen; ??? */

  de= Item->ptr + datalen;

  /* Store word together with '\0' */
  memcpy(Item->ptr, W->word, wordlen);

  /* Store ncoords */
  d= Item->ptr + wordlen;
  if (!(nbytes= udm_coord_put(ncoords,
                              (unsigned char*) d, (unsigned char*) de)))
    return UDM_ERROR;
  d+= nbytes;

  /* Store coords */
  for (i= 0; i < ncoords; i++)
  {
    nbytes= udm_coord_put(W[i].coord.pos,
                          (unsigned char*) d, (unsigned char*) de);
    if (!nbytes)
      continue;
    d+= nbytes;
  }
  Item->length= d - Item->ptr;
  return UDM_OK;
}


static udm_rc_t
UdmInvertedIndexCacheItemEncodeCoords(UDM_INVERTED_INDEX_CACHE_ITEM *Item, 
                                      UDM_CONSTWORD_HASH_DATA *data,
                                      const UDM_INVERTED_INDEX_WORD *W2,
                                      const UDM_INVERTED_INDEX_COORD *C,
                                      size_t ncoords,
                                      udm_pos_t section_size)
{
  size_t datalen, i, nbytes;
  char *d, *de;
  udm_pos_t prev;

  Item->url_id= data->url_id;
  Item->secno= C->coord.secno;
  datalen= W2->length + 1 + (ncoords + (section_size ? 1 : 0)) * 4 + 4;
  if (!(Item->ptr= (char*) UdmMemrootAlloc(&data->cache->coord_root, datalen)))
    return UDM_ERROR;

  de= Item->ptr + datalen;

  /* store word */
  memcpy(Item->ptr, W2->str, W2->length);
  Item->ptr[W2->length]= '\0';

  /* Store ncoords */
  d= Item->ptr + W2->length + 1;
  if (!(nbytes= udm_coord_put(ncoords + (section_size ? 1 : 0),
                              (unsigned char*) d, (unsigned char*) de)))
    return UDM_ERROR;
  d+= nbytes;

  for (prev=0, i= 0; i < ncoords; i++)
  {
    nbytes= udm_coord_put(C[i].coord.pos - prev,
                          (unsigned char*) d, (unsigned char*) de);
    prev= C[i].coord.pos;
    if (!nbytes)
      continue;
    d+= nbytes;
  }
  if (section_size)
  {
    nbytes= udm_coord_put(section_size - prev,
                          (unsigned char *) d, (unsigned char *) de);
    if (nbytes)
      d+= nbytes;
  }
  Item->length= d - Item->ptr;
  data->cache->coord_root.last_page_used_size-= (datalen - Item->length);
  /*
  printf("'%s' wordlen=%d datalen=%d datalen2=%d ncoords=%d\n",
         Item->data, wordlen, datalen, Item->datalen, ncoords);
  */
  return UDM_OK;
}



static udm_rc_t
UdmInvertedIndexCachePartAdd(UDM_AGENT *A,
                             UDM_INVERTED_INDEX_CACHE_PART *part,
                             urlid_t url_id, const UDM_WORD *W, size_t ncoords)
{
  if (UDM_OK != UdmInvertedIndexCachePartRealloc(A, part))
    return UDM_ERROR;

  if (UDM_OK != UdmInvertedIndexCacheItemEncode(&part->Item[part->nitems],
                                                url_id, W, ncoords))
    return UDM_ERROR;
  part->nitems++;
  return UDM_OK;
}


udm_rc_t
UdmInvertedIndexCacheAdd(UDM_AGENT *A,
                         UDM_CONSTWORD_HASH_DATA *data,
                         UDM_INVERTED_INDEX_CACHE *cache,
                         const UDM_WORD *W, size_t ncoords)
{
  /* TODO34: Turkish */
  size_t wlen= strlen(W->word);
  size_t crc= data->cnv.from->cset->crc32lcase(A->Conf->unidata,
                                               data->cnv.from,
                                               W->word, wlen,
                                               UDM_RECODE_HTML);
  size_t n= crc % cache->nitems;
  return UdmInvertedIndexCachePartAdd(A, &cache->Item[n], data->url_id, W, ncoords);
}


static udm_rc_t
UdmInvertedIndexCachePartAddCoords(UDM_AGENT *A,
                                   UDM_CONSTWORD_HASH_DATA *data,
                                   UDM_INVERTED_INDEX_CACHE_PART *part,
                                   UDM_INVERTED_INDEX_WORD *W2,
                                   const UDM_INVERTED_INDEX_COORD *C,
                                   size_t ncoords,
                                   udm_pos_t section_size)
{
  UDM_INVERTED_INDEX_CACHE_ITEM *Item;
  udm_rc_t rc;
  size_t ncoords_to_store;

  if (W2->length > A->Conf->WordParam.max_word_len ||
      W2->length < A->Conf->WordParam.min_word_len)
    return UDM_OK;

  /* TODO34-STAT */
  if (UDM_OK != UdmInvertedIndexCachePartRealloc(A, part))
    return UDM_ERROR;

  Item= &part->Item[part->nitems];

  ncoords_to_store= (data->cache->param.pair_limit &&
                     ncoords > data->cache->param.pair_limit) ? 1 : ncoords;
  rc= UdmInvertedIndexCacheItemEncodeCoords(Item, data, W2,
                                            C, ncoords_to_store,
                                            section_size);
  if (rc== UDM_OK)
    part->nitems++;
  return rc;
}


void
UdmInvertedIndexCacheInit(UDM_INVERTED_INDEX_CACHE *L, UDM_ENV *Env)
{
  size_t i;
  bzero(L, sizeof(UDM_INVERTED_INDEX_CACHE));
  L->param.unidata= Env->unidata;
  L->param.Env= Env;
  L->param.pair_limit= UdmVarListFindInt(&Env->Vars, "PairLimit", 0);
  L->param.cnvflags= UdmVarListFindBool(&Env->Vars, "StripAccents", UDM_FALSE) ?
                                        UDM_RECODE_HTML_OUT|UDM_RECODE_STRIP_ACCENTS :
                                        UDM_RECODE_HTML_OUT;
  L->param.save_section_size= UdmVarListFindInt(&Env->Vars, "SaveSectionSize", 1);
  for (L->param.aggregate_section_flags= 0, i= 0; i < Env->Sections.nvars; i++)
  {
    const UDM_VAR *Item= UdmVarListFindConstByIndex(&Env->Sections, i);
    L->param.aggregate_section_flags|= UdmVarFlags(Item);
  }
  UdmInvertedIndexWordDictionaryInit(&L->Words);
  UdmMemrootInit(&L->coord_root, 16*1024*1024);
}


void
UdmInvertedIndexCacheReset(UDM_INVERTED_INDEX_CACHE *L)
{
  size_t i;
  UdmInvertedIndexWordDictionaryReset(&L->Words);
  UdmMemrootReset(&L->coord_root);
  for (i= 0; i < L->nitems; i++)
    UdmInvertedIndexCachePartFree(&L->Item[i]);
}


void
UdmInvertedIndexCacheFree(UDM_INVERTED_INDEX_CACHE *L)
{
  UdmInvertedIndexCacheReset(L);
  UdmFree(L->Item);
  UdmInvertedIndexWordDictionaryFree(&L->Words);
  UdmMemrootFree(&L->coord_root);
}


static udm_rc_t
UdmInvertedIndexCacheRealloc(UDM_AGENT *A, UDM_INVERTED_INDEX_CACHE *L)
{
  if (L->nitems >= L->mitems)
  {
    size_t mitems2= L->mitems + INVERTED_INDEX_CACHE_PARTS;
    size_t nbytes= mitems2 * sizeof(UDM_INVERTED_INDEX_CACHE_PART);
    if (!(L->Item= (UDM_INVERTED_INDEX_CACHE_PART*) UdmRealloc(L->Item, nbytes)))
    {
      L->nitems= L->mitems= 0;
      UdmLog(A, UDM_LOG_ERROR,
             "UdmBlobCache2ListRealloc failed: %d bytes needed", (int) nbytes);
      return UDM_ERROR;
    }
    L->mitems= mitems2;
  }
  return UDM_OK;
}


udm_rc_t
UdmInvertedIndexCacheAddPart(UDM_AGENT *A, UDM_INVERTED_INDEX_CACHE *cache)
{
  if (UDM_OK != UdmInvertedIndexCacheRealloc(A, cache))
    return UDM_ERROR;
  UdmInvertedIndexCachePartInit(&cache->Item[cache->nitems]);
  cache->nitems++;
  return UDM_OK;
}


udm_rc_t
UdmInvertedIndexCacheAllocParts(UDM_AGENT *A,
                                UDM_INVERTED_INDEX_CACHE *cache, size_t n)
{
  size_t i;
  for (i= 0; i < n; i++)
  {
    if (UDM_OK != UdmInvertedIndexCacheAddPart(A, cache))
      return UDM_ERROR;
  }
  return UDM_OK;
}


static udm_rc_t
UdmInvertedIndexCacheAddCoords(UDM_AGENT *A,
                               UDM_CONSTWORD_HASH_DATA *data,
                               const UDM_INVERTED_INDEX_COORD *C, size_t ncoords,
                               udm_pos_t section_size)
{
  UDM_INVERTED_INDEX_CACHE *cache= data->cache;
  UDM_INVERTED_INDEX_WORD *W= &cache->Words.list.Item[C->word_id];
  size_t n= W->crc % cache->nitems;
  if (W->is_stopword)
    return UDM_OK;
  return UdmInvertedIndexCachePartAddCoords(A, data, &cache->Item[n],
                                             W, C, ncoords, section_size);
}


udm_rc_t
UdmInvertedIndexCoordListInit(UDM_INVERTED_INDEX_COORD_LIST *L, size_t mitems)
{
  L->nitems= 0;
  L->mitems= mitems;
  L->Item= (UDM_INVERTED_INDEX_COORD*) UdmMalloc(sizeof(UDM_INVERTED_INDEX_COORD) * mitems);
  return L->Item ? UDM_OK : UDM_ERROR;
}


static int
cmp_ii_coord(const UDM_INVERTED_INDEX_COORD *a,
             const UDM_INVERTED_INDEX_COORD *b)
{
  if (a->word_id != b->word_id)
    return a->word_id < b->word_id ? -1 : 1;
  if (a->coord.secno != b->coord.secno)
    return a->coord.secno < b->coord.secno ? -1 : 1;
  if (a->coord.pos != b->coord.pos)
    return a->coord.pos < b->coord.pos ? -1 : 1;
  return 0;
}


void
UdmInvertedIndexCoordListSort(UDM_INVERTED_INDEX_COORD_LIST *L)
{
  if (L->nitems)
    UdmSort(L->Item, L->nitems, sizeof(UDM_INVERTED_INDEX_COORD),
            (udm_qsort_cmp) cmp_ii_coord);
}


void
UdmInvertedIndexCoordListFree(UDM_INVERTED_INDEX_COORD_LIST *L)
{
  UdmFree(L->Item);
}


udm_rc_t
UdmConstWordListToInvertedIndexCoordList(UDM_INVERTED_INDEX_CACHE *cache,
                                         UDM_INVERTED_INDEX_COORD_LIST *CL,
                                         UDM_CONSTWORDLIST *CWL)
{
  udm_rc_t rc;
  size_t i;

  if (UDM_OK != (rc= UdmInvertedIndexCoordListInit(CL, CWL->nitems)))
    return rc;

  for (i= 0; i < CWL->nitems; i++)
  {
    UDM_CONSTWORD *Item= &CWL->Item[i];
    UDM_INVERTED_INDEX_HASH_WORD_REC rec;
    UDM_INVERTED_INDEX_HASH_WORD_KEY *key;
    UDM_INVERTED_INDEX_COORD *C= &CL->Item[CL->nitems++];
    rec.str.str= Item->str;
    rec.str.length= Item->length;
    rec.crc= Item->crc;
    if (!(key= (UDM_INVERTED_INDEX_HASH_WORD_KEY*) UdmHashPut(&cache->Words.hash, &rec)))
    {
      UdmInvertedIndexCoordListFree(CL);
      return UDM_ERROR;
    }
    C->word_id= key->word_id;
    C->coord= Item->coord;
  }
  return UDM_OK;
}


static int
clcmp_search(UDM_INVERTED_INDEX_COORD *c1, UDM_INVERTED_INDEX_COORD *c2)
{
  if (c1->word_id != c2->word_id)
    return c1->word_id < c2->word_id ? -1 : 1;
  if (c1->coord.secno != c2->coord.secno)
    return c1->coord.secno < c2->coord.secno ? -1 : 1;
  return 0;
}


udm_rc_t
UdmInvertedIndexCoordList2InvertedIndexCache(UDM_AGENT *A,
                                             UDM_INVERTED_INDEX_COORD_LIST *CL,
                                             UDM_CONSTWORDLIST *CWLWithPos,
                                             UDM_CONSTWORD_HASH_DATA *data,
                                             UDM_INVERTED_INDEX_CACHE *L)
{
  size_t i;
  UDM_INVERTED_INDEX_COORD *prev;
  if (!CL->nitems)
    return UDM_OK;
  for (prev= &CL->Item[0], i= 0; i <= CL->nitems; i++)
  {
    UDM_INVERTED_INDEX_COORD *C= &CL->Item[i];
    if (i == CL->nitems || clcmp_search(prev, C))
    {
      if (UDM_OK != UdmInvertedIndexCacheAddCoords(A, data, prev, C - prev,
                                                   data->cache->param.save_section_size ?
                                                   CWLWithPos->wordpos[prev->coord.secno] + 1 :
                                                   0))
        return UDM_ERROR;
      prev= C;
    }
  }
  return UDM_OK;
}


static size_t
UdmInvertedIndexWordListEstimateUsedMemory(const UDM_INVERTED_INDEX_WORD_LIST *List)
{
  return List->mitems * sizeof(List->Item[0]);
}


static size_t
UdmInvertedIndexCachePartsUsedMemory(const UDM_INVERTED_INDEX_CACHE *cache)
{
  size_t i, mitems; 
  for (i= mitems= 0; i < cache->nitems; i++)
  {
    mitems+= cache->Item[i].mitems;
  }
  return mitems * sizeof(UDM_INVERTED_INDEX_CACHE_ITEM);
}


#define M1 1048576e0
size_t
UdmInvertedIndexCacheEstimateUsedMemory(const UDM_INVERTED_INDEX_CACHE *cache)
{
  size_t total=
    UdmHashUsedMemorySize(&cache->Words.hash) +
    UdmInvertedIndexWordListEstimateUsedMemory(&cache->Words.list) +
    UdmMemrootAllocedMemory(&cache->Words.root) +
    UdmMemrootAllocedMemory(&cache->coord_root) +
    UdmInvertedIndexCachePartsUsedMemory(cache);
  /*
  fprintf(stderr, "rlist=%.1f alist=%.1f hash=%.1f parts=%.1f coords=%.1f total=%.1f rs=%.1f\n",
          (double) UdmMemrootAllocedMemory(&cache->Words.root) / M1,
          (double) UdmInvertedIndexWordListEstimateUsedMemory(&cache->Words.list) / M1,
          (double) UdmHashUsedMemorySize(&cache->Words.hash) / M1,
          (double) UdmInvertedIndexCachePartsUsedMemory(cache) / M1,
          (double) UdmMemrootAllocedMemory(&cache->coord_root) / M1,
          (double) total / M1,
          (double) UdmProcessCurrentResidentSize() / M1);
  */
  return total;
}
