/* Copyright (C) 2000-2015 Lavtech.com corp. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
*/

#include "udm_config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>

#ifdef HAVE_IO_H
#include <io.h>	/* for Win */
#endif
#ifdef HAVE_DIRECT_H
#include <direct.h> /* for Win */
#endif
#ifdef  HAVE_UNISTD_H
#include <unistd.h>
#endif
#ifdef HAVE_SYS_WAIT_H
#include <sys/wait.h>
#endif

#include <fcntl.h>
#include <sys/stat.h>
#include <errno.h>
#include <signal.h>

#include "udm_common.h"
#include "udm_utils.h"
#include "udm_parser.h"
#include "udm_xmalloc.h"
#include "udm_log.h"
#include "udm_vars.h"
#include "udm_wild.h"
#include "udm_signals.h"
#include "udm_http.h"

/* Bug 625: these macroses are not defined on some rare platforms */
#ifndef STDIN_FILENO
#define STDIN_FILENO 0
#endif
#ifndef STDOUT_FILENO
#define STDOUT_FILENO 1
#endif


UDM_API(udm_rc_t)
UdmParserAdd(UDM_PARSERLIST *List,UDM_PARSER *P)
{
  List->Parser= (UDM_PARSER*)(UdmRealloc(List->Parser,(List->nparsers+1)*sizeof(UDM_PARSER)));
  List->Parser[List->nparsers].from_mime= P->from_mime;
  List->Parser[List->nparsers].to_mime= P->to_mime;
  List->Parser[List->nparsers].cmd= P->cmd;
  List->Parser[List->nparsers].src= P->src ? P->src : NULL;
  List->nparsers++;
  return UDM_OK;
}


UDM_API(void)
UdmParserListFree(UDM_PARSERLIST *List)
{
  size_t i;
  for(i=0;i<List->nparsers;i++)
  {
    UDM_FREE(List->Parser[i].from_mime);
    UDM_FREE(List->Parser[i].to_mime);
    UDM_FREE(List->Parser[i].cmd);
    UDM_FREE(List->Parser[i].src);
  }
  UDM_FREE(List->Parser)
  List->nparsers=0;
}


UDM_API(UDM_PARSER *)
UdmParserFind(UDM_PARSERLIST *List,const char *mime_type)
{
  size_t i;
  UDM_ASSERT(mime_type != NULL);
  for(i=0;i<List->nparsers;i++)
  {
    if(!UdmWildCaseCmp(mime_type,List->Parser[i].from_mime))
      return &List->Parser[i];
  }
  return NULL;
}




#ifdef USE_PARSER

/* Parser1: from STDIN to STDOUT */

#ifdef WIN32
static char *parse1(UDM_AGENT * Agent, char *buf, size_t buflen, char *cmd, size_t maxlen)
{
  UdmLog(Agent,UDM_LOG_ERROR,"This type of parser isn't supported");
  return NULL;
}
#else

static void sighandler(int sign);
static void init_signals(void)
{
  /* Set up signals handler*/
  UdmSignal(SIGALRM, sighandler);
}


static void sighandler(int sign)
{
  switch(sign)
  {
   case SIGALRM:
      _exit(0);
      break;
    default:
      break;
  }
  init_signals();
}


static udm_rc_t
parse_stdin_to_stdout(UDM_AGENT * Agent, UDM_DOCUMENT *Doc, char *cmd)
{
  int wr[2];
  int rd[2];    
  pid_t pid;
  UDM_CONST_STR content;

  if (UdmHTTPBufContentToConstStr(&Doc->Buf, &content))
  {
    UdmLog(Agent, UDM_LOG_ERROR, "NULL content");
    return UDM_ERROR;
  }

  /* Create write and read pipes */
  if (pipe(wr) == -1)
  {
    UdmLog(Agent,UDM_LOG_ERROR,"Cannot make a pipe for writing");
    return UDM_ERROR;
  }
  if (pipe(rd) == -1)
  {
    UdmLog(Agent,UDM_LOG_ERROR,"Cannot make a pipe for reading");
    return UDM_ERROR;
  }    

  /* Fork a clild */
  if ((pid= fork()) == -1)
  {
    UdmLog(Agent,UDM_LOG_ERROR,"Cannot spawn a child");
    return UDM_ERROR;
  }

  if (pid > 0) /* Parent process */
  {
    /* Close the other pipe ends */
    close(wr[0]);
    close(wr[1]);
    close(rd[1]);   
    UdmHTTPBufTruncateContent(&Doc->Buf);
    UdmHTTPBufAppendFromFile(&Doc->Buf, rd[0]);
    close(rd[0]);
    wait(NULL);
  }
  else /* Child process */
  {
    /* Fork a clild */
    if ((pid= fork()) == -1)
    {
      UdmLog(Agent,UDM_LOG_ERROR,"Cannot spawn a child");
      return UDM_ERROR;
    }

    if (pid > 0) /* Parent process */
    {    
      /* Close other pipe ends */
      close(wr[0]);
      close(rd[0]);
      close(rd[1]);

      /* Send string to be parsed */
      write(wr[1], content.str, content.length);
      close(wr[1]);

      exit(0);
    }
    else /* Child process */
    {
      /* Close other pipe ends */
      close (wr[1]);
      close (rd[0]);

      /* Connect pipe to stdout */
      dup2(rd[1],STDOUT_FILENO);

      /* Connect pipe to stdin */
      dup2(wr[0],STDIN_FILENO);

      alarm((unsigned int) UdmVarListFindInt(&Agent->Conf->Vars, "ParserTimeOut", 300) );
      init_signals();

      system(cmd);
      exit(0);
    }
  }
  return UDM_OK;
}
#endif


/* Parser2: from FILE to STDOUT */
static udm_rc_t
parse_file_to_stdout(UDM_AGENT * Agent, UDM_DOCUMENT *Doc, char *cmd)
{
  FILE *f;
#ifdef WIN32
  f= _popen(cmd,"rb");
#else
  f= popen(cmd,"r");
#endif
  if (f)
  {
    int fd= fileno(f);
    UdmHTTPBufTruncateContent(&Doc->Buf);
    UdmHTTPBufAppendFromFile(&Doc->Buf, fd);
#ifdef WIN32
    _pclose(f);
#else
    pclose(f);
#endif
    return UDM_OK;
  }
  UdmLog(Agent,UDM_LOG_ERROR,"Error in popen() (parse2)");
  return UDM_ERROR;
}


/* Parser3: from FILE to FILE */
static udm_rc_t
parse_file_to_file(UDM_AGENT * Agent, UDM_DOCUMENT *Doc, char *cmd, char *to_file)
{
  int fd;

  UdmHTTPBufTruncateContent(&Doc->Buf);
  system(cmd);
  if ((fd= open(to_file,O_RDONLY|UDM_BINARY)))
  {
    UdmHTTPBufAppendFromFile(&Doc->Buf, fd);
    close(fd);
  }
  else
  {
    UdmLog(Agent,UDM_LOG_ERROR,"Can't open output file");
    return UDM_ERROR;
  }
  return UDM_OK;
}


/* Parser4: from STDIN to FILE */
static udm_rc_t
parse_stdin_to_file(UDM_AGENT * Agent, UDM_DOCUMENT *Doc, char *cmd, char *to_file)
{
  FILE *f;
  UDM_CONST_STR content;

  if (UdmHTTPBufContentToConstStr(&Doc->Buf, &content))
  {
    UdmLog(Agent, UDM_LOG_ERROR, "NULL content");
    return UDM_ERROR;
  }

#ifdef WIN32
  f= _popen(cmd,"wb");
#else
  f= popen(cmd,"w");
#endif
  if (f)
  {
    int fd= fileno(f);
    write(fd, content.str, content.length);
#ifdef WIN32
    _pclose(f);
#else
    pclose(f);
#endif
    return parse_file_to_file(Agent, Doc, cmd, to_file);
  }
  UdmLog(Agent,UDM_LOG_ERROR,"Error in popen() (parse_stdin_to_file)");
  return UDM_ERROR;
}


static udm_rc_t
parser_by_type(UDM_AGENT *Agent, UDM_DOCUMENT *Doc,
               char *cmd, int parser_type, char *out_file_name)
{
  switch(parser_type)
  {
#ifndef WIN32
    case 1: return parse_stdin_to_stdout(Agent, Doc, cmd); break;
#endif
    case 2: return parse_file_to_stdout(Agent, Doc, cmd); break;
    case 3: return parse_file_to_file(Agent, Doc, cmd, out_file_name); break;
    case 4: return parse_stdin_to_file(Agent, Doc, cmd, out_file_name); break;
  }
  return UDM_ERROR; /* Bad parser type, should never happen */
}


static size_t
build_tmp_file_name(UDM_AGENT *Agent,
                    char *str, size_t strsize,
                    const char *suffix)
{
  /* Build temp file names and command line */
#ifdef WIN32
  /*
   Windows generates something strange in tmpnam()
   So create the name ourself. Because some parsers
   might be compiled under DOS, the name should be
   DOS compatible.
  */
  return udm_snprintf(str, strsize, ".\\prstmp%d.in",Agent->handle);
#else
  return udm_snprintf(str, strsize, "%s/ind.%d.%d.in",
                      UdmVarListFindStr(&Agent->Conf->Vars, "TmpDir", UDM_TMP_DIR),
                      Agent->handle, getpid());
#endif
}


static udm_rc_t
parse_file(UDM_AGENT *Agent, UDM_PARSER *parser, UDM_DOCUMENT *Doc)
{
  char cmd[1024]="";
  char *arg1pos= strstr(parser->cmd,"$1");
  char *arg2pos= strstr(parser->cmd,"$2");
  int parser_type;
  char fn0[1024]= "";
  char fn1[1024]= "";
  char *fnames[2];
  const char *url= UdmVarListFindStr(&Doc->Sections,"URL","");
  udm_rc_t rc;

  fnames[0]= fn0;
  fnames[1]= fn1;
  build_tmp_file_name(Agent, fn0, sizeof(fn0), ".in");
  build_tmp_file_name(Agent, fn1, sizeof(fn1), ".out");

  if (strstr(parser->cmd, "${"))
  {
    UDM_DSTR dstr;
    UdmDSTRInit(&dstr, 1024);
    UdmDSTRParse(&dstr, parser->cmd, &Doc->Sections);
    UdmBuildParamStr(cmd, sizeof(cmd), UdmDSTRPtr(&dstr), fnames, 2);
    UdmDSTRFree(&dstr);
  }
  else
  {
    UdmBuildParamStr(cmd, sizeof(cmd), parser->cmd, fnames, 2);
  }

  if (arg1pos)
  {
    int fd;
    UDM_CONST_STR content;

    if (UdmHTTPBufContentToConstStr(&Doc->Buf, &content))
    {
      UdmLog(Agent, UDM_LOG_ERROR, "NULL content");
      return UDM_ERROR;
    }

    /* Create temporary file */
#ifdef WIN32
    umask((int)022);
#else
    umask((mode_t)022);
#endif
    fd= open(fnames[0],O_RDWR|O_CREAT|O_TRUNC|UDM_BINARY,UDM_IWRITE);
    if (fd < 0)
    {
      UdmLog(Agent,UDM_LOG_ERROR,"Can't open output file '%s'", fnames[0]);
      return UDM_ERROR;
    }
    /* Write to the temporary file */
    write(fd, content.str, content.length);
    close(fd);
  }

  if (arg1pos && arg2pos)
    parser_type= 3;
  else
  {
    if (arg1pos)
      parser_type= 2;
    else if (arg2pos)
      parser_type= 4; 
    else
      parser_type= 1;
  }

  UdmLog(Agent,UDM_LOG_EXTRA,"Starting external parser: '%s'",cmd);
  UdmSetEnv("UDM_URL",url);
  rc= parser_by_type(Agent, Doc, cmd, parser_type, fnames[1]);
  UdmUnsetEnv("UDM_URL");

  /* Remove temporary file */
  if (arg1pos) unlink(fnames[0]);
  if (arg2pos) unlink(fnames[1]);

  return rc;
}


udm_rc_t
UdmParserExec(UDM_AGENT * Agent,UDM_PARSER *P,UDM_DOCUMENT *Doc)
{
  if (P->src) /* Parse user defined variable */
  {
    UDM_DSTR dstr;
    UdmDSTRInit(&dstr, 1024);
    UdmDSTRParse(&dstr, P->src, &Doc->Sections);
    UdmHTTPBufPutContent(&Doc->Buf, UdmDSTRPtr(&dstr), UdmDSTRLength(&dstr));
    UdmDSTRFree(&dstr);
  }
  return parse_file(Agent, P, Doc);
}

#endif
