/* Copyright (C) 2000-2015 Lavtech.com corp. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
*/

#include "udm_config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>

#include "udm_common.h"
#include "udm_utils.h"
#include "udm_uniconv.h"
#include "udm_vars.h"
#include "udm_textlist.h"
#include "udm_parsexml.h"
#include "udm_hrefs.h"
#include "udm_sgml.h"
#include "udm_http.h"
#include "udm_lex.h"

typedef struct
{
  UDM_AGENT *Indexer;
  UDM_DOCUMENT *Doc;
  UDM_HREF Href;
  int body_sec;
  UDM_CONST_STR XMLDefaultSection;
  char *section_name;  /* TODO34: change to UDM_CONST_STR */
  char *secpath;       /* TODO34: change to UDM_CONST_STR */
  udm_bool_t is_sitemap;
} XML_PARSER_DATA;


/************ MY_XML **************/

/* TODO34: change XML path delimiter from '.' to '/' */
#define UDM_XML_PATH_DELIMITER  '/'
#define UDM_XML_ATTR_DELIMITER  '@'

static inline void
UdmConstTokenSkipLeadingSpaces(UDM_CONST_TOKEN *a)
{
  for ( ; a->str < a->end ; a->str++)
  {
    if (!UdmLexScannerIsSpace(a->str[0]))
      break;
  }
}


static inline void
UdmConstTokenSkipTrailingSpaces(UDM_CONST_TOKEN *a)
{
  for ( ; a->str < a->end ; a->end--)
  {
    if (!UdmLexScannerIsSpace(a->end[-1]))
      break;
  }
}


static inline void
UdmXMLParserSkipSpaces(UDM_XML_PARSER *p)
{
  UdmLexScannerSkipSpaces(&p->scanner);
}
static inline const char *
UdmXMLParserCur(UDM_XML_PARSER *p)
{
  return UdmLexScannerCur(&p->scanner);
}
static inline const char *
UdmXMLParserStr(UDM_XML_PARSER *p)
{
  return UdmLexScannerStr(&p->scanner);
}
static inline const char *
UdmXMLParserEnd(UDM_XML_PARSER *p)
{
  return UdmLexScannerEnd(&p->scanner);
}
static inline udm_bool_t
UdmXMLParserEOF(UDM_XML_PARSER *p)
{
  return (udm_bool_t) UdmLexScannerEOF(&p->scanner);
}
static inline void
UdmXMLParserShift(UDM_XML_PARSER *p)
{
  UdmLexScannerShift(&p->scanner);
}
static inline void
UdmXMLParserShiftN(UDM_XML_PARSER *p, size_t n)
{
  UdmLexScannerShiftN(&p->scanner, n);
}
static inline void
UdmXMLParserUnexpectedError(UDM_XML_PARSER *p,
                            const char *prefix, const char *wanted)
{
  udm_snprintf(p->errstr, sizeof(p->errstr), "%s: %s unexpected (%s wanted)",
               prefix, UdmLex2str(UdmXMLParserLastTokenType(p)), wanted);
}
static inline void
UdmXMLParserUnexpectedErrorNear(UDM_XML_PARSER *p,
                                const char *prefix, const char *wanted)
{
  size_t length= UdmXMLParserEnd(p) - UdmXMLParserCur(p);
  if (length > 32)
    length= 32;
  udm_snprintf(p->errstr, sizeof(p->errstr),
               "%s: %s unexpected (%s wanted) near '%.*s'",
               prefix, UdmLex2str(UdmXMLParserLastTokenType(p)), wanted,
               (int) length, UdmXMLParserCur(p));
}

static void UdmXMLNormText(UDM_CONST_TOKEN *a)
{
  UdmConstTokenSkipLeadingSpaces(a);
  UdmConstTokenSkipTrailingSpaces(a);
}


typedef enum
{
  UDM_XML_LEX_COMMENT,
  UDM_XML_LEX_CDATA,
  UDM_XML_LEX_PI,
  UDM_XML_LEX_XMLDECL,
  UDM_XML_LEX_OTHER,
  UDM_XML_LEX_EOF
} udm_xml_lex_t;


static udm_xml_lex_t
UdmXMLParserCheckSpecialSection(UDM_XML_PARSER *p, UDM_CONST_TOKEN *a)
{
  UdmXMLParserSkipSpaces(p);
  if (UdmXMLParserEOF(p))
  {
    a->str= UdmXMLParserEnd(p);
    a->end= UdmXMLParserEnd(p);
    return UDM_XML_LEX_EOF;
  }

  a->str= UdmXMLParserCur(p);
  a->end= UdmXMLParserCur(p);
  if (UdmXMLParserCur(p)[0] != '<')
    return UDM_XML_LEX_OTHER;

  if (!memcmp(UdmXMLParserCur(p), "<!--", 4))
  {
    for ( ; !UdmXMLParserEOF(p) ; UdmXMLParserShift(p))
    {
      if (!memcmp(UdmXMLParserCur(p), "-->", 3))
        break;
    }
    if (!memcmp(UdmXMLParserCur(p), "-->", 3))
      UdmXMLParserShiftN(p, 3);
    a->end= UdmXMLParserCur(p);
    return UDM_XML_LEX_COMMENT;
  }
  else if (!memcmp(UdmXMLParserCur(p), "<![CDATA[",9))
  {
    UdmXMLParserShiftN(p, 9);
    for ( ; !UdmXMLParserEOF(p) ; UdmXMLParserShift(p))
    {
      if (UdmXMLParserCur(p)[0] == ']' &&
          UdmXMLParserCur(p)[1] == ']' &&
          UdmXMLParserCur(p)[2] == '>')
      {
        UdmXMLParserShiftN(p, 3);
        a->end= UdmXMLParserCur(p);
        break;
      }
    }
    return UDM_XML_LEX_CDATA;
  }
  else if (UdmXMLParserCur(p)[1] == '?')
  {
    if (!strncasecmp(UdmXMLParserCur(p) + 2, UDM_CSTR_WITH_LEN("xml")))
      return UDM_XML_LEX_XMLDECL; /* Declaration */
    for ( ; !UdmXMLParserEOF(p); UdmXMLParserShift(p))
    {
      if (UdmXMLParserCur(p)[0] == '?' &&
          UdmXMLParserCur(p)[1] == '>')
      {
        UdmXMLParserShiftN(p, 2);
        a->end= UdmXMLParserCur(p);
        return UDM_XML_LEX_PI;
      }
    }
  }
  return UDM_XML_LEX_OTHER;
}


static void
UdmXMLScanLexToken(UDM_XML_PARSER *p, UDM_LEX_TOKEN *a)
{
  UdmXMLParserSkipSpaces(p);
  a->token.str= UdmXMLParserCur(p);
  if (!UdmLexScannerScanXMLIdentifier(&p->scanner, a))
  {
  }
  else if (!UdmLexScannerScanPunctuation(&p->scanner, a))
  {
  }
  else if (!UdmLexScannerScanString(&p->scanner, a) ||
           !UdmLexScannerScanChar(&p->scanner, a))
  {
    if (!(p->flags & UDM_XML_SKIP_TEXT_NORMALIZATION))
      UdmXMLNormText(&a->token);
  }
  else
  {
    a->type= UDM_LEX_UNKNOWN;
  }

#if 0
  fprintf(stderr, "LEX=%s='%.*s'\n",
          UdmLex2str(a->type), (int) (a->token.end-a->token.str), a->token.str);
#endif
}


static void
UdmXMLParserScan(UDM_XML_PARSER *p)
{
  UdmXMLScanLexToken(p, UdmXMLParserLastToken(p));
}


static udm_rc_t
UdmXMLValue(UDM_XML_PARSER *st, const char *str, size_t len)
{
#ifdef DEBUG_XML
  fprintf(stderr, "UdmXMLValue: %.*s\n", (int) len, str);
#endif
  return st->handler.value_action ?
         (st->handler.value_action)(st, str, len) : UDM_OK;
}


static udm_rc_t
UdmXMLEnterTagOrAttr(UDM_XML_PARSER *st,
                     const char *str, size_t len, int delimiter)
{
#ifdef DEBUG_XML
  fprintf(stderr, "UdmXMLEnter: %.*s\n", (int) len, str);
#endif
  if ((st->attrend - st->attr + len + 1) > sizeof(st->attr))
  {
    sprintf(st->errstr, "To deep XML");
    return UDM_ERROR;
  }
  st->attrend[0]= delimiter;
  st->attrend++;
  memcpy(st->attrend, str, len);
  st->attrend += len;
  st->attrend[0]= '\0';
  return st->handler.enter_action ?
         (st->handler.enter_action)(st, st->attr, st->attrend - st->attr) :
         UDM_OK;
}


static udm_rc_t
UdmXMLEnterTag(UDM_XML_PARSER *p, const char *str, size_t length)
{
  return UdmXMLEnterTagOrAttr(p, str, length, UDM_XML_PATH_DELIMITER);
}


static udm_rc_t
UdmXMLEnterAttr(UDM_XML_PARSER *p, const char *str, size_t length)
{
  return UdmXMLEnterTagOrAttr(p, str, length, UDM_XML_ATTR_DELIMITER);
}


static udm_rc_t
UdmXMLLeave(UDM_XML_PARSER *p, const char *str, size_t slen)
{
  char *e;
  size_t glen, delimiter= 0;
  udm_rc_t rc;

#ifdef DEBUG_XML
  fprintf(stderr, "UdmXMLLeave: %.*s\n", (int) slen, str);
#endif
  /* Find previous UDM_XML_PATH_DELIMITER or beginning */
  for(e= p->attrend; e > p->attr ; e--)
  {
    if (e[0] == UDM_XML_PATH_DELIMITER || e[0] == UDM_XML_ATTR_DELIMITER)
    {
      delimiter= 1;
      break;
    }
  }
  glen= p->attrend - e - delimiter;

  if (str && (slen != glen))
  {
    udm_snprintf(p->errstr, sizeof(p->errstr),
                 "'</%.*s>' unexpected ('</%.*s>' wanted)",
                 (int) slen, str, (int) glen, e + 1);
    return UDM_ERROR;;
  }

  rc= p->handler.leave_action ?
      (p->handler.leave_action)(p, p->attr, p->attrend - p->attr) : UDM_OK;

  *e= '\0';
  p->attrend= e;
  return(rc);
}



static inline udm_rc_t
UdmXMLValueUsingLastToken(UDM_XML_PARSER *st)
{
  return UdmXMLValue(st, UdmConstTokenStr(UdmXMLParserLastStrToken(st)),
                         UdmConstTokenLength(UdmXMLParserLastStrToken(st)));
}


static inline udm_rc_t
UdmXMLEnterTagUsingLastToken(UDM_XML_PARSER *st)
{
  return UdmXMLEnterTag(st, UdmConstTokenStr(UdmXMLParserLastStrToken(st)),
                            UdmConstTokenLength(UdmXMLParserLastStrToken(st)));
}


static inline udm_rc_t
UdmXMLEnterAttrUsingLastToken(UDM_XML_PARSER *st)
{
  return UdmXMLEnterAttr(st, UdmConstTokenStr(UdmXMLParserLastStrToken(st)),
                             UdmConstTokenLength(UdmXMLParserLastStrToken(st)));
}


static inline udm_rc_t
UdmXMLLeaveUsingLastToken(UDM_XML_PARSER *st)
{
  return UdmXMLLeave(st, UdmConstTokenStr(UdmXMLParserLastStrToken(st)),
                         UdmConstTokenLength(UdmXMLParserLastStrToken(st)));
}



static void
UdmXMLParserScanText(UDM_XML_PARSER *p)
{
  UDM_CONST_TOKEN a;
  a.str= UdmXMLParserCur(p);
  for ( ; !UdmXMLParserEOF(p) ; UdmXMLParserShift(p))
  {
    if (UdmXMLParserCur(p)[0] == '<')
      break;
  }
  a.end= UdmXMLParserCur(p);

  if (!(p->flags & UDM_XML_SKIP_TEXT_NORMALIZATION))
    UdmXMLNormText(&a);
  if (!UdmConstTokenIsEmpty(&a))
    UdmXMLValue(p, UdmConstTokenStr(&a), UdmConstTokenLength(&a));
}


/*
  Scan attributes. Assumes that the next token is already scanned in "lex".
*/
static udm_rc_t
UdmXMLParserScanAttributes(UDM_XML_PARSER *p)
{
  for ( ; UdmXMLParserLastTokenType(p) == UDM_LEX_IDENT ||
          UdmXMLParserLastTokenType(p) == UDM_LEX_STRING ||
          UdmXMLParserLastTokenType(p) == UDM_LEX_CHAR_LITERAL;
          UdmXMLParserScan(p))
  {
    UDM_CONST_TOKEN t= UdmXMLParserLastToken(p)->token;
    UdmXMLParserScan(p);
    if (UdmXMLParserLastTokenType(p) == UDM_LEX_EQ)
    {
      /* Scan attr=value */
      UdmXMLParserScan(p);
      if (UdmXMLParserLastTokenType(p) == UDM_LEX_IDENT ||
          UdmXMLParserLastTokenType(p) == UDM_LEX_STRING ||
          UdmXMLParserLastTokenType(p) == UDM_LEX_CHAR_LITERAL)
      {
        if ((UDM_OK != UdmXMLEnterAttr(p, UdmConstTokenStr(&t), UdmConstTokenLength(&t))) ||
            (UDM_OK != UdmXMLValueUsingLastToken(p)) ||
            (UDM_OK != UdmXMLLeave(p, UdmConstTokenStr(&t), UdmConstTokenLength(&t))))
          return UDM_ERROR;
      }
      else
      {
        UdmXMLParserUnexpectedError(p, "ScanAttributes", "ident or string");
        return UDM_ERROR;
      }
    }
    else if (UdmXMLParserLastTokenType(p) == UDM_LEX_IDENT)
    {
      if (UDM_OK != UdmXMLEnterAttrUsingLastToken(p) ||
          UDM_OK != UdmXMLLeaveUsingLastToken(p))
        return UDM_ERROR;
    }
    else if (UdmXMLParserLastTokenType(p) == UDM_LEX_STRING)
    {
      /* Do nothing */
    }
    else
      break;
  }
  return UDM_OK;
}


/*
  Skip the intSubset list of a DOCTYPE declaration.
  Assumes that the next token is already scanned to "lex".

  Example:

<!DOCTYPE book PUBLIC "-//OASIS//DTD DocBook XML V4.2//EN" [
  <!ENTITY install SYSTEM "install.xml">
  <!ENTITY general SYSTEM "general.xml">
  ]>
<book>...</book>

  This functions scans everything between [ and ],
  including the brackets.
*/
static udm_rc_t
UdmXMLParserScanIntSubset(UDM_XML_PARSER *p)
{
  if (UdmXMLParserLastTokenType(p) != UDM_LEX_LB)
    return UDM_OK;

  for (UdmXMLParserScan(p); ; UdmXMLParserScan(p))
  {
    switch (UdmXMLParserLastToken(p)->type)
    {
    case UDM_LEX_UNKNOWN:
    case UDM_LEX_EOF:
      UdmXMLParserUnexpectedErrorNear(p, "ScanIntSubset", "']'");
      return UDM_ERROR;

    case UDM_LEX_RB:
      UdmXMLParserScan(p);
      return UDM_OK;

    case UDM_LEX_EXCLAM:
    case UDM_LEX_HASH:
    case UDM_LEX_DOLLAR:
    case UDM_LEX_PERCENT:
    case UDM_LEX_AMPERSAND:
    case UDM_LEX_LP:
    case UDM_LEX_RP:
    case UDM_LEX_ASTERISK:
    case UDM_LEX_PLUS:
    case UDM_LEX_COMMA:
    case UDM_LEX_MINUS:
    case UDM_LEX_DOT:
    case UDM_LEX_SLASH:
    case UDM_LEX_COLON:
    case UDM_LEX_SEMICOLON:
    case UDM_LEX_LT:
    case UDM_LEX_LE:
    case UDM_LEX_EQ:
    case UDM_LEX_GT:
    case UDM_LEX_GE:
    case UDM_LEX_EQ_EQ:
    case UDM_LEX_NOT_EQ:
    case UDM_LEX_QUESTION:
    case UDM_LEX_AT:
    case UDM_LEX_LB:
    case UDM_LEX_CARET:
    case UDM_LEX_LCB:
    case UDM_LEX_VBAR:
    case UDM_LEX_BOOL_OR:
    case UDM_LEX_BOOL_AND:
    case UDM_LEX_RCB:
    case UDM_LEX_TILDE:
    case UDM_LEX_TEXT:

    case UDM_LEX_ESCAPED_STRING:
    case UDM_LEX_STRING:
    case UDM_LEX_CHAR_LITERAL:
    case UDM_LEX_IDENT:   
    case UDM_LEX_UINT:
    case UDM_LEX_IF:
    case UDM_LEX_ELSE:
    case UDM_LEX_WHILE:
    case UDM_LEX_DO:
    case UDM_LEX_FOR:
    case UDM_LEX_BREAK:
    case UDM_LEX_CONTINUE:
    case UDM_LEX_SIZEOF:
    case UDM_LEX_COUT:
    case UDM_LEX_AUTO:
    case UDM_LEX_CASE:
    case UDM_LEX_CONST:
    case UDM_LEX_DEFAULT:
    case UDM_LEX_ENUM:
    case UDM_LEX_EXTERN:
    case UDM_LEX_GOTO:
    case UDM_LEX_REGISTER:
    case UDM_LEX_RETURN:
    case UDM_LEX_STATIC:
    case UDM_LEX_STRUCT:
    case UDM_LEX_SWITCH:
    case UDM_LEX_TYPEDEF:
    case UDM_LEX_UNION:
    case UDM_LEX_VOLATILE:

    case UDM_LEX_CHAR:
    case UDM_LEX_DOUBLE:
    case UDM_LEX_FLOAT:
    case UDM_LEX_INT:
    case UDM_LEX_LONG:
    case UDM_LEX_SHORT:
    case UDM_LEX_SIGNED:
    case UDM_LEX_UNSIGNED:
      continue;

    case UDM_LEX_OPERATOR:
      continue;

    case UDM_LEX_COMMENT:
    case UDM_LEX_INC:
    case UDM_LEX_DEC:
    case UDM_LEX_MUL_EQ:
    case UDM_LEX_DIV_EQ:
    case UDM_LEX_REM_EQ:
    case UDM_LEX_INC_EQ:
    case UDM_LEX_DEC_EQ:
    case UDM_LEX_AND_EQ:
    case UDM_LEX_OR_EQ:
    case UDM_LEX_XOR_EQ:
    case UDM_LEX_LSHIFT_EQ:
    case UDM_LEX_LSHIFT:
    case UDM_LEX_RSHIFT:
    case UDM_LEX_RSHIFT_EQ:
      UDM_ASSERT(0);
      continue;
    }
  }
  return UDM_OK;
}


/*
  Scan the ETag body, everything between '<' and '>'.
  ETag ::= '</' Name S? '>'
*/
static udm_rc_t
UdmXMLParserScanETagBody(UDM_XML_PARSER *p)
{
  if (UdmXMLParserLastTokenType(p) !=  UDM_LEX_SLASH)
    return UDM_ERROR;
  UdmXMLParserScan(p);
  if (UdmXMLParserLastTokenType(p) != UDM_LEX_IDENT)
  {
    UdmXMLParserUnexpectedError(p, "ETagBody", "ident");
    return UDM_ERROR;
  }
  if (UDM_OK != UdmXMLLeaveUsingLastToken(p))
    return UDM_ERROR;
  UdmXMLParserScan(p);
  return UDM_OK;
}


/*
  Name (S Attribute)* S? 
*/
static udm_rc_t
UdmXMLParserScanTagNameAndAttributes(UDM_XML_PARSER *p)
{
  if (UdmXMLParserLastTokenType(p) != UDM_LEX_IDENT)
    return UDM_ERROR;
  if (UDM_OK != UdmXMLEnterTagUsingLastToken(p))
      return UDM_ERROR;
  UdmXMLParserScan(p);
  if (UDM_OK != UdmXMLParserScanAttributes(p))
    return UDM_ERROR;
  return UDM_OK;
}


/*
  Scan the body of a processing instruction (PI) or XMLDecl, i.e.
  everything between '<' and '>'.

  XMLDecl  ::= '<?xml' VersionInfo EncodingDecl? SDDecl? S? '?>'
  PI       ::= '<?' PITarget (S (Char* - (Char* '?>' Char*)))? '?>'
  PITarget ::= Name - (('X' | 'x') ('M' | 'm') ('L' | 'l'))
*/
static udm_rc_t
UdmXMLParserScanXMLDecl(UDM_XML_PARSER *p)
{
  if (UdmXMLParserLastTokenType(p) != UDM_LEX_QUESTION)
    return UDM_ERROR;

  UdmXMLParserScan(p);
  p->xmldecl= UDM_TRUE;

  if (UDM_OK != UdmXMLParserScanTagNameAndAttributes(p))
    return UDM_ERROR;

  if (UdmXMLParserLastTokenType(p) != UDM_LEX_QUESTION)
  {
    UdmXMLParserUnexpectedError(p, "ScanPIBody", "'?'");
    return UDM_ERROR;
  }
  if (UDM_OK != UdmXMLLeave(p, NULL, 0))
    return UDM_ERROR;
  p->xmldecl= UDM_FALSE;
  UdmXMLParserScan(p);
  return UDM_OK;
}


/*
  Scan a declaration body
*/
static udm_rc_t
UdmXMLParserScanDeclarationBody(UDM_XML_PARSER *p)
{
  if (UdmXMLParserLastTokenType(p) != UDM_LEX_EXCLAM)
    return UDM_ERROR;
  UdmXMLParserScan(p);
  if (UDM_OK != UdmXMLParserScanTagNameAndAttributes(p))
    return UDM_ERROR;
  if (UDM_OK != UdmXMLParserScanIntSubset(p))
    return UDM_ERROR;
  if (UDM_OK != UdmXMLLeave(p, NULL, 0))
    return UDM_ERROR;
  return UDM_OK;
}


/*
  Scan the body of STag or EmptyElemTag
  i.e. everything between '<' and '>'
  
  STag         ::=  '<' Name (S Attribute)* S? '>'
  EmptyElemTag ::=  '<' Name (S Attribute)* S? '/>'
*/
static udm_rc_t
UdmXMLParserScanSTagBody(UDM_XML_PARSER *p)
{
  if (UDM_OK != UdmXMLParserScanTagNameAndAttributes(p))
    return UDM_ERROR;

  if (UdmXMLParserLastTokenType(p) == UDM_LEX_SLASH)
  {
    if (UDM_OK != UdmXMLLeave(p, NULL, 0))
      return UDM_ERROR;
    UdmXMLParserScan(p);
  }
  return UDM_OK;
}


static udm_rc_t
UdmXMLParserScanTag(UDM_XML_PARSER *p)
{
  if (UdmXMLParserLastTokenType(p) != UDM_LEX_LT)
  {
    UdmXMLParserUnexpectedError(p, "ScanTag", "'<'");
    return UDM_ERROR;
  }
  UdmXMLParserScan(p);

  if (UDM_OK != UdmXMLParserScanETagBody(p) &&
      UDM_OK != UdmXMLParserScanSTagBody(p) &&
      UDM_OK != UdmXMLParserScanXMLDecl(p) &&
      UDM_OK != UdmXMLParserScanDeclarationBody(p))
  {
    if (!p->errstr[0])
      UdmXMLParserUnexpectedError(p, "ScanTag", "<Name> or '/' or '!' or '?'");
    return UDM_ERROR;
  }

  if (UdmXMLParserLastTokenType(p) != UDM_LEX_GT)
  {
    UdmXMLParserUnexpectedError(p, "ScanTag", "'>'");
    return UDM_ERROR;
  }
  return UDM_OK;
}


udm_rc_t
UdmXMLParserExec(UDM_XML_PARSER *p, const char *str, size_t len)
{
  p->attrend= p->attr;
  UdmLexScannerInit(&p->scanner, str, len);

  while (!UdmXMLParserEOF(p))
  {
    if (UdmXMLParserCur(p)[0] == '<')
    {
      switch (UdmXMLParserCheckSpecialSection(p, UdmXMLParserLastStrToken(p)))
      {
      case UDM_XML_LEX_EOF:
        UdmXMLParserLastToken(p)->type= UDM_LEX_EOF;
        UdmXMLParserUnexpectedError(p, "XMLParserExec", "tag body");
        return UDM_ERROR;

      case UDM_XML_LEX_COMMENT:
      case UDM_XML_LEX_PI:
        continue;

      case UDM_XML_LEX_CDATA:
        {
          UDM_CONST_TOKEN tmp= UdmXMLParserLastStrToken(p)[0];
          tmp.str+= 9;
          tmp.end-= 3;
          UdmXMLValue(p, tmp.str, tmp.end - tmp.str);
        }
        continue;
      case UDM_XML_LEX_XMLDECL:
      case UDM_XML_LEX_OTHER:
        break;
      }

      UdmXMLParserScan(p);
      if (UDM_OK != UdmXMLParserScanTag(p))
        return UDM_ERROR;
    }
    else
    {
      UdmXMLParserScanText(p);
    }
  }
  return UDM_OK;
}


void UdmXMLParserCreate (UDM_XML_PARSER *p)
{
  bzero((void*)p, sizeof(p[0]));
}


void UdmXMLParserFree (UDM_XML_PARSER *p)
{
}


void UdmXMLSetValueHandler (UDM_XML_PARSER *p, udm_rc_t (*action)(UDM_XML_PARSER *p, const char *s, size_t l))
{
  p->handler.value_action= action;
}


void UdmXMLSetEnterHandler (UDM_XML_PARSER *p, udm_rc_t (*action)(UDM_XML_PARSER *p, const char *s, size_t l))
{
  p->handler.enter_action= action;
}


void UdmXMLSetLeaveHandler (UDM_XML_PARSER *p, udm_rc_t (*action)(UDM_XML_PARSER *p, const char *s, size_t l))
{
  p->handler.leave_action= action;
}


void UdmXMLSetHandler(UDM_XML_PARSER *p, const UDM_XML_HANDLER *handler)
{
  p->handler= *handler;
}


void UdmXMLSetUserData (UDM_XML_PARSER *p, void *user_data)
{
  p->user_data= user_data;
}


const char *UdmXMLErrorString (UDM_XML_PARSER *p)
{
  return(p->errstr);
}


size_t UdmXMLErrorPos (UDM_XML_PARSER *p)
{
  const char *s, *beg= UdmXMLParserStr(p);
  for (s= beg; s < UdmXMLParserCur(p); s++)
  {
    if (s[0] == '\n')
      beg= s;
  }
  return UdmXMLParserCur(p) - beg;
}


size_t UdmXMLErrorLineno (UDM_XML_PARSER *p)
{
  size_t res= 0;
  const char *s;
  for (s= UdmXMLParserStr(p); s < UdmXMLParserCur(p); s++)
  {
    if (s[0]=='\n')
      res++;
  }
  return(res);
}

/************ /MY_XML **************/


static udm_rc_t
startElement(UDM_XML_PARSER *parser, const char *name, size_t l)
{
  XML_PARSER_DATA *D= (XML_PARSER_DATA*) parser->user_data;
  UDM_VARLIST *Vars= &D->Indexer->Conf->XMLEnterHooks;
  const char *val;
  
  UdmFree(D->section_name);
  D->section_name= udm_strndup(name, l);
  UdmFree(D->secpath);
  D->secpath= udm_strndup(name, l);
  if (Vars->nvars && (val= UdmVarListFindStr(Vars, D->secpath, NULL)))
  {
    if (!strcasecmp(val, "HrefVarInit"))
    {
      UdmVarListFree(&D->Href.HrefVars);
    }
    else if (!strcasecmp(val, "HrefInit"))
    {
      UdmHrefFree(&D->Href);
      UdmHrefInit(&D->Href);
    }
  }
  return UDM_OK;
}


static udm_rc_t
endElement(UDM_XML_PARSER *parser, const char *name, size_t l)
{
  XML_PARSER_DATA *D= (XML_PARSER_DATA*) parser->user_data;
  size_t i= l;

  if (D->Indexer->Conf->XMLLeaveHooks.nvars)
  {
    char *prevname= udm_strndup(name, l);
    const char *val;
    if (D->Href.url &&
        (val= UdmVarListFindStr(&D->Indexer->Conf->XMLLeaveHooks,
                                prevname, NULL)))
    {
      UDM_DOCUMENT *Doc= D->Doc;
      D->Href.Param.referrer= UdmVarListFindInt(&Doc->Sections, "Referrer-ID", 0);
      D->Href.Param.hops= 1 + UdmVarListFindInt(&Doc->Sections, "Hops", 0);
      D->Href.Param.link_source= UDM_LINK_SOURCE_XML;
      UdmHrefListAdd(&Doc->Hrefs, &D->Href);
    }
    UdmFree(prevname);
  }

  while (--i && name[i] != UDM_XML_PATH_DELIMITER);

  UdmFree(D->section_name);
  D->section_name= udm_strndup(name, i);
  UdmFree(D->secpath);
  D->secpath= udm_strndup(name, i);
  return UDM_OK;
}


static udm_rc_t
Text(UDM_XML_PARSER *parser, const char *s, size_t len)
{
  XML_PARSER_DATA *D= (XML_PARSER_DATA*) parser->user_data;
  UDM_DOCUMENT *Doc= D->Doc;
  UDM_CONST_TEXTITEM  ConstItem;
  UDM_TEXT_PARAM TextParam;
  UDM_VARLIST  *Vars= &D->Indexer->Conf->XMLDataHooks;
  const UDM_VAR *Sec;
  const char *val;
  size_t slen= 0;

  if (D->section_name == NULL)
    return UDM_OK;

  /* fprintf(stderr, "%s=%.*s\n", D->secpath, (int) len, s); */
  /* Process XMLDataHook commands */
  if ((val= UdmVarListFindStr(Vars, D->section_name, NULL)))
  {
    if (!strcasecmp(val, "HrefSet"))
    {
      UdmFree(D->Href.url);
      D->Href.url= udm_strndup(s, (size_t)len);
      UdmSGMLUnescape(D->Href.url);
    }
    else if (!strcasecmp(val, "HrefVarAdd"))
    {
      UdmVarListReplaceStrn(&D->Href.HrefVars, D->section_name, s, len);
    }
    else if (!strcasecmp(val, "HrefVarAppend"))
    {
      UDM_VAR *var;
      if ((var= UdmVarListFindVar(&D->Href.HrefVars, D->section_name)))
      {
        UdmVarAppendStrn(var, " ", 1);
        UdmVarAppendStrn(var, s, len);
      }
      else
      {
        UdmVarListReplaceStrn(&D->Href.HrefVars, D->section_name, s, len);
      }
    }
  }


  if (D->secpath &&
      (!strcasecmp(D->secpath, "/urlset@xmlns") ||
       !strcasecmp(D->secpath, "/sitemapindex@xmlns")))
  {
    D->is_sitemap= UDM_TRUE;
    D->Doc->Spider.robots.index= UDM_FALSE;
    UdmTextListFree(&Doc->TextList);
  }

  UdmTextParamInit(&TextParam, UDM_TEXTLIST_FLAG_HTML, 0);  
  UdmConstTextItemInit(&ConstItem);
  UdmConstStrSet(&ConstItem.text, s, len);
  if ((Sec= UdmVarListFind(&Doc->Sections, D->section_name)))
  {
    TextParam.secno= UdmVarSecno(Sec);
    ConstItem.section_name.str= D->section_name;
    ConstItem.section_name.length= D->section_name ? strlen(D->section_name) : 0;
  }
  else if (D->XMLDefaultSection.str && !D->is_sitemap)
  {
    TextParam.secno= D->body_sec;
    ConstItem.section_name= D->XMLDefaultSection;
  }
  else
  {
    TextParam.secno= 0;
    ConstItem.section_name.str= D->section_name;
    ConstItem.section_name.length= D->section_name ? strlen(D->section_name) : 0;
  }
  UdmTextListAddConst(&Doc->TextList, &ConstItem, &TextParam);

  if (D->secpath &&
      (slen= strlen(D->secpath)) >= 5 &&
      (!strncasecmp(&D->secpath[slen - 5], "/href", 5) ||
       !strcasecmp(D->secpath, "/sitemapindex/sitemap/loc") ||
       !strcasecmp(D->secpath, "/urlset/url/loc")))
  {
    UdmHrefFree(&D->Href);
    UdmHrefInit(&D->Href);
    D->Href.url= udm_strndup(s, (size_t)len);
    UdmSGMLUnescape(D->Href.url);
    D->Href.Param.referrer= UdmVarListFindInt(&Doc->Sections, "Referrer-ID", 0);
    D->Href.Param.hops= 1 + UdmVarListFindInt(&Doc->Sections, "Hops", 0);
    D->Href.Param.link_source= UDM_LINK_SOURCE_XML;
    UdmHrefListAdd(&Doc->Hrefs, &D->Href);
  }


  if (!strncasecmp(D->secpath, UDM_CSTR_WITH_LEN("/rss@encoding")) ||
      (parser->xmldecl &&
       !strncasecmp(D->secpath, UDM_CSTR_WITH_LEN("/xml@encoding"))))
  {
    char buf[64];
    if (len > 0 && len < sizeof(buf))
    {
      const char *csname;
      memcpy(buf, s, len);
      buf[len]= '\0';
      csname= UdmCharsetCanonicalName(buf);
      if (csname)
        UdmVarListReplaceStr(&Doc->Sections, "Meta-Charset", csname);
    }
  }
  return UDM_OK;
}


#if 0
static void Decl(void *userData, const XML_Char *version, const XML_Char *encoding, int standalone)
{
  XML_PARSER_DATA *D= userData;
  UDM_DOCUMENT *Doc= D->Doc;

  if (encoding != NULL)
    UdmVarListReplaceStr(&Doc->Sections,
                         "Meta-Charset", UdmCharsetCanonicalName(encoding));

}

static int EncHandler(void *encodingHandlerData, const XML_Char *name, XML_Encoding *info)
{
  UDM_AGENT *indexer= encodingHandlerData;
  UDM_CHARSET  *cs;
  size_t i;

  if (!(cs= UdmGetCharSet(name)))
  {
    return 0;
  }
  
  /* FIXME: rewrite this for multibytes encodings */
  if (cs->tab_to_uni == NULL) return 0; 

  info->convert= NULL;
  info->release= NULL;
  info->data= NULL;
  for(i= 0; i < 256; i++)
    info->map[i]= cs->tab_to_uni[i];

  return 1;
}
#endif


udm_rc_t
UdmXMLParse(UDM_AGENT *Indexer, UDM_DOCUMENT *Doc)
{
  udm_rc_t rc= UDM_OK;
  XML_PARSER_DATA Data;
  UDM_XML_PARSER parser;
  const char *XMLDefaultSection= UdmVarListFindStr(&Indexer->Conf->Vars, "XMLDefaultSection", NULL);
  const UDM_VAR *BSec= XMLDefaultSection ? UdmVarListFind(&Doc->Sections,XMLDefaultSection) : NULL;
  int body_sec= BSec ? UdmVarSecno(BSec) : 0;
  UDM_CONST_STR content;

  if (UdmHTTPBufContentToConstStr(&Doc->Buf, &content))
    return UDM_ERROR;

  UdmXMLParserCreate(&parser);
  bzero(&Data, sizeof(Data));
  Data.Indexer= Indexer;
  Data.Doc= Doc;
  Data.body_sec= body_sec;
  Data.XMLDefaultSection.str= XMLDefaultSection;
  Data.XMLDefaultSection.length= XMLDefaultSection ? strlen(XMLDefaultSection) : 0;

  UdmXMLSetUserData(&parser, &Data);
  UdmXMLSetEnterHandler(&parser, startElement);
  UdmXMLSetLeaveHandler(&parser, endElement);
  UdmXMLSetValueHandler(&parser, Text);

  if (UDM_OK != (rc= UdmXMLParserExec(&parser, content.str, content.length)))
  {
    char err[256];    
    udm_snprintf(err, sizeof(err), 
                 "XML parsing error: %s at line %d pos %d",
                  UdmXMLErrorString(&parser),
                  (int) UdmXMLErrorLineno(&parser),
                  (int) UdmXMLErrorPos(&parser));
    UdmVarListReplaceStr(&Doc->Sections, "X-Reason", err);
  }

  UdmXMLParserFree(&parser);
  UdmFree(Data.section_name);
  UdmFree(Data.secpath);
  UdmHrefFree(&Data.Href);
  return rc;
}
