/* Copyright (C) 2000-2015 Lavtech.com corp. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
*/

#include "udm_config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <errno.h>

#include "udm_common.h"
#include "udm_template.h"
#include "udm_server.h"
#include "udm_env.h"
#include "udm_conf.h"
#include "udm_db.h"
#include "udm_http.h"
#include "udm_parsehtml.h"
#include "udm_host.h"
#include "udm_contentencoding.h"
#include "udm_utils.h"
#include "udm_vars.h"
#include "udm_wild.h"
#include "udm_searchtool.h"
#include "udm_doc.h"
#include "udm_match.h"
#include "udm_sgml.h"
#include "udm_proto.h"
#include "udm_ctype.h"
#include "udm_lex.h"
#include "udm_prog.h"


void
UdmProgItemInit(UDM_PROG_ITEM *i)
{
  bzero((void*) i, sizeof(*i));
  i->cmdnum= UDM_PROG_HANDLED;
}


static void
UdmProgItemFree(UDM_PROG_ITEM *it)
{
  UDM_FREE(it->comment);
}


void UdmProgExecutorStateInitArgs(UDM_PROG_EXECUTOR_STATE *state)
{
  UDM_BZERO(&state->vreg, sizeof(state->vreg));
}


static void PrintComment(UDM_PROG_ITEM *it)
{
  printf("%s%s", it->comment ? " # " : "",
                 it->comment ? it->comment : "");
}


static void PrintJmp(UDM_PROG_ITEM *it, const char *name, int add)
{
  printf("%3s  %04X", name, (int) it->cmdarg.jump);
}


static void UdmProgItemPrintArg(const UDM_PROG_ARG *arg)
{
  if (arg->handler)
    printf(" handler=%s", arg->handler->type_name);
  if (arg->stack_offset || arg->handler)
    printf(" stack:%03X", (int) arg->stack_offset);
  if (arg->textdata_addr.length || arg->textdata_addr.offset)
  {
    printf(" textdata: %d:%d",
           (int) arg->textdata_addr.offset,
           (int) arg->textdata_addr.length);
  }
}


static void UdmProgItemPrintHandled(UDM_PROG_ITEM *it)
{
  printf("%s", it->cmd->name.str);
  UdmProgItemPrintArg(&it->item_args);
}


static void UdmProgItemPrint(UDM_PROG_ITEM *it)
{
  switch (it->cmdnum)
  {
  case UDM_PROG_JE:  PrintJmp(it, "je", (int) it->cmdarg.jump); break;
  case UDM_PROG_JNE: PrintJmp(it, "jne", (int) it->cmdarg.jump); break;
  case UDM_PROG_JMP: PrintJmp(it, "jmp", (int) it->cmdarg.jump); break;
  case UDM_PROG_NOT_IREG0:       printf("not  i0"); break;
  case UDM_PROG_INC_IREG0:       printf("inc  i0"); break;
  case UDM_PROG_DEC_IREG0:       printf("dec  i0"); break;
  case UDM_PROG_INC_DREG0:       printf("inc  d0"); break;
  case UDM_PROG_DEC_DREG0:       printf("dec  d0"); break;
  case UDM_PROG_MOV_CHR_PTR_S0_IREG0: printf("mov  CHR PTR s0, ireg0"); break;
  case UDM_PROG_MOV_INT_PTR_S0_IREG0: printf("mov  INT PTR s0, ireg0"); break;
  case UDM_PROG_MOV_DBL_PTR_S0_DREG0: printf("mov  DBL PTR s0, dreg0"); break;
  case UDM_PROG_MOV_IREG0_INT_PTR_STACK_OFFSET:
    printf("mov  i0,INT PTR [stack:%d]", (int) it->item_args.stack_offset);
    break;
  case UDM_PROG_MOV_IREG1_INT_PTR_STACK_OFFSET:
    printf("mov  i1,INT PTR [stack:%d]", (int) it->item_args.stack_offset);
    break;
  case UDM_PROG_MOV_DREG0_DBL_PTR_STACK_OFFSET:
    printf("mov  d0,DBL PTR [stack:%d]", (int) it->item_args.stack_offset);
    break;
  case UDM_PROG_MOV_DREG1_DBL_PTR_STACK_OFFSET:
    printf("mov  d1,DBL PTR [stack:%d]", (int) it->item_args.stack_offset);
    break;
  case UDM_PROG_INC_CHR_PTR_STACK_OFFSET:
    printf("inc  CHR PTR [stack:%d]", (int) it->cmdarg.offset);
    break;
  case UDM_PROG_DEC_CHR_PTR_STACK_OFFSET:
    printf("dec  CHR PTR [stack:%d]", (int) it->cmdarg.offset);
    break;
  case UDM_PROG_INC_INT_PTR_STACK_OFFSET:
    printf("inc  INT PTR [stack:%d]", (int) it->cmdarg.offset);
    break;
  case UDM_PROG_DEC_INT_PTR_STACK_OFFSET:
    printf("dec  INT PTR [stack:%d]", (int) it->cmdarg.offset);
    break;
  case UDM_PROG_INC_DBL_PTR_STACK_OFFSET:
    printf("inc  DBL PTR [stack:%d]", (int) it->cmdarg.offset);
    break;
  case UDM_PROG_DEC_DBL_PTR_STACK_OFFSET:
    printf("dec  DBL PTR [stack:%d]", (int) it->cmdarg.offset);
    break;
  case UDM_PROG_MUL_IREG0_IREG1: printf("mul  i0,i1"); break;
  case UDM_PROG_ADD_IREG0_IREG1: printf("add  i0,i1"); break;
  case UDM_PROG_SUB_IREG0_IREG1: printf("sub  i0,i1"); break;
  case UDM_PROG_AND_IREG0_IREG1: printf("and  i0,i1"); break;
  case UDM_PROG_OR_IREG0_IREG1:  printf("or   i0,i1"); break;
  case UDM_PROG_XOR_IREG0_IREG1: printf("xor  i0,i1"); break;
  case UDM_PROG_SHL_IREG0_IREG1: printf("shl  i0,i1"); break;
  case UDM_PROG_SHR_IREG0_IREG1: printf("shr  i0,i1"); break;
  case UDM_PROG_DIV_IREG0_IREG1: printf("div  i0,i1"); break;
  case UDM_PROG_REM_IREG0_IREG1: printf("rem  i0,i1"); break;

  case UDM_PROG_MUL_DREG0_DREG1: printf("mul  d0,d1"); break;
  case UDM_PROG_ADD_DREG0_DREG1: printf("add  d0,d1"); break;
  case UDM_PROG_SUB_DREG0_DREG1: printf("sub  d0,d1"); break;
  case UDM_PROG_DIV_DREG0_DREG1: printf("div  d0,d1"); break;

  case UDM_PROG_MOV_HREG0_0:     printf("mov  h0,0"); break;
  case UDM_PROG_MOV_SREG0_0:     printf("mov  s0,0"); break;

  case UDM_PROG_MOV_IREG0_0:     printf("mov  i0,0"); break;
  case UDM_PROG_MOV_IREG1_0:     printf("mov  i1,0"); break;
  case UDM_PROG_MOV_DREG0_0:     printf("mov  d0,0"); break;
  case UDM_PROG_MOV_DREG1_0:     printf("mov  d1,0"); break;
  case UDM_PROG_MOV_DREG0_IREG0: printf("mov  d0,i0"); break;
  case UDM_PROG_MOV_DREG1_IREG1: printf("mov  d1,i1"); break;
  case UDM_PROG_MOV_DREG1_DREG0: printf("mov  d1,d0"); break;
  case UDM_PROG_MOV_DREG1_IREG0: printf("mov  d1,i0"); break;
  case UDM_PROG_MOV_IREG1_DREG1: printf("mov  i1,d1"); break;
  case UDM_PROG_MOV_IREG0_DREG0: printf("mov  i0,d0"); break;

  case UDM_PROG_SWP_DREG0_DREG1: printf("swp  d0,d1"); break;
  case UDM_PROG_SWP_IREG0_IREG1: printf("swp  i0,i1"); break;

  case UDM_PROG_MOV_IREG1_IREG0: printf("mov  i1,i0"); break;

  case UDM_PROG_MOV_DREG0_DBL:   printf("mov  d0,%f", it->cmdarg.val_double); break;
  case UDM_PROG_MOV_DREG1_DBL:   printf("mov  d1,%f", it->cmdarg.val_double); break;
  case UDM_PROG_MOV_IREG0_INT:   printf("mov  i0,%d", it->cmdarg.val_int); break;
  case UDM_PROG_MOV_IREG1_INT:   printf("mov  i1,%d", it->cmdarg.val_int); break;
  case UDM_PROG_MOV_HREG0_PTR:
    printf("mov  h0, UdmValueHandler<%s>",
          ((const UDM_VALUE_HANDLER *)it->cmdarg.val_const_ptr)->type_name);
    break;
  case UDM_PROG_MOV_SREG0_STACK_OFFSET: printf("mov  s0,stack+%03X", (int) it->cmdarg.offset); break;
  case UDM_PROG_MOV_VREG0_STACK_OFFSET: printf("mov  v0,stack+%03X", (int) it->cmdarg.offset); break;
  case UDM_PROG_MOV_VREG1_STACK_OFFSET: printf("mov  v1,stack+%03X", (int) it->cmdarg.offset); break;
  case UDM_PROG_MOV_VREG2_STACK_OFFSET: printf("mov  v2,stack+%03X", (int) it->cmdarg.offset); break;
  case UDM_PROG_MOV_VREG3_STACK_OFFSET: printf("mov  v3,stack+%03X", (int) it->cmdarg.offset); break;
  case UDM_PROG_MOV_VREG4_STACK_OFFSET: printf("mov  v4,stack+%03X", (int) it->cmdarg.offset); break;
  case UDM_PROG_CMP_DREG0_DREG1_EQ: printf("cmp  d0,d1 [eq]"); break;
  case UDM_PROG_CMP_DREG0_DREG1_LT: printf("cmp  d0,d1 [lt]"); break;
  case UDM_PROG_CMP_DREG0_DREG1_GT: printf("cmp  d0,d1 [gt]"); break;
  case UDM_PROG_CMP_DREG0_DREG1_LE: printf("cmp  d0,d1 [le]"); break;
  case UDM_PROG_CMP_DREG0_DREG1_GE: printf("cmp  d0,d1 [ge]"); break;
  case UDM_PROG_CMP_DREG0_DREG1_NE: printf("cmp  d0,d1 [ne]"); break;

  case UDM_PROG_CMP_IREG0_IREG1_EQ: printf("cmp  i0,i1 [eq]"); break;
  case UDM_PROG_CMP_IREG0_IREG1_LT: printf("cmp  i0,i1 [lt]"); break;
  case UDM_PROG_CMP_IREG0_IREG1_GT: printf("cmp  i0,i1 [gt]"); break;
  case UDM_PROG_CMP_IREG0_IREG1_LE: printf("cmp  i0,i1 [le]"); break;
  case UDM_PROG_CMP_IREG0_IREG1_GE: printf("cmp  i0,i1 [ge]"); break;
  case UDM_PROG_CMP_IREG0_IREG1_NE: printf("cmp  i0,i1 [ne]"); break;

  case UDM_PROG_CALL:
  {
    int i;
    const UDM_FUNCTION *func= it->cmdarg.func;
    printf("call %s %.*s(",
           func->prototype.result.handler ?
           func->prototype.result.handler->type_name :
           "void",
           (int) func->name.length,
           func->name.str);
    for (i= 0; i < func->prototype.nargs; i++)
    {
      printf("%s%s", i > 0 ? "," : "",
                     func->prototype.args[i].handler->type_name);
    }
    printf(")");
    break;
  }

  case UDM_PROG_HANDLED:
    UdmProgItemPrintHandled(it); break;
  case UDM_PROG_NOP: printf("nop"); break;
  case UDM_PROG_EXIT: printf("exit"); break;
  }
  PrintComment(it);
  printf("\n");
}


void UdmProgInit(UDM_PROG *prg)
{
  bzero((void*)prg, sizeof(*prg));
  UdmDSTRInit(&prg->textdata, 1024);
}


void
UdmProgPrint(const UDM_PROG *p)
{
  size_t i;
  
  for (i=0 ; i < p->nitems ; i++)
  {
    UDM_PROG_ITEM *it= &p->Items[i];
    printf("%04X ", (int) i);
    UdmProgItemPrint(it);
  }
}


void UdmProgFree(UDM_PROG *prg)
{
  size_t i;
  for (i=0 ; i < prg->nitems; i++)
    UdmProgItemFree(&prg->Items[i]);
  UDM_FREE(prg->Items);
  UdmDSTRFree(&prg->textdata);
}


static int
UdmTemplateExecutorProcessOneCommand(UDM_PROG_EXECUTOR *executor,
                                     const UDM_PROG_ITEM *start,
                                     size_t nitems)
{
  UDM_PROG_EXECUTOR_STATE *state= &executor->state;  
  const UDM_PROG_ITEM *it= executor->state.curr;
  switch (it->cmdnum)
  {
    case UDM_PROG_JNE:
      if (state->reg.i[0])
        state->curr= start + it->cmdarg.jump;
      else
        state->curr++;
      return 0;
    
    case UDM_PROG_JE:
      if (!state->reg.i[0])
        state->curr= start + it->cmdarg.jump;
      else
        state->curr++;
      return 0;
    
    case UDM_PROG_JMP:
      state->curr= start + it->cmdarg.jump;
      return 0;

    case UDM_PROG_NOT_IREG0:       state->reg.i[0]= ~(state->reg.i[0]); break;
    case UDM_PROG_INC_IREG0:       state->reg.i[0]++; break;
    case UDM_PROG_DEC_IREG0:       state->reg.i[0]--; break;
    case UDM_PROG_INC_DREG0:       state->reg.d[0]++; break;
    case UDM_PROG_DEC_DREG0:       state->reg.d[0]--; break;
    case UDM_PROG_MOV_IREG0_INT_PTR_STACK_OFFSET:
      state->reg.i[0]= (((int *) (state->stack.Val.str + it->item_args.stack_offset))[0]);
      break;
    case UDM_PROG_MOV_IREG1_INT_PTR_STACK_OFFSET:
      state->reg.i[1]= (((int *) (state->stack.Val.str + it->item_args.stack_offset))[0]);
      break;
    case UDM_PROG_MOV_DREG0_DBL_PTR_STACK_OFFSET:
      state->reg.d[0]= (((double *) (state->stack.Val.str + it->item_args.stack_offset))[0]);
      break;
    case UDM_PROG_MOV_DREG1_DBL_PTR_STACK_OFFSET:
      state->reg.d[1]= (((double *) (state->stack.Val.str + it->item_args.stack_offset))[0]);
      break;
    case UDM_PROG_MOV_CHR_PTR_S0_IREG0:
      (((char *) (state->reg.s0))[0])= state->reg.i[0];
      break;
    case UDM_PROG_MOV_INT_PTR_S0_IREG0:
      (((int *) (state->reg.s0))[0])= state->reg.i[0];
      break;
    case UDM_PROG_MOV_DBL_PTR_S0_DREG0:
      (((double *) (state->reg.s0))[0])= state->reg.d[0];
      break;
    case UDM_PROG_INC_CHR_PTR_STACK_OFFSET:
      (((char *) (state->stack.Val.str + it->item_args.stack_offset))[0])++;
      break;
    case UDM_PROG_DEC_CHR_PTR_STACK_OFFSET:
      (((char *) (state->stack.Val.str + it->item_args.stack_offset))[0])--;
      break;
    case UDM_PROG_INC_DBL_PTR_STACK_OFFSET:
      (((double *) (state->stack.Val.str + it->item_args.stack_offset))[0])++;
      break;
    case UDM_PROG_DEC_DBL_PTR_STACK_OFFSET:
      (((double *) (state->stack.Val.str + it->item_args.stack_offset))[0])--;
      break;
    case UDM_PROG_INC_INT_PTR_STACK_OFFSET:
      (((int *) (state->stack.Val.str + it->item_args.stack_offset))[0])++;
      break;
    case UDM_PROG_DEC_INT_PTR_STACK_OFFSET:
      (((int *) (state->stack.Val.str + it->item_args.stack_offset))[0])--;
      break;
    case UDM_PROG_MOV_SREG0_STACK_OFFSET:
      state->reg.s0= state->stack.Val.str + it->cmdarg.offset;
      break;
    case UDM_PROG_MOV_VREG0_STACK_OFFSET:
      state->vreg[0].Data= state->stack.Val.str + it->cmdarg.offset;
      break;
    case UDM_PROG_MOV_VREG1_STACK_OFFSET:
      state->vreg[1].Data= state->stack.Val.str + it->cmdarg.offset;
      break;
    case UDM_PROG_MOV_VREG2_STACK_OFFSET:
      state->vreg[2].Data= state->stack.Val.str + it->cmdarg.offset;
      break;
    case UDM_PROG_MOV_VREG3_STACK_OFFSET:
      state->vreg[3].Data= state->stack.Val.str + it->cmdarg.offset;
      break;
    case UDM_PROG_MOV_VREG4_STACK_OFFSET:
      state->vreg[4].Data= state->stack.Val.str + it->cmdarg.offset;
      break;
    case UDM_PROG_MUL_IREG0_IREG1: state->reg.i[0]*=  state->reg.i[1]; break;
    case UDM_PROG_ADD_IREG0_IREG1: state->reg.i[0]+=  state->reg.i[1]; break;
    case UDM_PROG_SUB_IREG0_IREG1: state->reg.i[0]-=  state->reg.i[1]; break;
    case UDM_PROG_AND_IREG0_IREG1: state->reg.i[0]&=  state->reg.i[1]; break;
    case UDM_PROG_OR_IREG0_IREG1:  state->reg.i[0]|=  state->reg.i[1]; break;
    case UDM_PROG_XOR_IREG0_IREG1: state->reg.i[0]^=  state->reg.i[1]; break;
    case UDM_PROG_SHL_IREG0_IREG1: state->reg.i[0]<<= state->reg.i[1]; break;
    case UDM_PROG_SHR_IREG0_IREG1: state->reg.i[0]>>= state->reg.i[1]; break;

    case UDM_PROG_DIV_IREG0_IREG1:
       state->reg.i[0]= state->reg.i[1] ? state->reg.i[0] / state->reg.i[1] : 0;
       break;
    case UDM_PROG_REM_IREG0_IREG1:
       state->reg.i[0]= state->reg.i[1] ? state->reg.i[0] % state->reg.i[1] : 0;
       break;


    case UDM_PROG_MUL_DREG0_DREG1: state->reg.d[0]*=  state->reg.d[1]; break;
    case UDM_PROG_ADD_DREG0_DREG1: state->reg.d[0]+=  state->reg.d[1]; break;
    case UDM_PROG_SUB_DREG0_DREG1: state->reg.d[0]-=  state->reg.d[1]; break;
    case UDM_PROG_DIV_DREG0_DREG1:
       state->reg.d[0]= state->reg.d[1] ? state->reg.d[0] / state->reg.d[1] : 0;
       break;

    case UDM_PROG_MOV_HREG0_0:       state->reg.h0= NULL; break;
    case UDM_PROG_MOV_SREG0_0:       state->reg.s0= NULL; break;
    case UDM_PROG_MOV_IREG0_0:       state->reg.i[0]= 0;  break;
    case UDM_PROG_MOV_IREG1_0:       state->reg.i[1]= 0;  break;
    case UDM_PROG_MOV_DREG0_0:       state->reg.d[0]= 0;  break;
    case UDM_PROG_MOV_DREG1_0:       state->reg.d[1]= 0;  break;
    case UDM_PROG_MOV_DREG0_IREG0:   state->reg.d[0]= state->reg.i[0]; break;
    case UDM_PROG_MOV_DREG1_IREG1:   state->reg.d[1]= state->reg.i[1]; break;
    case UDM_PROG_MOV_DREG1_DREG0:   state->reg.d[1]= state->reg.d[0]; break;
    case UDM_PROG_MOV_DREG1_IREG0:   state->reg.d[1]= state->reg.i[0]; break;
    case UDM_PROG_MOV_IREG1_DREG1:   state->reg.i[1]= state->reg.d[1]; break;
    case UDM_PROG_MOV_IREG0_DREG0:   state->reg.i[0]= state->reg.d[0]; break;

    case UDM_PROG_SWP_DREG0_DREG1:
    {
      double tmp= state->reg.d[0];
      state->reg.d[0]= state->reg.d[1];
      state->reg.d[1]= tmp;
      break;
    }
    case UDM_PROG_SWP_IREG0_IREG1:
    {
      int tmp= state->reg.i[0];
      state->reg.i[0]= state->reg.i[1];
      state->reg.i[1]= tmp;
      break;
    }
    case UDM_PROG_MOV_IREG1_IREG0:  state->reg.i[1]= state->reg.i[0]; break;
    case UDM_PROG_MOV_IREG0_INT:    state->reg.i[0]= it->cmdarg.val_int; break;
    case UDM_PROG_MOV_IREG1_INT:    state->reg.i[1]= it->cmdarg.val_int; break;
    case UDM_PROG_MOV_DREG0_DBL:    state->reg.d[0]= it->cmdarg.val_double; break;
    case UDM_PROG_MOV_DREG1_DBL:    state->reg.d[1]= it->cmdarg.val_double; break;
    case UDM_PROG_MOV_HREG0_PTR:    state->reg.h0= it->cmdarg.val_const_ptr; break;
    case UDM_PROG_CMP_DREG0_DREG1_EQ:
       state->reg.i[0]= (state->reg.d[0] == state->reg.d[1]);
       break;
    case UDM_PROG_CMP_DREG0_DREG1_NE:
       state->reg.i[0]= (state->reg.d[0] != state->reg.d[1]);
       break;

    case UDM_PROG_CMP_DREG0_DREG1_LT:
       state->reg.i[0]= (state->reg.d[0] < state->reg.d[1]);
       break;

    case UDM_PROG_CMP_DREG0_DREG1_GT:
       state->reg.i[0]= (state->reg.d[0] > state->reg.d[1]);
       break;

    case UDM_PROG_CMP_DREG0_DREG1_LE:
       state->reg.i[0]= (state->reg.d[0] <= state->reg.d[1]);
       break;

    case UDM_PROG_CMP_DREG0_DREG1_GE:
       state->reg.i[0]= (state->reg.d[0] >= state->reg.d[1]);
       break;

    case UDM_PROG_CMP_IREG0_IREG1_EQ:
       state->reg.i[0]= (state->reg.i[0] == state->reg.i[1]);
       break;
    case UDM_PROG_CMP_IREG0_IREG1_NE:
       state->reg.i[0]= (state->reg.i[0] != state->reg.i[1]);
       break;

    case UDM_PROG_CMP_IREG0_IREG1_LT:
       state->reg.i[0]= (state->reg.i[0] < state->reg.i[1]);
       break;

    case UDM_PROG_CMP_IREG0_IREG1_GT:
       state->reg.i[0]= (state->reg.i[0] > state->reg.i[1]);
       break;

    case UDM_PROG_CMP_IREG0_IREG1_LE:
       state->reg.i[0]= (state->reg.i[0] <= state->reg.i[1]);
       break;

    case UDM_PROG_CMP_IREG0_IREG1_GE:
       state->reg.i[0]= (state->reg.i[0] >= state->reg.i[1]);
       break;

    case UDM_PROG_CALL:
      it->cmdarg.func->func(state);
      break;

    case UDM_PROG_HANDLED:
      it->cmd->exec(state);
      break;
    case UDM_PROG_NOP:
      break;
    case UDM_PROG_EXIT:
      state->curr= start + nitems;
      return state->reg.i[0];
  }
  state->curr++;
  return 0;
}


int
UdmProgExecutorExec(UDM_PROG_EXECUTOR *tmpl, const UDM_PROG *prg)
{
  int rc= 0;
  tmpl->state.prog= prg;
  for (tmpl->state.curr= prg->Items;
       tmpl->state.curr < &prg->Items[prg->nitems]; )
  {
    rc= UdmTemplateExecutorProcessOneCommand(tmpl, prg->Items, prg->nitems);
  }
  tmpl->state.prog= NULL;
  return rc;
}


void
UdmProgExecutorInit(UDM_PROG_EXECUTOR *executor,
                    FILE *stream, UDM_DSTR *dstr)
{
  executor->state.reg.s0= NULL;
  executor->state.reg.h0= NULL;
  bzero(&executor->state.globals, sizeof(executor->state.globals));
  executor->state.globals.stdout= stream;
  executor->state.fatal_error= UDM_FALSE;
  UdmProgExecutorStateInitArgs(&executor->state);
  UdmDSTRInit(&executor->state.stack, 1024);
  UdmDSTRAlloc(&executor->state.stack, 128*1024);
}


void
UdmProgExecutorFree(UDM_PROG_EXECUTOR *executor)
{
  UdmDSTRFree(&executor->state.stack);
}


static void CompileStackInit(UDM_PROG_COMPILE_STACK *st)
{
  bzero((void*) st, sizeof(*st));
}


static void CompileStackFree(UDM_PROG_COMPILE_STACK *st)
{
  UDM_FREE(st->Items);
}


void UdmProgVarListInit(UDM_PROG_VARLIST *List)
{
  UDM_BZERO(List, sizeof(*List));
}


void UdmProgVarListFree(UDM_PROG_VARLIST *List)
{
  size_t i;
  for (i= 0 ; i < List->nitems; i++)
    UdmFree(List->Item[i].name);
  UdmFree(List->Item);
}


udm_rc_t UdmProgVarListAdd(UDM_PROG_VARLIST *List,
                           const char *name,
                           const UDM_PROG_VAR_VALUE *Value)
{
  if (List->nitems >= List->mitems)
  {
    List->mitems+= 16;
    if (!(List->Item= (UDM_PROG_VAR2 *) UdmRealloc(List->Item,
                                                   List->mitems *
                                                   sizeof(UDM_PROG_VAR2))))
      return UDM_ERROR;
  }
  List->Item[List->nitems].name= UdmStrdup(name);
  List->Item[List->nitems].value= Value[0];
  List->nitems++;
  return UDM_OK;
}


void UdmProgVarListListInit(UDM_PROG_VARLISTLIST *List)
{
  UDM_BZERO(List, sizeof(*List));
}


void UdmProgVarListListFree(UDM_PROG_VARLISTLIST *List)
{
  size_t i;
  for (i= 0; i < List->nitems; i++)
    UdmProgVarListFree(&List->Item[i]);
  UdmFree(List->Item);
}


udm_rc_t UdmProgVarListListAdd(UDM_PROG_VARLISTLIST *List,
                               const UDM_PROG_VARLIST *Item)
{
  if (List->nitems >= List->mitems)
  {
    List->mitems+= 16;
    if (!(List->Item= (UDM_PROG_VARLIST *) UdmRealloc(List->Item,
                                                      List->mitems *
                                                      sizeof(UDM_PROG_VARLIST))))
      return UDM_ERROR;
  }
  List->Item[List->nitems++]= Item[0];
  return UDM_OK;
}


const UDM_PROG_VAR2 *UdmProgVarListFind(const UDM_PROG_VARLIST *List,
                                        const char *name)
{
  size_t i;
  for (i= 0; i < List->nitems; i++)
  {
    if (!strcmp(List->Item[i].name, name))
      return &List->Item[i];
  }
  return NULL;
}

                                        
void
UdmProgCompilerInit(UDM_PROG_COMPILER *compiler,
                    size_t lineno, UDM_PROG *prg,
                    const char *src, size_t srclen)
{
  const char *DebugProg= getenv("DebugProg");
  compiler->prg= prg;
  CompileStackInit(&compiler->st);
  compiler->comment_type= COMMENT_NONE;
  compiler->lineno= lineno;
  compiler->start= src;
  compiler->errstr[0]= '\0';
  compiler->pi_count= 0;
  compiler->current_loop_continue= 0;
  compiler->generate_debug_info= DebugProg &&
                                 udm_strntobool(DebugProg, strlen(DebugProg));
  UdmLexScannerInit(&compiler->scanner, NULL, 0);
  UdmProgVarListListInit(&compiler->Vars2);
}


void
UdmProgCompilerFree(UDM_PROG_COMPILER *compiler)
{
  CompileStackFree(&compiler->st);
}


UDM_PROG *
UdmProgCompilerGetProg(UDM_PROG_COMPILER *tmpl)
{
  return tmpl->prg;
}


udm_rc_t
UdmProgAdd(UDM_PROG *prg, UDM_PROG_ITEM *item)
{
  if (prg->nitems >= prg->mitems)
  {
    prg->mitems+= 64;
    prg->Items= (UDM_PROG_ITEM*) UdmRealloc(prg->Items,
                                            prg->mitems*sizeof(*item));
    if (!prg->Items)
      return UDM_ERROR;
  }
  prg->Items[prg->nitems]= item[0];
  prg->nitems++;
  return UDM_OK;
}


udm_rc_t
UdmProgAddArg0SimpleOp(UDM_PROG *prog, udm_tmpl_cmd_t op)
{
  UDM_PROG_ITEM i;
  UdmProgItemInit(&i);
  i.cmdnum= op;
  return UdmProgAdd(prog, &i);
}


udm_rc_t
UdmProgAddArg1SimpleOpDouble(UDM_PROG *prog, udm_tmpl_cmd_t op, double value)
{
  UDM_PROG_ITEM i;
  UdmProgItemInit(&i);
  i.cmdnum= op;
  i.cmdarg.val_double= value;
  return UdmProgAdd(prog, &i);
}


udm_rc_t
UdmProgAddArg1SimpleOpInt(UDM_PROG *prog, udm_tmpl_cmd_t op, int value)
{
  UDM_PROG_ITEM i;
  UdmProgItemInit(&i);
  i.cmdnum= op;
  i.cmdarg.val_int= value;
  return UdmProgAdd(prog, &i);
}


udm_rc_t
UdmProgAddArg0Simple(UDM_PROG *prog, UDM_PROG_CMD *cmd)
{
  UDM_PROG_ITEM i;
  UdmProgItemInit(&i);
  i.cmd= cmd;
  return UdmProgAdd(prog, &i);
}


udm_rc_t
UdmProgAddJmp(UDM_PROG *prog, udm_tmpl_cmd_t cmdnum, int addr)
{
  UDM_PROG_ITEM i;
  UdmProgItemInit(&i);
  i.cmdnum= cmdnum;
  i.cmdarg.jump= addr;
  return UdmProgAdd(prog, &i);
}


udm_rc_t
UdmProgAddValueHandler(UDM_PROG *prog, udm_tmpl_cmd_t cmdnum,
                       const UDM_VALUE_HANDLER *handler)
{
  UDM_PROG_ITEM i;
  UdmProgItemInit(&i);
  i.cmdnum= cmdnum;
  i.cmdarg.val_const_ptr= handler;
  return UdmProgAdd(prog, &i);
}


void *
UdmProgExecutorStateSetupReturnValue(UDM_PROG_EXECUTOR_STATE *state,
                                     const UDM_VALUE_HANDLER *handler)
{
  UDM_ASSERT(state->reg.s0);
  UDM_ASSERT(state->reg.h0 == handler);
  handler->Constructor(state->reg.s0, NULL, 0);
  return state->reg.s0;
}


void
UdmProgExecutorStateReturnStrn(UDM_PROG_EXECUTOR_STATE *state,
                               const char *str, size_t length)
{
  UDM_SECTION *ret= UdmProgExecutorStateSetupReturnValue(state, &UdmValueHandlerSimple);
  UdmValueHandlerSimple.SetStrn((char *) ret, str, length);
}
