/* Copyright (C) 2000-2015 Lavtech.com corp. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
*/

#include "udm_config.h"

#ifdef HAVE_SQL


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <ctype.h>
#include <time.h>

#ifdef WIN32
#include <time.h>
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifdef HAVE_SYS_TIME_H
#include <sys/time.h>
#endif

#include "udm_common.h"
#include "udm_utils.h"
#include "udm_spell.h"
#include "udm_robots.h"
#include "udm_db.h"
#include "udm_unicode.h"
#include "udm_unidata.h"
#include "udm_url.h"
#include "udm_log.h"
#include "udm_proto.h"
#include "udm_conf.h"
#include "udm_hash.h"
#include "udm_xmalloc.h"
#include "udm_boolean.h"
#include "udm_searchtool.h"
#include "udm_query.h"
#include "udm_server.h"
#include "udm_stopwords.h"
#include "udm_doc.h"
#include "udm_result.h"
#include "udm_vars.h"
#include "udm_agent.h"
#include "udm_store.h"
#include "udm_hrefs.h"
#include "udm_word.h"
#include "udm_db_int.h"
#include "udm_match.h"
#include "udm_indexer.h"
#include "udm_textlist.h"
#include "udm_parsehtml.h"

#ifdef HAVE_ZLIB
#include <zlib.h>
#endif


static udm_rc_t
UdmFetchCachedQuery(UDM_AGENT *A, UDM_DB *db,
                    UDM_QUERY *Query, const char *query, int *tm)
{
  udm_rc_t rc;
  UDM_SQLRES SQLRes;
  UDM_URLDATALIST *DataList= &Query->URLData;
  
  if (UDM_OK != (rc= UdmDBSQLQuery(A, db, &SQLRes, query)))
    return rc;
  if (UdmSQLNumRows(&SQLRes) == 1)
  {
    size_t i, nbytes;
    const char *p;
    DataList->nitems= UdmSQLLen(&SQLRes, 0, 0) / 12;
    nbytes= DataList->nitems * sizeof(UDM_URLDATA);
    DataList->Item= (UDM_URLDATA*) UdmMalloc(nbytes);
    bzero((void*) DataList->Item, nbytes);
    for (p= UdmSQLValue(&SQLRes, 0, 0), i= 0; i < DataList->nitems; i++)
    {
      UDM_URLDATA *D= &DataList->Item[i];
      D->url_id= udm_get_int4(p); p+= 4;
      D->score= udm_get_int4(p);  p+= 4;
      D->per_site= udm_get_int4(p); p+= 4;
    }
    if (UdmSQLNumCols(&SQLRes) >= 2)
    {
      UdmQueryFromXML(A, Query, UdmSQLValue(&SQLRes, 0, 1),
                       UdmSQLLen(&SQLRes, 0,1), A->Conf->lcs);
      if (UdmSQLNumCols(&SQLRes) >= 3)
        *tm= atoi(UdmSQLValue(&SQLRes,0,2));
    }
  }
  UdmSQLFree(&SQLRes);
  return rc;
}


static udm_rc_t
UdmLoadCachedQueryWords(UDM_AGENT *A, UDM_QUERY *Query,
                        UDM_DB *db, const char *pqid)
{
  char qbuf[128], bqid[32];
  char *tm, *end;
  int iid, itm;
  UDM_URLDATALIST *DataList= &Query->URLData;
  UDM_SQL_TOP_CLAUSE Top;
  
  bzero(DataList, sizeof(*DataList));
  if (!pqid)
    return UDM_OK;
  
  udm_snprintf(bqid, sizeof(bqid), "%s", pqid);
  if (!(tm= strchr(bqid, '-')))
    return UDM_OK;
  *tm++= '\0';
  iid= (int) strtoul(bqid, &end, 16);
  itm= (int) strtol(tm, &end, 16);
  
  UdmDBSQLTopClause(A, db, 1, &Top);
  udm_snprintf(qbuf, sizeof(qbuf),
               "SELECT %sdoclist FROM qcache "
               "WHERE id=%d AND tm=%d %s ORDER BY tm DESC %s",
               Top.top, iid, itm, Top.rownum, Top.limit);

  return UdmFetchCachedQuery(A, db, Query, qbuf, NULL);
}


const char QCACHEID[]=
"${q}.${pqid}.${SearchMode}.${orig_m}.${fl}.${wm}.${o}.${t}."
"${cat}.${ul}.${wf}.${g}.${tmplt}.${GroupBySite}.${site}."
"${type}.${sp}.${sy}.${dt}.${dp}.${dx}.${dm}.${dy}.${db}.${de}.${s}"
"${ss}.${us}.${su}.${sl.*}";


static int
QueryCacheID(UDM_AGENT *A, UDM_VARLIST *QueryVars)
{
  UDM_VARLIST *Vars= &A->Conf->Vars;
  const char *fmt= UdmVarListFindStr(Vars, "QueryCacheID", QCACHEID);
  UDM_DSTR d;
  int res;
  UdmDSTRInit(&d, 256);
  UdmDSTRParse(&d, fmt, Vars);
  res= (int) UdmStrHash32(UdmDSTRPtr(&d));
  /* TODO34: remove this, use only QueryVars below */
  UdmVarListReplaceStr(&A->Conf->Vars, "QueryCacheIDValue", UdmDSTRPtr(&d));
  UdmVarListReplaceStr(QueryVars, "QueryCacheIDValue", UdmDSTRPtr(&d));
  UdmDSTRFree(&d);
  return res;
}


udm_rc_t
UdmQueryCacheGetSQL(UDM_AGENT *A, UDM_DB *db, UDM_QUERY *Query)
{
  udm_bool_t use_qcache= UdmVarListFindBool(UdmSQLDBVars(db), "qcache", UDM_FALSE);
  int tm, id, qtm;
  udm_rc_t rc;
  char qbuf[128];
  UDM_SQL_TOP_CLAUSE Top;
  
  if (!use_qcache)
    return UDM_OK;
  
  if (UdmSQLDBMode(db) != UDM_SQLDBMODE_BLOB)
    return UDM_OK;
  if (UDM_OK != (rc= UdmBlobReadTimestamp(A, db, &tm, (int) time(0))))
    return rc;
  id= QueryCacheID(A, &Query->Res.Vars);

  UdmDBSQLTopClause(A, db, 1, &Top);

  udm_snprintf(qbuf, sizeof(qbuf),
               "SELECT %sdoclist, wordinfo, tm "
               "FROM qcache "
               "WHERE id=%d AND tm>=%d %s"
               "ORDER BY tm DESC%s", Top.top, id, tm, Top.rownum, Top.limit);
  rc= UdmFetchCachedQuery(A, db, Query, qbuf, &qtm);
  if (Query->URLData.nitems)
  {
    UdmLog(A, UDM_LOG_DEBUG,
           "Fetched from qcache %d documents, %d total found",
           (int) Query->URLData.nitems, (int) Query->stats.total_found);
    udm_snprintf(qbuf, sizeof(qbuf), "QCache:%08X-%08X", id, qtm);
    /* TODO34: remove this, use only Query->Res.Vars below */
    UdmVarListReplaceStr(&A->Conf->Vars, "ResultSource", qbuf);
    UdmVarListReplaceStr(&Query->Res.Vars, "ResultSource", qbuf);
    udm_snprintf(qbuf, sizeof(qbuf), "%08X-%08X", id, qtm);
    UdmVarListReplaceStr(&Query->Res.Vars, "qid", qbuf);
  }
  return UDM_OK;
}


static udm_rc_t
UdmExecuteUserCacheQuery(UDM_AGENT *A, UDM_DB *db, UDM_QUERY *Query)
{
  udm_rc_t rc= UDM_OK;
  const char *usercache= UdmVarListFindStr(UdmSQLDBVars(db), "usercache", "");
  const char *UserCacheQuery= UdmVarListFindStr(&A->Conf->Vars,
                                                "UserCacheQuery", NULL);

  if (UserCacheQuery && UserCacheQuery[0])
  {
    size_t i;
    UDM_DSTR dstr;
    UDM_VARLIST vars;
    UdmVarListInit(&vars);
    UdmVarListAddLst(&vars, &A->Conf->Vars, NULL, "*");
    UdmVarListReplaceInt(&vars, "total", Query->stats.total_found);
    UdmDSTRInit(&dstr, 64);
    for (i= 0; i < Query->URLData.nitems; i++)
    {
      UDM_URLDATA *Item= &Query->URLData.Item[i];
      UdmVarListReplaceInt(&vars, "url_id", Item->url_id);
      UdmVarListReplaceInt(&vars, "score", Item->score);
      UdmVarListReplaceInt(&vars, "rank", i);
      UdmDSTRParse(&dstr, UserCacheQuery, &vars);
      if (UDM_OK != (rc= UdmDBSQLQuery(A, db, NULL, UdmDSTRPtr(&dstr))))
        break;
      UdmDSTRReset(&dstr);
    }
    UdmDSTRFree(&dstr);
    UdmVarListFree(&vars);
    if (rc != UDM_OK)
      return rc;
  }


  /* Built-in UserCacheQuery */
  if (usercache[0] && strcasecmp(usercache, "no"))
  {
    size_t i;
    for (i= 0; i < Query->URLData.nitems; i++)
    {
      char qbuf[64];
      udm_snprintf(qbuf, sizeof(qbuf),
                   "INSERT INTO %s VALUES(%d, %d)",
                    usercache,
                    Query->URLData.Item[i].url_id,
                    Query->URLData.Item[i].score);
      if (UDM_OK != (rc= UdmDBSQLQuery(A, db, NULL, qbuf)))
        return rc;
    }
  }
  return rc;
}


/*
  TODO: reuse a function from cluster
*/
static udm_rc_t
UdmQueryToXML(UDM_QUERY *Query, UDM_DSTR *wwl)
{
  size_t i;
  UdmDSTRAppendf(wwl, "<result>"); 
  UdmDSTRAppendf(wwl, "<totalResults>%d</totalResults>",
                 (int) Query->stats.total_found);
  UdmDSTRAppendf(wwl,"<wordinfo>");
  for (i= 0; i < Query->Res.WWList.nwords; i++)
  {
    UDM_WIDEWORD *ww= &Query->Res.WWList.Word[i];
    UdmDSTRAppendf(wwl, "<word id='%d' order='%d' count='%d' len='%d' "
                         "origin='%d' weight='%d' match='%d' "
                         "secno='%d' phrlen='%d' phrpos='%d'>%s</word>",
                    (int) i, (int) ww->Param.order,
                    (int) ww->Param.count, (int) ww->Word.length,
                    (int) ww->Param.origin, ww->Param.weight,
                    (int) ww->Param.match_mode,
                    (int) ww->Param.secno, (int) ww->Param.phrlen, (int) ww->Param.phrpos,
                    ww->Word.str);
  }
  UdmDSTRAppendf(wwl, "</wordinfo></result>");
  return UDM_OK;
}


/*
  Collect coord information into dstr
*/
static udm_rc_t
UdmURLDataListEncode(UDM_AGENT *A, UDM_URLDATALIST *URLData,
                     UDM_DB *db, UDM_DSTR *buf)
{
  size_t i;
  for (i= 0; i < URLData->nitems; i++)
  {
    UDM_URLDATA *D= &URLData->Item[i];
    char tmp[4*3];
    udm_put_int4(D->url_id,   tmp + 0);
    udm_put_int4(D->score,    tmp + 4);
    udm_put_int4(D->per_site, tmp + 8);
    if (!db)
    {
      UdmDSTRAppend(buf, tmp, 12);
    }
    else if (UdmSQLDBType(db) == UDM_DB_PGSQL)
    {
      size_t len= UdmDBSQLBinEscStr(A, db, buf->Val.str + buf->Val.length, buf->size_alloced, tmp, 12);
      buf->Val.length+= len;
    }
    else
      UdmDSTRAppendHex(buf, tmp, 12);
  }
  return UDM_OK;
}


udm_rc_t
UdmQueryCachePutSQL(UDM_AGENT *Indexer, UDM_DB *db, UDM_QUERY *Query)
{
  udm_bool_t prevcache= UdmVarListFindBool(UdmSQLDBVars(db), "qcache", UDM_FALSE);
  udm_rc_t rc= UDM_OK;
  char idbuf[64];
  UDM_DSTR buf, wwl, bindata;
  size_t nbytes= Query->URLData.nitems * 24, wwl_length_escaped;
  int id, tm= time(0);
  const char *prefix= (UdmSQLDBFlags(db) & UDM_SQL_HAVE_STDHEX) ? "X'" : 
                      (UdmSQLDBFlags(db) & UDM_SQL_HAVE_BLOB_AS_HEX) ? "'" : "0x";
  const char *suffix= (UdmSQLDBFlags(db) & UDM_SQL_HAVE_STDHEX) ? "'"  : 
                      (UdmSQLDBFlags(db) & UDM_SQL_HAVE_BLOB_AS_HEX) ? "'" : "";
  udm_timer_t ticks= UdmStartTimer();


  /* Put information into a user-defined cache */
  /* TODO: add its own debug logging inside this func: */
  if (UDM_OK != (rc= UdmExecuteUserCacheQuery(Indexer, db, Query)))
    return rc;
   
  if (!prevcache)
    return UDM_OK;

  UdmLog(Indexer, UDM_LOG_DEBUG,
         "Start UdmQueryCachePut %d documents", (int) Query->URLData.nitems);
    
  if (UdmSQLDBType(db) == UDM_DB_PGSQL)
  {
    prefix= "'";
    suffix= "'";
  }
  
  id= QueryCacheID(Indexer, &Query->Res.Vars);
  sprintf(idbuf, "%08X-%08X", id, tm);

  UdmDSTRInit(&wwl, 256);
  UdmQueryToXML(Query, &wwl);

  wwl_length_escaped= UdmDSTRLength(&wwl) * 5;

  UdmDSTRInit(&bindata, 2048);
  UdmDSTRInit(&buf, 256);
  UdmDSTRRealloc(&buf, nbytes + 128 + wwl_length_escaped);
  UdmDSTRAppendf(&buf,
                 "INSERT INTO qcache (id, tm, doclist, wordinfo) "
                 "VALUES (%d, %d, ", id, tm);

  if (UdmSQLDBFlags(db) & UDM_SQL_HAVE_BIND_BINARY)
  {
    int bindtype= UDM_SQLTYPE_LONGVARBINARY;
    
    UdmDSTRAppendSTR(&buf, UdmDBSQLParamPlaceHolder(db, 1));
    UdmDSTRAppendSTR(&buf, ",");
    UdmDSTRAppendSTR(&buf, UdmDBSQLParamPlaceHolder(db, 2));
    UdmDSTRAppendSTR(&buf, ")");

    UdmURLDataListEncode(Indexer, &Query->URLData, NULL, &bindata);
    
    if (UDM_OK != (rc= UdmDBSQLPrepare(Indexer, db, UdmDSTRPtr(&buf))) ||
        UDM_OK != (rc= UdmDBSQLBindParameter(Indexer, db, 1,
                                             UdmDSTRPtr(&bindata),
                                             (int) UdmDSTRLength(&bindata),
                                             bindtype)) ||
        UDM_OK != (rc= UdmDBSQLBindParameter(Indexer, db, 2,
                                             UdmDSTRPtr(&wwl),
                                             (int) UdmDSTRLength(&wwl),
                                             UDM_SQLTYPE_VARCHAR)) ||
        UDM_OK != (rc= UdmDBSQLExecute(Indexer, db)) ||
        UDM_OK != (rc= UdmDBSQLStmtFree(Indexer, db)))
      goto ret;
  }
  else
  {
    UdmDSTRAppendSTR(&buf, prefix);
    
    /* Append coord information */
    UdmURLDataListEncode(Indexer, &Query->URLData, db, &buf);

    UdmDSTRAppendSTR(&buf, suffix);

    /* Append word information */
    UdmDSTRAppend(&buf, ",'", 2);
    UdmDBSQLEscStr(Indexer, db, buf.Val.str + buf.Val.length,
                   UdmDSTRPtr(&wwl), UdmDSTRLength(&wwl)); /* WWList XML*/
    nbytes= strlen(UdmDSTRPtr(&buf) + UdmDSTRLength(&buf));
    buf.Val.length+= nbytes;
    UdmDSTRAppend(&buf, "')", 2);
    
    if (UDM_OK != (rc= UdmDBSQLQuery(Indexer, db, NULL, UdmDSTRPtr(&buf))))
      goto ret;
  }

  UdmVarListReplaceStr(&Query->Res.Vars, "qid", idbuf);

ret:

  UdmDSTRFree(&wwl);
  UdmDSTRFree(&buf);
  UdmDSTRFree(&bindata);

  UdmLog(Indexer, UDM_LOG_DEBUG, "Stop  UdmQueryCachePut: %.2f",
         UdmStopTimer(&ticks));

  return rc;
}


udm_rc_t
UdmApplyCachedQueryLimit(UDM_AGENT *query, UDM_URLSCORELIST *ScoreList, UDM_DB *db)
{
  UDM_QUERY CachedQuery;
  UDM_URLDATALIST *CachedData= &CachedQuery.URLData;
  const char *pqid= UdmVarListFindStr(&query->Conf->Vars, "pqid", NULL);
  UdmQueryInit(&CachedQuery);
  if (pqid && UDM_OK == UdmLoadCachedQueryWords(query, &CachedQuery, db, pqid))
  {
    UdmLog(query, UDM_LOG_DEBUG,
           "Start applying pqid limit: %d docs", (int) CachedData->nitems);
    if (CachedData->nitems)
    {
      size_t i, to;
      UdmURLDataListSort(CachedData);
      for (to= 0, i= 0; i < ScoreList->nitems; i++)
      {
        if (UdmURLDataListSearch(CachedData, ScoreList->Item[i].url_id))
        {
          if (to != i)
            ScoreList->Item[to]= ScoreList->Item[i];
          to++;
        }
      }
      ScoreList->nitems= to;
    }
    else
    {
      ScoreList->nitems= 0;
    }
    UdmLog(query, UDM_LOG_DEBUG,
           "Stop applying pqid limit: %d docs", (int) ScoreList->nitems);
  }
  UdmQueryFree(&CachedQuery);
  return UDM_OK;
}

#endif /* HAVE_SQL */
