/* Copyright (C) 2000-2015 Lavtech.com corp. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
*/

#include "udm_config.h"

#include <stdio.h>
#include <sys/types.h>
#include <stdlib.h>
#include <string.h>

#include "udm_common.h"
#include "udm_word.h"
#include "udm_doc.h"
#include "udm_utils.h"
#include "udm_result.h"
#include "udm_parsehtml.h"
#include "udm_parsexml.h"
#include "udm_vars.h"
#include "udm_searchtool.h"
#include "udm_boolean.h"

#include "udm_db.h" /* for UdmDocAction */


void UdmResultInit(UDM_RESULT *Res)
{
  bzero((void*)Res, sizeof(UDM_RESULT));
}


UDM_API(void)
UdmResultFree(UDM_RESULT * Res)
{
  size_t i;
  if(!Res)return;
  UdmVarListFree(&Res->Vars);
  UdmWideWordListFree(&Res->WWList);
  if(Res->Doc)
  {
    for(i=0;i<Res->num_rows;i++)
    {
      UdmDocFree(&Res->Doc[i]);
    }
    UdmFree(Res->Doc);
  }
  bzero((void*)Res, sizeof(*Res));
}


UDM_API(size_t) UdmResultNumRows(UDM_RESULT *Res)
{
  return Res->num_rows;
}


UDM_API(size_t)
UdmResultFirst(UDM_RESULT *Res)
{
  return (size_t) UdmVarListFindInt(&Res->Vars, "first", 1) - 1;
}


UDM_API(size_t)
UdmResultLast(UDM_RESULT *Res)
{
  return (size_t) UdmVarListFindInt(&Res->Vars, "last", 1) - 1;
}


UDM_API(size_t)
UdmResultTotalFound(UDM_RESULT *Res)
{
  return (size_t) UdmVarListFindInt(&Res->Vars, "total", 0);
}


UDM_API(udm_rc_t)
UdmResultToVarList(UDM_VARLIST *Vars, const UDM_RESULT *Res)
{
  char search_time[100];
  /* Convert milliseconds to seconds */
  udm_snprintf(search_time, sizeof(search_time), "%.3f",
               ((double) UdmVarListFindInt(&Res->Vars, "SearchTime", 0))/1000);
  if (UDM_OK != UdmVarListReplaceStr(Vars, "SearchTime", search_time))
    return UDM_ERROR;
  return UdmVarListReplaceLst(Vars, &Res->Vars, NULL, "*");
}

/***************************************************************/
udm_rc_t
UdmResultAppendAndInit(UDM_RESULT *Res)
{
  size_t nbytes= (Res->num_rows + 1) * sizeof(UDM_DOCUMENT);
  UDM_DOCUMENT *tmp;
  if (!(tmp= UdmRealloc(Res->Doc, nbytes)))
    return UDM_ERROR;
  Res->Doc= tmp;
  UdmDocInit(&Res->Doc[Res->num_rows]);
  Res->num_rows++;
  return UDM_OK;
}


udm_rc_t
UdmResultAppendDoc(UDM_RESULT *Res, UDM_DOCUMENT *Doc)
{
  size_t nbytes= (Res->num_rows + 1) * sizeof(UDM_DOCUMENT);
  UDM_DOCUMENT *tmp;
  if (!(tmp= UdmRealloc(Res->Doc, nbytes)))
    return UDM_ERROR;
  Res->Doc= tmp;
  Res->Doc[Res->num_rows]= Doc[0];
  Res->num_rows++;
  return UDM_OK;
}


/*****************************************************************/
#define UDM_DF_SIZE 64

typedef struct
{
  int state;
  UDM_AGENT *Agent;
  UDM_WIDEWORD WW;
  UDM_DOCUMENT Doc;
  UDM_QUERY *Query;
  UDM_CHARSET *cs;
  uint4       score;
  uint4       per_site;
  urlid_t     site_id;
  char date_format[UDM_DF_SIZE];
} RES_PARSER_DATA;


static udm_rc_t
PD_ReplaceOrAppendStrn_hl0(RES_PARSER_DATA *D,
                        const char *name,
                        const char *val, size_t len)
{
  return UdmVarListReplaceOrAppendStrn(&D->Doc.Sections, name, val, len, 0);
}


static udm_rc_t
PD_ReplaceOrAppendStrn_hl1(RES_PARSER_DATA *D,
                           const char *name,
                           const char *val, size_t len)
{
  return UdmVarListReplaceOrAppendStrn(&D->Doc.Sections, name, val, len, 1);
}


static udm_rc_t
PD_ReplaceStrn(RES_PARSER_DATA *D,
               const char *name,
               const char *val, size_t len)
{
  return UdmVarListReplaceStrn(&D->Doc.Sections, name, val, len) ? UDM_OK : UDM_ERROR;
}


static udm_rc_t
PD_ReplaceLastModified(RES_PARSER_DATA *D,
                       const char *name,
                       const char *val, size_t len)
{
  time_t last_mod_time;
  char tmp[UDM_MAXTIMESTRLEN];
  len= len >= sizeof(tmp) ? sizeof(tmp) - 1 : len;
  memcpy(tmp, val, len);
  tmp[len]= '\0';
  last_mod_time= UdmHttpDate2Time_t(tmp);
  UdmVarListReplaceInt(&D->Doc.Sections, "Last-Modified-Timestamp", (int) last_mod_time);
  if ((len= strftime(tmp, sizeof(tmp)-1, D->date_format, localtime(&last_mod_time))))
    tmp[len]= '\0';
  else
    UdmTime_t2HttpStr(last_mod_time, tmp, sizeof(tmp));
  UdmVarListReplaceStr(&D->Doc.Sections, "Last-Modified", tmp);
  return UDM_OK;
}


typedef udm_rc_t
(*section_handler)(RES_PARSER_DATA *D,
                   const char *name,
                   const char *val, size_t len);




#define RES_MISC 0

struct udm_res_section_st
{
  int        state;
  size_t     length;
  const char *str;
  const char *section_name;
  section_handler handler;
};

#define UDM_RESSEC_TOTAL_RESULTS       100

#define UDM_RESSEC_WORD                1000
#define UDM_RESSEC_WORD_ID             1001
#define UDM_RESSEC_WORD_ORDER          1002
#define UDM_RESSEC_WORD_COUNT          1003
#define UDM_RESSEC_WORD_ORIGIN         1004
#define UDM_RESSEC_WORD_WEIGHT         1005
#define UDM_RESSEC_WORD_MATCH          1006
#define UDM_RESSEC_WORD_SECNO          1007
#define UDM_RESSEC_WORD_PHRLEN         1008
#define UDM_RESSEC_WORD_PHRPOS         1009
#define UDM_RESSEC_WORD_WORD           1010

#define UDM_RESSEC_ITEM                2000
#define UDM_RESSEC_ITEM_DESCR          2002
#define UDM_RESSEC_ITEM_SCORE          2005

#define UDM_RESSEC_ITEM_PERSITE        2011
#define UDM_RESSEC_ITEM_DESCR_B        2014

#define SECDEF(x)  ((size_t) (sizeof(x) - 1)), (x)
#define WSL "/rss/channel/mnoGoSearch:WordStatList"
#define WSI "/rss/channel/mnoGoSearch:WordStatList/mnoGoSearch:WordStatItem"
#define RCI(x) SECDEF("/rss/channel/item/" x)

static struct udm_res_section_st res_sec[]=
{
  {UDM_RESSEC_WORD,           SECDEF("/result/wordinfo/word"),        NULL, NULL},
  {UDM_RESSEC_WORD_ID,        SECDEF("/result/wordinfo/word@id"),     NULL, NULL},
  {UDM_RESSEC_WORD_ORDER,     SECDEF("/result/wordinfo/word@order"),  NULL, NULL},
  {UDM_RESSEC_WORD_COUNT,     SECDEF("/result/wordinfo/word@count"),  NULL, NULL},
  {UDM_RESSEC_WORD_ORIGIN,    SECDEF("/result/wordinfo/word@origin"), NULL, NULL},
  {UDM_RESSEC_WORD_WEIGHT,    SECDEF("/result/wordinfo/word@weight"), NULL, NULL},
  {UDM_RESSEC_WORD_MATCH,     SECDEF("/result/wordinfo/word@match"),  NULL, NULL},
  {UDM_RESSEC_WORD_SECNO,     SECDEF("/result/wordinfo/word@secno"),  NULL, NULL},
  {UDM_RESSEC_WORD_PHRLEN,    SECDEF("/result/wordinfo/word@phrlen"), NULL, NULL},
  {UDM_RESSEC_WORD_PHRPOS,    SECDEF("/result/wordinfo/word@phrpos"), NULL, NULL},
  {UDM_RESSEC_TOTAL_RESULTS,  SECDEF("/result/totalResults"),         NULL, NULL},
  
  {UDM_RESSEC_WORD,           SECDEF(WSI),           NULL, NULL},
  {UDM_RESSEC_WORD_ID,        SECDEF(WSI "@id"),     NULL, NULL},
  {UDM_RESSEC_WORD_WORD,      SECDEF(WSI "@word"),   NULL, NULL},
  {UDM_RESSEC_WORD_ORDER,     SECDEF(WSI "@order"),  NULL, NULL},
  {UDM_RESSEC_WORD_COUNT,     SECDEF(WSI "@count"),  NULL, NULL},
  {UDM_RESSEC_WORD_ORIGIN,    SECDEF(WSI "@origin"), NULL, NULL},
  {UDM_RESSEC_WORD_WEIGHT,    SECDEF(WSI "@weight"), NULL, NULL},
  {UDM_RESSEC_WORD_MATCH,     SECDEF(WSI "@match"),  NULL, NULL},
  {UDM_RESSEC_WORD_SECNO,     SECDEF(WSI "@secno"),  NULL, NULL},
  {UDM_RESSEC_WORD_PHRLEN,    SECDEF(WSI "@phrlen"), NULL, NULL},
  {UDM_RESSEC_WORD_PHRPOS,    SECDEF(WSI "@phrpos"), NULL, NULL},

  {UDM_RESSEC_TOTAL_RESULTS,  SECDEF("/rss/channel/openSearch:totalResults"), NULL, NULL},
  
  {UDM_RESSEC_ITEM,           SECDEF("/rss/channel/item"),         NULL, NULL},
  {UDM_RESSEC_ITEM_SCORE,     SECDEF("/rss/channel/item/score"),   NULL, NULL},
  {UDM_RESSEC_ITEM_PERSITE,   SECDEF("/rss/channel/item/persite"), NULL, NULL},

  /* Duplicate: */
  {0, RCI("updated"),         "Last-Modified",    PD_ReplaceLastModified},
  {0, RCI("pubDate"),         "Last-Modified",    PD_ReplaceLastModified},
  {0, RCI("last-modified"),   "Last-Modified",    PD_ReplaceLastModified},

  {0, RCI("title"),           "title",            PD_ReplaceOrAppendStrn_hl0},
  {0, RCI("title/b"),         "title",            PD_ReplaceOrAppendStrn_hl1},
  
  /* Duplicate: */
  {0, RCI("body"),            "body",             PD_ReplaceOrAppendStrn_hl0},
  {0, RCI("description"),     "body",             PD_ReplaceOrAppendStrn_hl0},
  {0, RCI("description/b"),   "body",             PD_ReplaceOrAppendStrn_hl1},
  
  /* Duplicate: */
  {0, RCI("link"),            "url",              PD_ReplaceStrn},
  {0, RCI("url"),             "url",              PD_ReplaceStrn},
  
  {0, RCI("id"),              "id",               PD_ReplaceStrn}, 
  {0, RCI("content-length"),  "Content-Length",   PD_ReplaceStrn},
  {0, RCI("content-type"),    "Content-Type",     PD_ReplaceStrn},
  {0, RCI("cached"),          "stored_href",      PD_ReplaceStrn},
  
  /* Duplicate: */
  {0, RCI("CachedCopyBase64"),"CachedCopyBase64", PD_ReplaceStrn},
  {0, RCI("cached-content"),  "CachedCopyBase64", PD_ReplaceStrn}, 
  
  {0, RCI("tag"),             "tag",              PD_ReplaceStrn},
  {0, RCI("crc32"),           "crc32",            PD_ReplaceStrn},
  {0, RCI("charset"),         "charset",          PD_ReplaceStrn},
  {0, RCI("status"),          "status",           PD_ReplaceStrn},

  {0, 0, NULL, NULL, NULL}
};


static struct udm_res_section_st *res_sec_find(const char *attr, size_t len)
{
  struct udm_res_section_st *s;
  for (s= res_sec; s->str; s++)
  {
    if (len == s->length && !strncasecmp(attr, s->str, len))
      return s;
  }
  return NULL;
}

                    
static udm_rc_t
ResFromXMLEnter(UDM_XML_PARSER *parser, const char *name, size_t l)
{
  RES_PARSER_DATA *D= (RES_PARSER_DATA*) parser->user_data;
  struct udm_res_section_st *st= res_sec_find(parser->attr,
                                              parser->attrend - parser->attr);
  D->state= st ? st->state : 0;
  if (D->state == UDM_RESSEC_WORD)
  {
    UdmWideWordInit(&D->WW);
    D->WW.Param.origin= UDM_WORD_ORIGIN_QUERY;
  }
  if (D->state == UDM_RESSEC_ITEM)
  {
    char dbuf[128];
    UdmDocInit(&D->Doc);
    snprintf(dbuf, 128, "%.5f", (float) 0);
    UdmVarListReplaceStr(&D->Doc.Sections, "Pop_Rank", dbuf);
  }
  return UDM_OK;
}


static udm_rc_t
ResFromXMLAddDocHook(UDM_XML_PARSER *parser)
{
  RES_PARSER_DATA *D= (RES_PARSER_DATA*) parser->user_data;
  size_t nbytes;
  D->Query->URLData.nitems++;
  D->Query->Res.num_rows++;
  nbytes= D->Query->Res.num_rows * sizeof(UDM_DOCUMENT);
  D->Query->Res.Doc= (UDM_DOCUMENT*) UdmRealloc(D->Query->Res.Doc, nbytes);
  D->Query->Res.Doc[D->Query->Res.num_rows-1]= D->Doc;
  bzero((void*)&D->Doc, sizeof(UDM_DOCUMENT));
  
  nbytes= D->Query->Res.num_rows * sizeof(UDM_URLDATA);
  D->Query->URLData.Item= (UDM_URLDATA*) UdmRealloc(D->Query->URLData.Item, nbytes);
  bzero((void*)&D->Query->URLData.Item[D->Query->Res.num_rows-1], sizeof(UDM_URLDATA));
  D->Query->URLData.Item[D->Query->Res.num_rows-1].url_id= D->Query->URLData.nitems-1;
  D->Query->URLData.Item[D->Query->Res.num_rows-1].score= D->score;
  D->Query->URLData.Item[D->Query->Res.num_rows-1].per_site= D->per_site;
  D->Query->URLData.Item[D->Query->Res.num_rows-1].site_id= D->site_id;
  
  D->score= 0;
  D->per_site= 0;
  D->site_id= 0;
  return UDM_OK;
}

static udm_rc_t
ResFromXMLAddDocHookImport(UDM_XML_PARSER *parser)
{
  RES_PARSER_DATA *Data= (RES_PARSER_DATA*) parser->user_data;
  UDM_DOCUMENT *D= &Data->Doc;
  UdmDocAction(Data->Agent, D, UDM_DOCCMD_RESTOREDATA);
  UdmVarListFree(&D->Sections);
  return UDM_OK;
}

static udm_rc_t
ResFromXMLLeave(UDM_XML_PARSER *parser, const char *name, size_t l)
{
  RES_PARSER_DATA *D= (RES_PARSER_DATA*) parser->user_data;
  struct udm_res_section_st *st= res_sec_find(parser->attr,
                                              parser->attrend - parser->attr);
  D->state= st ? st->state : 0;

  if (D->state == UDM_RESSEC_WORD)
  {
    if (!D->WW.Word.str)
    {
      D->WW.Word.str= UdmStrdup("<empty>");
      D->WW.Word.length= 7;
    }
    UdmWideWordListAddForStat(&D->Query->Res.WWList, &D->WW);
    UdmWideWordFree(&D->WW);
  }
  if (D->state == UDM_RESSEC_ITEM)
  {
    if (D->Query)
      ResFromXMLAddDocHook(parser);
    else
      ResFromXMLAddDocHookImport(parser);
  }
  /* fprintf(stderr, "leave: len=%d '%s'\n", l, name);*/
  return(UDM_OK);
}


static udm_rc_t
ResFromXMLValue(UDM_XML_PARSER *parser, const char *s, size_t len)
{
  RES_PARSER_DATA *D= (RES_PARSER_DATA*) parser->user_data;
  struct udm_res_section_st *st= res_sec_find(parser->attr,
                                              parser->attrend - parser->attr);
  if (!st)
  {
    /* Add user defined tags */
    if (!strncasecmp(parser->attr, UDM_CSTR_WITH_LEN("/rss/channel/item/")))
      UdmVarListReplaceStrn(&D->Doc.Sections, parser->attr + 18, s, len);
    return UDM_OK;
  }
  
  if (st->handler)
  {
    st->handler(D, st->section_name, s, len);
    return UDM_OK;
  }
  D->state= st->state;
  switch (D->state)
  {
    case  UDM_RESSEC_WORD         :
      UdmFree(D->WW.Word.str);
      D->WW.Word.str= udm_strndup(s, len);
      D->WW.Word.length= len;
      break;
    case  UDM_RESSEC_WORD_WORD    :
      UdmFree(D->WW.Word.str);
      D->WW.Word.str= udm_strndup(s, len);
      D->WW.Word.length= len;
      break;
    case  UDM_RESSEC_WORD_ID      :
      break;
    case  UDM_RESSEC_WORD_ORDER   :
      D->WW.Param.order= atoi(s);
      break;
    case  UDM_RESSEC_WORD_COUNT   :
      D->WW.Param.count= atoi(s);
      break;
    case  UDM_RESSEC_WORD_ORIGIN  :
      D->WW.Param.origin= atoi(s); /* TODO34: check if valid */
      break;
    case  UDM_RESSEC_WORD_WEIGHT  :
      D->WW.Param.weight= atoi(s);
      break;
    case  UDM_RESSEC_WORD_MATCH   :
      D->WW.Param.match_mode= atoi(s); /* TODO34: check if valid */
      break;
    case  UDM_RESSEC_WORD_SECNO   :
      D->WW.Param.secno= atoi(s);
      break;
    case  UDM_RESSEC_WORD_PHRLEN  :
      D->WW.Param.phrlen= atoi(s);
      break;
    case  UDM_RESSEC_WORD_PHRPOS  :
      D->WW.Param.phrpos= atoi(s);
      break;
    case UDM_RESSEC_ITEM_PERSITE:
      D->per_site= udm_strntod(s, len);
      break;
    case UDM_RESSEC_ITEM_SCORE:
      D->score= udm_strntod(s, len) * 1000 + 0.5;
      break;
    case UDM_RESSEC_TOTAL_RESULTS:
      D->Query->stats.total_found= atoi(s);
      break;
  }
  /*fprintf(stderr, "UdmXMLValue: st=%d '%.*s' name='%s'\n", D->state, len, s, parser->attr);*/
  return UDM_OK;
}


udm_rc_t
UdmQueryFromXML(UDM_AGENT *A, UDM_QUERY *Query,
                const char *str, size_t length, UDM_CHARSET *cs)
{
  udm_rc_t rc= UDM_OK;
  RES_PARSER_DATA Data;
  UDM_XML_PARSER parser;
  const char *date_format= UdmVarListFindStr(&A->Conf->Vars, "DateFormat",
                                             "%a, %d %b %Y, %X %Z");
  UdmXMLParserCreate(&parser);
  parser.flags |= UDM_XML_SKIP_TEXT_NORMALIZATION;
  bzero(&Data, sizeof(Data));
  Data.Agent= A;
  Data.Query= Query;
  Data.cs= cs;
  udm_snprintf(Data.date_format, UDM_DF_SIZE, "%s", date_format);

  UdmXMLSetUserData(&parser, &Data);
  UdmXMLSetEnterHandler(&parser, ResFromXMLEnter);
  UdmXMLSetLeaveHandler(&parser, ResFromXMLLeave);
  UdmXMLSetValueHandler(&parser, ResFromXMLValue);

  if (UDM_OK != (rc= UdmXMLParserExec(&parser, str, length)))
  {
    char err[256];    
    udm_snprintf(err, sizeof(err), 
                 "XML parsing error: %s at line %d pos %d",
                  UdmXMLErrorString(&parser),
                  (int) UdmXMLErrorLineno(&parser),
                  (int) UdmXMLErrorPos(&parser));
  }

  UdmXMLParserFree(&parser);
  return rc;
}
