/* Copyright (C) 2000-2015 Lavtech.com corp. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
*/

#include "udm_config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#include <sys/types.h>
#ifdef HAVE_SYS_PARAM_H
#include <sys/param.h>
#endif
#ifdef HAVE_SYS_SOCKET_H
#include <sys/socket.h>
#endif
#ifdef HAVE_NETDB_H
#include <netdb.h>
#endif
#ifdef HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif
#ifdef HAVE_ARPA_INET_H
#include <arpa/inet.h>
#endif
#include <errno.h>

#include "udm_common.h"
#include "udm_utils.h"
#include "udm_proto.h"
#include "udm_services.h"
#include "udm_agent.h"
#include "udm_db.h"
#include "udm_doc.h"
#include "udm_result.h"
#include "udm_sdp.h"
#include "udm_xmalloc.h"
#include "udm_searchtool.h"
#include "udm_vars.h"
#include "udm_word.h"
#include "udm_log.h"
#include "udm_host.h"
#include "udm_http.h"
#include "udm_url.h"
#include "udm_store.h"
#include "udm_query.h"

/*
#define DEBUG_SDP
*/

/*
#define DEBUG_SEARCH
*/

typedef struct
{
  UDM_QUERY Query;
  UDM_VARLIST Vars;  /* Connection params: Host, port, trackquery, etc */
  int errcode;
  char errstr[UDM_ERRSTR_SIZE];
} UDM_SDDB;


static void
UdmSDDBInit(UDM_SDDB *db)
{
  bzero((void*) db, sizeof(UDM_SDDB));
}

static void
UdmSDDBFree(UDM_SDDB *db)
{
  UdmQueryFree(&db->Query);
  UdmVarListFree(&db->Vars);
}


static inline UDM_SDDB *
UdmSDDB(UDM_DB *db)
{
  return (UDM_SDDB*) db->specific;
}


static inline UDM_VARLIST *
UdmSDDBVars(UDM_DB *db)
{
  return &UdmSDDB(db)->Vars;
}


/*
  Cache document list, so UDM_QUERYCMD_DOCINFO is handled from cache,
  without having to fetch a new HTTP response from a node.
*/
static void
UdmResultTakeoverDocList(UDM_QUERY *To, UDM_RESULT *From)
{
  UdmResultFree(&To->Res);
  UdmResultInit(&To->Res);
  To->Res.Doc= From->Doc;
  To->Res.num_rows= From->num_rows;
  From->Doc= NULL;
  From->num_rows= 0;
}


static udm_rc_t
UdmQueryAddDocInfoSearchd(UDM_AGENT *A, UDM_DB *db, UDM_QUERY *Query)
{
  size_t i;
  for (i= 0; i < UdmResultNumRows(&Query->Res); i++)
  {
    UDM_URLDATA *Data= &Query->URLData.Item[i + Query->stats.first];
    size_t Doc_dbnum= UDM_COORD2DBNUM(Data->score);

    if (&A->Conf->DBList.Item[Doc_dbnum] == db)
    {
      size_t id, num= Data->url_id;
      UDM_RESULT *SrcRes= A->Conf->DBList.nitems == 1 ? &Query->Res : &UdmSDDB(db)->Query.Res;
      UDM_VARLIST *DstSections= &Query->Res.Doc[i].Sections;
      id= UdmVarListFindInt(&SrcRes->Doc[num].Sections, "id", 0);
      Data->url_id= id;
      if (A->Conf->DBList.nitems > 1)
      {
        UdmVarListReplaceLst(DstSections, &SrcRes->Doc[num].Sections, NULL, "*");
      }
      else if (A->Conf->DBList.nitems == 1)
      {
        if (Query->stats.first)
        {
          UdmVarListFree(DstSections);
          UdmVarListReplaceLst(DstSections, &SrcRes->Doc[num].Sections, NULL, "*");
        }
      }
      UdmVarListReplaceInt(DstSections, "id", id);
      UdmVarListReplaceInt(DstSections, "DBOrder", num + 1);
    }
  }
  return UDM_OK;
}


static udm_rc_t
UdmVarListLog2(UDM_AGENT *A,UDM_VARLIST *V,int l,const char *pre)
{
  size_t h;
  if (UdmNeedLog(A->Conf, l))
  {
    for(h=0; h < V->nvars; h++)
    {
      const UDM_VAR *v= UdmVarListFindConstByIndex(V, h);
      UDM_CONST_STR strbuf, *str= UdmVarGetConstStr(v, &strbuf);
      UdmLog(A, l, "%s.%s: %.*s", pre, UdmVarName(v), (int) str->length, str->str);
    }
  }
  return UDM_OK;
}


/*
static char *
UdmParseOpenSearchAddr(UDM_VARLIST *Vars, const char *src)
{
  char *res;
  UDM_DSTR d;
  UdmDSTRInit(&d, 1024);
  UdmDSTRParse(&d, src, Vars);
  res= UdmStrdup(UdmDSTRPtr(&d));
  UdmDSTRFree(&d);
  return res;
}
*/

static udm_rc_t
UdmMakeNodeQueryString(UDM_AGENT *A, UDM_DSTR *dstr, const char *dbaddr)
{
  size_t i;
  const UDM_VARLIST *Vars= &A->Conf->Vars;
  size_t ps= UdmVarListFindInt(Vars, "ps", 10);
  size_t np= UdmVarListFindInt(Vars, "np", 0);
  size_t offs= UdmVarListFindInt(Vars, "offs", 0);
  int group_by_site= UdmVarListFindBool(Vars, "GroupBySite", UDM_FALSE) &&
                     (UdmVarListFindStr(Vars, "site", "")[0] == '\0');
  int group_by_site_factor= group_by_site ? 3 : 1;
  size_t newps= ((offs + ps * (np + 1)) * group_by_site_factor);

  UdmDSTRAppendSTR(dstr, dbaddr);
  if (!strchr(dbaddr, '?'))
    UdmDSTRAppend(dstr, "?", 1);
  for (i= 0; i < Vars->nvars; i++)
  {
    const UDM_VAR *Var= &Vars->Var[i][0];
    UDM_CONST_STR tmp, *str;
    /**
      Suppress all offset related variables, "ps" is calculated below.
      Suppress dbnum, it should be only in the fron-end query string
      and should not be in the node query string.
    */
    if (strncasecmp(UdmVarName(Var), UDM_CSTR_WITH_LEN("query.")) ||
        !strncasecmp(UdmVarName(Var), UDM_CSTR_WITH_LEN("query.ps")) ||
        !strncasecmp(UdmVarName(Var), UDM_CSTR_WITH_LEN("query.np")) ||
        !strncasecmp(UdmVarName(Var), UDM_CSTR_WITH_LEN("query.offs")) ||
        !strncasecmp(UdmVarName(Var), UDM_CSTR_WITH_LEN("dbnum")) ||
        !(str= UdmVarGetConstStr(Var, &tmp)))
      continue;
    UdmDSTRAppend(dstr, "&", 1);
    UdmDSTRAppendSTR(dstr, UdmVarName(Var) + 6);
    UdmDSTRAppend(dstr, "=", 1);
    UdmDSTRAppendURLEncode(dstr, str->str, str->length);
  }
  if (newps)
  {
    UdmDSTRAppend(dstr, UDM_CSTR_WITH_LEN("&ps="));
    UdmDSTRAppendf(dstr, "%d", (int) newps);
  }
  return UDM_OK;
}


static udm_rc_t
UdmFindWordsSD(UDM_AGENT *A, UDM_DB *db,
               UDM_QUERY *Query, udm_bool_t takeover_doclist)
{
  UDM_DOCUMENT Inc;
  const char *host= UdmVarListFindStr(UdmSDDBVars(db), "DBHost", "localhost");
  const char *dbaddr= UdmVarListFindStr(UdmSDDBVars(db), "DBAddr", "");
  int port= UdmVarListFindInt(UdmSDDBVars(db), "DBPort", 80);
  udm_rc_t rc;
  udm_timer_t ticks;
  UDM_CONST_STR content;

  UdmDocInit(&Inc);
  if (!UdmHTTPBufPtr(&Inc.Buf) &&
      UDM_OK != UdmHTTPBufAlloc(&Inc.Buf, UDM_MAXDOCSIZE))
    return UDM_ERROR;
  Inc.Spider.read_timeout= UdmVarListFindInt(&A->Conf->Vars, "ReadTimeOut", UDM_READ_TIMEOUT);

  UDM_GETLOCK(A, UDM_LOCK_CONF);
  if (strncasecmp(dbaddr, UDM_CSTR_WITH_LEN("file:")))
  {
    UDM_DSTR dstr;
    UdmDSTRInit(&dstr, 256);
    UdmMakeNodeQueryString(A, &dstr, dbaddr);
    UdmURLParse(&Inc.CurURL, UdmDSTRPtr(&dstr));
    UdmLog(A, UDM_LOG_ERROR, "DBAddr: %s", UdmDSTRPtr(&dstr));
    UdmDSTRFree(&dstr);
  }
  else
  {
    if (strchr(dbaddr, '?') || strchr(dbaddr, '&'))
    {
      char *tmp= UdmStrdup(dbaddr), *param;
      if (!tmp)
        return UDM_ERROR;
      if ((param= strchr(tmp, '?')) ||
          (param= strchr(tmp, '&')))
        *param= '\0';
      UdmURLParse(&Inc.CurURL, tmp);
      UdmFree(tmp);
    }
    else
    {
      UdmURLParse(&Inc.CurURL, dbaddr);
    }
  }
  UDM_RELEASELOCK(A, UDM_LOCK_CONF);

  if (!strcmp(Inc.CurURL.schema, "http"))
  {
    UdmVarListReplaceStr(&Inc.RequestHeaders, "Host", host);
    Inc.connp.hostname= (char*)UdmStrdup(host);
    Inc.connp.port= port;
    
    if (UDM_OK != (rc= UdmHostLookup2(A, &A->Conf->Hosts, &Inc.connp)))
    {
      sprintf(A->Conf->errstr, "Host lookup failed: '%s'", host);
      return rc;
    }
  }
  
  ticks= UdmStartTimer();
  rc= UdmGetURL(A, &Inc);
  UdmLog(A, UDM_LOG_DEBUG, "Received searchd response: %.2f", UdmStopTimer(&ticks));
  
  if (UDM_OK != rc)
    return rc;
  
  UdmParseHTTPResponse(A, &Inc);
  if (UdmHTTPBufContentToConstStr(&Inc.Buf, &content))
    return UDM_ERROR;
  UdmVarListLog2(A, &Inc.Sections, UDM_LOG_DEBUG, "Response");
  
  UdmLog(A, UDM_LOG_DEBUG, "Start parsing results");
  ticks= UdmStartTimer();
  
  UdmQueryFromXML(A, Query, content.str, content.length, A->Conf->lcs);
  UdmDocFree(&Inc);
  UdmLog(A, UDM_LOG_DEBUG, "Stop parsing results: %.2f", UdmStopTimer(&ticks));
  UdmLog(A, UDM_LOG_DEBUG, "searchd: %d rows, %d totalResults",
                           (int) UdmResultNumRows(&Query->Res),
                           (int) Query->stats.total_found);
  if (A->Conf->DBList.nitems > 1 && takeover_doclist)
    UdmResultTakeoverDocList(&UdmSDDB(db)->Query, &Query->Res);
  return UDM_OK;
}


static udm_rc_t
UdmCachedCopyUnpack(UDM_DOCUMENT *Doc, const char *sval, size_t l)
{
#ifdef HAVE_ZLIB
  return UdmHTTPBufBase64Decode(&Doc->Buf, sval, l);
#endif
  return UDM_ERROR;
}


static udm_rc_t
UdmDocActionSD(UDM_AGENT *A, UDM_DB *db, UDM_DOCUMENT *D, udm_doccmd_t cmd)
{
  udm_rc_t rc= UDM_OK;
  if (cmd == UDM_DOCCMD_GET_CACHED_COPY)
  {
    UDM_QUERY Query;
    UdmQueryInit(&Query);
    /*
       UDM_LOCK_CONF mutex is locked here,
       UdmFindWordsSearchd expects unlocked mutex.
       Let's unlock it and then lock again.
    */
    UDM_RELEASELOCK(A, UDM_LOCK_CONF);
    rc= UdmFindWordsSD(A, db, &Query, UDM_FALSE); /*Keep Docs in Query*/
    UDM_GETLOCK(A, UDM_LOCK_CONF);
    if (UdmResultNumRows(&Query.Res))
    {
      UDM_DOCUMENT *tmpDoc= &Query.Res.Doc[0];
      const UDM_VAR *cc= UdmVarListFind(&tmpDoc->Sections, "CachedCopyBase64");
      if (cc && !UdmHTTPBufContent(&D->Buf))
      {
        UDM_CONST_STR valuebuf, *value= UdmVarGetConstStr(cc, &valuebuf);
        if (UDM_OK == (rc= UdmHTTPBufAlloc(&D->Buf, UDM_MAXDOCSIZE)))
          UdmCachedCopyUnpack(D, value->str, value->length);
        UdmVarListAddLst(&D->Sections, &tmpDoc->Sections, NULL, "*");
      }
    }
    UdmResultFree(&Query.Res);
  }
  return rc;
}


static udm_rc_t
not_implemented(UDM_DB *db, const char *action, int code)
{
  udm_snprintf(UdmSDDB(db)->errstr, sizeof(UdmSDDB(db)->errstr),
               "%s(%d): not implemented", action, code);
  return UDM_ERROR;
}


static udm_rc_t
UdmDBActionSD(UDM_AGENT *A, UDM_DB *db, udm_dbcmd_t cmd)
{
  return not_implemented(db, "DBAction", cmd);
}


static udm_rc_t
UdmTrackSD(UDM_AGENT *A, UDM_DB *db, UDM_QUERY *Query)
{
  /*
   Be silent if trackquery is not set "yes", complain otherwise.
   TODO34: Make it possible to track into a different (e.g. SQL) database.
  */
  udm_bool_t trackquery= UdmVarListFindBool(UdmSDDBVars(db), "trackquery", UDM_FALSE);
  return trackquery ? not_implemented(db, "ResultAction", UDM_QUERYCMD_TRACK) : UDM_OK;
}


static udm_rc_t
UdmServerActionSD(UDM_AGENT *A, UDM_DB *db, UDM_SERVERLIST *Srv, udm_srvcmd_t cmd)
{
  return not_implemented(db, "ServerAction", cmd);
}


static udm_rc_t
UdmQueryWordForms(UDM_AGENT *A, UDM_DB *db, UDM_QUERY *Query)
{
  const char *sql= UdmVarListFindStr(&A->Conf->Vars, "SQLWordForms", NULL);
  if (sql) /* TODO34 */
  {
    udm_snprintf(UdmSDDB(db)->errstr, sizeof(UdmSDDB(db)->errstr),
                 "SQLWordForms is not supported by this DBAddr type");
    UdmSDDB(db)->errcode= 1;
    return UDM_ERROR;
  }
  return UDM_OK;
}


static udm_rc_t
UdmHrefActionSD(UDM_AGENT *A, UDM_DB *db, UDM_HREF *Href, udm_hrefcmd_t cmd)
{
  return not_implemented(db, "HrefAction", cmd);
}


static udm_bool_t
UdmSDDBKnownSchema(const char *schema)
{
  return UDM_TEST(!strcasecmp(schema, "searchd") ||
                  !strcasecmp(schema, "http") ||
                  !strcasecmp(schema, "file"));
}


static udm_rc_t
UdmDBSetAddrSDFromURL(UDM_DB *db, UDM_URL *addr, const char *dbaddr)
{
  if (!(db->specific= UdmMalloc(sizeof(UDM_SDDB))))
    return UDM_ERROR;

  UdmSDDBInit(UdmSDDB(db));
  db->dbhandler= &udm_dbhandler_searchd;
  UdmVarListReplaceStr(UdmSDDBVars(db), "DBAddr", dbaddr);
  UdmVarListReplaceStr(UdmSDDBVars(db), "DBHost", addr->hostname);
  if (addr->port)
    UdmVarListReplaceInt(UdmSDDBVars(db), "DBPort", addr->port);
  return UDM_OK;
}


static udm_rc_t
UdmDBSetAddrSD(UDM_DB *db, const char *dbaddr)
{
  udm_rc_t rc= UDM_NOTARGET;
  UDM_URL addr;
  UdmURLInit(&addr);
  if (!UdmURLParse(&addr, dbaddr) && addr.schema &&
       UdmSDDBKnownSchema(addr.schema))
    rc= UdmDBSetAddrSDFromURL(db, &addr, dbaddr);
  UdmURLFree(&addr);
  
  return rc;
}


static udm_rc_t
UdmDBCloseSD(UDM_DB *db)
{
  if (db->specific)
  {
    UdmSDDBFree(UdmSDDB(db));
    UDM_FREE(db->specific);
  }
  return UDM_OK;
}


static udm_rc_t
UdmDBInfoSD(UDM_DB *db, void *dst, size_t dstlen, size_t *bytes_written,
            udm_dbinfo_t info)
{
  switch (info)
  {
    case UDM_DBINFO_IS_THREAD_SAFE:
      if (!dstlen)
        return UDM_ERROR;
      ((udm_bool_t *) dst)[0]= UDM_TRUE;
      *bytes_written= 1;
      return UDM_OK;
    case UDM_DBINFO_ERRCODE:
      if (dstlen < sizeof(int))
        return UDM_ERROR;
      ((int *)dst)[0]= UdmSDDB(db)->errcode;
      *bytes_written= sizeof(int);
      return UDM_OK;
    case UDM_DBINFO_ADDR:
    {
      const char *dbaddr= UdmVarListFindStr(UdmSDDBVars(db), "DBAddr", "<noaddr>");
      if (dstlen < 1)
        return UDM_ERROR;
      *bytes_written= udm_snprintf((char*) dst, dstlen, "%s", dbaddr);
      return UDM_OK;
    }
    case UDM_DBINFO_ERRSTR:
    {
      if (dstlen < 1)
        return UDM_ERROR;
      *bytes_written= udm_snprintf((char*) dst, dstlen, "%s", UdmSDDB(db)->errstr);
      return UDM_OK;
    }
  }
  return UDM_ERROR;
}


static udm_rc_t
UdmQueryActionSD(UDM_AGENT *A, UDM_DB *db,
                 UDM_QUERY *Query, udm_querycmd_t cmd)
{
  switch (cmd)
  {
    case UDM_QUERYCMD_FINDWORDS: return UdmFindWordsSD(A, db, Query, UDM_TRUE);
    case UDM_QUERYCMD_DOCINFO:   return UdmQueryAddDocInfoSearchd(A, db, Query);
    case UDM_QUERYCMD_TRACK:     return UdmTrackSD(A, db, Query);
    case UDM_QUERYCMD_SUGGEST:   return UDM_OK;
    case UDM_QUERYCMD_CLONES:    return UDM_OK;
    case UDM_QUERYCMD_WORDFORMS: return UdmQueryWordForms(A, db, Query);
    case UDM_QUERYCMD_TARGETS:
    case UDM_QUERYCMD_CLEAR:
    case UDM_QUERYCMD_INDEX:
    case UDM_QUERYCMD_EXPIRE:
    case UDM_QUERYCMD_REFERERS:
    case UDM_QUERYCMD_DUMPDATA:
    case UDM_QUERYCMD_REWRITE_URLDATA:
    case UDM_QUERYCMD_REWRITE_LIMITS:
    case UDM_QUERYCMD_REWRITE_POPULARITY:
    case UDM_QUERYCMD_EXPORT:
    case UDM_QUERYCMD_WORDSTAT:
    case UDM_QUERYCMD_STATISTICS:
      return not_implemented(db, "QueryAction", cmd);
  }
  return UDM_ERROR;
}


const UDM_DBHANDLER udm_dbhandler_searchd=
{
  UdmDBSetAddrSD,      /*udm_rc_t (*Init)          (UDM_DB *db, const char *addr);*/
  UdmDBCloseSD,        /*udm_rc_t (*Close)         (UDM_DB *db); */
  UdmDBInfoSD,         /*udm_rc_t (*Info)          (UDM_DB *db, void *dst, size_t dstlen, size_t *bytes_written, udm_dbhandler_info_t info);*/
  UdmQueryActionSD,    /*udm_rc_t (*QueryAction)   (UDM_AGENT *A, UDM_DB *db, UDM_QUERY *Query, udm_querycmd_t cmd);*/
  UdmDBActionSD,       /*udm_rc_t (*DBAction)      (UDM_AGENT *A, UDM_DB *db, udm_dbcmd_t cmd);*/
  UdmDocActionSD,      /*udm_rc_t (*DocumentAction)(UDM_AGENT *A, UDM_DB *db, UDM_DOCUMENT *D, udm_doccmt_t cmd);*/
  UdmHrefActionSD,     /*udm_rc_t (*HrefAction)    (UDM_AGENT *A, UDM_DB *db, UDM_HREF *Href, udm_hrefcmd_t cmd);*/
  UdmServerActionSD,   /*udm_rc_t (*ServerAction)  (UDM_AGENT *A, UDM_DB *db, UDM_SERVERLIST *Srv, udm_srvcmd_t cmd);*/
};
