/* Copyright (C) 2000-2015 Lavtech.com corp. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
*/

#include "udm_config.h"

#ifdef HAVE_SQL

/*
#define DEBUG_SQL
*/

#define DEBUG_ERR_QUERY



#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#include "udm_common.h"
#include "udm_utils.h"
#include "udm_vars.h"
#include "udm_sqldbms.h"
#include "udm_xmalloc.h"

#if (HAVE_PGSQL)
#include <libpq-fe.h>
#endif

#ifdef WIN32
#include <process.h>
#endif

static char udm_hex_digits[]= "0123456789ABCDEF";


void DecodeHexStr (const char *src, UDM_STR *dst, size_t size)
{
  if (!(dst->str= (char*) UdmMalloc(size / 2 + 1)))
  {
    dst->length= 0;
    return;
  }
  dst->length= UdmHexDecode(dst->str, src, size);
  dst->str[dst->length]= '\0';
}


const char *
UdmSQLTypeToStr(udm_sqltype_t type)
{
  switch (type)
  {
    case UDM_SQLTYPE_LONGVARBINARY : return "LONGVARBINARY";
    case UDM_SQLTYPE_LONGVARCHAR   : return "LONGVARCHAR";
    case UDM_SQLTYPE_VARCHAR       : return "VARCHAR";
    case UDM_SQLTYPE_INT32         : return "INT";
    case UDM_SQLTYPE_UNKNOWN       : break;
  }
  return "UNKNOWN_TYPE";
}



/***************************************************************/


udm_rc_t
UdmSQLResFreeGeneric(UDM_SQLRES *res)
{
  size_t i;
  size_t nitems;
  if(res)
  {
    if(res->Items)
    {
      nitems = res->nCols * res->nRows;
      for(i=0;i<nitems;i++)
      {
        if (res->Items[i].str)
          UDM_FREE(res->Items[i].str);
      }
      UDM_FREE(res->Items);
    }
  }
  return UDM_OK;
}




/***********************************************************************/

/*
 *   Wrappers for different databases
 *
 *   UdmDBEscStr();
 *   UdmSQLBinEscStr();
 *   UdmSQLQuery();
 *   UdmSQLValue();
 *   UdmSQLLen();
 *   UdmSQLFree();
 *   UdmSQLClose();
 */  


udm_rc_t
UdmSQLDropTableIfExists(UDM_SQL *db, const char *name)
{
  char qbuf[128];
  udm_rc_t rc;
  int have_if_exists= db->flags & UDM_SQL_HAVE_DROP_IF_EXISTS;
  const char *if_exists= have_if_exists ? "IF EXISTS " : "";
  
  if (db->DBType == UDM_DB_MSSQL)
  {
    /*
    Another option:
    IF OBJECT_ID(N'tempdb..#temptable', N'U') IS NOT NULL 
    DROP TABLE #temptable;
    */
    udm_snprintf(qbuf, sizeof(qbuf),
                 "IF EXISTS "
                 "(SELECT TABLE_NAME FROM INFORMATION_SCHEMA.TABLES "
                 "WHERE TABLE_NAME='%s') "
                 "DROP TABLE %s", name, name);
    return UdmSQLQuery(db, NULL, qbuf);
  }
  if (!have_if_exists)
    db->flags|= UDM_SQL_IGNORE_ERROR;
  udm_snprintf(qbuf, sizeof(qbuf), "DROP TABLE %s%s", if_exists, name);
  rc= UdmSQLQuery(db, NULL, qbuf);
  if (!have_if_exists)
    db->flags^= UDM_SQL_IGNORE_ERROR;

  /*
    TODO:
    
    SQLite3:  DROP TABLE IF EXISTS t1; (DONE)
    Sybase ?  Perhaps simimalr to MSSQL
    Oracle ?
    Ibase  ?
    Mimer  ?
    DB2    ?
  */
  return rc;
}


/*
 Bind type for long varchar columns, like urlinfo.sval
*/
udm_sqltype_t
UdmSQLLongVarCharBindType(UDM_SQL *db)
{
  udm_sqldbtype_t dbtype= db->DBType;
  int bindtype= 
      db->DBDriver == UDM_DBAPI_ORACLE8 ? UDM_SQLTYPE_LONGVARCHAR :
         (dbtype == UDM_DB_MSSQL   ||
          dbtype == UDM_DB_SYBASE  ||
          dbtype == UDM_DB_SQLITE3 ||
          dbtype == UDM_DB_MONETDB ?
                  UDM_SQLTYPE_VARCHAR : UDM_SQLTYPE_LONGVARCHAR);
#ifdef HAVE_UNIXODBC
  if (dbtype == UDM_DB_SYBASE && db->DBDriver == UDM_DBAPI_ODBC)
    bindtype= UDM_SQLTYPE_LONGVARCHAR;
#endif
  return bindtype;
}


udm_rc_t
UdmSQLTableTruncateOrDelete(UDM_SQL *db, const char *name)
{
  char qbuf[128];
  if (db->flags & UDM_SQL_HAVE_TRUNCATE)
    udm_snprintf(qbuf, sizeof(qbuf), "TRUNCATE TABLE %s", name);
  else
    udm_snprintf(qbuf, sizeof(qbuf), "DELETE FROM %s", name);
  return UdmSQLQuery(db,NULL,qbuf);
}


static size_t
UdmSQLEscStrMonet(UDM_SQL *db, char *to, const char *from, size_t len)
{
  char *to0= to;
  /* Escape single quote with single quote, backslash with backslash */
  for ( ; len && *from; from++, len--)
  {
    switch (*from)
    {
      case '\\':
      case '\'':
       /* Note that no break here!*/
       *to++= *from;
      default:
       *to++= *from;
    }
  }
  *to= '\0';
  return to - to0;
}


static size_t
UdmSQLEscStrStandard(UDM_SQL *db, char *to,const char *from,size_t len)
{
  char *to0= to;
  for ( ; len && *from; from++, len--)
  {
    switch (*from)
    {
      case '\'':
       /* Note that no break here!*/
       *to++= *from;
      default:
       *to++= *from;
    }
  }
  *to= '\0';
  return to - to0;
}


static size_t
UdmSQLEscStrPgSQL(UDM_SQL *db, char *to,const char *from,size_t len)
{
  char *to0= to;
  udm_sqldbtype_t DBType= db->DBType;
  const char *fend= from + len;

  while (*from && from < fend)
  {
#ifdef WIN32
    /*
      A workaround against a bug in SQLExecDirect
      in PostgreSQL ODBC driver. It produces
      an error when meets a question mark in a string
      constant. For some reasons question marks are
      considered as parameters. This should not happen
      with SQLExecDirect.
      Let's escape question mark using \x3F notation.
    */
    if (DBType == UDM_DB_PGSQL &&
        (*from == '?' ||
         *from == '{' ||
         *from == '}'))
    {
      *to++= '\\';
      *to++= 'x';
      *to++= '3';
      *to++= 'F';
      from++;
      continue;
    }
#endif
    /*
    "ODBC escape convert error" is returned when '{' or '}' are not escaped.
    */
    if (DBType == UDM_DB_PGSQL && (*from == '{' || *from == '}'))
    {
      *to++= '\\';
      *to++= 'x';
      *to++= '7';
      *to++= udm_hex_digits[*from & 0x0F];
      from++;
      continue;
    }
    switch(*from)
    {
      case '\'':
      case '\\':
        *to='\\';to++;
      default:*to=*from;
    }
    to++;from++;
  }
  *to= '\0';
  return to - to0;
}


size_t
UdmSQLEscStrGeneric(UDM_SQL *db, char *to,const char *from,size_t len)
{
  switch (db->DBType)
  {
    case UDM_DB_MYSQL:
      return UdmSQLEscStrPgSQL(db, to, from, len);
    case UDM_DB_MONETDB:
      return UdmSQLEscStrMonet(db, to, from, len);
    case UDM_DB_PGSQL:
      /*
        Older version also did PgSQL-style escaping for:
        UDM_DB_SOLID, UDM_DB_VIRT, UDM_DB_ORACLE7.
        Restore PgSQL-style escaping if there are trobles with 
      */
      return (db->version < 90000) ?
             UdmSQLEscStrPgSQL(db, to, from, len) :
             UdmSQLEscStrStandard(db, to, from, len);
    default:
      return UdmSQLEscStrStandard(db, to, from, len);
  }
}


size_t
UdmSQLEscStr(UDM_SQL *db, char *to, const char *from, size_t len)
{
  UDM_ASSERT(from);
  UDM_ASSERT(to);
  return db->handler.EscStr(db, to, from, len);
}


static int
UdmSQLEscStrLength(UDM_SQL *db, size_t srclen)
{
  return ((db->DBType == UDM_DB_PGSQL) ? 4 : 2) * srclen + 1;
}


char *
UdmSQLEscStrAlloc(UDM_SQL *db, const char *src, size_t srclen)
{
  char *dst;
  if(!src ||
     !(dst= (char*) UdmMalloc(UdmSQLEscStrLength(db, srclen))))
    return NULL;
  db->handler.EscStr(db, dst, src, srclen);
  return dst;
}


/*
  SQL-Escape string to DSTR
  dstr - a pointer to a initialized DSTR
*/
udm_rc_t
UdmSQLEscDSTR(UDM_SQL *db, UDM_DSTR *dstr, const char *src, size_t srclen)
{
  udm_rc_t rc;
  if (UDM_OK != (rc= UdmDSTRAlloc(dstr, UdmSQLEscStrLength(db, srclen))))
    return rc;
  dstr->Val.length= db->handler.EscStr(db, dstr->Val.str, src, srclen);
  return rc;
}



static char
dangerous_character[256]=
{
/*00*/  1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
/*10*/  1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
/*20*/  0,1,1,1,1,0,1,1,1,1,0,0,0,0,0,0,  /*  !"#$%&'()*+,-./ */
/*30*/  0,0,0,0,0,0,0,0,0,0,0,1,1,0,1,0,  /* 0123456789:;<=>? */
/*40*/  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,  /* @ABCDEFGHIJKLMNO */
/*50*/  0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,0,  /* PQRSTUVWXYZ[\]^_ */
/*60*/  1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,  /* `abcdefghijklmno */
/*70*/  0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,  /* pqrstuvwxyz{|}~  */
/*80*/  3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,
/*90*/  3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,
/*A0*/  3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,
/*B0*/  3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,
/*C0*/  3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,
/*D0*/  3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,
/*E0*/  3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,
/*F0*/  3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3
};


/*
  Escape simple strings, not containing dangerous characters.
  Used for things like user limit names, user score names, etc.
*/
char *
UdmSQLEscStrSimple(UDM_SQL *db, char *dst, const char *src, size_t len)
{
  size_t multiply= 1;
  const char *srcend= src + len;
  char *dst0;
  /*
    Backslash, quote and double quote character are not allowed,
    to avoid SQL injection.
  */
  UDM_ASSERT(!dangerous_character['%']); /* Need this for "indexer -u url% */
  UDM_ASSERT(!dangerous_character['?']); /* "indexer -u http://host/?a=b"  */
  UDM_ASSERT(!dangerous_character['=']); /* "indexer -u http://host/?a=b"  */

  UDM_ASSERT(dangerous_character['"']);
  UDM_ASSERT(dangerous_character['\'']);
  UDM_ASSERT(dangerous_character['\\']);
  
  if (!dst && !(dst= (char*) UdmMalloc(len * multiply + 1)))
    return NULL;
  for ( dst0= dst, srcend= src + len; src < srcend; src++)
  {
    *dst++= dangerous_character[(unsigned char) *src] ? '?' : *src;
  }
  *dst= '\0';
  return  dst0;
}


/*
  Escape a byte for PgSQL bytea encoding
*/
static inline size_t
UdmSQLBinEscCharForPg(char *dst, unsigned int ch)
{
  if (ch  >= 0x20 && ch <= 0x7F && ch != '\'' && ch != '\\')
  {
    *dst= ch;
    return 1;
  }
  
  dst[4]= udm_hex_digits[ch & 0x07]; ch /= 8;
  dst[3]= udm_hex_digits[ch & 0x07]; ch /= 8;
  dst[2]= udm_hex_digits[ch & 0x07];
  dst[1]= '\\';
  dst[0]= '\\';
  return 5;
}


size_t
UdmSQLBinEscStr(UDM_SQL *db, char *dst, size_t dstlen, const char *src0, size_t len)
{
  const unsigned char *src= (const unsigned char*) src0;
  UDM_ASSERT(dst != NULL);
  
  if (db->DBType == UDM_DB_PGSQL)
  {
    char *dst0;
    for (dst0= dst ; len > 0 ; len--, src++)
      dst+= UdmSQLBinEscCharForPg(dst, *src);
    *dst= '\0';
    return dst - dst0;
  }
  UdmSQLEscStr(db, dst, src0, len);
  return 0;
}


udm_rc_t
UdmSQLBegin(UDM_SQL *db)
{
  return db->handler.Begin(db);
}

udm_rc_t
UdmSQLCommit(UDM_SQL *db)
{
  return db->handler.Commit(db);
}

udm_rc_t
UdmSQLExecDirect(UDM_SQL *db, UDM_SQLRES *R, const char *q)
{
  return db->handler.ExecDirect(db, R, q);
}

udm_rc_t
UdmSQLPrepare(UDM_SQL *db, const char *q)
{
  return db->handler.Prepare(db, q);
}

udm_rc_t
UdmSQLExecute(UDM_SQL *db)
{
  return db->handler.Exec(db);
}

udm_rc_t
UdmSQLStmtFree(UDM_SQL *db)
{
  return db->handler.StmtFree(db);
}

udm_rc_t
UdmSQLBindParameter(UDM_SQL *db, int pos, const void *data, int size, udm_sqltype_t type)
{
  return db->handler.Bind(db, pos, data, size, type);
}

udm_rc_t
UdmSQLRenameTable(UDM_SQL *db, const char *from, const char *to)
{
  return db->handler.RenameTable(db, from, to);
}

udm_rc_t
UdmSQLCopyStructure(UDM_SQL *db, const char *from, const char *to)
{
  return db->handler.CopyStructure(db, from, to);
}

udm_rc_t
UdmSQLLockOrBegin(UDM_SQL *db, const char *param)
{
  return db->handler.LockOrBegin(db, param);
}

udm_rc_t
UdmSQLUnlockOrCommit(UDM_SQL *db)
{
  return db->handler.UnlockOrCommit(db);
}


UDM_API(udm_rc_t)
UdmSQLQuery(UDM_SQL *db, UDM_SQLRES *SQLRes, const char * query)
{
  UDM_SQLRES res;
  
  /* FIXME: call UdmSQLFree at exit if SQLRes = NULL */
  if (! SQLRes) SQLRes = &res;
  
  db->handler.Query(db, SQLRes, query);

  if (db->errcode && (db->flags & UDM_SQL_IGNORE_ERROR))
    db->errcode= 0;
  
#ifdef DEBUG_ERR_QUERY_XXX
  if (db->errcode)
    fprintf(stderr, "{%s:%d} Query: %s\n\n", file, line, query);
#endif

  return db->errcode ? UDM_ERROR : UDM_OK;
}


UDM_API(size_t)
UdmSQLNumRows(UDM_SQLRES * res)
{
  return(res?res->nRows:0);
}

size_t UdmSQLNumCols(UDM_SQLRES * res)
{
  return(res?res->nCols:0);
}

UDM_API(const char *)
UdmSQLValue(UDM_SQLRES * res,size_t i,size_t j){

#if HAVE_PGSQL
  if (res->db->DBDriver == UDM_DBAPI_PGSQL && !res->Items)
    return(PQgetvalue((PGresult*)res->specific,(int)(i),(int)(j)));
#endif
  
  if (i<res->nRows)
  {
    size_t offs=res->nCols*i+j;
    return res->Items[offs].str;
  }
  else
  {
    return NULL;
  }
}


UDM_API(udm_rc_t)
UdmSQLFetchRowSimple (UDM_SQL *db, UDM_SQLRES *res, UDM_STR *buf)
{
  size_t j;
  size_t offs = res->nCols * res->curRow;

  if (res->curRow >= res->nRows)
    return UDM_ERROR;
  
  for (j = 0; j < res->nCols; j++)
  {
    buf[j]= res->Items[offs + j];
  }
  res->curRow++;
  
  return(UDM_OK);
}


UDM_API(udm_rc_t)
UdmSQLStoreResultSimple(UDM_SQL *db, UDM_SQLRES *res)
{
  return UDM_OK;
}


UDM_API(udm_rc_t)
UdmSQLFreeResultSimple(UDM_SQL *db, UDM_SQLRES *res)
{
  if (res->Fields)
  {
    size_t i;
    for(i=0;i<res->nCols;i++)
    {
      /*
      printf("%s(%d)\n",res->Fields[i].sqlname,res->Fields[i].sqllen);
      */
      UDM_FREE(res->Fields[i].sqlname);
    }
    UDM_FREE(res->Fields);
  }
  
#if HAVE_PGSQL
  if (res->db->DBDriver == UDM_DBAPI_PGSQL)
  {
    PQclear((PGresult*)res->specific);
    /*
      Continue to UdmSQLResFreeGeneric() to
      free allocated memory if we have bytea datatype
    */
  }
#endif

  UdmSQLResFreeGeneric(res);

  return UDM_OK;
}


UDM_API(void)
UdmSQLFree(UDM_SQLRES * res)
{
  res->db->handler.FreeResult(res->db, res);
}


size_t UdmSQLLen(UDM_SQLRES * res,size_t i,size_t j)
{
  size_t offs=res->nCols*i+j;
#if HAVE_PGSQL
  if (res->db->DBDriver == UDM_DBAPI_PGSQL && !res->Items)
    return PQgetlength((PGresult*)res->specific, i, j);
#endif
  return res->Items[offs].length;
}


void
UdmSQLValueToConstStr(UDM_CONST_STR *str,
                      UDM_SQLRES *res, size_t row, size_t col)
{
  str->str= UdmSQLValue(res, row, col);
  str->length= UdmSQLLen(res, row, col);
}


udm_rc_t
UdmSQLClose(UDM_SQL *db)
{
  udm_rc_t rc;
  if (!db->connected)
    return UDM_OK;
  rc= db->handler.Close(db);
  db->connected= UDM_FALSE;
  return rc;
}


void
UdmSQLResListInit(UDM_SQLRESLIST *List)
{
  bzero((void*)List, sizeof(*List));
}


udm_rc_t
UdmSQLResListAdd(UDM_SQLRESLIST *List, UDM_SQLRES *Res)
{
  size_t nbytes= (List->nitems + 1) * sizeof(UDM_SQLRES);
  if (!(List->Item= (UDM_SQLRES*) UdmRealloc(List->Item, nbytes)))
    return UDM_ERROR;
  List->Item[List->nitems]= Res[0];
  List->nitems++;
  return UDM_OK;
}


void
UdmSQLResListFree(UDM_SQLRESLIST *List)
{
  size_t i;
  for (i= 0; i < List->nitems; i++)
  {
    UdmSQLFree(&List->Item[i]);
  }
  UdmFree(List->Item);
  UdmSQLResListInit(List);
}


udm_rc_t
UdmSQLQueryOneRowInt(UDM_SQL *db, int *res, const char *qbuf)
{
  UDM_SQLRES sqlRes;
  udm_rc_t rc;
  if (UDM_OK != (rc= UdmSQLQuery(db, &sqlRes, qbuf)))
    return rc;
  if (UdmSQLNumRows(&sqlRes) < 1)
  {
    rc= UDM_ERROR;
    *res= 0;
    udm_snprintf(db->errstr, sizeof(db->errstr),
                 "Query should have returned one row");
  }
  else
    *res= UDM_ATOI(UdmSQLValue(&sqlRes, 0, 0));
  UdmSQLFree(&sqlRes);
  return rc;
}


static const char*
odbc_params[UDM_SQL_MAX_BIND_PARAM]=
{
  "?","?","?","?","?","?","?","?",
  "?","?","?","?","?","?","?","?",
  "?","?","?","?","?","?","?","?",
  "?","?","?","?","?","?","?","?",
  "?","?","?","?","?","?","?","?",
  "?","?","?","?","?","?","?","?",
  "?","?","?","?","?","?","?","?",
  "?","?","?","?","?","?","?","?",
};


static const char*
oracle_params[UDM_SQL_MAX_BIND_PARAM]=
{
   ":1", ":2", ":3", ":4", ":5", ":6", ":7", ":8", ":9",
  ":10",":11",":12",":13",":14",":15",":16",":17",":18",":19",
  ":20",":21",":22",":23",":24",":25",":26",":27",":28",":29",
  ":30",":31",":32",":33",":34",":35",":36",":37",":38",":39",
  ":40",":41",":42",":43",":44",":45",":46",":47",":48",":49",
  ":50",":51",":52",":53",":54",":55",":56",":57",":58",":59",
  ":60",":61",":62",":63",":64"
};


static const char*
pgsql_params[UDM_SQL_MAX_BIND_PARAM]= 
{
   "$1", "$2", "$3", "$4", "$5", "$6", "$7", "$8", "$9",
  "$10","$11","$12","$13","$14","$15","$16","$17","$18","$19",
  "$20","$21","$22","$23","$24","$25","$26","$27","$28","$29",
  "$30","$31","$32","$33","$34","$35","$36","$37","$38","$39",
  "$40","$41","$42","$43","$44","$45","$46","$47","$48","$49",
  "$50","$51","$52","$53","$54","$55","$56","$57","$58","$59",
  "$60","$61","$62","$63","$64"
};


const char *
UdmSQLParamPlaceHolder(UDM_SQL *db, size_t i)
{
  UDM_ASSERT(i < UDM_SQL_MAX_BIND_PARAM);
  if (db->handler.Bind == UdmSQLBindGeneric)
    return odbc_params[i - 1];
  if (db->DBDriver == UDM_DBAPI_ORACLE8)
    return oracle_params[i - 1];
  if (db->DBDriver == UDM_DBAPI_PGSQL)
    return pgsql_params[i - 1];
  return odbc_params[i - 1];
}


/*
  Generic prepared statement API for libraries
  not supporting their own prepared statements.
*/

#define UDM_GENERIC_MAX_BIND_PARAM UDM_SQL_MAX_BIND_PARAM


typedef struct generic_sql_ps_st
{
  char *sql;
  int nParams;
  int paramTypes[UDM_GENERIC_MAX_BIND_PARAM];
  const void *paramValues[UDM_GENERIC_MAX_BIND_PARAM];
  int paramLengths[UDM_GENERIC_MAX_BIND_PARAM];
} UDM_PS;


static UDM_PS *
alloc_ps(UDM_SQL *db)
{
  if (db->ps)
    return (UDM_PS*) db->ps;
  if ((db->ps= (UDM_PS*) UdmMalloc(sizeof(UDM_PS))))
    return (UDM_PS*) db->ps;
  return NULL;
}


udm_rc_t
UdmSQLPrepareGeneric(UDM_SQL *db, const char *query)
{
  UDM_PS *ps= alloc_ps(db);
  
  if (!ps)
    return UDM_ERROR;

  ps->nParams= 0;

  /*
    TODO: add automatic connecting here
  */
  if (!(ps->sql= UdmStrdup(query)))
  {
    udm_snprintf(db->errstr, sizeof(db->errstr),
                 "UdmSQLPrepareGeneric: could not allocate memory");
    return UDM_ERROR;
  }
  return UDM_OK;
}


udm_rc_t
UdmSQLBindGeneric(UDM_SQL *db, int pos, const void *data, int size, udm_sqltype_t type)
{
  UDM_PS *ps= (UDM_PS*) db->ps;
  UDM_ASSERT(ps->sql);

  if (!ps)
    return UDM_ERROR;

  if (ps->nParams < pos)
    ps->nParams= pos;   
  pos--;
  ps->paramValues[pos]= data;
  ps->paramLengths[pos]= (int) size;
  ps->paramTypes[pos]= type;
  return UDM_OK;
}


static size_t
prepared_query_length(const char *sql, UDM_PS *param)
{
  int i;
  size_t len= strlen(param->sql) + 1;
  
  for (i= 0; i < param->nParams; i++)
  {
    switch (param->paramTypes[i])
    {
      case UDM_SQLTYPE_LONGVARBINARY:
      case UDM_SQLTYPE_LONGVARCHAR  :
      case UDM_SQLTYPE_VARCHAR:
        len+= param->paramLengths[i] < 0 ?
              4 :/* NULL in ODBC*/
              param->paramLengths[i] * 10 + 10; /* TODO: proper constant */
        break;
      case UDM_SQLTYPE_INT32:
        len+= 21;
        break;
    }
  }
  return len;
}


static size_t
prepared_query_add_params(UDM_SQL *db, char *dst, size_t dstlen,
                          UDM_PS *param, size_t param_num)
{
  char *dst0= dst;
  int is_bin= (param->paramTypes[param_num] == UDM_SQLTYPE_LONGVARBINARY);
  size_t srclen= param->paramLengths[param_num];
  const char *src= (const char *) param->paramValues[param_num];

  if (!srclen)
  {
    /* Note, srclen is never 0 when type is UDM_SQLTYPE_INT32 */
    if (db->DBType == UDM_DB_MIMER)
      *dst++= 'X';
    *dst++= '\'';
    *dst++= '\'';
    goto ret;
  }

  switch (param->paramTypes[param_num])
  {
    case UDM_SQLTYPE_LONGVARBINARY:
    case UDM_SQLTYPE_LONGVARCHAR  :
    case UDM_SQLTYPE_VARCHAR:
      if (is_bin && (db->flags & UDM_SQL_HAVE_0xHEX))
      {
        /*
          MSSQL:   0x20C883 notation  (0xHEX)
          Sybase:  0x20C883 notation  (0xHEX)
          Access:  0x20C883 notation  (0xHEX)
        */
        *dst++= '0';
        *dst++= 'x';
        dst+= UdmHexEncode(dst, src, srclen);
      }
      else if (is_bin && (db->flags & UDM_SQL_HAVE_STDHEX))
      {
        /*
          SQLite3: X'20C883'          (STDHEX)
        */
        *dst++= 'X';
        *dst++= '\'';
        dst+= UdmHexEncode(dst, src, srclen);
        *dst++= '\'';
      }
      else if (is_bin && (db->DBType == UDM_DB_ORACLE8))
      {
        if (param->paramLengths[param_num] < 0) /* Oracle via ODBC */
        {
          strcpy(dst, "NULL");
          dst+= 4;
        }
        else
        {
          /* Oracle: '20C883', i.e. binary notation by default */
          *dst++= '\'';
          dst+= UdmHexEncode(dst, src, srclen);
          *dst++= '\'';
        }
      }
      else
      {
        /*
          TODO:
          PgSQL:   E'\x20\xC8\x83' notation
          SQLite:  ???
        */
        if (db->DBType == UDM_DB_PGSQL && db->version >= 80101)
          *dst++= 'E';
        *dst++= '\'';
        if (is_bin)
          UdmSQLBinEscStr(db, dst, dstlen, src, srclen); /* TODO: get rid of strlen below*/
        else
          UdmSQLEscStr(db, dst, src, srclen);
        dst+= strlen(dst);
        *dst++= '\'';
      }
      break;
    case UDM_SQLTYPE_INT32:
      return sprintf(dst, "%d", *((const int*) src));
      break;
  }

ret:
  *dst= '\0';
  return dst - dst0;
}


udm_rc_t
UdmSQLExecGeneric(UDM_SQL *db)
{
  UDM_PS *ps= (UDM_PS*) db->ps;
  char *qbuf, *dst;
  const char *src;
  size_t qlen= prepared_query_length(ps->sql, ps);
  size_t param_num= 0;
  udm_rc_t rc;
  UDM_SQLRES SQLRes;
  
  if (!(qbuf= (char*) UdmMalloc(qlen)))
  {
    udm_snprintf(db->errstr, sizeof(db->errstr),
                 "UdmSQLExecGeneric: Failed to allocated buffer %d bytes",
                 (int) qlen);
    return UDM_ERROR;
  }
  
  for (src= ps->sql, dst= qbuf; *src; src++)
  {
    if (*src == '?')
    {
      size_t len= prepared_query_add_params(db, dst, qlen, ps, param_num);
      param_num++;
      dst+= len;
    }
    else
    {
      *dst++= *src;
    }
  }
  *dst= '\0';

  rc= UdmSQLExecDirect(db, &SQLRes, qbuf);

  UdmSQLFree(&SQLRes);
  UdmFree(qbuf);

  return rc;
}


udm_rc_t
UdmSQLStmtFreeGeneric(UDM_SQL *db)
{
  UDM_PS *ps= (UDM_PS*) db->ps;
  UDM_ASSERT(ps->sql);
  UDM_FREE(ps->sql);
  UDM_FREE(db->ps);
  return UDM_OK;
}


/*

Prepared statement API

Name         Native  Generic None Default
----         ------  ------- ---- -------
Oracle       Yes     No      No   Generic
MSSQL        TODO    ?       ?    None
Sybase       TODO    ?       ?    None
MySQL        Yes     Yes     Yes  Native|Generic (client version dependent)
PgSQL        Yes     Yes     Yes  Native|Generic (client version dependend)
IBase        Yes     No      No   Native
SQLite       TODO    Yes     No   Generic
SQLite3      TODO    Yes     Yes  Generic

ODBC-DB2     Yes     No      No   Native
ODBC-MIMER   Yes     No      No   Native
ODBC-ORACLE  Yes     No      No   Native
ODBC-MSQQL   Yes     ?       Yes  None
ODBC-SYBASE  Yes     ?       Yes  None
ODBC-MYSQL   Yes     Yes     Yes  Native
ODDB-IBASE   Yes     No      No   Native

ODBC-CACHE   Yes     ?       ?    Native
ODBC-VIRT    Yes     ?       ?    Native
ODBC-Solid   ?       ?       ?    Native
ODBC-SapDB   ?       ?       ?    Native
ODBC-ACCESS  ?       ?       ?    Native

*/


udm_rc_t
UdmSQLLockOrBeginGeneric(UDM_SQL *db, const char *query)
{
  return db->handler.Begin(db);
}

udm_rc_t
UdmSQLUnlockOrCommitGeneric(UDM_SQL *db)
{
  return db->handler.Commit(db);
}

#endif /* HAVE_SQL */
