/* Copyright (C) 2000-2015 Lavtech.com corp. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
*/

#include "udm_config.h"

#ifdef HAVE_SQL

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#include "udm_common.h"
#include "udm_utils.h"

#include "udm_db.h"
#include "udm_sqldbms.h"
#include "udm_db_int.h"
#include "udm_vars.h"
#include "udm_log.h" /* For UDM_LOG_INFO */

/*
  Set connection to the given database.
  newnum is the database number.
  1 means the first DBAddr.
*/
static udm_rc_t
SetConnection(UDM_ENV *Env, UDM_SQLMON_PARAM *prm, int newnum)
{
  char msg[255];
  if (newnum >= 1 && newnum <= (int) Env->DBList.nitems)
  {
    prm->currdbnum= newnum - 1;
    sprintf(msg, "Connection changed to #%d", (int) prm->currdbnum);
    prm->iohandler.prompt(&prm->iohandler, UDM_MSG_INFO, msg);
    return UDM_OK;
  }
  else
  {
    sprintf(msg,"Wrong connection number %d", newnum);
    prm->iohandler.prompt(&prm->iohandler, UDM_MSG_ERROR, msg);
    return UDM_ERROR;
  }
}


static udm_rc_t
UdmDisplaySQLQuery(UDM_SQLMON_PARAM *prm, UDM_SQLRES *sqlres)
{
  udm_rc_t rc= UDM_OK;
#ifdef HAVE_SQL     
  size_t i, j;

  if (sqlres->nCols && (prm->flags & UDM_SQLMON_DISPLAY_FIELDS))
  {
    UDM_DSTR dstr;
    UdmDSTRInit(&dstr, 128);
    for (i= 0; i < sqlres->nCols; i++)
    {
      if (i > 0)
        UdmDSTRAppendf(&dstr, "\t");
      UdmDSTRAppendf(&dstr, "%s", sqlres->Fields ?
                            sqlres->Fields[i].sqlname : "<NONAME>");
    }
    prm->iohandler.prompt(&prm->iohandler, UDM_MSG_DATA, UdmDSTRPtr(&dstr));
    UdmDSTRFree(&dstr);
    prm->iohandler.prompt(&prm->iohandler, UDM_MSG_EOL, "");
  }
  
  for (i=0;i<sqlres->nRows;i++)
  {
    for (j=0;j<sqlres->nCols;j++)
    {
      const char *v=UdmSQLValue(sqlres,i,j);
      UDM_DSTR dstr;
      UdmDSTRInit(&dstr, 1024);
      if (j>0)
        UdmDSTRAppendf(&dstr, "\t");
      if (j < 10 && (prm->colflags[j] & udm_sqlmon_column_flag_binary))
      {
        const char* s, *e;
        UdmDSTRAppendf(&dstr, "0x");
                 
        for (s=v, e= s + UdmSQLLen(sqlres, i, j); s < e; s++)
          UdmDSTRAppendf(&dstr, "%02X", (int) (unsigned char) s[0]);
      }
      else if (j < 10 && (prm->colflags[j] & udm_sqlmon_column_flag_semibinary))
      {
        const char *s, *e;
        for (s= v, e= s + UdmSQLLen(sqlres, i, j) ; s < e; s++)
        {
          if (*s == '\r')
            UdmDSTRAppendf(&dstr, "\\r");
          else if (*s == '\n')
            UdmDSTRAppendf(&dstr, "\\n");
          else if (*s == '\\')
            UdmDSTRAppendf(&dstr, "\\");
          else if (*s < 0x20)
            UdmDSTRAppendf(&dstr, "\\x%02X", (int) (unsigned char) *s);
          else
            UdmDSTRAppendf(&dstr, "%c", *s);
        }
      }
      else
      {
        UdmDSTRAppendf(&dstr, "%s", v ? v : "NULL");
      }
      prm->iohandler.prompt(&prm->iohandler, UDM_MSG_DATA, UdmDSTRPtr(&dstr));
      UdmDSTRFree(&dstr);
    }
    prm->iohandler.prompt(&prm->iohandler, UDM_MSG_EOL, "");
  }
     
#endif
  return rc;
}



/*
  Execute an SQL query, or a build-in command.
*/
static udm_rc_t
RunCommand(UDM_AGENT *A, UDM_ENV *Env, UDM_SQLMON_PARAM *prm, const char *str)
{
  udm_rc_t rc= UDM_OK;
  prm->iohandler.prompt(&prm->iohandler, UDM_MSG_COMMAND, str);
  
  if (!strncasecmp(str,"connection",10))
  {
    int newnum= atoi(str + 10);
    SetConnection(Env, prm, newnum + 1);
  }
  else if (!strcasecmp(str,"fields=off"))
  {
    prm->flags=0;
  }
  else if (!strcasecmp(str,"fields=on"))
  {
    prm->flags=1;
  }
  else if (!strncasecmp(str,"colflags",8))
  {
    int colnum= atoi(str+8);
    int colval= atoi(str+10);
    if (colnum>=0 && colnum<10)
      prm->colflags[colnum]= (udm_sqlmon_column_flag_t) colval;
  }
  else
  {
    udm_rc_t res;
    UDM_SQLRES sqlres;
    UDM_DB *db= &Env->DBList.Item[prm->currdbnum];
    prm->nqueries++;
    bzero((void*)&sqlres,sizeof(sqlres));
    res= UdmDBSQLQuery(A, db, &sqlres, str);
    if (res!=UDM_OK)
    {
      prm->nbad++;
      rc= UDM_ERROR;
      prm->iohandler.prompt(&prm->iohandler, UDM_MSG_ERROR, UdmDBSQLError(db));
    }
    else
    {
      prm->ngood++;
      res= UdmDisplaySQLQuery(prm, &sqlres);
    }
    UdmSQLFree(&sqlres);
  }
  return rc;
}


static size_t
RemoveTrailingSpaces(UDM_SQLMON_PARAM *prm,
                     char *s, size_t slen,
                     size_t *space_count)
{
  char *e= s + slen;
  for ( space_count[0]= 0 ; e > s ; *--e= '\0')
  {
    int ch= e[-1];
    if (ch == ' ' || ch == '\t' || ch == '\r')
    {
      space_count[0]++;
    }
    else if (ch == '\n')
    {
      space_count[0]++;
      prm->lineno++;
    }
    else
      break;
  }
  return e - s;
}

/*

  UdmSQLMonitor()
  
  - Executes SQL queries arriving from some virtual input stream
  - Displays query results to some virtual output streams
  
  The incoming strings are checked against:
    
    - SQL comments, i.e. lines starting with "--".
      The comment lines are skipped.
    
    - A set of end-of-query markers, i.e. 
      * MSSQL   style: 'GO' keyword written in the beginning
      * mysql   style: ';'  character written in the end of the string
      * minisql style: '\g' characters written in the end of the string
      
      As soon as a end-of-query marker has found, current collected query
      is executed.
    
    - In other case, the next line is concatenated to the previously
      collected data thus helping to run multi-line queries.
  
  
  The particular ways of handling input and output streams are
  passed in the UDM_SQLMON_PARAM argument. This approach makes 
  it possible to use UdmSQLMonitor() for many purposes:
  
  - executing a bundle of queries written in a text file.
  - running a command-line SQL monitor ("mysql", "pgsql" or "sqlplus" alike),
    i.e. executing queries arriving from the keyboard.
    
  - running a GUI SQL monitor,
    i.e. executing querues arriving from the keyboard in some GUI application.
  
  
  UDM_SQLMON_PARAM has got this structure:
  
  typedef struct udm_sqlmon_param_st
  {
    int flags;
    FILE *infile;
    FILE *outfile;
    char *(*gets)(struct udm_sqlmon_param_st *prm, char *str, size_t size);
    int (*prompt)(struct udm_sqlmon_param_st *, const char *msg);
  void* tag_ptr;
  } UDM_SQLMON_PARAM;
  
  
  gets():     a function to get one string from the incoming stream.
  prompt():   a function to display some prompts:
                - an invitation to enter data, for example: "SQL>".
                - an error message
  
  flags:      sets various aspects of the monitor behaviour:
              the only one so far:
                - whether to display field names in the output

  infile:     if the caller wants to input SQL queries from some FILE stream,
              for example, from STDIN or from a textual file opened via fopen(),
              it can use this field for this purpose
              
  outfile:    if the caller wants to output results to some FILE stream,
              for example, to STDOUT or to a textial file opened via fopen(),
              it can use this field for this purpose

  user_data:  extra pointer parameter for various purposes.
  
  To start using UdmSQLMonitor() one have to set
    - gets()
    - diplay()
    - prompt()
  
  If diplay() or prompt() use some FILE streams, the caller should
  also set infile and outfile with appropriative stream pointers.
  
*/


#define CHUNK_SIZE 64*1024


UDM_API(udm_rc_t)
UdmSQLMonitor(UDM_AGENT *A, UDM_ENV *Env, UDM_SQLMON_PARAM *prm)
{
  udm_rc_t rc= UDM_OK;
  int DBLimit= UdmVarListFindInt(&Env->Vars, "DBLimit", 0);
  UDM_DSTR qbuf;

  UdmDSTRInit(&qbuf, 256);

  if (DBLimit && UDM_OK != (rc= SetConnection(Env, prm, DBLimit)))
    return rc;
  
  while(1)
  {
    int  exec=0;
    size_t new_chunk_len, space_count;
    char *snd;
    
    UdmDSTRRealloc(&qbuf, UdmDSTRLength(&qbuf) + CHUNK_SIZE);
    
    if (qbuf.size_alloced > 512 * 1024 * 1024)
    {
      char errstr[160];
      rc= UDM_ERROR;
      udm_snprintf(errstr, sizeof(errstr), "Too long query: %s", UdmDSTRPtr(&qbuf));
      prm->iohandler.prompt(&prm->iohandler, UDM_MSG_ERROR, errstr);
      break;
    }

    snd= qbuf.Val.str + qbuf.Val.length;
    if (!prm->iohandler.gets(&prm->iohandler, snd, CHUNK_SIZE))
      break;

    /* Remove trailing spaces/tabs/NLs */
    if (!(new_chunk_len= RemoveTrailingSpaces(prm, snd, strlen(snd), &space_count)))
      continue;

    if (!strncmp(snd, "--seed=", 7))
    {
      udmhash32_t seed= atoi(snd + 7);
      SetConnection(Env, prm, UdmDBNumBySeed(Env, seed) + 1);
    }
    
    /*
      Skip comments
      TODO: detect long comments which do not fit into CHUNK_SIZE
      TODO: detect semicolon correctly, only when outside a string
    */
    if(snd[0]=='#' || !strncmp(snd, "--", 2))
      continue;
    
    /* Append the newly loaded chunk into qbuf */
    qbuf.Val.length+= new_chunk_len;

    snd= snd + new_chunk_len;
   
    if (prm->flags & UDM_SQLMON_DONT_NEED_SEMICOLON)
    {
      exec= 1;
    }
    else if (snd[-1] == ';' &&
             /*
               If ';' follows an empty comment ('/' + '**' + '/' + ';'),
               then don't treat ';' as a query end and keep collecting
               the query.
               This is needed for PostgreSQL's CREATE FUNCTION:

               CREATE FUNCTION func()
               RETURNS TRIGGER AS $$
               BEGIN
                 IF (...) THEN INSERT INTO t1 VALUES (NEW.*)<COMMENT>;
                 ELSEIF (...) THEN INSERT INTO t2 VALUES (NEW,*)<COMMENT>;
                 END IF<COMMENT>;
                 RETURN NULL<COMMENT>;
               END<COMMENT>; $$
               LANGUAGE plsql;
             */
             (snd - 3 < qbuf.Val.str || snd[-4] != '*' || snd[-3]!= '*' || snd[-2] != '/'))
    {
      exec=1;
      *--snd='\0';
      UdmDSTRShrinkLast(&qbuf);
    }
    else if(snd - 2 >= qbuf.Val.str && snd[-1]=='g' && snd[-2]=='\\')
    {
      exec=1;
      snd-=2;
      *snd='\0';
      qbuf.Val.length-= 2;
    }
    else if(snd-2 >= qbuf.Val.str && strchr("oO", snd[-1]) && strchr("gG",snd[-2]))
    {
      exec=1;
      snd-=2;
      *snd='\0';
      qbuf.Val.length-= 2;
    }
    
    if(exec)
    {
      rc= RunCommand(A, Env, prm, UdmDSTRPtr(&qbuf));
      UdmDSTRReset(&qbuf);
      if (prm->mode == udm_sqlmon_mode_interactive)
        prm->lineno= 0;
    }
    else
    {
      if (space_count)
        UdmDSTRAppend(&qbuf, " ", 1);
    }
  }
  UdmDSTRFree(&qbuf);
  return rc;
}


#endif /* HAVE_SQL */
