/* Copyright (C) 2000-2015 Lavtech.com corp. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
*/

#include <udm_config.h>

#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include <sys/types.h>
#ifdef   HAVE_UNISTD_H
#include <unistd.h>
#endif
#ifdef HAVE_ZLIB
#include <zlib.h>
#endif

#include "udm_common.h"
#include "udm_store.h"
#include "udm_utils.h"
#include "udm_vars.h"
#include "udm_parsehtml.h"
#include "udm_unicode.h"
#include "udm_unidata.h"
#include "udm_searchtool.h"
#include "udm_query.h"
#include "udm_doc.h"
#include "udm_db.h"
#include "udm_result.h"
#include "udm_parser.h"
#include "udm_http.h"
#include "udm_guesser.h"
#include "udm_contentencoding.h"

/********************************************************************/
static udm_rc_t
UdmHTTPBufAppendEncodingDecodeWithTimer(UDM_HTTPBUF *Buf,
                                        const UDM_CONST_STR *content,
                                        udm_content_encoding_t encoding,
                                        udm_timer_t *timer)
{
  udm_timer_t ticks= UdmStartTimer();
  udm_rc_t rc= UdmHTTPBufAppendEncodingDecode(Buf, content->str, content->length, encoding);
  *timer+= UdmStartTimer() - ticks;
  return rc;
}


udm_rc_t
UdmDocSetFromCachedHTTPResponse(UDM_DOCUMENT *Doc,
                                const char *src, size_t srclen,
                                size_t max_doc_size,
                                udm_timer_t *decoding_time)
{
  size_t content_offset;
  const char *ce;
  udm_rc_t rc= UDM_OK;
  UDM_CONST_STR content;

  UdmHTTPBufFree(&Doc->Buf);
  if (UdmHTTPBufAlloc(&Doc->Buf, max_doc_size + 1))
    return UDM_ERROR;
  if (!src || !srclen)
    return UDM_OK;
  if (!(content_offset= UdmHTTPFindContent(src, srclen)))
    return UDM_OK;
  content.str= src + content_offset;
  content.length= srclen - content_offset;
  UdmHTTPBufAppend(&Doc->Buf, src, content_offset);
  Doc->Buf.content= Doc->Buf.buf + content_offset; /* TODO34 */
  UdmHTTPBufParseHeadersForCachedCopy(&Doc->Sections, src, content_offset);
  if ((ce= UdmVarListFindStr(&Doc->Sections, "Cached.Content-Encoding", NULL)))
  {
    udm_content_encoding_t encoding= UdmContentEncodingID(ce);
    rc= UdmHTTPBufAppendEncodingDecodeWithTimer(&Doc->Buf, &content,
                                                encoding, decoding_time);
  }
  else /* Not encoded content */
  {
    UdmHTTPBufAppend(&Doc->Buf, content.str, content.length);
  }
  return rc;
}


#include "udm_log.h"


UDM_CHARSET *UdmDocDetectCachedCharset(UDM_AGENT *A,
                                       UDM_DOCUMENT *Doc,
                                       urlid_t url_id)
{
  UDM_CHARSET *doccs, *hdrcs;
  doccs= UdmVarListFindCharset(&Doc->Sections, "Meta-Charset", NULL);
  hdrcs= UdmVarListFindCharset(&Doc->Sections, "Cached.Charset", NULL);
  if (doccs == NULL)
  {
    doccs= hdrcs ? hdrcs : &udm_charset_latin1;
  }
  else
  {
    if (hdrcs != NULL && hdrcs != doccs)
    {
      UdmLog(A, UDM_LOG_DEBUG,
             "Warning: URLID=%d: Charset conflict: Headers=%s, Meta=%s\n",
             url_id, hdrcs->name, doccs->name);
      doccs= hdrcs;
    }
  }
  return doccs;
}


/******************* Cached Copy routines ***************************/

typedef struct udm_cached_copy_helper_st
{
  UDM_DOCUMENT *Doc;
  UDM_QUERY    Query;
  UDM_DSTR     dstr;
  UDM_HIGHLIGHT_CONV hconv_bcs;
  UDM_AGENT   *Agent;
} UDM_CACHED_COPY_HELPER;


static void
UdmCachedCopyHelperInit(UDM_CACHED_COPY_HELPER *helper,
                        UDM_AGENT *Agent, UDM_DOCUMENT *Doc)
{
  bzero(helper, sizeof(*helper));
  UdmQueryInit(&helper->Query);
  UdmDSTRInit(&helper->dstr, 1024);
  helper->Agent= Agent;
  helper->Doc= Doc;
}


static void
UdmCachedCopyHelperFree(UDM_CACHED_COPY_HELPER *helper)
{
  UdmResultFree(&helper->Query.Res);
  UdmQueryFree(&helper->Query);
  UdmDSTRFree(&helper->dstr);
}


static udm_rc_t
UdmCachedCopyCommentHandler(UDM_HTML_PARSER *parser)
{
  UDM_CACHED_COPY_HELPER *param= (UDM_CACHED_COPY_HELPER *) parser->user_data;
  UdmDSTRAppend(&param->dstr, parser->lasttok.str, parser->lasttok.length);
  return UDM_OK;
}


static udm_rc_t
UdmCachedCopyTextHandler(UDM_HTML_PARSER *parser)
{
  UDM_CACHED_COPY_HELPER *param= (UDM_CACHED_COPY_HELPER *) parser->user_data;
  UDM_WIDEWORDLIST *WWList= &param->Query.Res.WWList;
  UDM_SNIPPETCHUNK_STAT dummy;

  if (!parser->lasttok.length)
    return UDM_OK;

  if (parser->state.title || parser->state.script ||
      parser->state.comment || parser->state.style)
    WWList= NULL;

  UdmHlConvertExtWithConv(param->Agent, &param->dstr, &dummy, WWList,
                          parser->lasttok.str, parser->lasttok.length,
                          &param->hconv_bcs);
  return UDM_OK;
}


static udm_rc_t
UdmParseCachedCopy(UDM_CACHED_COPY_HELPER *param, const char *src, size_t srclen)
{
  UDM_HTML_PARSER tag;
  UdmHTMLParserInit(&tag);
  UdmHTMLParserSetUserData(&tag, param);
  UdmHTMLParserSetCommentHandler(&tag, UdmCachedCopyCommentHandler);
  UdmHTMLParserSetTagHandler(&tag, UdmCachedCopyCommentHandler);
  UdmHTMLParserSetTextHandler(&tag, UdmCachedCopyTextHandler);
  return UdmHTMLParserExec(&tag, src, srclen);
}


static void
UdmPrepareQueryString(UDM_ENV *Env, UDM_DOCUMENT *Doc)
{
  const char *qstring;

  if ((qstring= UdmVarListFindStr(&Env->Vars, "ENV.QUERY_STRING", NULL)))
  {
    /*
      Remove dbnum from query string.
      Note: dbnum is passed when a cached copy is displayed in cluster.
    */
    if (!strncmp(qstring, "dbnum=", 6))
    {
      char tmp[1024];
      for (qstring+= 6 ; (qstring[0] >= '0' && qstring[0] <= '9') || qstring[0]=='&' ; qstring++);
      udm_snprintf(tmp, sizeof(tmp), "%s", qstring);
      UdmVarListReplaceStr(&Doc->Sections, "ENV.QUERY_STRING", tmp);
      UdmVarListReplaceStr(&Env->Vars, "ENV.QUERY_STRING", tmp);
    }
    else
    {
      UdmVarListReplaceStr(&Doc->Sections, "ENV.QUERY_STRING", qstring);
    }
  }
}


static void
UdmCachedCopyFromTextSource(UDM_AGENT *A,
                            UDM_CACHED_COPY_HELPER *param,
                            const char *src, size_t length)
{
  UDM_SNIPPETCHUNK_STAT dummy;
  UdmHlConvertExtWithConv(A, &param->dstr, &dummy, &param->Query.Res.WWList,
                          src, length,
                          &param->hconv_bcs);
}


static void
UdmCachedCopyMake(UDM_AGENT *Agent, UDM_CACHED_COPY_HELPER *param,
                  UDM_VARLIST *Vars)
{
  UDM_CHARSET *cs;
  UDM_CONST_STR content;
  udm_content_type_t ct;
  const char *content_type= UdmVarListFindStr(&param->Doc->Sections, "Cached.Content-Type", NULL);

#ifdef USE_PARSER
  if (content_type)
  {
    UDM_PARSER   *Parser;
    if ((Parser= UdmParserFind(&Agent->Conf->Parsers, content_type)))
    {
      content_type= Parser->to_mime ? Parser->to_mime : "text/html";
    }
  }
#endif
  if (UdmHTTPBufContentToConstStr(&param->Doc->Buf, &content))
    return;

  cs= UdmVarListFindCharset(&param->Doc->Sections, "Cached.Charset", &udm_charset_latin1);
  UdmVarListReplaceStr(&param->Doc->Sections, "Charset", cs->name);
  UdmVarListReplaceStr(Vars, "Charset", cs->name);
  UdmExcerptConvInitFromEnv(&param->hconv_bcs, Agent->Conf->lcs, cs, cs,
                            Agent->Conf);

  if (!content_type || !(ct= UdmContentTypeByName(content_type)))
    ct= UdmGuessContentType(content.str, content.length, UDM_CONTENT_TYPE_TEXT_PLAIN);

  switch (ct)
  {
    case UDM_CONTENT_TYPE_TEXT_HTML:
    case UDM_CONTENT_TYPE_TEXT_XML:
      UdmParseCachedCopy(param, content.str, content.length);
      break;

    case UDM_CONTENT_TYPE_MESSAGE_RFC822:
      UdmMessageRFC822CachedCopy(Agent, &param->Query, param->Doc, &param->dstr);
      fprintf(stderr, "CS=%s\n", UdmVarListFindStr(&param->Doc->Sections, "Charset", "NOTFOUND"));
      if (Vars != &param->Doc->Sections) /* TODO34: get rid of this */
        UdmVarListReplaceLst(Vars, &param->Doc->Sections, NULL, "Charset");
      break;

    case UDM_CONTENT_TYPE_TEXT_PLAIN:
      UdmDSTRAppend(&param->dstr, UDM_CSTR_WITH_LEN("<pre>\n"));
      UdmCachedCopyFromTextSource(Agent, param, content.str, content.length);
      UdmDSTRAppend(&param->dstr, UDM_CSTR_WITH_LEN("</pre>\n"));
      break;

    case UDM_CONTENT_TYPE_DOCX:
      UdmVarListReplaceStr(Vars, "Charset", "utf-8");
      UdmDOCXCachedCopy(Agent, &param->Query, param->Doc, &param->dstr);
      break;

    case UDM_CONTENT_TYPE_TEXT_RTF:
      UdmDSTRAppend(&param->dstr, UDM_CSTR_WITH_LEN("<span style=\"white-space:pre-wrap\">"));
      if (UDM_OK == UdmRTFCachedCopy(Agent, &param->Query, param->Doc, &param->dstr))
      {
        UDM_CHARSET *tmpcs= UdmVarListFindCharset(&param->Doc->Sections, "Meta-Charset",
                                                  UdmGetCharSet("cp1252"));
        UdmVarListReplaceStr(Vars, "Charset", tmpcs->name);
        UdmDSTRAppend(&param->dstr, UDM_CSTR_WITH_LEN("</span>\n"));
      }
      else
      {
        UdmDSTRReset(&param->dstr);
      }
      break;

    case UDM_CONTENT_TYPE_AUDIO_MPEG:
    case UDM_CONTENT_TYPE_HTDB:
    case UDM_CONTENT_TYPE_UNKNOWN:
      break;
  }
}


/*
  Fetch cached copy from the database,
  highlight and store into the "document" variable
  of Agent->Env->Vars;
  The URL of the needed document must be stored in Agent->Env-Vars.
*/
udm_rc_t
UdmDocCachedCopyGet(UDM_AGENT *Agent, UDM_DOCUMENT *Doc, UDM_VARLIST *Vars)
{
  UDM_CACHED_COPY_HELPER param;
  UDM_ENV *Env= Agent->Conf;

  UdmCachedCopyHelperInit(&param, Agent, Doc);

  UdmQueryInit(&param.Query);
  UdmQueryPrepare(Agent, &param.Query);
  UdmPrepareQueryString(Agent->Conf, param.Doc);
  UdmDocAction(Agent, Doc, UDM_DOCCMD_GET_CACHED_COPY);
  UdmVarListReplaceLst(&Env->Vars, &Doc->Sections, NULL, "*");

  /* Highlight and Display */
  UdmCachedCopyMake(Agent, &param, Vars);
  UdmVarListReplaceStrn(Vars, "document",
                        UdmDSTRPtr(&param.dstr), UdmDSTRLength(&param.dstr));

  UdmCachedCopyHelperFree(&param);
  return UDM_OK;
}

udm_rc_t
UdmCachedCopyGet(UDM_AGENT *Agent)
{
  udm_rc_t rc;
  UDM_ENV *Env= Agent->Conf;
  UDM_DOCUMENT Doc;
  UdmDocInit(&Doc);
  UdmVarListReplaceStr(&Doc.Sections, "URL", UdmVarListFindStr(&Env->Vars, "URL", "0"));
  UdmVarListReplaceStr(&Doc.Sections, "dbnum", UdmVarListFindStr(&Env->Vars, "dbnum", "0"));
  rc= UdmDocCachedCopyGet(Agent, &Doc, &Agent->Conf->Vars);
  UdmDocFree(&Doc);
  return rc;
}
