/* Copyright (C) 2000-2015 Lavtech.com corp. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
*/

#include "udm_config.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include "udm_common.h"
#include "udm_utils.h"
#include "udm_textlist.h"
#include "udm_unicode.h"
#include "udm_word.h"


static inline udm_rc_t
UdmTextListRealloc(UDM_TEXTLIST *tlist)
{
  if (tlist->nitems >= tlist->mitems)
  {
    tlist->mitems+= 16*1024;
    tlist->Item= (UDM_TEXTITEM*) UdmRealloc(tlist->Item,
                                            tlist->mitems * sizeof(UDM_TEXTITEM));
    if (!tlist->Item)
    {
      tlist->nitems= 0;
      tlist->mitems= 0;
      return UDM_ERROR;
    }
  }
  return UDM_OK;
}


UDM_API(void)
UdmTextListAdd(UDM_TEXTLIST *tlist, const UDM_TEXTITEM *item)
{
  if(!item->str)return;

  if (UDM_OK != UdmTextListRealloc(tlist))
    return; /* TODO34: return error code  */

  tlist->Item[tlist->nitems].str = (char*)UdmStrdup(item->str);
  tlist->Item[tlist->nitems].href = item->href ? (char*)UdmStrdup(item->href) : NULL;
  tlist->Item[tlist->nitems].section_name = item->section_name ? (char*)UdmStrdup(item->section_name) : NULL;
  tlist->Item[tlist->nitems].Param= item->Param;
  tlist->nitems++;
  return;
}


UDM_API(void)
UdmTextListAddConst(UDM_TEXTLIST *tlist,
                    const UDM_CONST_TEXTITEM *item,
                    const UDM_TEXT_PARAM *TextParam)
{
  UDM_TEXTITEM *I;
  /*
  if (!item->str)
    return;
  */
  if (UDM_OK != UdmTextListRealloc(tlist))
    return; /* TODO34: return error code  */

  I= &tlist->Item[tlist->nitems];
  I->str= UdmConstStrDup(&item->text);
  I->href= item->href.str ? UdmConstStrDup(&item->href) : NULL;
  I->section_name= item->section_name.str ? UdmConstStrDup(&item->section_name) : NULL;
  I->Param= *TextParam;
  tlist->nitems++;
  return;
}


UDM_API(void)
UdmTextListAppend(UDM_TEXTLIST * tlist,
                                          const UDM_TEXTITEM *item)
{
  size_t oldlen, newlen;
  UDM_TEXTITEM *I;

  if(!item->str)return;
  
  if (item->href || !tlist->nitems)
  {
    UdmTextListAdd(tlist, item);
    return;
  }
  
  I= &tlist->Item[tlist->nitems-1];
  
  oldlen= strlen(I->str);
  newlen= oldlen + strlen(item->str) + 1;
  I->str= (char*)UdmRealloc(I->str, newlen);
  strcpy(I->str + oldlen, item->str);
  return;
}


UDM_API(void)
UdmTextListFree(UDM_TEXTLIST *tlist)
{
  size_t i;

  for(i=0;i<tlist->nitems;i++)
  {
    UDM_FREE(tlist->Item[i].str);
    UDM_FREE(tlist->Item[i].href);
    UDM_FREE(tlist->Item[i].section_name);
  }
  UDM_FREE(tlist->Item);
  tlist->nitems= 0;
  tlist->mitems= 0;
}


UDM_API(void)
UdmTextListInit(UDM_TEXTLIST *tlist)
{
  bzero(tlist, sizeof(UDM_TEXTLIST));
}


static udm_rc_t
UdmParseTextListItem(const UDM_TEXTITEM *text,
                     UDM_WORD_SCANNER *scanner,
                     int cnvflags,
                     UDM_CONSTWORDLIST *CWL)
{
  /*fprintf(stderr, "flags=%d text=%s\n", cnvflags, text->str);*/
  if (text->Param.flags & UDM_TEXTLIST_FLAG_HTML)
    cnvflags|= UDM_RECODE_HTML_IN;
  return UdmConstWordListAddString(scanner, cnvflags, CWL, text->Param.secno,
                                   text->str, strlen(text->str)); /* TODO34: strlen */
}


udm_rc_t
UdmTextListToConstWordList(const UDM_TEXTLIST *TextList,
                           UDM_UNIDATA *unidata, UDM_CHARSET *cs, int cnvflags,
                           UDM_CONSTWORDLIST *CWL)
{
  size_t i;
  UDM_WORD_SCANNER scanner;
  if (!cs->cset->getword)
    return UDM_OK;
  UdmWordScannerInit(&scanner, unidata, cs);
  for (i= 0; i < TextList->nitems; i++)
  {
    UDM_TEXTITEM *Item= &TextList->Item[i];
    if (Item->Param.secno)
      UdmParseTextListItem(&TextList->Item[i], &scanner, cnvflags, CWL); /* TODO34: check rc */
  }
  return UDM_OK;
}


udm_rc_t
UdmTextItemNormalizeDecimal(UDM_TEXTITEM *Item) /* TODO34: recognize multiple decimal entries */
{
  char decimal[128];
  UdmNormalizeDecimal(decimal, sizeof(decimal), Item->str);
  UdmFree(Item->str);
  Item->str= UdmStrdup(decimal);
  return UDM_OK;
}


void
UdmTextParamInit(UDM_TEXT_PARAM *Param,
                 udm_textlist_flag_t flags, udm_secno_t secno)
{
  Param->flags= flags;
  Param->secno= secno;
}

void
UdmTextItemInit(UDM_TEXTITEM *Item, udm_textlist_flag_t flags)
{
  bzero(Item, sizeof(*Item));
  Item->Param.flags= flags;
}


void
UdmConstTextItemInit(UDM_CONST_TEXTITEM *Item)
{
  bzero(Item, sizeof(*Item));
}
